/*  This file is part of PaceMaker Estimator 2.0.
     PaceMaker Estimator 2.0 is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
     PaceMaker Estimator 2.0 is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with  PaceMaker Estimator 2.0. If not, see <https://www.gnu.org/licenses/>.
*/

#include "mainwindow.h"
#include "ui_dialogheaderofestimate.h"
#include <string.h>

DialogHeaderOfEstimate::DialogHeaderOfEstimate(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::DialogHeaderOfEstimate)
{
    ui->setupUi(this);

/*
 * I cannot use font2.
 * I have to use the setting in
 * void MainWindow::dialogChooseFont() menuEdit.cpp
 */
    //MainWindow* parent2 = qobject_cast<MainWindow*>(this->parent());
    // ((MainWindow *)parent) instead of parent2
    // would have been simpler. Feb 18, 2023. I am a learner bee.

    ui->New->setFont( ((MainWindow *)parent)->globalFont2 );
    ui->Open->setFont( ((MainWindow *)parent)->globalFont2 );
    ui->SaveEstimate->setFont( ((MainWindow *)parent)->globalFont2 );
    ui->cancelPushButton->setFont( ((MainWindow *)parent)->globalFont2 );
    ui->okPushButton->setFont( ((MainWindow *)parent)->globalFont2 );

    blankHeaderFields();

    return;
}

DialogHeaderOfEstimate::~DialogHeaderOfEstimate()
{
    MainWindow *parent = qobject_cast<MainWindow*>(this->parent());

    delete ui;
    parent->oneHeader = nullptr;

    return;
}

void DialogHeaderOfEstimate::on_New_clicked()  // same as File->New
{
    MainWindow *parent = qobject_cast<MainWindow*>(this->parent());

    if ( parent == 0)
    {
        ;//qDebugA() << "header parent is 0";
    }
    else
    {
        parent->dialogNewFilename();
    }
    return;
}

void DialogHeaderOfEstimate::on_Open_clicked() // same as File->Open
{
    //parent()->dialogOpenFilename;
    //ui->parent()->;
    //((MainWindow*)(parent()))->

    MainWindow *parent = qobject_cast<MainWindow*>(this->parent());
    if ( parent == 0 )
    {
        ;//qDebugA() << "header parent is 0";
    }
    else
    {
        parent->dialogOpenFilename();
    }
    return;
}

void DialogHeaderOfEstimate::on_SaveEstimate_clicked()  // same as File->Save
{
    MainWindow *parent = qobject_cast<MainWindow*>(this->parent());
    if (parent == 0)
    {
        ;//qDebugA() << "header parent is 0";
    }
    else
    {
        /*Mar 6 this bug is obvious in retrospect:
         * dialogSaveFilename_sub( false ) fails to save the header
         * because: estimateWasChanged == false
         * although we typed into the header! of course!
         * This would work too as a preliminary:
         *estimateWriteHeader( &(parent->estimateFilename) );
         * but
         *estimateWasChanged = true;
         * should work instead.
         */
        parent->estimateWasChanged = true;
        if( parent->estimateFilename.isEmpty() )
        {
            parent->estimateFilename = parent->estimateDirectory + "/" + "noname";
        }
        parent->dialogSaveFilename_sub( false );

    }

    return;
}

void DialogHeaderOfEstimate::on_cancelPushButton_clicked()
{
    ;//qDebugA() << "cancelPushButton, ignore the header changes";

    this->setVisible( false );

    // reload the header, canceling/ignoring changes.
    MainWindow *parent = qobject_cast<MainWindow*>(this->parent());
    if( parent != nullptr )
       estimateReadHeader( &(parent->estimateFilename) );

    return;
}

void DialogHeaderOfEstimate::on_okPushButton_clicked() // save/go away button
{
    MainWindow *parent = qobject_cast<MainWindow*>(this->parent());
    if (parent == 0)
    {
        ;//qDebugA() << "dia header parent is 0";
    }
    else
    {
        if( parent->estimateFilename.isEmpty() ) // ask filename first
            parent->dialogNewFilename_sub( false ); // do not clear est
        estimateWriteHeader( &(parent->estimateFilename) );
        this->setVisible( false );
    }
    return;
}

/*
char qtymathHLongChar[20];  // not [4]. memory trashed bug happens
long qtymathH=0L, *qtymathHLongpt;
 */

bool DialogHeaderOfEstimate::estimateWriteHeader( QString *filename )
{
    bool ret = false;
    bool existingFile=false;
    qint64 er,readsz,xc;
    MainWindow *parent = qobject_cast<MainWindow*>(this->parent());

    if( existsAndNotEstimateMaker( (*filename ) ) )
    {
        *filename = "cancelled";
        ui->filenameLineEdit->setText( *filename );
        parent->estimateWasChanged = false;
        ret = false;
    }
    else
    {
        QFile fp( (*filename) );

        existingFile = fp.exists();

        this->setWindowTitle( QDir::toNativeSeparators(*filename) );
        ui->filenameLineEdit->setText( stripFilename2( filename ) );

        //memset( buf,' ', 80); // fill with spaces.
        //buf[80] = '\0';

        headstr[0] = 'F';// this is a 1987 char to detect estimate maker filenames
        snprintfreverse( ui->clientLineEdit, headstr+90, 29 );
        snprintfreverse( ui->addressLineEdit, headstr+170, 29 );
        snprintfreverse( ui->cityLineEdit, headstr+250, 14 );
        snprintfreverse( ui->postalCodeLineEdit, headstr+272, 8 );
        snprintfreverse( ui->siteLineEdit, headstr+330, 29 );

        snprintfreverse( ui->dateLineEdit, headstr+49, 29 );
        snprintfreverse( ui->projectLineEdit, headstr+129, 29 );
        snprintfreverse( ui->businessPhoneLineEdit, headstr+209, 29 );
        snprintfreverse( ui->homePhoneLineEdit, headstr+289, 29 );
        snprintfreverse( ui->contactNameLineEdit, headstr+369, 29 );
        snprintfreverse( ui->emailLineEdit, headstr+369+25, 29 );

        if( fp.open(QIODeviceBase::ReadWrite) == false)
            ret =  false;
        else
        {
            if( fp.write( headstr, 401) == 401)
            {
                setstr[0]=timecode;
                setstr[1]=(char)qdec;
                setstr[2]=(char)mdec;
                setstr[3]=(char)tdec;
                setstr[4]=(char)ttdec;
                setstr[5]=(char)tmdec;

                if( fp.write( setstr, 6) == 6)
                {
                    //If offset 407 was not yet written by estimateWrite()
                    if( existingFile == false )
                    {
                        /*
                        qtymathHLongpt = (long *)qtymathHLongChar;
                        qtymathH=0L;
                        *qtymathHLongpt = qtymathH;
                        *
                        if( fp.write( qtymathHLongChar, 4) == 4)
                        */
                        // placeholder for new file:
                        if( fp.write( "\0\0\0\0", 4) == 4)
                        {
                            ret = true;
                        }
                    }
                    else
                    {
                    // written by estimateWrite(), so do not overwrite it.
                        ret = true;
                    }
                }
            }
            fp.close();
        }
    }

    if( ret == false )
    {
        ;//qDebugA() << "Estimate header write failed";
       // QMessageBox msgBox;
       // msgBox.setText("Estimate header write failed.");
       // msgBox.exec();
    }

    return( ret );
}

bool DialogHeaderOfEstimate::existsAndNotEstimateMaker( QString filename )
{
    QMessageBox msgBox;
    bool ret = false;
    char buf[1];

    if( filename.isEmpty() )
    {
        //;//qDebugA() << "blank filename: skip estimate header write";
        //QMessageBox msgBox;
        //msgBox.setText("That filename is a blank.");
        //msgBox.exec();

        return(ret);
    }
    else if( filename.contains(".config") || filename.contains(".lib") || filename.contains("cancelled") )
    {
        return(true);
    }

    QFile fp(filename);

    if( fp.open(QIODeviceBase::ReadOnly) == false )
          ret = false; // let user decide since not able to open.
    else
    {
        if( fp.read( buf, 1) == 1 )
        {
            if( buf[0] != 'F') // all estimate maker files start with F
            {
                msgBox.setText(filename+" is not an Estimator file. Cancel is recommended.");
                msgBox.setInformativeText("Do you really want to erase that file?");
          msgBox.setStandardButtons( QMessageBox::Cancel | QMessageBox::Save );
                msgBox.setDefaultButton(QMessageBox::Cancel);
                int ans = msgBox.exec();
                switch (ans)
                {
                    case QMessageBox::Cancel:
                        ret = true;
                        break;

                    case QMessageBox::Save:
                        ret = false;
                        break;

                    default:
                        ret = true;

                        // should never be reached
                        break;
                }
            }
        }
        else
        {
            ret = false; // A 0 bytes file is acceptable to overwrite.
        }
        fp.close();
    }

    return(ret);
}

char * DialogHeaderOfEstimate::stripEnd( char *ee, size_t sz )
{
    char *bb = ee;
    //size_t sz;  // the old header had a fixed 30 char length for each field.

    for( ; *ee != '\0' && sz>0; --sz) // find end of string
        ++ee;

    // if empty string was sent to us, do nothing

    // strip spaces from end, quitting at not-a-space or quit at beginning

    for( --ee; ee>=bb && *ee == ' ' ; --ee )
        *ee = '\0';

    return(bb);
}

void DialogHeaderOfEstimate::blankHeaderFields()
{
    ui->clientLineEdit->setText("");
    ui->addressLineEdit->setText("");
    ui->cityLineEdit->setText("");
    ui->postalCodeLineEdit->setText("");
    ui->siteLineEdit->setText("");

    ui->dateLineEdit->setText("");
    ui->projectLineEdit->setText("");
    ui->homePhoneLineEdit->setText("");
    ui->businessPhoneLineEdit->setText("");
    ui->contactNameLineEdit->setText("");
    ui->emailLineEdit->setText("");
}

bool DialogHeaderOfEstimate::estimateReadHeader( QString *filename )
{
    bool ret = false;
    unsigned block, lg;
    char buf[51];

    qint64 er,readsz,xc;

    MainWindow *parent = qobject_cast<MainWindow*>(this->parent());

    memset( headstr, ' ', 401);
    blankHeaderFields();

#ifdef OLDWAY1987
    if( (*filename).isEmpty() )
    {
        QMessageBox msgBox;
        msgBox.setText("That filename is blank.");
        msgBox.exec();
        return(ret);
    }
#else
    if( existsAndNotEstimateMaker( *filename ) )
    {
        *filename = "cancelled";
        ui->filenameLineEdit->setText( *filename );
        parent->estimateWasChanged = false;

        return(ret);
    }
#endif

    QFile fp( *filename );

    ui->filenameLineEdit->setText( *filename );
    this->setWindowTitle( stripFilename2( filename ));

    if( fp.open(QIODeviceBase::ReadOnly) == false)
        ret = false;
    else
    {
        er=fp.read( buf, 1);
        if(er == 1 && buf[0]!='F')
        {
            fp.close();
            ;//qDebugA() << "This file:"+(*filename)+" exists and is not an estimate.";

            return(ret);
        }
        else if( er == 1 )
            headstr[0]=buf[0];

        if( er == 1)
          er = fp.read( headstr+1, 400);

        if( er == 400 )
        {
            ui->clientLineEdit->setText( stripEnd(headstr+90, 29) );
            ui->addressLineEdit->setText(stripEnd(headstr+170, 29) );
            ui->cityLineEdit->setText(stripEnd(headstr+250, 14) );
            ui->postalCodeLineEdit->setText(stripEnd(headstr+272, 8) );
            ui->siteLineEdit->setText(stripEnd(headstr+330, 29) );

            ui->dateLineEdit->setText(stripEnd(headstr+49, 29) );
            ui->projectLineEdit->setText(stripEnd(headstr+129, 29) );
            ui->businessPhoneLineEdit->setText(stripEnd(headstr+209, 29) );
            ui->homePhoneLineEdit->setText(stripEnd(headstr+289, 29) );
            ui->contactNameLineEdit->setText(stripEnd(headstr+369, 29) );
            ui->emailLineEdit->setText(stripEnd(headstr+369+25, 29) );

            if( fp.read( setstr, 6) == 6 )
            {
                timecode=setstr[0];
                qdec=(int)setstr[1];
                mdec=(int)setstr[2];
                tdec=(int)setstr[3];
                ttdec=(int)setstr[4];
                tmdec=(int)setstr[5];

                if(timecode=='M'||timecode=='m') // minutes or hours for the estimate units
                {
                    //headinga=heading1;
                    timecode='M';
                }
                else
                {
                    //headinga=heading2;
                    timecode='H';
                }

                if(qdec<0||qdec>5)
                    qdec=0;
                if(mdec<0||mdec>4)
                    mdec=2;
                if(tdec<0||tdec>4)
                    tdec=2;
                if(ttdec<0||ttdec>5)
                    ttdec=1;
                if(tmdec<0||tmdec>5)
                    tmdec=2;

                /* offset 407 is read in estimateRead() instead of here.
                 *
                qtymathHLongpt = (long *)qtymathHLongChar;
                if( fp.read( qtymathHLongChar, 4) == 4)
                {
                    qtymathH = *qtymathHLongpt;
                }
                */
                ret = true;

                ;//qDebugA() << "estimateReadHeader() is ret=true.";
            }
        }
        fp.close();
    }

    return( ret );
}

QString DialogHeaderOfEstimate::stripFilename2( QString *ss )
{
    QDir dd;
    int sepint;

    MainWindow *parent = qobject_cast<MainWindow*>(this->parent());

    parent->linuxMsWindowsSlashFixes(ss);
    /* Qt does correct file opens with \ or / no matter which OS you are in
     * but the next line of code will not work if we did not correct
     * the ss, above, to have the right slash type which matches dd.separator().
     */
    sepint = ss->lastIndexOf( dd.separator() ); // '\' or '/'

    if( sepint != -1 && (ss->size() > (sepint+1)) )
    {
        *ss = ss->sliced( sepint+1 );
        ;//qDebugA() << "strip2: sliced stripFilename of its dir" << *ss;
    }
    else
        ;//qDebugA() << "strip2: not sliced stripFilename of its dir" << *ss;

    return ( *ss );
}


/************* Below are the write to disk functions **************/

// I wrote this so I can use all the buffer to lineEdits sprintf
// statements when I do the reverse copy of the lineedits to buffer.
// This way I can have fewer typing mistakes and if I change the
// read function then I can copy and paste into the write function
// and only change the sprintf() to sprintfreverse();
void DialogHeaderOfEstimate::snprintfreverse( QLineEdit *arg1, char *arg2, size_t sz )
{
    unsigned yy;
    char *pt;

    /* Same as sprintf( arg2, "%-*.*s", sz, sz, arg1 );
     * except arg1 is not a QLineEdit() but a string
     * and arg2 is not '\0' terminated by my function here:
     */

    strncpy( arg2, arg1->text().toStdString().c_str(), sz );

    // pad out with spaces
    for( yy=0, pt=arg2; yy<sz; ++yy, ++pt)
        if( *pt == '\0' )
            *pt = ' ';

    return;
}

