/*  This file is part of PaceMaker Estimator 2.0.
    PaceMaker Estimator 2.0 is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
    PaceMaker Estimator 2.0 is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with PaceMaker Estimator 2.0. If not, see <https://www.gnu.org/licenses/>.
*/

#include "estlineedithistory.h"
#include "mainwindow.h"

estLineEditHistory::estLineEditHistory( MainWindow *xx, long targ, long src, int typeOfAction )
{
    QString ss;
    unsigned yy;

    saveTarg = targ;
    saveSrc = src;
    saveTypeOfAction = typeOfAction;
    saveParent = xx;

    /* archive the first state for that line so undo/redo estUndoCommand works.
     * See estUndoCommand.cpp where this estLineEditHistory object data is used.
     */
    //if( xx->estRow[targ] != nullptr )
    this->oldestrefcode = xx->estRow[targ]->refcode;

    for( yy=1; yy<=7; ++yy)
    {        
        // this probably doesn't happen since I stopped doing yy=0
        if( xx->estRow[targ]->col[yy] == nullptr )
        {
            this->oldestcol[yy] = "";
            this->oldfQB = nullptr;
        }
        else
        {
            this->oldestcol[yy] = xx->estRow[targ]->col[yy]->text();
            this->oldfQB = xx->estRow[targ]->fileQtyBlock;
        }

        // these are QLineEdit pointers to one estimate line
        this->estRowActual[yy] = xx->estRow[targ]->col[yy];
    }

    this->fileQtyBlockActual = &(xx->estRow[targ]->fileQtyBlock);
    this->estRowActualrefcode = &(xx->estRow[targ]->refcode);

    switch( typeOfAction )
    {
        case TYPELibEst:
            preferredScrollIntoViewQle = xx->estRow[targ]->col[1];

        // archive the second state for that line
        this->estrefcode = xx->libRow[ src ]->refcode;

        ;//qDebugA() << "Add a line from lib to est for this row:" << src;
        if( (ss = xx->libRow[ src ]->col[1]->text()) == nullptr )
            ss = "";
        ;//qDebugA() <<ss;
        this->estcol[1] = ss;

        if( (ss = xx->libRow[ src ]->col[2]->text()) == nullptr )
            ss = "";
        ;//qDebugA() <<ss;
        this->estcol[2] = ss;

    // ERASE estRow col[3], which is Qty  because this is a fresh library item to est side
        this->estcol[3] = "";
        this->fQB = nullptr;

    // Column 3 lib goes to 4 est. It's not a mistake.
        if( ( ss = xx->libRow[ src ]->col[3]->text()) == nullptr )
            ss = "";
        ;//qDebugA() <<ss;
        this->estcol[4] = ss;

        if( ( ss = xx->libRow[ src ]->col[4]->text()) == nullptr )
            ss = "";
        ;//qDebugA() <<ss;
        this->estcol[5] = ss;

        this->estcol[6] = ""; // two totals columns are uncalculated blanks
        this->estcol[7] = ""; // because there is no Qty value yet.

        break;

        case TYPEDel:
            preferredScrollIntoViewQle = xx->estRow[targ]->col[1];

        /* done by .redo(), so skip all this
         *  xx->delLineEstSide(targ);

            // archive the second state for that line
            // This only does one line for an example. There are 998 of them.
            for( yy=1; yy<=7; ++yy)
            {
                if( targ < ESTROWCOUNT-1 )
                {
                    this->estcol[yy] = xx->estRow[targ+1]->col[yy]->text();
                    this->fQB = xx->estRow[targ+1]->fileQtyBlock;
                }
            }
         */
        break;

        case TYPEIns:
            preferredScrollIntoViewQle = xx->estRow[src]->col[1];

        /* set src to desired row, and targ=ESTROWCOUNT-1
         * before calling this function because we are doing
         * last row and negative increments for Ins, not positive.
         */
        /* done by .redo(), so skip all this.
         * xx->insertBlankLineEstSide(src);

            // archive the second state for that line
            // This only does one line for an example. There are 998 of them.
            for( yy=1; yy<=7; ++yy)
            {
                if( src > 1 )
                {
                    this->estcol[yy] = xx->estRow[src-1]->col[yy]->text();
                    this->fQB = xx->estRow[src-1]->fileQtyBlock;
                }
            }
         */
        break;

        case TYPEmenuEdit:
            preferredScrollIntoViewQle = xx->estRow[targ]->col[1];

            // archive the second state for that line
            this->estrefcode = xx->copyE->estRow[ src ]->refcode;

            ;//qDebugA() << "Add a line from lib to est for this row:" << src;
            if( (ss = xx->copyE->estRow[ src ]->col[1]->text()) == nullptr )
                ss = "";
            ;//qDebugA() <<ss;
            this->estcol[1] = ss;

            if( (ss = xx->copyE->estRow[ src ]->col[2]->text()) == nullptr )
                ss = "";
            ;//qDebugA() <<ss;
            this->estcol[2] = ss;

            // Column 3 lib goes to 4 est. It's not a mistake.
            if( ( ss = xx->copyE->estRow[ src ]->col[3]->text()) == nullptr )
                ss = "";
            ;//qDebugA() <<ss;
            this->estcol[3] = ss;

            if( ( ss = xx->copyE->estRow[ src ]->col[4]->text()) == nullptr )
                ss = "";
            ;//qDebugA() <<ss;
            this->estcol[4] = ss;

            if( ( ss = xx->copyE->estRow[ src ]->col[5]->text()) == nullptr )
                ss = "";
            ;//qDebugA() <<ss;
            this->estcol[5] = ss;

            this->estcol[6] = ""; // two totals columns are uncalculated blanks
            this->estcol[7] = ""; // because there is no Qty value yet.

            this->fQB = xx->copyE->estRow[src]->fileQtyBlock;
            if( this->fQB != nullptr )
            {
                for(int bb=0;bb<17;++bb)
                {
                    this->fQB->block[bb] = xx->copyE->estRow[src]->fileQtyBlock->block[bb];  // 680 chars 17x40
                }
                //unused I think: this->fQB->refcode = copyE->estRow[rr]->fileQtyBlock->refcode;
            }
            this->estrefcode = xx->copyE->estRow[src]->refcode;

        break;

        default:
            preferredScrollIntoViewQle = xx->estRow[targ]->col[1];

        break;
    }

    xx->estimateWasChanged = true;

    return;
}
