/*  This file is part of PaceMaker Estimator 2.0.
    PaceMaker Estimator 2.0 is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
    PaceMaker Estimator 2.0 is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with PaceMaker Estimator 2.0. If not, see <https://www.gnu.org/licenses/>.
*/

#include "mainwindow.h"
#include "estundocommand.h"

estUndoCommand::estUndoCommand( estLineEditHistory *xx, QUndoCommand *parent )
                : QUndoCommand(parent)
{
    /* "archive" is the est line qlineedit pointers,
     * which point to
     * the estimate, and it has the before and after
     * text for each column of that estimate line.
     */
    archive = xx;
}

void estUndoCommand::undo()
{
    unsigned yy;

    if( archive->saveTypeOfAction == TYPEDel )
    {
        archive->saveParent->insertBlankLineEstSide(archive->saveTarg);
    }
    else if( archive->saveTypeOfAction == TYPEIns )
    {
        archive->saveParent->delLineEstSide(archive->saveSrc);
    }

    *(archive->estRowActualrefcode) = archive->oldestrefcode;
    *archive->fileQtyBlockActual = archive->oldfQB;

    // Copy the "before" text data into the qlineedits
    for( yy=1; yy<=7; ++yy )
    {
        archive->estRowActual[yy]->selectAll();
        archive->estRowActual[yy]->insert( archive->oldestcol[yy] );

        /* I use insert() instead because it allows undo/redo interior to
           each QLineEdit whereas setText() flushes all that history.
        archive->estRowActual[yy]->setText( archive->oldestcol[yy] );
        */


        //oldfQB is used for TYPEDel TYPEIns because they
        //lost one est line above from a Del and one at the last est line=999
        // from an Ins. Those must be recovered.
    }
    /* Since March 2023 I added subtotal/total.
     * There is a smarter way to do this, but let's do it the stupid
     * error free way for now:
     */
    archive->saveParent->setDRAGtotalsBottomSubtotalRow();
    archive->saveParent->searchForSubtotalTotalUpdateThem( 1 );
    //archive->estRowActual[1]->setReadOnly(false); in delLineEstSide instead

    archive->saveParent->estimateWasChanged = true;
}

void estUndoCommand::redo()
{
    unsigned yy;

    if( archive->saveTypeOfAction == TYPEDel )
    {
        archive->saveParent->delLineEstSide(archive->saveTarg);
    }
    else if( archive->saveTypeOfAction == TYPEIns )
    {
        archive->saveParent->insertBlankLineEstSide(archive->saveSrc);
    }
    else  // TYPELibEst
    {
        *(archive->estRowActualrefcode) = archive->estrefcode;
        *archive->fileQtyBlockActual = archive->fQB;

        // Copy the "after" text data into the qlineedits
        for( yy=1; yy<=7; ++yy )
        {
            archive->estRowActual[yy]->selectAll();
            archive->estRowActual[yy]->insert( archive->estcol[yy] );

            //archive->estRowActual[yy]->setText( archive->estcol[yy] );

            //fQB is not used for TYPEDel TYPEIns because they
            //copy from +1 line above (Ins) or -1 line below (Del).
            //See keyInsDelRows.cpp for what I mean.
            //One fQB is insufficient. I need 998 lines moved.
            // I did understand I could add 998 items to my estlineedithistory
            // object but I did not bother because the estimate itself
            // has all those items, one line down, or one line up.

            // When I wrote the first version for TYPELibEst only,
            // I did not understand that QUndoCommand is
            // executing .redo() immediately.
            // You cannot tell there is a double action because it
            // is acting on the same row for TYPELibEst, but for
            // TYPEDel and TYPEIns it did double scroll ups and scroll downs.
            // That is when I realized I never entirely understood QUndoCommand
            // until now, Mar 4, 2023. I commented out the estlineedithistory.cpp
            // Del and Ins commands so as to not do doubles, but that damaged
            // the value stored in archive->fQB because it is not the AFTER
            // scroll like it MUST BE. Do not worry. The code works anyway.
            // (edit:change it anyway)
            // The middle part in the switch statement
            // is not supposed to touch the est lines directly at all.
            // Instead it should all be erased and all the loading of
            // "second state" should be directly from the lib side into the
            // estcol[] variables and then the .redo() will immediately write
            // to the est lines, which is what it does anyway, right now.
            // I am going to change it. I write a note here
            // for next time I do QUndoCommand. I will have some
            // documentation. I will never remember this quirk about redo().

       /* !!!! Qt is mislabelled redo(). It is both: immediately-do and redo. */

            /* Here is the coding which works but is incorrect,
             * as described above. Look at the new estlinehistory.cpp (Mar5)
             * and see it also works, and is, instead, correct.
             * (Mar4):
                    switch( typeOfAction )
                    {
        case TYPELibEst:
            preferredScrollIntoViewQle = xx->estRow[targ]->col[1];

        ;//qDebugA() << "Add a line from lib to est for this row:" << src;
        if( (ss = xx->libRow[ src ]->col[1]->text()) == nullptr )
            ss = "";
        ;//qDebugA() <<ss;
        xx->estRow[targ]->col[1]->selectAll();
        xx->estRow[targ]->col[1]->insert( ss );

        if( (ss = xx->libRow[ src ]->col[2]->text()) == nullptr )
            ss = "";
        ;//qDebugA() <<ss;
        xx->estRow[targ]->col[2]->selectAll();
        xx->estRow[targ]->col[2]->insert( ss );

        if( (ss = xx->libRow[ src ]->col[3]->text()) == nullptr )
            ss = "";
        ;//qDebugA() <<ss;

    // ERASE estRow col[3], which is Qty  because this is a fresh library item to est side
        xx->estRow[targ]->col[3]->clear();
        xx->estRow[targ]->fileQtyBlock = nullptr;

        xx->estRow[targ]->col[4]->selectAll();
        xx->estRow[targ]->col[4]->insert( ss ); // this is the corresponding library column

        if( ( ss = xx->libRow[ src ]->col[4]->text()) == nullptr )
            ss = "";
        ;//qDebugA() <<ss;
        xx->estRow[targ]->col[5]->selectAll();
        xx->estRow[targ]->col[5]->insert(ss);

        xx->estRow[targ]->col[6]->clear(); // totals are unknown until Qty is typed
        xx->estRow[targ]->col[7]->clear();

        break;

        case TYPEDel:
            preferredScrollIntoViewQle = xx->estRow[targ]->col[1];

        // done by .redo(), so skip this: xx->delLineEstSide(targ); // set targ = src before calling
        break;

        case TYPEIns:
            preferredScrollIntoViewQle = xx->estRow[src]->col[1];

         * set src to desired row, and targ=ESTIMATEROWCOUNT-1
         * before calling this function because last row needs recovery for Ins.
         *
         done by .redo(), so skip this: xx->insertBlankLineEstSide(src); break;

             *
             */
        }
        /* Since March 2023 I added subtotal/total.
         * There is a smarter way to do this, but let's do it the stupid
         * error free way for now:
         */
        archive->saveParent->setDRAGtotalsBottomSubtotalRow();
        archive->saveParent->searchForSubtotalTotalUpdateThem( 1 );
        //archive->estRowActual[1]->setReadOnly(false); in delLineEstSide instead
    }
    archive->saveParent->estimateWasChanged = true;
}

void estUndoCommand::delete_estLineEditHistory()
{
    delete archive;
}

