/*  This file is part of PaceMaker Estimator 2.0.
    PaceMaker Estimator 2.0 is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
    PaceMaker Estimator 2.0 is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with PaceMaker Estimator 2.0. If not, see <https://www.gnu.org/licenses/>.
*/

#include "mainwindow.h"
#include "qthread.h"
#include "qtimer.h"
#include "libundocommand.h"
#include "estundocommand.h"
#include <QDrag>
#include <QMimeData>
#include "totalsubtotal.h"

/* https://doc.qt.io/qt-6/qobject.html#installEventFilter
 *
 * eventFilter() are not all the events. Look inside mainwindow.h
 * for events functions in addition to the ones called by connect() in load2.cpp.
 * They are coded by Qt Creator menus and look invisible, as far as searching
 * through these .cpp and .h files go.
 * Look for their names in mainwindow.h under "private slots" and use Qt Creator
 * to research how those slot methods get called.
 */
static unsigned countspontaneousEvents = 3333;
static unsigned thresholdSpont=590; //this worked on my PC for font change

static QDialog *QDia = nullptr;
static QLabel *msgBox3;

void MainWindow::waitLoading(unsigned on, unsigned setLimit ) // start at 0
{
    if( on == 0 ) // reset countspontaneousEvents to start it
    {
        countspontaneousEvents = 0;
        if( setLimit != 0 )
            thresholdSpont = setLimit;

       /* Bug: If the user changed fonts then the geometry
        * is wrong and the label doubles like cross-eyed.
        * (I don't know why. It does not happen anylonger.)
        */
        if( !QDia )
        {
            // StaysOnTop blocks keyboard input on other windows!
            //QDia = new QDialog( this, Qt::WindowStaysOnTopHint);
            //QDia->setModal(false); and ignores modal false.

            QDia = new QDialog(nullptr); // use nullptr for BIG CENTRED
            //QDia = new QDialog(this); // use "this" to specify size below

            //QDia->setParent();
            QDia->setWindowTitle("* wait *");
            QDia->setSizeGripEnabled(true);
            //QDia->setFixedSize( ui->scrollArea_4->geometry().x()/2, ((ui->scrollArea_4->geometry().x()/2)*1000)/1618L );
            msgBox3 = new QLabel("Wait\nWhile\nLoading",QDia);
            msgBox3->setFrameStyle(QFrame::Panel | QFrame::Sunken);
        }
        QDia->show();
        QDia->raise();
        // bad idea: this->hide();

        /* Wait for GUI to fill text area by processing events
         * before doing combox reading from disk in mainwindow.cpp.
         */
        QThread::sleep(1L);
        // I get an earlier White screen when doing dialogFontChange
        // with this sleep here.
        // I guess there is a separate thread that continues here
        // before the show() is actually finished, above.
        // The parent (below here) will waste cpus forming a frame
        // if I do not sleep. Apparently, those cpus delay the
        // white in my show() if I do not sleep 1s.
        QApplication::processEvents(QEventLoop::AllEvents);

        //QDia->setWindowState(Qt::WindowState::WindowActive);
        //QDia->setWindowState( (windowState() & ~Qt::WindowMinimized) | Qt::WindowActive);

       /* This is a long late SLOT(accept()). It is a failsafe in case
        * spontaneous() never reaches number of countspontaneousEvents,
        * and never erases the wait sign the normal way.
        * I cannot image why this would be needed, though;
        * spontaneous() always happens? A blinking cursor generates them.
        */
        QTimer::singleShot( 140000, Qt::CoarseTimer, QDia, SLOT(accept()) );
    }
    else if( on == thresholdSpont ) //number of countspontaneousEvents
    {
        /* goes with bad idea below
            this->show();
            this->raise();
         */

        if( QDia )
        {
            // accept() closes the QDialog wait.
            QTimer::singleShot( 1000, Qt::CoarseTimer, QDia, SLOT(accept()) );
        }
    }
    /* bad idea: The event spontaneous() signals get turned off.
    else if( on < thresholdSpont )
    {
        this->hide();
        QDia->show();
        QDia->raise();
    }
    */

    return;
}

static long Ev=0L;
bool MainWindow::eventFilter(QObject *obj, QEvent *event)
{
    estLineEditHistory *eH;
    estUndoCommand *command;
    QChar cc;
    QMessageBox msgBox;
    int idx;

    //;//qDebugA() << "Ev" << ++Ev;
    //MainWindow *parent = qobject_cast<MainWindow*>(this->parent());

    //This all worked: ;//qDebugA() << "e";   only when I did example: ui->est33->installEventFilter( this ) in MainWindow(), above.

    //;//qDebugA() << "mylinedit was sending events" << event->type() ;

    //if(event->type() == QEvent::StatusTip )
    // <-this can mean the statusTip() changed by Entering. I know this occurs
    // when mouse enters a widget
    // I dropped it because it gets triggered 2x: keyboard and mouse, of course!

    if( event->type() == QEvent::KeyPress)
    {
        QKeyEvent *ke = static_cast<QKeyEvent *>(event);

        if( ke->text().isNull() == true )
            cc = '\0';
        else
        {
            QString ss = ke->text().toStdString().c_str();
            cc = ss.data()[0];
        }

        if( cc == '*') // this is the same as select library drop-menu
        {
            if( /*keyboardnow.libest == 2 &&*/ keyboardnow.col == 1 )
            {
                msgBox.setText("Double click library code to switch library lists\nis the new way. Typing *WD was the 1987 program.");
                msgBox.exec();
                // do not let * be typed into the libcode.
                return(true);
            }
        }
        else if( cc == '=') // libcode restriction for subtotal/total
        {
            if( /*keyboardnow.libest == 2 &&*/ keyboardnow.col == 1 )
            {
                // do not let = be typed into the libcode.
                return(true);
            }
        }
        else if (ke->key() == Qt::Key_Insert)
        {
            if( isFullRow( (keyboardnow.libest == 2)?estRow[ESTROWCOUNT-1] : libRow[LIBROWCOUNT-1]) )
            {
                ;
            }
            else
            {
                getrowcol_ofcursors( &keyboardnow.libest, &keyboardnow.row, &keyboardnow.rowL, &keyboardnow.rowE, &keyboardnow.col, &keyboardnow.colL, &keyboardnow.colE);

            if( keyboardnow.libest == 2 )
            {
                //MainWindow::insertBlankLineEstSide(keyboardnow.row);

                eH = new estLineEditHistory(
                    this, ESTROWCOUNT-1, keyboardnow.row, TYPEIns);
                command = new estUndoCommand( eH );
                est_undo_stack.push(command);

                show_linenums_showHighlights( &keyboardnow.litLibButton, &keyboardnow.litEstButton, keyboardnow.rowL, keyboardnow.row );
            }
            else
            {
                //MainWindow::insertBlankLineLibSide(keyboardnow.row);

                libLineEditHistory *eH = new libLineEditHistory(
                    this, LIBROWCOUNT-1, keyboardnow.row, TYPEIns);
                libUndoCommand *command = new libUndoCommand( eH );
                lib_undo_stack.push(command);
            }
            }
        }
        else if (ke->key() == Qt::Key_Delete)
        {
            getrowcol_ofcursors( &keyboardnow.libest, &keyboardnow.row, &keyboardnow.rowL, &keyboardnow.rowE, &keyboardnow.col, &keyboardnow.colL, &keyboardnow.colE);

            if( keyboardnow.libest == 2 )
            {
                /* new March 2023. Never let = == == libcodes get into the undo
                 * stack because those are subtotal/total blocks of the estimate
                 * and you must not recover one line of these. They are blocks.
                 * User can drag/drop a new one, instead.
                 */
                if( estRow[keyboardnow.row]->col[1]->text().indexOf("=") >= 0 )
                    delLineEstSide(keyboardnow.row);
                else
                {
                    eH = new estLineEditHistory(
                        this, keyboardnow.row, keyboardnow.row, TYPEDel );
                    command = new estUndoCommand( eH );
                    est_undo_stack.push(command);
                }
                show_linenums_showHighlights( &keyboardnow.litLibButton, &keyboardnow.litEstButton, keyboardnow.rowL, keyboardnow.row );
            }
            else
            {
                //MainWindow::delLineLibSide(keyboardnow.row);
                libLineEditHistory *eH = new libLineEditHistory(
                    this, keyboardnow.row, keyboardnow.row, TYPEDel );
                libUndoCommand *command = new libUndoCommand( eH );
                lib_undo_stack.push(command);
            }
        }
        else
        {
            if( keyboardnow.col == 1 )
            {
                if( cc.isLetterOrNumber() == true )
                {
                    if( keyboardnow.libest == 2 )
                    {
                        cc = cc.toUpper();

                        // do search the library as we type and pass chars normally.
                        searchLibrary(
                            QString("%1%2").arg(estRow[keyboardnow.row]->col[1]->text()).arg(cc).toUpper()
                            );
                    }
                }
                else if( cc.isPrint() == true )
                {
                    // discard punctuation and spaces ... .
                    return(true);
                }
                else
                {
                    ;//backspace, ctrl-Z, et cetera, will go through normally.
                }
            }
            else if( keyboardnow.libest == 2
                     && keyboardnow.col == 3 )
            {
                if( cc == '\x11' )
                {
                    //;//qDebugA() << cc ;

                    // 17 ctrl-Q (old Qty hot key 1987)

                    on_qtymath_pressed();
                    on_qtymath_released();
                }
                else if( cc == '\x01' ) // ctrl-a ctrl-A Ctrl-A Ctrl-a
                {
                    /* Insert a version
                     * the About -> actionQtLicence
                     *     qtmsg.setWindowTitle("PaceMaker Estimator V2.0");
                     * for future use. It could track date of download
                     * by modifying the totalhours.png file date:
                     * unix command: touch totalhours.png
                     *
                     * get modify-date of file
                     * insert date into title thusly:
                     */
 QFileInfo *executable=new QFileInfo(QCoreApplication::applicationDirPath()+"/PaceMakerEstimator");
                    ;//qDebugA( << QCoreApplication::applicationDirPath()+"/PaceMakerEstimator";
                    if( executable->exists() )
                    {
                            QDate qdate=executable->lastModified(QTimeZone::UTC).date(); setqtmsgWindowTitle( QString("v2.0%1%2%3 \x41uthor: \x44\x61\x76\x69\x64\x20\x50\x61\x63\x65, Canada national").arg(qdate.year()).arg(qdate.month()).arg(qdate.day()));
                    }
                    actionQtLicence();
                }
            }
        }
    }
    else if( event->type() == QEvent::MouseButtonDblClick )
    {
        if( keyboardnow.libest == 2 && keyboardnow.col == 1 )
        {
            /* double click on libcode on est side triggers library search
             * load this library list (by refcode). Is trimmed() needed?
             * Yes.
             * I see the refcode was estimateRead() in menuNewOpen.cpp as
             * padded with spaces, like this: "0 " example.
             * old:estRow[keyboardnow.row]->refcode.remove('*').trimmed();
             */

            // This is the library filename:
            if( estRow[keyboardnow.row]->refcode.isEmpty() == false )
            {
                for( idx=0; idx < ui->comboBox->count(); ++idx)
                {
                    QVariant vv;
                    vv = ui->comboBox->itemData(idx,Qt::UserRole);
                    QString *ss = (QString *)vv.data();
     ;//qDebugA() << ss <<"combo,filename"<< estRow[keyboardnow.row]->refcode;

                    if( *ss == estRow[keyboardnow.row]->refcode )
                    {
                        //ui->comboBox->currentIndex();
                        loadByIndexLibrarynameOfComboboxLibs(idx);
                        break;
                    }
                }
            }
            searchLibrary(
                estRow[keyboardnow.row]->col[1]->text().toUpper()
                         );
            //estRow[keyboardnow.row]->col[1]->text()
        }
    }
#ifdef OMITsinceMar7_2024
    /*outside,system event==true*/
    else if (event->type() == QEvent::Enter ) // not Key_Enter
    {
        // This means mouse entered the widget; not clicked on it.

        //qDebug() << "Hi to Enter StatusTip (mouse enters)" << mousenow.row;

        /*works perfectly but TOO SLOW therefore I decided not
         *to keep track of row by mouse:
         *
         *    getrowcol_ofcursors( &keyboardnow.libest, &keyboardnow.row, &keyboardnow.rowL, &keyboardnow.rowE, &keyboardnow.col, &keyboardnow.colL, &keyboardnow.colE);

         */
    }
#endif
    else if (event->type() == QEvent::FocusIn ) //clicked on it
    {
        getrowcol_ofcursors( &keyboardnow.libest, &keyboardnow.row, &keyboardnow.rowL, &keyboardnow.rowE, &keyboardnow.col, &keyboardnow.colL, &keyboardnow.colE);

        //clear libcode srchlist and unhighlight libcode
        searchLibrary(nullptr);

        /* I moved this to show_linenums_showHighlights()
         * because FocusOut was turning off the qle frame when
         * focus went away to the qtymath widget,
         * but I only wanted it to turn off
         * for focus changes in MainWindow.
         *
        if( qtymathRequestsQtyColBoxHighlight == true
                && keyboardnow.col == 3 )
        {
            QLineEdit *qle = (QLineEdit *)obj;
            qle->setFrame(globalFrame==false?true:false);
        }
         */

        if( keyboardnow.libest == 2 )
        {
                show_linenums_showHighlights( &keyboardnow.litLibButton, &keyboardnow.litEstButton, keyboardnow.rowL,keyboardnow.row );
            //if( keyboardnow.col == 1 )
              //MainWindow::searchLibrary(nullptr);//clear libcode srchlist
        }
        else
            show_linenums_showHighlights( &keyboardnow.litLibButton, &keyboardnow.litEstButton, keyboardnow.row,keyboardnow.rowE );
        //;//qDebugA() << "FocusIn event (keyboard) on row:" << keyboardnow.row;

        /* This is only setting the name of a drop menu inside Edit */
        QString selected;
        if( keyboardnow.row==0 )
            selected = "Click a Row";
        else
            selected = QString("Clicked Row %1").arg(keyboardnow.row);
        ui->actionSelected_Row->setText( selected );
    }
    else if (event->type() == QEvent::FocusOut )//clicked other
    {
        /* I moved this to show_linenums_showHighlights()
         * because it was turning off the qle frame when
         * focus went away to the qtymath widget,
         * but I only wanted it to turn off
         * for focus changes in MainWindow widget.
         *
        if( qtymathRequestsQtyColBoxHighlight == true
                && keyboardnow.col == 3 )
        {
            QLineEdit *qle = (QLineEdit *)obj;
            qle->setFrame(globalFrame);
        }
         */

        //;//qDebugA() << "FocusOut event (keyboard) on row:" << keyboardnow.row;

        // I know by try/error that FocusOut triggers first and editingFinished second.
        // If I TAB or mouse out of a square, which has a number and needs doCalc() then
        // the error was that the keyboardnow.row got set to 0 before doCalc() was called
        // and so doCalc always failed. So I moved it into here so it matters-not, which
        // event calls it first. It only runs once, if keyboardnow.col is already set to 0
        // by a previous call to doCalc.

        // Also, my next experiment proved that docalc() damages
        // QLineEdit's undo() redo() function. When I comment
        // docalc() out then undo/redo works for EVERY
        // qlineedit with extreme depth of memory. wow.
        // Ctrl-Z and Shift-Ctrl-Z also work (even when doCalc() is running, by the way)
        // I presume the fact that docalc() is setting some
        // QLineEdits, (including the description?? (doubt it))
        // is causing a damage to the undo/redo depth.
        // Solution? Do doCalc QLineEdit.setText()s ONLY
        // for when a change has occurred in Qty/HrsEach/CostEach
        // Did I do easy way and never checkif anything changed?
        // Feb1, 2023 DP. I checked doCalc() validateField()
        // and nothing is setText unless you touch col 3,4, or 5.
        // Edit: I was wrong.
        // The validate.cpp did setText()
        // all of the time and that ERASES undo-stack history!
        // So I changed validateFields() to not do setText()
        // for cases where validateFields() need not do a change.
        // BUG FIXED! undo/redo works perfectly unless
        // you type letters into the Qty field, for example,
        // which causes a setText() to remove nonDigits.

        // Edit: I learned from stackexchange: clear() or selectAll()
        // and insert(), instead of setText(), will maintain
        // the undo/redo stack. It works! See validate.cpp
        // qlineedit.setText()
        //    "Setting this property clears the selection,
        //     clears the undo/redo history, moves the cursor to the end of the line and resets the modified property to false. The text is not validated when inserted with .
        // qlineedit.clear()
        //    "Clears the contents of the line edit".
        // qlineedit.insert()
        //    "Deletes any selected text, inserts newText,
        //     and validates the result. If it is valid,
        //     it sets it as the new contents of the line edit."

        if( keyboardnow.libest == 2)
            doCalc();

        /* I moved this to docalc.cpp because qtymath can change an estimate
         * and it calls doCalc(); I think FocusOut might happen anyway.
         * But I moved estimateWasChanged = true to doCalc is safe, so not doing
         * it here too:
         *
            if( keyboardnow.libest == 2 ) // est side is 2
                estimateWasChanged = true;
         */

        historySave_ZeroSetKeyboardCursor();

        //show_linenums_showHighlights( &keyboardnow.litLibButton, &keyboardnow.litEstButton, mousenow.rowL, mousenow.rowE );
        //I was sending 0s which come up as ? and ?
        //show_linenums_showHighlights( &keyboardnow.litLibButton, &keyboardnow.litEstButton, keyboardnow.row, keyboardnow.row );
        // historical values have capital L or E and newest values have no capitals.
        show_linenums_showHighlights( &keyboardnow.litLibButton, &keyboardnow.litEstButton, keyboardnow.rowL, keyboardnow.rowE );
    }
    else if (event->type() == QEvent::Drop )
    {
        /* Note re: search"on_TotalHours_pressed" and think about dragging it:
         * We are not going to let QLineEdit accept the drag text. Instead, we are
         * to notice the row number, insert a blank line, and do our work on
         * the new line. If we do not intercept these drop events then QLineEdit
         * has its own drag accept function which will work with this event, like
         * inserting the mimeData text into the QLineEdit, for example.
         */
        /* There could be a second purpose for using Leave Drop drag-events.
         * If the Leave and Drop is from/to a QLineEdit() instead of TotalHours_pressed,
         * then I want to highlight the libRow[] or estRow[] start/finish for a
         * "Select" and prepare for "Copy" and "Paste" or "Cut" in a menu.
         * How do I distinguish this from a Drag from button on_TotalHours_pressed()?
         *
         * Third: A superior idea: When I let go of the button
         * QEvent::MouseButtonRelease
         * and the displayed Status row number differs from
         * the keyboard row number. that observation is useful for doing
         * select, as in "copy/paste", instead of using drag events.
         * Is it superior to drag? The mouse wheel and down arrow will
         * let you move down beyond the visible screen and thus scroll
         * down to highlight more lines. Maybe drag does that too,
         * although drag->exec() will not do scroll. I tried it on
         * on_TotalHours_pressed. So I am going with See QEvent::MouseButtonRelease
         */
        return(true);/* Without passing event or else QLineEdit will handle
                      * the signal and that is bad because I want the signal
                      * for my uses. See different return() below.
                      */
    }
#ifdef UNUSED
    else if (event->type() == QEvent::StatusTip )
    {
        /*
         * The status tip in the bottom left region of MainWindow display
         * does those tip displays without me coding them anywhere.
         * I checked connect() and SLOTS and I did not code them here, nor anywhere.
         * I setText() them in load2() and they run on their own.
         *    if( parseRowFromStatusTip( &row, QLineEdit estRow or libRow )==true )
         */
        //int row;
        //if( parseRowFromStatusTip( &row, qle ) == true )

        /* This event is a request therefore the currentMessage is the old message! */
        qDebug() << statusBar()->currentMessage();
        // ui->centralwidget->focusWidget()->statusTip();

    }
#endif
    else if( event->type() == QEvent::MouseButtonRelease )
    {
        getrowcol_ofcursors( &keyboardnow.libest, &keyboardnow.row, &keyboardnow.rowL, &keyboardnow.rowE, &keyboardnow.col, &keyboardnow.colL, &keyboardnow.colE);

        //getrowcol_ofcursors();  These fetch row for keyboard

        /* getrowcol...() is fetching two sides for keyboard
         * and it is coding 2 row variables for those 2 sides but not
         * all of the time, just when the side is involved. The
         * row col names are scratch pad, no side specified.
         *
         *
         * It also sets a third variable called: keyboardnow.libest
         * which is lib=1 est=2.
         *
         * Only think about the keyboard cursor now, not mouse cursor.
         *
         * MainWindow->statusBar->currentMessage(); is late by
         * showing previous message when I put
         * it in (event->type() == QEvent::StatusTip )
         * which does a request event, not a post-request event.
         * The mouse-hover involves StatusTip().
         *
         * qDebug() << statusBar()->currentMessage();
         */
        //qDebug() << "statusBar says" << statusBar()->currentMessage();
        //qDebug() << "statusBar central" << ui->centralwidget->focusWidget()->statusTip();
        //qDebug() << keyboardnow.row << keyboardnow.libest;
        if( keyboardnow.libest == 2 ) // keyboard cursor is blinking on 1=lib, 2=est
        {
            if( keyboardnow.row  )
                ;//qDebug() << "statusTip estside: keyboardrow/mouserow" << keyboardnow.row;
        }
        else if( keyboardnow.libest == 1 ) // keyboard cursor is blinking on 1=lib, 2=est.
        {
            if( keyboardnow.row )
                ;//qDebug() << "statusTip libside: keyboardrow/mouserow" << keyboardnow.row;
        }
    }
    else if( countspontaneousEvents<=thresholdSpont+30 && event->spontaneous() )
    {
    /*  This works but it does 183 and then 533 happens.
     *  This is for WAIT-until-loaded indicator at dialog font change
     *  and at slow Qt start up, since I have 8000+ QLineEdits.
     *  I do not know why counting these works as a tool to time
     *  when to remove the wait dialog of waitLoading();
     */
        //fprintf( stderr,"\rspontaneous() event..cpp: %d", (thresholdSpont - (countspontaneousEvents++)) );
        fprintf( stderr,"\rspontaneous() event..cpp: %d", ++countspontaneousEvents );
        //fprintf( stderr,"." );
        fflush(stderr);
        waitLoading( countspontaneousEvents, 0 );
    }
    //else if( event->type() == QEvent::PolishRequest )
    //{
    //  This works too early. It happens at the beginning of the show().
    //  The description says it happens at the end. 8991 is the count.
    //  I was trying to create a WAIT until loaded indicator.
    //  This PaceMakerEstimator takes 32 seconds to start; ie. slow.
    //    ;//qDebugA() << ++countspontaneousEvents;
    //}

    // pass the event on to the parent class

    return( QMainWindow::eventFilter(obj, event) );
}

void MainWindow::getrowcol_ofcursors(short *libest, long *row, long *rowL, long *rowE, short *col, short *colL, short *colE)
{       /*
         * When a new click comes in, it is always stored in
         * example: row (for any lib or est side clicks),
         * never to lit, but yes copied to rowL or rowE.
         *
         *Why? How do you shift a highlighted button to another button?
         *You turn off the old and turn on the new. So you need two variables.
         *Highlights are the red border of the left/right buttons,
         *down the centre of the screen. I use the word lit.
         *
         * If you have two variables then one function can record clicks
         * in the event...() function and they might not be needed, but
         * they are always in row or copied to rowE, for example.
         *
         * A second function can query rowE, and a third highlight function
         * can be fed both litEstButton ... or Lib... as arguments
         * for lib and est sides, and turn off and turn on and shift new
         * rowE, or rowL to now highlighted row, for both sides.
         *
         * See eventMouseKeyboard.cpp, eventFilter(),
         * getrowcol_ofcursors(), and show_linenums_showHighlights().
         *
         * Think of row and col as a scratch pad
         * of the last click on either lib or est side; unspecified side.
         *
         * rowE and rowL are specified est or lib sides and useful
         * information of last known blinking cursor on either side,
         * but might not be blinking now.
         *
         * litEstButton litLibButton are historical and in use as a
         *      button highlight, waiting to be extinguished.
         */

    QString tipstatus = nullptr;
    QString ss;
    long xx, totallines, endpoint;
    unsigned yy,slowway=0;
    QLineEdit *qle = nullptr;

    /* default is failure with row or col == 0 until parsing.
     * 0 means no blinking keyboard cursor; no cell clicked as yet.
     * rows are 1 through ESTROWCOUNT-1.
     */
    *libest = 2;
    *row = 0;
    *col = 0;

    tipstatus = ui->centralwidget->focusWidget()->statusTip();
    /* tipstatus is preset at startup with a name which says
     * row and col of each QlineEdit */

    if( tipstatus == nullptr )
        tipstatus = "keyboard statusTip() failed.";

    // easy peasy

#ifdef OMITsinceMar7_2024
    else if( mousekeyboard == 'm' )  // M is for mouse!
    {
        /* The routine does work but it is slow and it was only being used
         * for libToEst() and estToLib() to decide where the mouse floated
         * to decide on a row. However, we can force user to
         * click on the row to select one and then we do not need
         * this slow routine, which is doing a dumb search.
         * I cannot afford slow processing on old computers.
         * It was called on EVERY mouse movement in event...()
         * QEVENT::ENTER.
         */

        mknow = &mousenow;

        /* fails: tipstatus = ui->centralwidget->statusTip();
         * MainWindow->statusBar->currentMessage(); is late by showing previous
         * message when I put it in (event->type() == QEvent::StatusTip )
         * which does a request event, not a post-request event.
         *
         * qDebug() << statusBar()->currentMessage();
         */

        // I have no way to find the statusTip which is displayed bottom left of the MainWindow container for mouse-over.
        // I tried every name. focusWidget() is under the keyboard cursor.
        // What is the equivalent command for under the mouse cursor?
        // tipstatus = ui->centralwidget->statusTip(); ? No.
        // So do a slow for() loop serial search instead but the data is library/estimate sides
        // and by rows and cols. That is FOUR for() loops!

        // Start at whichever side has the keyboard cursor.
#ifndef THISWORKS
        // Instead of starting at line 1 and going to 11000 rows, a faster idea is
        // to start at the current keyboardrow and search starting a screen full above it;
        // a total search of maybe only 280 boxes instead of 11,000 x 11 =~ 121000 boxes.
        // 11,000 rows would slow the mouse movements as you near the end of the estimate.
        //
        // This new idea assumes keyboardrow is always going to have a valid FocusIn on screen.
        // I will keep that in mind when I do scrolling and paging.
        // I do not want FocusIn to be on an off screen square, anyway.
        // So this assumption could only fail with keyboard cursor missing; ie. a visible error.
        bool done=false;
        totallines = estimateLinesVisible(keyboardnow.row, keyboardnow.col);
        xx = 1 + keyboardnow.row - totallines;
        endpoint = keyboardnow.row + totallines;
        if( endpoint > ESTROWCOUNT )
            endpoint = ESTROWCOUNT;
        if(xx < 1) // It is ok to start at lines too high, but let's not be ridiculously high.
            xx=1;
        if( keyboardnow.row == 0 )
        {
            ;//;//qDebugA() << "The keyboard FocusIn is missing. Programming error? Continue with slow mouse labels";
        }

       slowway=0;
       do {
        ++slowway;
        for( ; xx < endpoint; ++xx)
        {
             for( yy=1; yy<MAXCOL_libest; ++yy)
             {
                if( estRow[xx]->col[yy] == nullptr )
                     continue;
                qle = estRow[xx]->col[yy];  // is the QLineEdit
                if( qle->underMouse() == true )
                {
                    tipstatus = qle->statusTip();
                    done = true;
                    break;
                }
             }
             if( done == true )
                break;

             // cost library side is pt_lib...
             for( yy=1; yy<=4; ++yy)
             {
                if( libRow[xx]->col[yy] == nullptr )
                      continue;
                qle = libRow[xx]->col[yy];  // is the QLineEdit
                if( qle->underMouse() == true )
                {
                    tipstatus = qle->statusTip();
                    done = true;
                    break;
                }
             }
             if( done == true )
                break;
        }
        if( done == true)
            break;
        // We have to try again the slow way, for example if the keyboard was on the est side
        // and the lib side was scrolled down 500 lines; ie. out of range of our fast guess.
        // We would have missed it. If the keyboard FocusIn is in view, slowway never happens.
        endpoint = ESTROWCOUNT;
        xx=1;
       }while( slowway == 1 );  // happens only once, if at all.

#elif  THISWORKSTHESAMEWAY_but_slower
            // Instead of starting at line 1 and going to 11000 rows, a faster idea is
            // to start at the current toprow and calculate
            // screen height pixels/28 pixels per char and that gives maximum visible rows.
            // Next, set the for() loop from toprow which can be kept up to date
            // whenever scroll or page event occurs. toprow is always 1 when program starts.
            // keyboardnow.toprow+height/28pixels is a total search of maybe only 280 boxes
            bool done;
            for( done=false,xx=1; xx<ESTROWCOUNT; ++xx)
            {
                 for( yy=1; yy<MAXCOL_libest; ++yy)
                 {
                    if( estRow[xx]->col[yy] == nullptr )
                         continue;
                    qle = estRow[xx]->col[yy];  // is the QLineEdit
                    if( qle->underMouse() == true )
                    {
                        tipstatus = qle->statusTip();
                        done = true;
                        break;
                    }
                 }
                 if( done == true )
                    break;

                 // cost library side is pt_lib...
                 for( yy=1; yy<=4; ++yy)
                 {
                    if( libRow[xx]->col[yy] == nullptr )
                          continue;
                    qle = libRow[xx]->col[yy];  // is the QLineEdit
                    if( qle->underMouse() == true )
                    {
                        tipstatus = qle->statusTip();
                        done = true;
                        break;
                    }
                 }
                 if( done == true )
                    break;
            }

#elif  JUSTATEST
        // small test on one square. underMouse() works! (setFocus() keybrd to square works too).
         qle = estRow[3]->col[3];  // is the QLineEdit
         if( qle->underMouse() == true )
             tipstatus = qle->statusTip();
#endif
         if( tipstatus == nullptr )
            tipstatus = "mouse statusTip() failed.";
    }
#endif


    if( tipstatus != nullptr )
    {
        // parse it. example: lib9994 means cost-library, row 999, col 4,
        // but I did not exactly say lib9994 in the statusTip().
        // 999 is there. 4 is "a title", like "Quantity"

        ss = estcol1;
        ss.truncate(3);
        if( tipstatus.indexOf( ss ) == 0 )
        {
            *libest = 2;
            ss = tipstatus.sliced(OFFSET);
 //;//qDebugA() << "est slice14 is" << ss ;
            ss.truncate(DIGITSforROW);
            bool ok;
            *row = ss.toLong(&ok, 10);
            if( ok == true )  //this && cannot happen unless a  set the #defines
                                    //&& tipstatus.size() > OFFSET2  so I removed it.
            {
                ss = tipstatus.sliced(OFFSET2);
  //;//qDebugA() << "slice21 is" << ss ;
                if(      ss.indexOf ( estcol1.sliced(OFFSET2) ) == 0 )
                    *col = 1;
                else if( ss.indexOf ( estcol2.sliced(OFFSET2) ) == 0 )
                    *col = 2;
                else if( ss.indexOf ( estcol3.sliced(OFFSET2) ) == 0 )
                    *col = 3;
                else if( ss.indexOf ( estcol4.sliced(OFFSET2) ) == 0 )
                    *col = 4;
                else if( ss.indexOf ( estcol5.sliced(OFFSET2) ) == 0 )
                    *col = 5;
                else if( ss.indexOf ( estcol6.sliced(OFFSET2) ) == 0 )
                    *col = 6;
                else if( ss.indexOf ( estcol7.sliced(OFFSET2) ) == 0 )
                    *col = 7;
                else if( ss.indexOf( esttolib.sliced(OFFSET2) ) == 0 )
                {
                    // focus is on the button;
                    *col = 1; // assume this because verticalSlider() needs it
                    ;//qDebugA() << "est button" << mknow->currentEstbuttonrow;
               }
            }
        }

        ss = libcol1;
        ss.truncate(3);
        if( tipstatus.indexOf ( ss ) == 0 )
        {
            *libest = 1;
            ss = tipstatus.sliced(OFFSET);
            ss.truncate(DIGITSforROW);
   //;//qDebugA() << "lib slice14 is" << ss ;
            bool ok;
            *row = ss.toLong(&ok, 10);

            /*
            if( ok == true )
                ;//qDebugA() << mknow->row << "and ok is true";
            else
                ;//qDebugA() << mknow->row << "and ok is false";
             */

            if( ok==true )  // && cannot happen unless a monkey set the #defines--> && tipstatus.size() > OFFSET2 )
            {
                ss = tipstatus.sliced(OFFSET2);
     //;//qDebugA() << "slice21 is" << ss ;
                if(      ss.indexOf ( libcol1.sliced(OFFSET2) ) == 0 )
                    *col = 1;
                else if( ss.indexOf ( libcol2.sliced(OFFSET2) ) == 0 )
                    *col = 2;
                else if( ss.indexOf ( libcol3.sliced(OFFSET2) ) == 0 )
                    *col = 3;
                else if( ss.indexOf ( libcol4.sliced(OFFSET2) ) == 0 )
                    *col = 4;
                else if( ss.indexOf( libtoest.sliced(OFFSET2) ) == 0 )
                {
                    // focus is on the button; uuuu
                    *col = 4; // verticalSlider() needs a col so assume this.
                    ;//qDebugA() << "lib button" << mknow->currentLibbuttonrow;
                }
            }
        }

        // Now we know the row col for either the mouse cursor or the keyboard cursor.
        // *libest = 1 or 2
        // *row == 0 if the above failed
        // *col == 0 if the above failed

        // Failed idea:
        // Next, set toprow bottomrow which are just above/below cursor,
        // Cursors are on screen (unless scrolled out).
        // toprow is above them and bottomrow is below them.
        //Edit Dec 16, 2022. Now I know remembering toprow bottomrow is not needed
        // because ensureWidgetVisible() is easier instead.
    }

        //;//qDebugA() << "row/col of kybrd cursor QLineEdit box" << tipstatus  << mknow->row << "," << mknow->col;
        if( keyboardnow.libest == 2 )
        {
            *rowE = *row;
            *colE = *col;
        }
        else
        {
            *rowL = *row;
            *colL = *col;
        }

#ifdef OMITsinceMar7_2024
    else
    {
        //;//qDebugA() << "row/col of mouse cursor QLineEdit box" << tipstatus  << mknow->row << "," << mknow->col;
        if( mousenow.libest == 2 )
        {
            mousenow.rowE = mousenow.row; // save history for verticalSlider use.
            mousenow.colE = mousenow.col;
        }
        else
        {
            mousenow.rowL = mousenow.row; // save history for verticalSlider use.
            mousenow.colL = mousenow.col;
        }
    }
#endif

    return;
}

short MainWindow::estimateLinesVisible(long row, short col)
{
    int hh;
    int li;
    int totallines;
    int qpos, xx;

    if( row < 1 || col < 1)
           return (0); // unknown screensize.

#ifdef FAILEDat_IsVisible

        for( xx=mknow->row; --xx>=1; )
        {
            if( mknow->libest == 1 )
            {
                if( libRow[xx]->col[mknow->col] == nullptr )
                    continue;
                else
                    qle = libRow[xx]->col[mknow->col];
            }
            else
            {
                if( estRow[xx]->col[mknow->col] == nullptr )
                    continue;
                else
                    qle = estRow[xx]->col[mknow->col];
            }
            // debug temporary test:
            qle = estRow[xx]->col[3];

    ;//qDebugA() << "above, we got to " << xx;

            if( qle->isVisible() == true )
                    mknow->toprow = xx;
            else
                break;

            // Next, let's save the current visible toprow number because it is useful for fast searches
            // and for keeping the keyboard cursor visible and not do typing while scrolled off screen.

        }
        for( xx=mknow->row; ++xx<ESTROWCOUNT; )
        {
            if( mknow->libest == 1 )
            {
                if( libRow[xx]->col[mknow->col] == nullptr )
                    continue;
                else
                    qle = libRow[xx]->col[mknow->col];
            }
            else
            {
                if( estRow[xx]->col[mknow->col] == nullptr )
                    continue;
                else
                    qle = estRow[xx]->col[mknow->col];
            }
            // debug temporary test:
            qle = estRow[xx]->col[3];

    ;//qDebugA() << "below, we got to " << xx;

            if( qle->isVisible() == true )
                    mknow->bottomrow = xx;
            else
                break;

            // Next, let's save the current visible toprow number because it is useful for fast searches
            // and for keeping the keyboard cursor visible and not do typing while scrolled off screen.

        }
#else
        // Edit Dec 15, 2022: This idea below does know total LineEdits on the visible screen
        // but fails to find the bottomrow and toprow because the current cursor row position
        // is only known by all widgets by the full window, not the scrollable part.
        // I looked everywhere. I give up.
        // Next, consider not knowing since the Qt philosophy is full windows.
        // Do we need to know toprow bottomrow? I only wanted to know to stop typing
        // into an out of view line widget, which I tested and it is possible.
        // In other words, failed too:

        // Apparently, isVisible() returns true even when a widget is scrolled out of frame.
        // I looked for a QScrollArea method to disable this feature. I cannot find it.
        // So here is a new idea which is also faster:

        // QWidget of my parent is *QScrollArea::widget() or nullptr if none.
        // QScrollArea height divided by QLineEdit height = totallines on screen.
        // bottomrow-toprow = totallines and p.(height) of ScrollArea.height is knowable
        // pos.y() of estRow[xx]->col[3] (QLineEdit) is also known.
        // So (p.height-pos.y) / p.height = (totallines-xx) / totalines
        //      see: QPoint p = MainWindow::mapFromGlobal(QCursor::pos()); << p.x() << p.y();
        //          I think you do not need mapGlobal part. pos() works with/without_it on widgets.
        // You have an algebra equation (or fractions).
        // Solve for xx where xx is the number of lines down from the top.
        // Add xx to row and you get the toprow. (bottomrow = toprow+totallines).

        // This is probably faster than repeating loops, like blindfolded feeling.
        // It reads four data items and only does integer math; no for() loop.

        /* These are unchanged as I scroll or shrink the frame.
         * That means they are the imaginary FULL window sizes, not the viewing area.
         * So they are useless for calculating the visible toprow, bottomrow.
         * ;//qDebugA() << "hheight is:" << ui->scrollAreaWidgetContents->height();
         */
        //;//qDebugA() << "rectheight is:" << ui->scrollAreaWidgetContents->rect().height();
        //;//qDebugA() << "QLineEdit height is:" << estRow[mknow->row]->col[mknow->col]->pos();
        //;//qDebugA() << "scrollframe height is:" << ui->scrollAreaWidgetContents->frameSize();
        // Best choice:
        //;//qDebugA() << "frame height is:" << ui->scrollArea->frameGeometry().height();
        //;//qDebugA() << "QLineEdit frame height is:" << estRow[mknow->row]->col[mknow->col]->frameGeometry();
        //same as viewport-height: int hh = ui->scrollArea->frameGeometry().height();
        hh = ui->scrollArea->viewport()->height();
        li = estRow[row]->col[col]->frameGeometry().height();
        totallines = hh/li;
//;//qDebugA() << "totallines:" << totallines << "scrollableheight:" << hh << "QLineEdit height:" << li;
    //qpos = estRow[row]->col[col]->y();
          // Fails to find y() relative to visible top edge. y() is relative to full scrollArea.
          // I have to use scrollbar? viewport?
        //;//qDebugA() << ui->scrollAreaWidgetContents->minimumHeight() ;
        //;//qDebugA() << ui->scrollAreaWidgetContents->maximumHeight() ;
        // I cannot determine the actual scrolled-window value for qpos.
        // so try viewport(). It takes me days to learn the words and inheritance of Qt widgets.
        //
        // int qpos = ui->scrollArea->viewport()-> ??  no

   //;//qDebugA() << "scrollable viewport:" << qpos << "y()." ;
        // xx = totallines - ((totallines * (hh-qpos)) / hh);

        //;//qDebugA() << "keyboard cursor is " << xx << "lines below the unscrolled top,"
                 //<< " but turns negative as we scroll down to other lines because "
                    //"qpos can be 10,000 while hh is 557 constant.";
        //Failed ideas.

        //;//qDebugA() << "toprow:" << mknow->toprow << " bottomrow:" << mknow->bottomrow;
#endif
        return( totallines );
}

QLineEdit * MainWindow::currentKeyboardLineEdit()
{
    QLineEdit *qle = nullptr;

    if( keyboardnow.row == 0 || keyboardnow.col == 0 ) // maybe user did not click on a square yet
        qle = nullptr;
    else if( keyboardnow.libest == 2 )
    {
        ;//qDebugA() << "est side doing keyboard cursor";
        qle = estRow[ keyboardnow.row ]->col[ keyboardnow.col ];
    }
    else if( keyboardnow.libest == 1 )
    {
        ;//qDebugA() << "lib side doing keyboard cursor";
        qle = libRow[ keyboardnow.row ]->col[ keyboardnow.col ];
    }

    return ( qle );
}

void MainWindow::clearCurrentKeyboardCursorFocus()
{
    QLineEdit *qle = nullptr;

    if( keyboardnow.libest == 2 )
    {
        qle = estRow[ keyboardnow.row ]->col[ keyboardnow.col ];

        // Do not do =0 because a signal is emitted and needs these as non-zero:
        //keyboardnow.row = keyboardnow.col = 0;
    }
    else if( keyboardnow.libest == 1 )
    {
        qle = libRow[ keyboardnow.row ]->col[ keyboardnow.col ];

        // Do not do this because a signal is emitted and needs these as non-zero:
        //keyboardnow.row = keyboardnow.col = 0;
    }

    if( qle != nullptr )
    {
        qle->clearFocus();
    // do this in the FocusOut event instead: qle->setFrame(globalFrame);
    }
    return;
}

/* Shows row number in mainwindow just below File Settings About Edit
 * and puts a red frame setDefault() around the arrow buttons of
 * keyboard cursor row. rowl is library side and rowe is estimate side.
 * 0 means off, >0 is on.
 */
void MainWindow::show_linenums_showHighlights( long *litLibButton, long *litEstButton, long rowl, long rowe )
{
    QString ss;
    QString blank = "?  ";

    //VERY SLOW: ;//qDebugA() << "Label_2 label_7 libcodeEstsidepushbutton linenum in titles" << rowl << " " << rowe;

    // erase old button highlights, if any.
    eraseHighlights( litLibButton, litEstButton);

    if( rowl == 0L )
    {
        ui->rowLibkbd->setText(blank);
    }
    else
    {
        ss.setNum( rowl, 10);
        ui->rowLibkbd->setText(ss);
        libRow[rowl]->pushbutton->setDefault(true);
        *litLibButton = rowl;
    }

    if( rowe == 0L )
    {
        ui->rowEstkbd->setText(blank);
    }
    else
    {
        ss.setNum( rowe, 10);
        ui->rowEstkbd->setText(ss);

        QString refcode = estRow[rowe]->refcode.leftJustified(2, ' ');
        if( refcode.isNull() )
            refcode="  ";

        //too slow: setStyleSheet(" QPushButton {background-color: rgb(90%,81%,73%);border-style: outset;border-width: 1px;border-radius: 3px;border-color: black;font: bold 16px;padding: 2px; } QPushButton:pressed {background-color: rgb(56%, 40%, 29%);border-style: inset;} QPushButton:default {border-color: red;}");
        //moved to load2.cpp ui->libcodeEstside->setStyleSheet("button:pressed {background-color: rgb(56%, 40%, 29%)}");
        ui->libcodeEstside->setText(" "+refcode+" ");

        // This is not for here. See menuEdit.cpp where we load a Cost Library
        //      ui->libcode2char->setText(" "+refcode+" ");

        estRow[rowe]->pushbutton->setDefault(true);
        *litEstButton = rowe;
    }

    // Highlight the Qty box if the qtymath() requests it.
    if( qtymathRequestsQtyColBoxHighlight == true )
    {
        if( saveQtyMathHighlightRow != rowe )
        {
            // turn off old one because row changed.
          if( saveQtyMathHighlightRow > 0 ) // 0 is no old one
            //estRow[saveQtyMathHighlightRow]->col[3]->setFrame(globalFrame);
            estRow[saveQtyMathHighlightRow]->col[3]->setStyleSheet( "background-color: palette()" );


          if( rowe > 0 ) // 0 is no new one
            //estRow[rowe]->col[3]->setFrame(globalFrame==false?true:false);
            estRow[rowe]->col[3]->setStyleSheet("background-color: beige");

            // save the last highlight in case we have to turn off
            saveQtyMathHighlightRow = rowe;
        }
    }

    return;
}

void MainWindow::eraseHighlights( long *rowLib, long *rowEst)
{
    // erase old button highlights, if any.
    if( *rowLib )
    {
        libRow[ *rowLib ]->pushbutton->setDefault(false);
        *rowLib = 0L;
    }

    if( *rowEst )
    {
        estRow[ *rowEst ]->pushbutton->setDefault(false);
        *rowEst = 0L;
    }
}

void MainWindow::historySave_ZeroSetKeyboardCursor()
{
    if( keyboardnow.libest == 2 )
    {
        keyboardnow.rowE = keyboardnow.row; // save history maybefor verticalSlider use.
        keyboardnow.colE = keyboardnow.col;
    }
    else
    {
        keyboardnow.rowL = keyboardnow.row; // save history maybefor verticalSlider use.
        keyboardnow.colL = keyboardnow.col;
    }
    keyboardnow.row = keyboardnow.col = 0; // 0 means no cursor; Focussed Out.
}

/* How to find next and previous widget in the line, which is ui->lib3424 or ui->est3421, for examples.
 * def keyPressEvent(self, event):
    if event.key() == QtCore.Qt.Key_Tab:
        self.parent().focusNextChild()
    elif event.key() == QtCore.Qt.Key_Backtab:
        self.parent().focusPrevChild()
 *
 * https://stackoverflow.com/questions/23498848/how-to-focus-on-a-next-previous-widget
 */
void MainWindow::tabRightKeyMove()
{

}
void MainWindow::tabLeftKeyMove()
{

}

/* Feb 17, 2023:
 * I mastered the instances and heritance concepts
 * and how to share data scopes by Feb 2023,
 * as shown in my qtymath.cpp and qtymath.h
 * and so I did fix my oneHeader syntax problems. I used the
 * archiveQtyFrame and oneHeader variable is moved into mainwindow.h "public:",
 * and casting a QWidget * to a Mainwindow * helped.
 * I was using external variables, which is usually inappropriate for object programming.
 */

void MainWindow::on_pushHeaderButton_released()
{
    if ( oneHeader == nullptr )
    {
        oneHeader = new DialogHeaderOfEstimate(this);

        //;//qDebugA() << "Header button released and it is meant to be BLANK because it is new.";
    }

    //else
        //;//qDebugA() << "Header button released and it is meant to be FILLED because it is not new.";

    /* https://forum.qt.io/topic/74902/how-to-change-size-of-qdialog
     *
     * QFont font2 = property("font").value<QFont>();
     */

     if( oneHeader->isVisible() ) // up down toggle for each press so user does not have to mouse to Discard button; double click instead.
         oneHeader->setVisible( false );
     else
     {
        oneHeader->resize( 40 * globalFont2.pointSize(), 13 * globalFont2.pointSize());

        oneHeader->show();
        oneHeader->raise();
        oneHeader->activateWindow();
     }

    return;
}

#define LOADfromdata
#ifdef LOADfromdata
/* https://stackoverflow.com/questions/47414607/how-to-include-data-object-files-images-etc-in-program-and-access-the-symbol
 * see convertPngToC.cc utility and totalhours.png
 * totalhours.inc
 * struct { const char *data; int size; } totalHoursPng = {
 * must be changed to something like this, instead of struct {}:
 * QByteArray totalHoursPng( QByteArray totalHoursPng(
"\x89\x50\x4e\x47\x0d\x0a\x1a\x0a\x00\x00\x00\x0d\x49\x48\x44\x52"
    "\x00\x00\x01\xde\x00\x00\x00\x6f\x08\x02\x00\x00\x00\xcd\x1d\x36"
    "\xf0\x00\x00\x00\xd0\x65\x58\x49\x66\x49\x49\x2a\x00\x08\x00\x00"
    .
    .
    .
        , 50038 );
 */
#include "totalhours.inc"
#endif

// Ctrl-S ctrl-S is set in mainwindow.ui
// Ctrl-T ctrl-t is set in mainwindow.ui
void MainWindow::on_TotalHours_pressed()
{
    long row;
    QDrag *drag = new QDrag(this);
    QMimeData *mimeData = new QMimeData;
    QPixmap *iconPixmap = new QPixmap;
    QPixmap iconPixmapScaled;

#ifdef LOADfromdata
    if( !iconPixmap->loadFromData( totalHoursPng, nullptr, Qt::AutoColor) )
    {
        // failed to show image to user, but drag still works with an image missing error.
    }
#else
    if( !iconPixmap->load( QCoreApplication::applicationDirPath()+"/totalhours.png") )
    {
        /* try to create the missing totalhours.png */
        QFile fOut(QCoreApplication::applicationDirPath()+"/totalhours.png");
        //QFile fOut("/tmp/totalhours.png");
        if(fOut.open(QIODeviceBase::ReadWrite) )
        {
            // See convertPngToC.cc utility which created this image data:
            fOut.write( totalHoursPng.data, totalHoursPng.size);
            fOut.close();

            if( !iconPixmap->load( QCoreApplication::applicationDirPath()+"/totalhours.png") )
            {
                // failed to show image to user, but drag still works with an image missing error.
            }
        }
    }
#endif

    int trail = globalFont2.pointSize();
    iconPixmapScaled=iconPixmap->scaled(24*trail, 7*trail, Qt::IgnoreAspectRatio, Qt::FastTransformation);

    mimeData->setText("Total Hours");
    drag->setMimeData(mimeData);
    drag->setPixmap(iconPixmapScaled);

    drag->setHotSpot(QPoint(drag->pixmap().width()/2,
                            0                         ));
        //drag->pixmap().height()));

    if (drag->exec(Qt::MoveAction) != Qt::MoveAction)
    {
        QLineEdit *trg = (QLineEdit *)(drag->target());
        if( parseRowFromStatusTip( &row, trg ) == true )
        {
            insertTotalSubtotal( row );
            //doInsertHoursTotalSubtotal( row );
            //doInsertCostTotalSubtotal( row );
        }
    }
    return;
}

// Ctrl-S ctrl-S is set in mainwindow.ui
// Ctrl-T ctrl-t is set in mainwindow.ui
void MainWindow::on_TotalCost_pressed()
{
    long row;
    QDrag *drag = new QDrag(this);
    QMimeData *mimeData = new QMimeData;
    QPixmap *iconPixmap = new QPixmap;
    QPixmap iconPixmapScaled;

#ifdef LOADfromdata
    if( !iconPixmap->loadFromData( totalHoursPng, nullptr, Qt::AutoColor) )
    {
        // failed to show image to user, but drag still works with an image missing error.
    }
#else
    if( !iconPixmap->load( QCoreApplication::applicationDirPath()+"/totalhours.png") )
    {
        /* try to create the missing totalhours.png */
        QFile fOut(QCoreApplication::applicationDirPath()+"/totalhours.png");
        //QFile fOut("/tmp/totalhours.png");
        if(fOut.open(QIODeviceBase::ReadWrite) )
        {
            // See convertPngToC.cc utility which created this image data:
            fOut.write( totalHoursPng.data, totalHoursPng.size);
            fOut.close();

            if( !iconPixmap->load( QCoreApplication::applicationDirPath()+"/totalhours.png") )
            {
                // failed to show image to user, but drag still works with an image missing error.
            }
        }
    }
#endif

    int trail = globalFont2.pointSize();
    iconPixmapScaled=iconPixmap->scaled(24*trail, 7*trail, Qt::IgnoreAspectRatio, Qt::FastTransformation);

    mimeData->setText("Total Cost");
    drag->setMimeData(mimeData);
    drag->setPixmap(iconPixmapScaled);

    drag->setHotSpot(QPoint(drag->pixmap().width()/2,
                            0                         ));
        //drag->pixmap().height()));

    if (drag->exec(Qt::MoveAction) != Qt::MoveAction)
    {
        QLineEdit *trg = (QLineEdit *)(drag->target());
        if( parseRowFromStatusTip( &row, trg ) == true )
        {
            insertTotalSubtotal( row );
            //doInsertCostTotalSubtotal( row );
            //doInsertHoursTotalSubtotal( row );
        }
    }
    return;
}

/* I reduced the code to one type of descriptionSUBTOTAL block.
 * Originally I wrote an Hours one and a Costs one.
 * Now I am putting them both into one block and the DRAG off
 * of either menu creates the same one.
 * The order is always Hours and then Costs. I was reversing the
 * order depending on which one,
 * thus I had two functions. Now I have one.
 * doInsertCostTotalSubtotal( int row ) has been omitted
 * doInsertHoursTotalSubtotal( int row ) has been omitted
 * insertTotalSubtotal() is the new one.
 *
 * The estRow[rr]->col[1]->setReadOnly(true) WAS a nightmare
 * because it is not moved by TYPEIns and TYPEDel undo/redo.
 * I resolved to not push these onto the undo/redo stack
 * and Ins/Del keys are modified in keyInsDelRows.cpp for = == == special cases
 * and see Qt::Key_Delete, above.
 */
bool MainWindow::insertTotalSubtotal( int row )
{
    double prefHoursMultcurrent = prefHoursMultiplier;
    double prefCostsMultcurrent = prefCostsMultiplier;
    double sumH=0;
    double sumC=0;
    QString ss;
    estLineEditHistory *eH;
    estUndoCommand *command;
    bool overlap=false;
    QMessageBox msgBox;
    int rr;
    int saveStartRow = row;
    QString insufficient = "Insufficient room. Estimate";
    QString tabEst = " tab 1"; // future feature has multiple tabs of 231 lines.

    if( row > 0 ) // insert a blank line first of all (by using TYPEIns)
    {
        /*Sum the Hours column. Omit === libcodes because they are subtotal/totals.
         *Insert === markers in the inserted blank line and insert items.
         */

        // skip "=" lines
        for( rr=row; rr > 1; --rr)
        {
            if( estRow[rr]->col[1]->text().indexOf("=") < 0 )
                break;
        }
        // Now you are on a not= line and it could be row 1, not row 0, not < 0.

        for( ; rr >= 1; --rr)
        {
            ;//qDebugA( << "\n" << estRow[rr]->col[2]->text();

            // quit at the first descriptionSUBTOTAL because that one summed everything above it.
            if( estRow[rr]->col[2]->text().indexOf( descriptionSUBTOTAL ) == 0 )
                break;

            ss = estRow[rr]->col[1]->text();
            if( *ss.data() != '=' )
            {
                ss = estRow[rr]->col[6]->text();
                sumH += ss.toDouble();

                ss = estRow[rr]->col[7]->text();
                sumC += ss.toDouble();
            }
        }

        ss = estRow[row]->col[1]->text();
        if( *ss.data() == '=' )
        {
            overlap = true;

            // go to top of =
            do{
                if( row <= 1 )
                    break;
                else
                    --row;

                ss = estRow[row]->col[1]->text();

            }while( *ss.data() == '=' );

            // work down through all the ==... lines and quit at uniqueID_TOTALSblock which is descriptionSUBTOTAL
            do
            {
                if( row >= ESTROWCOUNT-1 )
                {
                    break;
                }
                else
                    ++row;

                // grab the prefHoursMultcurrent while we are passing by
                if( estRow[row]->col[2]->text().indexOf( SUBTOTALHOURSDESCRIPTION ) >= 0 )
                    prefHoursMultcurrent = estRow[row]->col[3]->text().toDouble();

                // grab the prefCostsMultcurrent while we are passing by
                else if( estRow[row]->col[2]->text().indexOf( SUBTOTALCOSTSDESCRIPTION ) >= 0 )
                    prefCostsMultcurrent = estRow[row]->col[3]->text().toDouble();

                ss = estRow[row]->col[1]->text();
            }while( ss.indexOf( uniqueID_TOTALSblock ) != 0 );  // there is only one of these in the block

            if( !is1987subtotal(row) )
            {
                // We are at ===, descriptionSUBTOTAL so go back to the top.
                for( int ee=6; --ee>=0; )
                //go up 6 rows (they should be = == or === in libcode col)
                {
                    ss = estRow[row]->col[1]->text();
                    if( *ss.data() == '=' )
                        --row;
                    else // should never happen
                    {
                        overlap = false;
                        break;
                    }
                }
                /* for() is the same as row -= 6;  but checks
                 * for '=' lines, which are always true, but maybe
                 * I made a mistake. So I check instead of doing
                 * row = row -6; which is how I originally
                 * more simply coded it.
                 */
            }
            else
            {
                //row is the top of the old 4 row subtotal block style, not 6.
                overlap = false;
            }
        }

        if( overlap == false )
        {
            /* If the line above is === then insert an extra blank line
             * to separate subtotal/total blocks
             */
            if( row-1 >= 1 && estRow[row-1]->col[1]->text().indexOf("=")>=0 )
            {
                if( insertBlankLineEstSide(row) == false )
                    return(false);
                if( row<ESTROWCOUNT-1)
                    ++row;
                else // end of estimate means we have insufficient room to continue
                {
                    msgBox.setText( insufficient + tabEst );
                    msgBox.exec();
                    return(false);
                }
            }

            // need ten blank lines for block of =, ==, === subtotal total area
            // We have 1 and we need 9 more
            if( row+9 > ESTROWCOUNT-1)
            {
                msgBox.setText( insufficient + tabEst );
                msgBox.exec();
                return(false);
            }

            for(rr=10;rr>=1;--rr)
            {
                if( insertBlankLineEstSide(row) == false )
                    return(false); // this function inserts blank lines without moving row
            }
        }

        for( rr=1;rr<=7;++rr )
        {
            if( row >= ESTROWCOUNT-1 )
            {
                break;
            }
            else
                ++row; // so we are sitting on top of where the first === should go.

            /* set col1 col2 ++row to no edit */
            estRow[row]->col[1]->setReadOnly(true);
            estRow[row]->col[2]->setReadOnly(true);

            estRow[row]->col[2]->clear();
            estRow[row]->col[2]->setText( " " );

            estRow[row]->col[3]->clear();
            estRow[row]->col[3]->setText( " " );

            estRow[row]->col[4]->clear();
            estRow[row]->col[4]->setText( " " );

            estRow[row]->col[5]->clear();
            estRow[row]->col[5]->setText( " " );

            if( rr==1 || rr==5 )
            {
                estRow[row]->col[1]->clear();
                estRow[row]->col[1]->setText( "===" );

                estRow[row]->col[6]->clear();
                estRow[row]->col[6]->setText( "-------------" );
                estRow[row]->col[7]->clear();
                estRow[row]->col[7]->setText( "-------------" );
            }
            else if( rr==2 || rr==3 || rr==4 )
            {
                estRow[row]->col[1]->clear();
                estRow[row]->col[1]->setText( "== " );
                if(rr==2)
                {
                    estRow[row]->col[6]->clear();
                    ss = QString("%1").arg(sumH, 6, 'f', 2);
                    estRow[row]->col[6]->setText( ss );
                    estRow[row]->col[7]->clear();
                }
                else if( rr==3 )
                {
                    estRow[row]->col[2]->clear();
                    estRow[row]->col[2]->setText( SUBTOTALHOURSDESCRIPTION );

                    /* grab this multiplier from the preferences or init values
                     * if none was grabbed from the Qty box above.
                     */
                    estRow[row]->col[3]->clear();
                    ss = QString("%1").arg(prefHoursMultcurrent, 6, 'f', 2);
#ifdef MULTIPLIER_colours
                    estRow[row]->col[3]->setStyleSheet("background-color: orange" );
#endif
                    estRow[row]->col[3]->setText( ss );

                    estRow[row]->col[6]->clear();
                    ss = QString("x %1").arg(prefHoursMultcurrent, 6, 'f', 2);
#ifdef MULTIPLIER_colours
                    estRow[row]->col[6]->setStyleSheet("background-color: orange" );
#endif
                    estRow[row]->col[6]->setText( ss );

                    estRow[row]->col[7]->clear();
                    ss = QString("%1").arg(sumC, 6, 'f', 2);
                    estRow[row]->col[7]->setText( ss );
                }
                else if( rr==4 )
                {
                    estRow[row]->col[2]->clear();
                    estRow[row]->col[2]->setText( SUBTOTALCOSTSDESCRIPTION );

                    /* grab this multiplier from the preferences or init values
                     * if none was grabbed from the Qty box above.
                     */
                    estRow[row]->col[3]->clear();
                    ss = QString("%1").arg(prefCostsMultcurrent, 6, 'f', 2);
#ifdef MULTIPLIER_colours
                    estRow[row]->col[3]->setStyleSheet("background-color: orange" );
#endif
                    estRow[row]->col[3]->setText( ss );

                    estRow[row]->col[6]->clear();
                    estRow[row]->col[7]->clear();
                    ss = QString("x %1").arg(prefCostsMultcurrent, 6, 'f', 2);
#ifdef MULTIPLIER_colours
                    estRow[row]->col[7]->setStyleSheet("background-color: orange" );
#endif
                    estRow[row]->col[7]->setText( ss );
                }
            }
            else if( rr==6 )
            {
                // this line is "=  descriptionSUBTOTAL
                estRow[row]->col[1]->clear();
                estRow[row]->col[1]->setText( uniqueID_TOTALSblock );

                estRow[row]->col[2]->clear();
                estRow[row]->col[2]->setText( descriptionSUBTOTAL );

                // grab this multiplier from the preferences if none in the Qty box
                estRow[row]->col[3]->clear();
                estRow[row]->col[3]->setText( " " );

                estRow[row]->col[6]->clear();
                ss = QString("%1").arg(sumH*prefHoursMultcurrent, 6, 'f', 2);
                estRow[row]->col[6]->setText( ss );

                estRow[row]->col[7]->clear();
                ss = QString("%1").arg(sumC*prefCostsMultcurrent, 6, 'f', 2);
                estRow[row]->col[7]->setText( ss );
                ;//qDebugA() << "doInsert did the sumC TIMES prefCostsMultcurrent" << sumC*prefCostsMultcurrent;
            }
            else if( rr==7 )
            {
                estRow[row]->col[1]->clear();
                estRow[row]->col[1]->setText( "===" );

                estRow[row]->col[6]->clear();
                estRow[row]->col[6]->setText( "-------------" );
                estRow[row]->col[7]->clear();
                estRow[row]->col[7]->setText( "-------------" );
            }
        }

        // add two more lines for descriptionTOTAL sometimes. (below descriptionSUBTOTAL)
        ;//qDebugA() << "testbottomrowsubtot" << row << bottomSubtotalRow;
        if( row<ESTROWCOUNT-1 && row >= bottomSubtotalRow )
        // do this only if this is the lowest one because the ones above are only subtotals
        {   // I do not mean row <= ESTROWCOUNT-1 because 2 rows needed

            bottomSubtotalRow = row;
            ;//qDebugA() << "test SETS" << row << bottomSubtotalRow;

            /* change === under the subtotal
             *
             *     to ---    <-- and this is the value of bottomSubtotalRow
             *     (new total here)
             *        ===
             */
            estRow[row]->col[6]->clear();
            estRow[row]->col[6]->setText( "-------------" );
            estRow[row]->col[7]->clear();
            estRow[row]->col[7]->setText( "-------------" );

            ++row;
            /* set two more to no edit */
            int xx=2;
            for( rr=row; --xx>=0 && rr<=ESTROWCOUNT-1; ++rr )
            {
                estRow[rr]->col[1]->setReadOnly(true);
                estRow[rr]->col[2]->setReadOnly(true);
            }


            /* Add up all the above subtotals, and insert their sum into this line
             * which is called descriptionTOTAL.
             */
            double dd1 = 0.0, dd2 = 0.0;
            QString ss;

            //[6] column
            for( int xx = row-1; xx > 0; --xx )
            {
                if( estRow[xx]->col[2]->text().indexOf( descriptionSUBTOTAL ) == 0 )
                {
                    ss = estRow[xx]->col[6]->text().replace( "x"," ");

                    dd1 = dd1 + ss.toDouble();
                }
            }
            estRow[row]->col[6]->clear();
            ss = QString("%1").arg(dd1, 6, 'f', 2);
            estRow[row]->col[6]->setText( ss );

            //------------

            //[7] column
            for( int xx = row-1; xx > 0; --xx )
            {
                if( estRow[xx]->col[2]->text().indexOf( descriptionSUBTOTAL ) == 0 )
                {
                    ss = estRow[xx]->col[7]->text().replace( "x"," ");

                    dd2 = dd2 + ss.toDouble();
                }
            }
            estRow[row]->col[7]->clear();
            ss = QString("%1").arg(dd2, 6, 'f', 2);
            estRow[row]->col[7]->setText( ss );
            //-------------

            estRow[row]->col[1]->clear();
            estRow[row]->col[1]->setText( "== " );

            estRow[row]->col[2]->clear();
            estRow[row]->col[3]->clear();
            estRow[row]->col[4]->clear();
            estRow[row]->col[5]->clear();

            QLocale clocale;
            estRow[row]->col[2]->setText( descriptionTOTAL + QString("%1%L2").arg(clocale.currencySymbol(clocale.CurrencySymbol)).arg(dd1+dd2, 6, 'f', 2) );

            if( row >= ESTROWCOUNT-1 )
            {
                return(false);
            }
            else
                ++row;
            estRow[row]->col[1]->clear();
            estRow[row]->col[1]->setText( "===" );

            estRow[row]->col[2]->clear();
            estRow[row]->col[3]->clear();
            estRow[row]->col[4]->clear();
            estRow[row]->col[5]->clear();

            estRow[row]->col[6]->clear();
            estRow[row]->col[6]->setText( "========" );
            estRow[row]->col[7]->clear();
            estRow[row]->col[7]->setText( "========" );

            // insert a blank line because I saw a touching of the next line sometimes.
            if( insertBlankLineEstSide(row) == false )
                return(false); // this function inserts blank lines without moving row
        }

        /* Now delete ALL the = segments which have descriptionTOTAL
         * because there can be only one descriptionTOTAL in an estimate
         */
        row -= 8; // start above us and work backwards up the estimate. See --row.
        do{
            if( row<=1 )
                break;
            else
                --row;
            if( estRow[row]->col[2]->text().indexOf( descriptionTOTAL ) == 0 )
            {
                estRow[row]->col[1]->clear(); //erase the "===" libcode first or else the entire block gets deleted by TYPEDel
                delLineEstSide(row);
                estRow[row]->col[1]->clear(); //erase the "===" libcode first or else the entire block gets deleted by TYPEDel
                delLineEstSide(row);
            }
        }while(1);

        /* We finished creating data into both the overlap == false and overlap == true cases
         * and now we must update the old subtotal blocks which are down the estimate
         * if we did a not-overlap; ie. a new inserted block of subtotals.
         */
        if( overlap == false )
        {
        /* This interesting case is where we inserted a new subtotal block.
         * What about the subtotal block below us. We have to update that one
         * like we would with a DRAG to overlap. We could write the code here
         * but instead I made a recursive call, searchForSubtotalTotalUpdateThem()
         * which calls insertTotalSubtotal(), which is us in overlap mode,
         * which has the correct code already written. How did I know this would work?
         * The delLineEstSide( long row) function was doing this call to update the block below
         * after deleting a subtotal block above!
         *
         * Also this is not an endless loop because searchForSubtotalTotalUpdateThem()
         * only calls us if( overlap == true ), not ==false; eg. like inside this comment area.
         */

            searchForSubtotalTotalUpdateThem( saveStartRow ); //this new block.
            // It also stupidly goes to the end of the estimate whereas we could quit at the first one
            // if any. Improve it later for better speed, some day, by quitting early.
        }

    }

    return(true);
}

bool MainWindow::parseRowFromStatusTip( long *row, QLineEdit *trg )
{
    bool ret = false;

    //drag->target(); // the drop's QLineEdit object
    QString ss = estcol1; // see mainwindow.h
    ss.truncate(3);

    if( trg == nullptr )
        ;
    else
    {
        QString tipstatus = trg->statusTip();
;//qDebugA() << "Drag: found statusTip()" << tipstatus;

        if( tipstatus.isEmpty() )
            ;
        else if( tipstatus.indexOf( ss ) == 0 ) //Est is the first 3 chars
        {
           ss = tipstatus.sliced(OFFSET);
           ss.truncate(DIGITSforROW);
           *row = ss.toLong(&ret, 10);
           if( ret == true )
           {
;//qDebugA() << "Drag: found the row" << *row;

           }
        }
    }

    return(ret);
}

/* I changed this to a load library pushButton, instead of a label.
 * It is the same as event Dbl above; ie. where double click on a library code,
 * except the library code is the libcodeEstside name instead of the estimate line.
 */
void MainWindow::on_libcodeEstside_clicked()
{
    int idx;

    // This is the library filename: ui->libcodeEstside->setText

        for( idx=0; idx < ui->comboBox->count(); ++idx)
        {
            QVariant vv;
            vv = ui->comboBox->itemData(idx,Qt::UserRole);
            QString *ss = (QString *)vv.data();

            //;//qDebugA() << ss <<"libcodeEstside refcode button filename"<< ui->libcodeEstside->text().trimmed();
            if( *ss == ui->libcodeEstside->text().trimmed() )
            {
                //ui->comboBox->currentIndex();
                loadByIndexLibrarynameOfComboboxLibs(idx);
                break;
            }
        }

}

bool MainWindow::is1987subtotal( long rr )
{
    bool is1987 = false;
    QString ss;

    // detect old format subtotals/totals from date 1987
    if( estRow[rr]->col[2]->text().indexOf( descriptionSUBTOTAL1987 ) == 0 )
    {
        is1987 = true;

        /* 1987 is a special case. Want to delete its =   ==  === and
         * do no overlap because the format is new and different.
         * So if I delete =
         * then inserting a subtotal at this top of 4 blank rows will be excellent.
         * This function can be deleted and will have no effect on 2023 files.
         */

            // work down through all = lines setting them to blanks.
            do
            {
                    estRow[rr]->col[1]->clear();
                    estRow[rr]->col[2]->clear();
                    estRow[rr]->col[3]->clear();
                    estRow[rr]->col[4]->clear();
                    estRow[rr]->col[5]->clear();
                    estRow[rr]->col[6]->clear();
                    estRow[rr]->col[7]->clear();

                    estRow[rr]->col[1]->setReadOnly(false);
                    estRow[rr]->col[2]->setReadOnly(false);

                if( rr >= ESTROWCOUNT-1 )
                {
                    break;
                }
                else
                    ++rr;

                ss = estRow[rr]->col[1]->text();
            }while(*ss.data() == '=');
    }

    return(is1987);
}

bool MainWindow::isFullRow( struct QLineEditArray *qlea )
{
    bool ret=false;
    QMessageBox msgBox;

    if( qlea == nullptr )
        ret = true;
    else
    {
        for( int cc=1; cc<MAXCOL_libest; ++cc )
        {
            if( qlea->col[cc] == nullptr )//unused columns of lib are nullptr
                ;
            else if( !qlea->col[cc]->text().simplified().isEmpty() )
            {
                ret = true; // quit on the first used column of row.
                break;
            }
        }
    }

    if( ret == true )
    {
        msgBox.setText("Full. Scroll to the bottom.\nDelete the last row.");
        msgBox.setStandardButtons(QMessageBox::NoButton);
        msgBox.setStandardButtons(QMessageBox::Cancel);
        msgBox.exec();
    }

    return(ret);
}
