/*  This file is part of PaceMaker Estimator 2.0.
    PaceMaker Estimator 2.0 is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
    PaceMaker Estimator 2.0 is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with PaceMaker Estimator 2.0. If not, see <https://www.gnu.org/licenses/>.
*/

#include "mainwindow.h"
//#include "totalsubtotal.h"

/* Edit March 4, 2023: I gave up on Ins going into stack. Del is in stack
 * but it is not too usable because you have to Ins a blank line manually
 * and then press Undo, for each line or else you have no writing space.
 * I got bored. It needs a special Undo that presses the Ins for you,
 * but not for TYPELibEst, only TYPEDelIns.
 *
 * Conclusion for now: Del stack works and Ins stack is not in use
 * but Ins key does work. See eventMouseKeyboard.cpp for more notes.
 *
 * I moved this UNDO_REDOSTACK code to eventsMouseKeyboard.cpp
 * Understand, my first QUndoCommand of lib->est and est->lib buttons
 * envelopes before and after the button action and were written inside
 * estlineedithistory and liblineedithistory as constructor code.
 *
 * So Ins/Del actions must also be inside, (not outside like below),
 * See estlineedithistory.cpp and liblineedithistory.cpp.
 * and called from a keyboard event() in eventMouseKeyboard.cpp.
 *
 * I am not interested in starting a new QUndoCommand stack.
 * So I am putting these with the first QUndoCommand stack of which I
 * had no good example, except I found video (see video copy inside subdir),
 * and I did not comprehend the syntax in the Qt6 document for this feature.
 * Now I do.
 *
 * These Ins/Del worked perfectly when I directly called
 * them from eventMouseKeyboard.cpp but now I am moving
 * where they are called so this Undo/Redo feature is added for Ins/Del.
 * See estlineedithistory.cpp and liblineedithistory.cpp.
 * I am not changing this code below. It also could be moved
 * out of the est->lib lib->est command stack easily, for some future reason.
 * #define UNDO_REDOSTACK_notIntendedToWork_justNotes
 */

bool MainWindow::insertBlankLineEstSide( long row)
{
    unsigned xx, rw;
    bool ret = true;

    /* March 2023: refuse to Ins in the middle of subtotal/total block */
    if( row > 1 && row<=ESTROWCOUNT-1
                && estRow[row]->col[1]->text().indexOf("=") >= 0
                && estRow[row-1]->col[1]->text().indexOf("=") >= 0 )
        ret = false;
    else if( isFullRow( estRow[ESTROWCOUNT-1] ) == true )
        ret = false;
    else if( row > 0)
    {
        for( rw=ESTROWCOUNT-1; rw>row; --rw)
        {
            for( xx=1; xx<=7; ++xx )
            {
             // see rw-1. transfer from a line lower, but not beyond line limit.

                    estRow[rw]->col[xx]->selectAll();
                    estRow[rw]->col[xx]->insert( estRow[rw-1]->col[xx]->text() );
                    estRow[rw]->fileQtyBlock = estRow[rw-1]->fileQtyBlock;
                    estRow[rw]->refcode = estRow[rw-1]->refcode;
            }
            if( estRow[rw]->col[1]->text().indexOf("=") < 0 )
            {
                estRow[rw]->col[1]->setReadOnly(false);
                estRow[rw]->col[2]->setReadOnly(false);
            }
            else
            {
                estRow[rw]->col[1]->setReadOnly(true);
                estRow[rw]->col[2]->setReadOnly(true);
            }
        }

        // delete one line at "row" of estimate
        for( xx=1; xx<=7; ++xx )
            estRow[row]->col[xx]->clear();

        estRow[row]->col[1]->setReadOnly(false);
        estRow[row]->col[2]->setReadOnly(false);

        estRow[row]->fileQtyBlock = nullptr;
        estRow[row]->refcode = "";

#ifdef UNDO_REDOSTACK_notIntendedToWork_justNotes
        estLineEditHistory *xx = new estLineEditHistory( this, targ, src, TYPEInsDel );
        // transfer a pointer to the object:
        estUndoCommand *command = new estUndoCommand( xx );
        // no delete because the stack needs it.
        //delete xx;   Instead, delete *archive variable later
        // See clearLibEst.cpp and clearEstStack().

        est_undo_stack.push(command);
#endif

        /* xxx there was a case where it was too big. (might be fixed Jan 2024)
         * I think it has to do with undo() and drag/drop adding blank rows.
         * If it is too small it means the descriptionTOTAL will be disallowed.
         */
        ++bottomSubtotalRow; // this is the last known descriptionTOTAL.
        ;//qDebugA() << "++ insertBlankLine SETS" << bottomSubtotalRow;

        /* Since March 2023 I added subtotal/total.
         * There is a smarter way to do this, but let's do it the stupid
         * error-free way for now:
         * EDIT: Cannot do this because drag/drop calls this function
         * while starting = == === types of lines and you cannot
         * do this Update until all the 8+ lines are done.
         * The symptom is an endless loop but I am not going to fix it
         * since you should not do updates for every line that is added.
         */
        //setDRAGtotalsBottomSubtotalRow();
        //searchForSubtotalTotalUpdateThem( 1 );
    }
    return(ret);
}

bool MainWindow::insertBlankLineLibSide( long row)
{
    unsigned xx, rw;
    bool ret = true;

    if( isFullRow( libRow[LIBROWCOUNT-1] ) == true )
        ret = false;
    else if( row > 0)
    {
        for( rw=ESTROWCOUNT-1; rw>row; --rw)
        {
            for( xx=1; xx<=4; ++xx )
            {
             // see rw-1. transfer from a line lower, but not beyond line limit.

                    libRow[rw]->col[xx]->selectAll();
                    libRow[rw]->col[xx]->insert( libRow[rw-1]->col[xx]->text() );
            }
        }
        // delete one line at "row" of estimate
        for( xx=1; xx<=4; ++xx )
            libRow[row]->col[xx]->clear();

#ifdef UNDO_REDOSTACK_notIntendedToWork_justNotes
        libLineEditHistory *xx = new libLineEditHistory( this, targ, src, Del );
        // transfer a pointer to the object:
        libUndoCommand *command = new libUndoCommand( xx );
        // no delete because the stack needs it.
        //delete xx;   Instead, delete *archive variable later
        // See clearLibEst.cpp and clearLibStack().

        lib_undo_stack.push(command);
#endif

    }
    return(ret);
}

void MainWindow::delLineEstSide( long row)
{
    unsigned xx, rw;
    bool blockDelete = false;

    if( row > 0)
    {
        /* If this is a block of = == === subtotal/totals then find
         * the move up incrementally to the top = of the block.
         */
        if( estRow[row]->col[1]->text().indexOf("=") >= 0 )
        {
            blockDelete = true;
            if( row > 1 )
            {
                if( estRow[row-1]->col[1]->text().indexOf("=") >= 0 )
                    --row;

                /* special case: where since March 2023 I have added a subtotal/total
                 * function and I set the libcode to = or == or ===
                 * and I set the readonly(true). This readonly MUST be undone.
                 * User is deleting a subtotal/total area and it must be a normal line.
                 */
                for( rw=row; rw<ESTROWCOUNT; ++rw)
                {
                    if( estRow[rw]->col[1]->text().indexOf("=") < 0 )
                        break;
                    else
                    {
                        for( xx=1; xx<=5; ++xx )
                            estRow[rw]->col[xx]->setReadOnly(false);

#ifdef MULTIPLIER_colours
                        // the orange and blue highlights have to be set back to normal palette()
                        estRow[rw]->col[3]->setStyleSheet("background-color: palette()" );
                        estRow[rw]->col[6]->setStyleSheet("background-color: palette()" );
                        estRow[rw]->col[7]->setStyleSheet("background-color: palette()" );
#endif
                    }
                }
            }
        }
#ifdef LAZY
        I found the strays were in eventMouseKeyboard.cpp, where I clear() ===s
                and also Del and Ins (and undo and redo sometimes) all
                do TYPEDel and TYPEIns. Those scroll
                multiple sections of = == === up and down, not just one section!,
                without moving the setReadOnly(false);setReadOnly(true); up/dn.
        See also  insertBlankLineEstSide() which scans the entire estimate too.

        else  // normal non-"===" types of est lines
        {
            /* This should not be required, but I see stray "true"s which
             * might be from normal undo/redo in estundocommand.cpp
             * conflicting with the Ins/Del types of line changes because
             * they share the same QUNCOMMAND stack.
             * Instead of modifying estundocommand.cpp, I am just setting
             * these to normal here, although they should already be normal.
             */
            for( xx=1; xx<=5; ++xx )
                estRow[row]->col[xx]->setReadOnly(false);
        }
#endif

        // pre-March 2023 and post-March 2023:
        for( rw=row; rw<ESTROWCOUNT-1; ++rw)
        {
            for( xx=1; xx<=7; ++xx )
            {
             // see rw+1. transfer from a line lower, but not beyond line limit.

                    estRow[rw]->col[xx]->selectAll();
                    estRow[rw]->col[xx]->insert( estRow[rw+1]->col[xx]->text() );
                    estRow[rw]->fileQtyBlock = estRow[rw+1]->fileQtyBlock;
                    estRow[rw]->refcode = estRow[rw+1]->refcode;
            }
            if( estRow[rw]->col[1]->text().indexOf("=") < 0 )
            {
                estRow[rw]->col[1]->setReadOnly(false);
                estRow[rw]->col[2]->setReadOnly(false);

                //estRow[rw]->col[3]->setStyleSheet("background-color: palette()" );
                //estRow[rw]->col[6]->setStyleSheet("background-color: palette()" );
                //estRow[rw]->col[7]->setStyleSheet("background-color: palette()" );
            }
            else
            {
                estRow[rw]->col[1]->setReadOnly(true);
                estRow[rw]->col[2]->setReadOnly(true);

#ifdef MULTIPLIER_colours
                if( estRow[rw]->col[2]->text().indexOf(SUBTOTALHOURSDESCRIPTION) >= 0 )
                {
                    //estRow[rw]->col[3]->setStyleSheet("background-color: orange" );
                    ;//estRow[rw+1]->col[6]->setStyleSheet("background-color: orange" );
                }
                else if( estRow[row]->col[2]->text().indexOf(SUBTOTALCOSTSDESCRIPTION) >= 0 )
                {
                    //estRow[rw]->col[3]->setStyleSheet("background-color: blue" );
                    ;//estRow[rw+1]->col[7]->setStyleSheet("background-color: blue" );
                }
#endif
            }
        }
        // delete one line on the end of estimate
        for( xx=1; xx<=7; ++xx )
            estRow[ESTROWCOUNT-1]->col[xx]->clear();

        estRow[ESTROWCOUNT-1]->fileQtyBlock = nullptr;
        estRow[ESTROWCOUNT-1]->refcode = "";
        estRow[ESTROWCOUNT-1]->col[1]->setReadOnly(false);
        estRow[ESTROWCOUNT-1]->col[2]->setReadOnly(false);

        estRow[ESTROWCOUNT-1]->col[3]->setStyleSheet("background-color: palette()" );
        estRow[ESTROWCOUNT-1]->col[6]->setStyleSheet("background-color: palette()" );
        estRow[ESTROWCOUNT-1]->col[7]->setStyleSheet("background-color: palette()" );

#ifdef UNDO_REDOSTACK_notIntendedToWork_justNotes
        estLineEditHistory *xx = new estLineEditHistory( this, targ, src, TYPEInsDel );
        // transfer a pointer to the object:
        estUndoCommand *command = new estUndoCommand( xx );
        // no delete because the stack needs it.
        //delete xx;   Instead, delete *archive variable later
        // See clearLibEst.cpp and clearEstStack().

        est_undo_stack.push(command);
#endif
        --bottomSubtotalRow; // this is the last known descriptionTOTAL. If it goes negative it is ignored.
        ;//qDebugA() << "-- delLine SETS" << bottomSubtotalRow;

        if( blockDelete == true && estRow[row]->col[1]->text().indexOf("=") >= 0 )
        {
            delLineEstSide(row); // del the next one too because it is a = == ===
        }
        else
        {
            /* Since March 2023 I added subtotal and subtotal+total summing function
             * which is done in void MainWindow::insertTotalSubtotal( int row )
             * There is a smarter way to do this, but let's do it the stupid
             * error-free way for now, which starts at the top, row 1.
             */
            setDRAGtotalsBottomSubtotalRow();
            searchForSubtotalTotalUpdateThem( 1 );
        }
    }
    return;
}

void MainWindow::delLineLibSide( long row)
{
    unsigned xx, rw;

    if( row > 0)
    {
        for( rw=row; rw<LIBROWCOUNT-1; ++rw)
        {
            for( xx=1; xx<=4; ++xx )
            {
             // see rw+1. transfer from a line lower, but not beyond line limit.

                    libRow[rw]->col[xx]->selectAll();
                    libRow[rw]->col[xx]->insert( libRow[rw+1]->col[xx]->text() );

            }
        }
        // delete one line on the end of library
        for( xx=1; xx<=4; ++xx )
            libRow[LIBROWCOUNT-1]->col[xx]->clear();

#ifdef UNDO_REDOSTACK_notIntendedToWork_justNotes
        libLineEditHistory *xx = new libLineEditHistory( this, targ, src, TYPEInsDel );
        // transfer a pointer to the object:
        libUndoCommand *command = new libUndoCommand( xx );
        // no delete because the stack needs it.
        //delete xx;   Instead, delete *archive variable later
        // See clearLibEst.cpp and clearLibStack().

        lib_undo_stack.push(command);
#endif
    }
}

void MainWindow::searchForSubtotalTotalUpdateThem( int rr )
{
    // search for any subtotal = areas and update their totals.

    do{

        bool done = false;

        ;//qDebugA() << "bottomSubtotalRow:" << bottomSubtotalRow;
        for( ; rr<=bottomSubtotalRow && rr<=ESTROWCOUNT-1; ++rr )
        {
            if( done==false && estRow[ rr ]->col[1]->text().indexOf("=") >= 0 )
            {
                done = true;

                ;//qDebugA() << "bottomSubtotalRow:" << bottomSubtotalRow <<"row"<<rr;

                /* I combined Hours and Costs into one subtotal/total inserter
                 * therefore the same function gets called and overwrites both Hours and Costs.
                 * so omit the if() statements and start on any row.
                 *
                if( estRow[ rr ]->col[2]->text().indexOf(SUBTOTALCOSTSDESCRIPTION) >= 0)
                {
                    // this is also called by drag and drop the QCOMMAND button
                    if( insertTotalSubtotal( rr ) == false )
                        return;
                }

                if( estRow[ rr ]->col[2]->text().indexOf(SUBTOTALHOURSDESCRIPTION) >= 0)
                {*/

                if( insertTotalSubtotal( rr ) == false ) // this is also called by drag and drop the QCOMMAND button
                    return;
            }
            // gone past the = segment
            if( done == true &&
                estRow[ rr ]->col[1]->text().indexOf("=") < 0 )
            {
                break; // finished this block
            }
        }
        // find the next block
    }while( rr<=bottomSubtotalRow && rr<=ESTROWCOUNT-1 );

    return;
}

