
/*  This file is part of PaceMaker Estimator 2.0.
    PaceMaker Estimator 2.0 is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
    PaceMaker Estimator 2.0 is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with PaceMaker Estimator 2.0. If not, see <https://www.gnu.org/licenses/>.
*/

#include "liblineedithistory.h"
#include "mainwindow.h"

libLineEditHistory::libLineEditHistory( MainWindow *xx, long targ, long src, int typeOfAction )
{
    QString ss;
    unsigned yy;

    saveTarg = targ;
    saveSrc = src;
    saveTypeOfAction = typeOfAction;
    saveParent = xx;

    /* archive the first state for that line so undo/redo libUndoCommand works.
     * See libUndoCommand.cpp where this libLineEditHistory object data is used.
     */
    for( yy=1; yy<=4; ++yy)
    {
        // this probably doesn't happen since I stopped doing yy=0
        if( xx->libRow[targ]->col[yy] == nullptr )
            this->oldlibcol[yy] = "";
        else
            this->oldlibcol[yy] = xx->libRow[targ]->col[yy]->text();

        // these are QLineEdit pointers to one library line
        libRowActual[yy] = xx->libRow[targ]->col[yy];
    }

    switch( typeOfAction )
    {
        case TYPEEstLib:
        preferredScrollIntoViewQle = xx->libRow[targ]->col[4];

    ;//qDebugA() << "Add a line from est to lib for this row:" << src;

    if( (ss = xx->estRow[ src ]->col[1]->text()) == nullptr )
        ss = "";
    ;//qDebugA() <<ss;
    xx->libRow[targ]->col[1]->selectAll(); // only send upper case to library libcode
    xx->libRow[targ]->col[1]->insert( ss.toUpper() );

    if( (ss = xx->estRow[ src ]->col[2]->text()) == nullptr )
        ss = "";
    ;//qDebugA() <<ss;
    xx->libRow[targ]->col[2]->selectAll();
    xx->libRow[targ]->col[2]->insert( ss );

    // do not use (EST->LIB) estRow col[3] because LIB has no Qty

    if( (ss = xx->estRow[ src ]->col[4]->text()) == nullptr )
            ss = "";
    ;//qDebugA() <<ss;

    xx->libRow[targ]->col[3]->selectAll();  // est col 4 goes to lib col 3
    xx->libRow[targ]->col[3]->insert( ss );

    if( ( ss = xx->estRow[ src ]->col[5]->text()) == nullptr )
            ss = "";
    ;//qDebugA() <<ss;
    xx->libRow[targ]->col[4]->selectAll(); // est col 5 goes to lib col 4
    xx->libRow[targ]->col[4]->insert(ss );

    break;

    case TYPEDel:
        preferredScrollIntoViewQle = xx->libRow[targ]->col[4];
        // done by .redo(), so skip this: xx->delLineLibSide(targ); // set targ = src before calling
    break;

    case TYPEIns:
        preferredScrollIntoViewQle = xx->libRow[src]->col[4];
        /* set src to desired row, and targ=LIBROWCOUNT-1
         * before calling this function because last row needs recovery for Ins.
         */
        // done by .redo(), so skip this: xx->insertBlankLineLibSide(src);
    break;

    case TYPEmenuEdit:
        preferredScrollIntoViewQle = xx->libRow[targ]->col[4];

        ;//qDebugA() << "Add a line from est to lib for this row:" << src;

        if( (ss = xx->copyL->libRow[ src ]->col[1]->text()) == nullptr )
            ss = "";
        ;//qDebugA() <<ss;
        xx->libRow[targ]->col[1]->selectAll(); // only send upper case to library libcode
        xx->libRow[targ]->col[1]->insert( ss.toUpper() );

        if( (ss = xx->copyL->libRow[ src ]->col[2]->text()) == nullptr )
            ss = "";
        ;//qDebugA() <<ss;
        xx->libRow[targ]->col[2]->selectAll();
        xx->libRow[targ]->col[2]->insert( ss );

        // do not use (EST->LIB) estRow col[3] because LIB has no Qty

        if( (ss = xx->copyL->libRow[ src ]->col[3]->text()) == nullptr )
            ss = "";
        ;//qDebugA() <<ss;

        xx->libRow[targ]->col[3]->selectAll();
        xx->libRow[targ]->col[3]->insert( ss );

        if( ( ss = xx->copyL->libRow[ src ]->col[4]->text()) == nullptr )
            ss = "";
        ;//qDebugA() <<ss;
        xx->libRow[targ]->col[4]->selectAll();
        xx->libRow[targ]->col[4]->insert(ss );

    break;

    default:
        preferredScrollIntoViewQle = xx->libRow[targ]->col[4];
    break;
    }

    // archive the second state for that line
    for( yy=1; yy<=4; ++yy)
    {
        // this probably doesn't happen since I stopped doing yy=0
        if( xx->libRow[targ]->col[yy] == nullptr )
            this->libcol[yy] = "";
        else
            this->libcol[yy] = xx->libRow[targ]->col[yy]->text();
    }

    return;
}

