
/*  This file is part of PaceMaker Estimator 2.0.
    PaceMaker Estimator 2.0 is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
    PaceMaker Estimator 2.0 is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with PaceMaker Estimator 2.0. If not, see <https://www.gnu.org/licenses/>.
*/

#include "libundocommand.h"

libUndoCommand::libUndoCommand( libLineEditHistory *xx, QUndoCommand *parent )
                    : QUndoCommand(parent)
{
    /* "archive" is the lib line qlineedit pointers,
     * which point to
     * the library side, and it has the before and after
     * text for each column of that lib line.
     */
    archive = xx;
}

void libUndoCommand::undo()
{
    unsigned yy;

    if( archive->saveTypeOfAction == TYPEDel )
    {
        archive->saveParent->insertBlankLineLibSide(archive->saveTarg);
    }
    else if( archive->saveTypeOfAction == TYPEIns )
    {
        archive->saveParent->delLineLibSide(archive->saveSrc);
    }

    for( yy=1; yy<=4; ++yy )
    {
        archive->libRowActual[yy]->selectAll();
        archive->libRowActual[yy]->insert( archive->oldlibcol[yy] );
        //archive->libRowActual[yy]->setText( archive->oldlibcol[yy] );
    }

}

void libUndoCommand::redo()
{
    unsigned yy;

    if( archive->saveTypeOfAction == TYPEDel )
    {
        archive->saveParent->delLineLibSide(archive->saveTarg);
    }
    else if( archive->saveTypeOfAction == TYPEIns )
    {
        archive->saveParent->insertBlankLineLibSide(archive->saveSrc);
    }
    else
    {
        for( yy=1; yy<=4; ++yy )
        {
            archive->libRowActual[yy]->selectAll();
            archive->libRowActual[yy]->insert( archive->libcol[yy] );
            //archive->libRowActual[yy]->setText( archive->libcol[yy] );
        }
    }

}

void libUndoCommand::delete_libLineEditHistory()
{
    delete archive;
}

/* See this in libTOest.cpp redo/do buttons, instead
 * because ui is not available to this object.
 *
void libUndoCommand::scroll_est_into_view( QLineEdit *qle )
{
    int hh,ww;

    hh = qle->frameGeometry().height();
    ww = qle->frameGeometry().width();
    ui-> scrollArea->ensureWidgetVisible( qle, ww, hh);
 }

void libUndoCommand::scroll_lib_into_view( QLineEdit *qle )
{
    int hh,ww;

    hh = qle->frameGeometry().height();
    ww = qle->frameGeometry().width();
    ui->scrollArea_2->ensureWidgetVisible( qle, ww, hh);
}
 */
