
/*  This file is part of PaceMaker Estimator 2.0.
    PaceMaker Estimator 2.0 is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
    PaceMaker Estimator 2.0 is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with PaceMaker Estimator 2.0. If not, see <https://www.gnu.org/licenses/>.
*/

#include "mainwindow.h"

#include <QApplication>
#include <QLocale>
#include <QTranslator>

int main(int argc, char *argv[])
{
    long xx;
    short yy;
    QString eD = nullptr;

    QApplication aa(argc, argv);

    QTranslator translator;
    const QStringList uiLanguages = QLocale::system().uiLanguages();
    for (const QString &locale : uiLanguages) {
        const QString baseName = "PaceMakerEstimator_" + QLocale(locale).name();
        if (translator.load(":/i18n/" + baseName)) {
            aa.installTranslator(&translator);
            break;
        }
    }

#ifdef Experiment
    QLabel *msgBox3;
    QDialog *QDia;

        // StaysOnTop blocks keyboard input on other windows!
        //QDia = new QDialog( this, Qt::WindowStaysOnTopHint);
        //QDia->setModal(false); and ignores modal false.
        QDia = new QDialog(nullptr);
        //QDia->setParent();
        QDia->setWindowTitle("* wait *");
        QDia->setSizeGripEnabled(true);
        //QDia->setGeometry( QDia->x(), QDia->y(), ui->scrollArea_4->geometry().x()/2, ((ui->scrollArea_4->geometry().x()/2)*1000)/1618L );
        //QDia->setFixedSize( ui->scrollArea_4->geometry().x()/2, ((ui->scrollArea_4->geometry().x()/2)*1000)/1618L );
        msgBox3 = new QLabel("Wait\nWhile\nLoading",QDia);
        msgBox3->setFrameStyle(QFrame::Panel | QFrame::Sunken);

        QDia->show();
        QDia->raise();
        aa.processEvents(QEventLoop::AllEvents);

#else
    // See waitLoading() in mainwindow.cpp. It fails when in main.cpp.
#endif

    // check if the user wants to specify a data subdirectory.
    for( int xx=1; xx<argc; ++xx)
    {
        if( strncmp( argv[xx], "estDir=",7) == 0 )
        {
            eD = QString("%1").arg(argv[xx]).sliced(7).simplified();
            ;//qDebugA() << "eD" << eD;
            QDir dir;
            dir.setPath( eD );
            if( !dir.exists() )
                eD = nullptr;
            break;
        }
    }

    /* This solves a bug which I saw on a 2009 1/10th slow linux laptop:
     * qtHow to avoid SIGPIPE (due to a timeout ) when debugging an X11 program - Stack Overflow.pdf
     * ICEauthority IO error handler. Putting this after MainWindow() does not work.
     */
    //aa.processEvents(QEventLoop::AllEvents);
    /* causes a truncation to 367 est rows,
     * which is the same as when I set Fixed(width,height) in load2.cpp
     * So I cannot fix this. Just do not run this program on a slow computer.
     */
    MainWindow ww(nullptr, eD);

   /* I moved all the junk back to mainwindow.cpp. I thought
    * there is a way to not show anything until load2() and waitLoading()
    * completes here in main.cpp. However, all my ideas failed because
    * apparently the signals are not activated until ww.show()
    * and ww.raise() and waitLoading() needs those signals. I thought
    * I could put ww.show() and ww.raise() inside waitLoading(). That idea
    * failed with NOTHING showing and PaceMakerEstimator was running
    * endlessly waiting for signals, apparently.
    *
    * I moved the junk back to mainwindow.cpp
    * Feb 14, 2024.
    * Maybe hide() put inside waitLoading() will work?
    *   No, because hide() turns off the spontaneous() signals
    *   which I am counting. It nicely hid mainwindow though.
    */
    ww.show();
    ww.raise();

    return aa.exec();
}
