
/*  This file is part of PaceMaker Estimator 2.0.
    PaceMaker Estimator 2.0 is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
    PaceMaker Estimator 2.0 is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with PaceMaker Estimator 2.0. If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include "menuEdit.h"
#include "qcoreevent.h"
#include "qprinter.h"
#include "qtymath.h"
#include "qtymathhidden.h"

#include "dialogheaderofestimate.h"

#include "qtymathhidden.h"
#include "qundostack.h"
#include "ui_mainwindow.h"
#include <QMainWindow>
#include <QKeyEvent>
#include <QMessageBox>
#include <QFileDialog>
#include <QDirIterator>
#include <QInputDialog>

//#define ARTIFICIALtrial_TESTallROWs

// this does nothing: #define QT_NO_DEBUG_OUTPUT
//#define QT_NO_INFO_OUTPUT
//#define QT_NO_WARNING_OUTPUT

/*
 * See menuEdit.h
 * I think LIBROWCOUNT should equal ESTROWCOUNT (note April 2023)
 * only if I forgetfully used ESTROWCOUNT in for() loops
 * while doing libRow[]. I checked all the libRow code
 * if you want LIBROWCOUNT different than ESTIMATE ROW size.
 * Edit:
 * I found one more error which I did fix March 10 2024
 *                 libLineEditHistory *eH = new libLineEditHistory(
                    this, ESTROWCOUNT-1, keyboardnow.row, TYPEIns);
    I changed it to LIBROWCOUNT. estLineEditHistory is ESTROWCOUNT.
 */

  /*typical for linux:*/
#define DOTCONFIG ".config"
#define ESTIMATEMAKER "PaceMakerEstimator"
#define ESTCONFIG "PaceMakerEstimator.config"

//Used with ui->TotalHours ui->TotalCost
#define DRAGtotalhours "\n%1"
#define DRAGtotalcosts "\n%1"
//#define DRAGtotalhours "\n  %1"
//#define DRAGtotalcosts "\n  %1"
//#define DRAGtotalhours "  %1\n  Qty*Hrs"
//#define DRAGtotalcosts "  %1\n  Qty*Cost"

/* These are used in the StatusTip and later in the parsing.
 * Alignment matters. So I centralised them as #defines to edit (and catch errors) easily.
 */
#define OFFSET 14
#define OFFSET2 21
#define DIGITSforROW 6
#define LIBFIELD1 "Library  line 1     [library item code]"
#define LIBFIELD2 "Library  line 1     [description]"
#define LIBFIELD3 "Library  line 1     [item cost for each]"
#define LIBFIELD4 "Library  line 1     [item hours labour for each]"

#define BUTTONE   "Estimate line 1     [insert line into library]";
#define BUTTONL   "Library  line 1     [overwrite library line to estimate]";

#define ESTFIELD1 "Estimate line 1     [library item code]"
#define ESTFIELD2 "Estimate line 1     [description]"
#define ESTFIELD3 "Estimate line 1     [quantity]"
#define ESTFIELD4 "Estimate line 1     [item cost for each]"
#define ESTFIELD5 "Estimate line 1     [item hours labour for each]"
#define ESTFIELD6 "Estimate line 1     [total hours for the quantity]"
#define ESTFIELD7 "Estimate line 1     [total cost for the quantity ]"


#define FROMDECONSTRUCTOR true
#define TYPELibEst 1
#define TYPEEstLib 1
#define TYPEDel 2
#define TYPEIns 3
#define TYPEmenuEdit 4

QT_BEGIN_NAMESPACE
namespace Ui { class MainWindow; }
QT_END_NAMESPACE

class MainWindow : public QMainWindow
{
    Q_OBJECT

public:
    MainWindow(QWidget *parent = nullptr, QString eD = nullptr);
    ~MainWindow();

    bool globalFrame = true; // I like all QLineEdits() framed now but I might change my mind

    void dialogNewFilename();
    bool dialogNewFilename_sub( bool clearEst );
    void dialogOpenFilename();
    void dialogOpenFilename_sub();
    void dialogSaveFilename();
    void dialogSaveAsFilename();
    bool dialogSaveFilename_sub( bool fromDeconstructor );
    void mainQuit();
    void dialogHeaderofEstimate_estimateHeaderToBuffer( MainWindow* parent );

    bool insertBlankLineEstSide( long row);
    bool insertBlankLineLibSide( long row);
    void delLineEstSide( long row);
    void delLineLibSide( long row);

    bool libraryRead( QString *filename, QString *partialText, bool comboBoxTextOnly );
    bool libraryWrite( QString *filename, QString partialText );
    qsizetype searchLibrary( QString libcode );
    long searchLibrarySlowly( QString libcode );
    bool currentLibSortON = false;

    QString estimateDirectory = nullptr;
    QString estDotlib =       "est.lib";
    void linuxMsWindowsSlashFixes( QString *eD );

    QString estimateFilename = nullptr;
    bool estimateWasChanged = false;
    QString libraryFilename = nullptr;
    QString libraryDescription = nullptr;
    void libraryEsttoLib();
    void libraryLibtoEst();
    void waitLoading( unsigned on, unsigned setLimit );

    QUndoStack lib_undo_stack, est_undo_stack;

 // QLineEdits:

    /* See mainwindow.cpp comments about why I have to point at 11,000+ object names for
     * Qt Creator.
     * Also, I did this as a structure instead of a two dimensional array
     * in case I think of other characteristics to add for each row, like special
     * purpose rows flagging subtotals/totals.
     */

    // If you know the row and col, you can look up the QLineEdit object here:
    struct QLineEditArray {
                        qtymathHidden *fileQtyBlock;
                        QString refcode;//two chars says *WD or *B, listname
                                        //fileQtyBlock has it too because
                        // that block can be saved to the est.lib too,
                        // (Jan 2024: I think I never implemented fileQtyBlock on lib side)
                        // and it is one filename per 3 char libcode like this:
                        // WD.B08 or B_.S24 in the case of *B listname.

                        QLineEdit *col[MAXCOL_libest];
                        QPushButton *pushbutton;
            } *libRow[ESTROWCOUNT], *estRow[ESTROWCOUNT];

 /* New February 2024: After completing the entire project, six months of work,
  * I installed it on a 14 year old laptop and it was 4 minutes to start up.
  * It was too slow. I read in stackexchange that I put too many widgets in one layout.
  * I have 11 per row, plus 2 buttons times 800 lines = 10,400.
  * Each one has two events. See Model_View Programming _ Qt Widgets 6.6.2.pdf
  * So I reduced the lines to 231 instead of 800 and performance is very good.
  * However, the old 8088 MSDOS program of 1987 had almost 800 and was very fast as text-only
  * and had the same functionality without a mouse.
  * The computer had 0.640 megabytes of memory and look up the CPU specs and laugh.
  *
  * So here I introduce a new abstraction layer with the concept of Tabs
  * which you see at bottom of estimate side: Tab 1 through Tab 10.
  * These will be loaded and unloaded into the QLineEdits so the 231 lines of QLineEdits
  * can be reused. The entire estimate will be 10x231=2310 lines
  * and the performance will not be sluggish, like if I attempted putting them all
  * in a layout.
  *
  * The event slot function is called: on_tabWidget_currentChanged(index0to9)
  * in tabs_plusEstReadWrite.cpp. The estimateRead() estimateWrite() functions
  * will be changed because those will be rewritten to load this new structure
  * of 2310 lines, instead of loading the 231 QLineEdits directly, which was
  * pre-Feb2024. I might move them from menuEdit.cpp to tab.cpp.
  *
  * Secondly, a load and unload function will have to swap the inactive/active Tab's data
  * out-of/into the 231 QLineEdits. It is not too complicated to follow
  * one new structure, one new event SLOT, and two new functions and see estimateRead()
  * modified as two steps to load an estimate instead of one step. Also estimateWrite()
  * is two steps now. The steps are: read the filename estimate into Tabs and then write
  * one Tab to QLineEdits. The steps are: write the currentQLinesEdits into one Tab and
  * write all the Tabs into filename.
  *
  * void copyQLineEditstoTab();
    void copyTabtoQLineEdits(); are the two new functions.
  *
  * New Tab array is easily understandable as parallel to the struct QLineEditArray {}
  * but as strings and data and no QLineEdits since it is not a visible layout component
  * and no library items and no buttons. Also I kept the same names so I can modify the
  * old code by swapping estRow to estTabRow.
  *
  * Look for it in tabs_plusEstReadWrite.cpp
  *
  typedef struct estTab {
        qtymathHidden *fileQtyBlock;
        QString refcode;
        QString col[MAXCOL_libest];
    } *estTabRow[ESTROWCOUNT];

After I finished Tab by March 2024, I realised I could have made them
all QLineEdits and initialised them in MainWindow.cpp/load2.cpp constructor
as separate and they would run fast as long as they are not in a layout.
I did not have to use QStrings. Why would this be an advantage?
The structures would be in the same place here. Anything else?
Maybe the pointers could be swapped in the layout instead
of doing tab....cpp unload...() load...() . It is not much
different on second thought, but I see the layout could have been
a window on one big single estimate and no tabs at all, just start
and finish pointers, estRow[454] to estRow[685] for example,
assigned into the layout. That is what Qt's window-model said.
Model_View Programming _ Qt Widgets 6.6.2.pdf
It could be a moving window. My tab idea was less coding error likelihood.
  */

 // Keyboard data, no mouse cursor data:

 /* 1=libest=library side, 2=libest=estimate side. [row][col].
  * This is the current keyboard and can be used to look up the QLineEdit
  * using the struct above.
  *
  * mouse cursor and keyboard cursor locales can be different row/col from each other.
  * getrowcol_ofcursors() loads and show_linenums_showHighlights() are recorded by
  * eventsMouseKeyboard.cpp upon keyboard FocusIn Out events. Mouse row/col is ignored.
  * This is a record of where, so I know what
  * the user is trying do if an action is called, like page up page down and
  * I, as programmer, say to myself "which side, estimate or library? Which box?".
  * Now I know by checking here:
  */
  struct keyboard_now {

        long litLibButton=0L, litEstButton=0L;

        /*initialize as est side, 2, although it is unknown until
         *user does something.
         *Which side is the keyboard cursor clicked on now? lib=1,est=2
         */
        short libest=2;

        // See ensureWidgetVisible() to know what is visible.

        long row=0L, rowE=0L, rowL=0L;// 3 keyboard rows information.
        short col=0, colE=0, colL=0;

        /* 0 means no blinking keyboard cursor; no cell clicked as yet.
         * rows are 1 through ESTROWCOUNT-1
         */
        /*
         * When a new click comes in, it is always stored in
         * example: row (for any lib or est side clicks),
         * never to lit, but yes copied to rowL or rowE.
         *
         * Why? How do you shift a highlighted button to another button?
         * You turn off the old and turn on the new. So you need two variables.
         * Highlights are the red border of the left/right buttons,
         * down the centre of the screen. I use the word lit.
         *
         * If you have two variables then one function can record clicks
         * in the event...() function and they might not be needed, but
         * they are always in row or copied to rowE, for example.
         *
         * A second function can query rowE, and a third highlight function
         * can be fed both litEstButton ... or Lib... as arguments
         * for lib and est sides, and turn off and turn on and shift new
         * rowE, or rowL to now highlighted row, for both sides.
         *
         * See eventMouseKeyboard.cpp, eventFilter(),
         * getrowcol_ofcursors(), and show_linenums_showHighlights().
         *
         * Think of row and col as a scratch pad
         * of the last click on either lib or est side; unspecified side.
         *
         * rowE and rowL are specified est or lib sides and useful
         * information of last known blinking cursor on either side,
         * but might not be blinking now.
         *
         * litEstButton litLibButton are historical and in use as a
         *      button highlight, waiting to be extinguished.
         */
    } keyboardnow;

  estMenuEdit *copyE = nullptr;
  libMenuEdit *copyL = nullptr;
  long prefSelectCount=1;

      // See ensureWidgetVisible() for know what is visible in a scroll area.

    // Also, the mouse has to visit one side and the other for verticalSlider()
    // to lock the two windows out of sync, like for example line 1 one side, line 8
    // the other side, via mouse wheel scrolling and clicking on a QLineEdit
    // I gave up on mouse cursor. It is too slow, by linear/binary search, to detect
    // the location of the mouse cursor. So everything is keyboard cursor.

    // used in libTOest.cpp
    void scroll_est_into_view( QLineEdit *qle );
    void scroll_lib_into_view( QLineEdit *qle );
    //void scroll_est_into_view( QLineEditArray *qlea );
    //void scroll_lib_into_view( QLineEditArray *qlea );

    // set in load2() and in readSettings() and used in dialogheaderofestimate.cpp
    QFont globalFont2;
    QFont qtyDetailBoxMonospaceFont;

    // save variable here so as to be able to kill it, if pressed again.
    // Only one is allowed to be alive at a time because only one calc box is visible at a time.
    qtymath *archiveQtyFrame = nullptr;
    bool qtymathRequestsQtyColBoxHighlight = false;
    long saveQtyMathHighlightRow = 0L;
    long bottomSubtotalRow = -1L; // see eventMouseKeyboard.cpp for subtotals/total

    DialogHeaderOfEstimate *oneHeader = nullptr;

    Ui::MainWindow *ui;

    bool insertTotalSubtotal( int nowrow );
    void searchForSubtotalTotalUpdateThem( int row );
    void qleToUpper( QLineEdit *qle );
    bool writeSettings( QString usually_homePath );
    bool readSettings(QFont *font );
    void zerosetDRAGtotalsBottomSubtotalRow();
    void setDRAGtotalsBottomSubtotalRow();

public slots:
    void doCalc();
    void load2( QFont font );
    bool isFullRow( struct QLineEditArray *qlea );

    // estimate lines undo/redo
    void on_bundoEst_pressed(); //Moved from private to public because this one is used in LibtoEst() to prevent need for double-undo (special case; comment it out and recompile to see what happens with Undo if it is not in)
    void on_bredoEst_pressed();
    // lib lines undo/redo
    void on_bundoLib_pressed();
    void on_bredoLib_pressed();

private slots:

    void on_comboBox_currentIndexChanged(int index);

    void on_verticalSlider_valueChanged(int value);

    void getrowcol_ofcursors(short *libest,
                             long *row, long *rowL, long *rowE,
                             short *col, short *colL,short *colE); // fetches the current row/col of both cursors

    // jump from square to square forward and backward with keys instead of mouse:
    void tabRightKeyMove();
    void tabLeftKeyMove();

    void clearLib( QString defaultLibcode2char );
    void clearEst();
    bool loadComboboxLibs( QString );
    void loadByIndexLibrarynameOfComboboxLibs(int xx);
    QString cleanQString(QString ss, int sz);

    void makeNewLibrary();

    void on_sortOnce_triggered();
    void on_sortAlways_triggered();
    void on_actionLibrariesDelete_triggered();
    void on_actionLibrariesRecover_triggered();

    void calc_one_estline( long row );

#ifdef AllThisIsNowDoneIn_doInsertHoursTotalSubtotal
    void calc_subtotalTotal( QString multiplier,
                             QString multiplicand,
                             QLineEdit *product );
#endif

    void validateField();
    void show_linenums_showHighlights( long *litLibButton, long *litEstButton, long rowl, long rowe );

    void eraseHighlights( long *litLibButton, long *litEstButton );

    short estimateLinesVisible( long, short);

    QLineEdit* currentKeyboardLineEdit();
    void clearCurrentKeyboardCursorFocus();
    void historySave_ZeroSetKeyboardCursor();
    QString stripFilename( QString *ss );

    void on_clearSlider_pressed();

    void dialogChooseFont();
    void action_xHours();
    void action_xCosts();
    void dialogChooseEstimateDirectory();

    void dialogChooseSelectCount();
    void setEditSelected_Row();
    void EditCopy( );
    void EditCut( );
    void EditInsert( );
    void EditPaste( );
    void EditUndo( );
    void EditRedo( );

    void actionQtLicence();
    void setqtmsgWindowTitle( QString title );

    void on_pushHeaderButton_released();

    bool estimateRead( QString *filename );
    bool estimateWrite( QString *filename );

    void dialogPrint();
    bool dialogPrint_sub(QPainter *painter, QPrinter *printer, int page, int countOfPrints, int whichtt );

    void sprintfreverse( QLineEdit *arg1, char *arg2, char *arg3 );
    void sprintfreverseTabversion( QString arg1, char *arg2, char *arg3 );

    char *stripEnd( char *ee );


    void clearLibStack();
    void clearEstStack();


    void on_qtymath_released();
    void on_qtymath_pressed();
    float tediously_meticulously1987compatibility( char *str, qtymath *qtytemp);

    //void on_comboBox_editTextChanged(const QString &arg1);

    void on_TotalHours_pressed();
    void on_TotalCost_pressed();
    bool parseRowFromStatusTip( long *row, QLineEdit *trg );

    void on_libcodeEstside_clicked(); // this is now a button, left of Description

    void on_tabWidget_currentChanged(int index);
    void clearTabPages();
    bool is1987subtotal( long row );

protected:
        bool eventFilter( QObject *obj, QEvent *event);
private:

    int lastSliderValue=0;

    // #defines had trouble so I switched to assigning them to strings here:
    QString esttolib= BUTTONE;
    QString libtoest= BUTTONL;

    QString libcol1 = LIBFIELD1;
    QString libcol2 = LIBFIELD2;
    QString libcol3 = LIBFIELD3;
    QString libcol4 = LIBFIELD4;

    QString estcol1 = ESTFIELD1;
    QString estcol2 = ESTFIELD2;
    QString estcol3 = ESTFIELD3;
    QString estcol4 = ESTFIELD4;
    QString estcol5 = ESTFIELD5;
    QString estcol6 = ESTFIELD6;
    QString estcol7 = ESTFIELD7;

    QString aboutTitle="PaceMaker Estimator Author and Licence";

};

#endif // MAINWINDOW_H

