
/*  This file is part of PaceMaker Estimator 2.0.
    PaceMaker Estimator 2.0 is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
    PaceMaker Estimator 2.0 is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with PaceMaker Estimator 2.0. If not, see <https://www.gnu.org/licenses/>.
*/

#include "mainwindow.h"


#include <QTimer>
#include <QThread>
#include <QErrorMessage>
#include <QDialog>

/* I think FROMDECONSTRUCTOR serves no purpose as of Mar 7, 2023
 * and I could remove it soon. I was varying messaging.
 */
#define NotFromDeconstructor false

void MainWindow::dialogNewFilename()
{
    QMessageBox msgBox;

    if( estimateWasChanged == true )
    {
            if( dialogSaveFilename_sub( NotFromDeconstructor ) == false )
            {
                msgBox.setText("Try another name or the storage is full.\nThat name failed to save.");
                msgBox.setParent(oneHeader);
                msgBox.exec();
                return;
            }
            //else // the current estimate is adopted as the new estimate.
               // return;  New was failing to happen. User had to press it again.
    }

    if( dialogNewFilename_sub( true ) == false )
    {
           msgBox.setText("Failed to create new estimate.");
           msgBox.exec();
    }
    //else
       // estimateWasChanged = false;

    return;
}

bool MainWindow::dialogNewFilename_sub( bool eraseEst )
{
    QDir dd;
    QString ss;
    bool ret = false;

    // save the current active one in case somebody wrote something and we must not lose data.
    if( !libraryFilename.isEmpty() )
        libraryWrite( &libraryFilename, libraryDescription );

    // I think you can set the QInputDialog for multiple inputs or else
    //  better: https://doc.qt.io/qt-6/qfiledialog.html
    // I used the standard dialogs Example dialogs.cpp.

    //QString select = tr("Text files (*)");

    QString fileName = QFileDialog::getSaveFileName(this,
                                tr("New Estimate File Name"),
                                estimateDirectory,
                                tr("Estimate Files (*);;Text Files (*.txt)")
                                                    );

    if( !fileName.isEmpty() ) //estimateDirectory is the global name in MainWindow.h
    {
        if( eraseEst == true )
        {
            clearEst();
            //oneHeader->blankHeaderFields(); moved to inside clearEst();
        }

        ss = dd.filePath( fileName );
        linuxMsWindowsSlashFixes(&ss);

        int sepint = ss.lastIndexOf( dd.separator() ); // '\' or '/'
        if( sepint != -1)
            ss.truncate( sepint );
        estimateDirectory = ss;

        ;//qDebugA() << "User selected directory:" << estimateDirectory;
        ;//qDebugA() << "User selected filename:" << fileName;

        dd.mkpath( estimateDirectory );
        if( QFileInfo::exists(ss) && ss!=estimateDirectory )
        {
            QString oldEstimateDirectory = estimateDirectory;
            estimateDirectory = ss;
            // discern homePath()
            int sepint = oldEstimateDirectory.lastIndexOf( dd.separator() ); // '\' or '/'
            if( sepint != -1)
            {
                oldEstimateDirectory.truncate( sepint );
                writeSettings(oldEstimateDirectory);

                /* reload libraries like we do at mainwindow.cpp startup */
                loadComboboxLibs( estimateDirectory + "/" + estDotlib );
                loadByIndexLibrarynameOfComboboxLibs(0);
            }
        }
        dd.setCurrent( estimateDirectory );

        ss = estimateDirectory + "/" + estDotlib;
        if( !QFileInfo::exists(ss))
        {
            dd.mkdir( ss ); // make a est.lib library separate from estimates.
            loadComboboxLibs( ss ); // clears and reloads the list if any is there
            loadByIndexLibrarynameOfComboboxLibs(0);
        }

        if( oneHeader == nullptr )
            oneHeader = new DialogHeaderOfEstimate(this);

        // delete the file if it already exists
        remove( fileName.toStdString().c_str() );
        // create an empty estimate file
        ret = oneHeader->estimateWriteHeader( &fileName );
        if( ret == true )
            estimateWasChanged = false;

        // this is the global name in MainWindow.h
        estimateFilename = fileName;
    }
    else
        ret = true;  //user must have cancelled so return no error.

    return(ret);
}


void MainWindow::dialogOpenFilename()
{
    QString filename;
    QString ss;
    QMessageBox msgBox;

    if( estimateWasChanged == true )
    {
        if( dialogSaveFilename_sub( NotFromDeconstructor ) == false )
        {
            msgBox.setText("Try another name or the storage is full.\nThat name failed to save.");
            msgBox.setParent(oneHeader);
            msgBox.exec();
            return;
        }
        //else  I commented this out because Open in the menu was not happening after auto-save above
        // and user would have to press open again. That was confusing.
        //{
        //    return;
        //}
    }

    // save the current active one in case somebody wrote something and we must not lose data.
    if( !libraryFilename.isEmpty() )
        libraryWrite( &libraryFilename, libraryDescription );

    //https://doc.qt.io/qt-6/qfiledialog.html

    //std::cout << file.openFile()<< std::endl;

    QDir dd;
    QFileDialog qfd(this);

    qfd.setFilter( QDir::Files|QDir::Writable ); // |QDir::NoDotAndDotDot );
    //qfd.setFileMode(QFileDialog::ExistingFile);

    filename = qfd.getOpenFileName( this,
        tr("Open Estimate"), estimateDirectory,
                        tr("Estimate Files (*);;Text Files (*.txt)"));

    if( !filename.isEmpty() )   // this is the global name in MainWindow.h
    {
        qDebug() << "bigtrade name:" << filename;
        clearEst();

        ss = dd.filePath( filename );
        linuxMsWindowsSlashFixes(&ss);

        int sepint = ss.lastIndexOf( dd.separator() ); // '\' or '/'
        if( sepint != -1)
            ss.truncate( sepint );

        //qDebug() << "xUser open selected directory:" << ss;
        //qDebug() << "xUser open selected filename:" << filename;

        dd.mkpath( ss );
        if( QFileInfo::exists(ss) && ss!=estimateDirectory )
        {
            QString oldEstimateDirectory = estimateDirectory;
            estimateDirectory = ss;

qDebug() << "User changed subdirectory" << estimateDirectory << oldEstimateDirectory;

            // discern homePath()
            int sepint = oldEstimateDirectory.lastIndexOf( dd.separator() ); // '\' or '/'
            if( sepint != -1)
            {
                oldEstimateDirectory.truncate( sepint );
                writeSettings(oldEstimateDirectory);

                /* reload libraries like we do at mainwindow.cpp startup */
                loadComboboxLibs( estimateDirectory + "/" + estDotlib );
                loadByIndexLibrarynameOfComboboxLibs(0);
            }
        }
        dd.setCurrent( estimateDirectory );
        ss = estimateDirectory + "/" + estDotlib;
        if( !QFileInfo::exists(ss))
        {
            dd.mkdir( ss ); // make a est.lib library separate from estimates.
            loadComboboxLibs( ss ); // clears and reloads the list if any is there
            loadByIndexLibrarynameOfComboboxLibs(0);
        }

        if( estimateRead( &filename ) == true )
        {
            ;//qDebugA() << "Success reading:" << filename;
        }
        else
            ;//qDebugA() <<"Error reading:" << filename;

        // this is the global name in MainWindow.h
        estimateFilename = filename;
    }

    return;
}

void MainWindow::dialogSaveFilename()
{
    // Do not destroy the oneHeader because save to disk needs it.

    dialogSaveFilename_sub( false );
}

#define SUCCESSFULLYSAVED "\n Successfully SAVED:\n   "+estimateFilename
// These are all close but failed in various ways:
//#define QTIMERMSG3 QThread::msleep(2000);QTimer::singleShot( 3000, Qt::PreciseTimer, &msgBox3, SLOT(done(QDialog::Accepted )) )
//#define QTIMERMSG3 msgBox3.showMessage(ss);QTimer::singleShot( 3000, Qt::PreciseTimer, &msgBox3, SLOT(done(QDialog::Accepted )) )
//#define QTIMERMSG3 msgBox3.setModal(false);msgBox3.setText(ss);msgBox3.show();
//#define QTIMERMSG3 msgBox2.setModal(false);msgBox2.setText(ss);msgBox2.show();
//#define QTIMERMSG3 msgBox2.setText(ss);msgBox2.exec();
//#define QTIMERMSG3 msgBox3 = new QMessageBox(this);msgBox3->setText(ss);msgBox3->exec();QThread::sleep(2);delete msgBox3;
//#define QTIMERMSG3 msgBox3 = new QMessageBox(this);msgBox3->setText(ss);msgBox3->exec();QThread::sleep(2);delete msgBox3;
//#define QTIMERMSG3 msgBox3 = new QLabel(ss,this,Qt::WindowStaysOnTopHint);msgBox3->setFrameStyle(QFrame::Panel | QFrame::Sunken);msgBox3->show();//QThread::sleep(2);delete msgBox3;
//#define QTIMERMSG3 QDia = new QDialog(this);QDia->setGeometry( QDia->x(), QDia->y(), 400, 33);msgBox3 = new QLabel(ss,QDia,Qt::WindowStaysOnTopHint);msgBox3->setFrameStyle(QFrame::Panel | QFrame::Sunken);QDia->show();QThread::sleep(2);delete QDia;  this sleep fails to fill in the label
//#define QTIMERMSG3 QDia = new QDialog(this);QDia->setWindowTitle("***")/*QDia->setGeometry( 0, 0, ui->scrollArea_4->geometry().x()/2, ((ui->scrollArea_4->geometry().x()/2)*1000)/1618L )*/;msgBox3 = new QLabel(ss,QDia);msgBox3->setFrameStyle(QFrame::Panel | QFrame::Sunken);QDia->show();QDia->raise();QTimer::singleShot( 2000, Qt::CoarseTimer, QDia, SLOT(accept()) );
#define QTIMERMSG3 QDia = new QDialog(this);QDia->setWindowTitle("***");QDia->setFixedSize( ui->scrollArea_4->geometry().x()/2, ((ui->scrollArea_4->geometry().x()/2)*1000)/1618L );msgBox3 = new QLabel(ss,QDia);msgBox3->setFrameStyle(QFrame::Panel | QFrame::Sunken);QDia->show();QDia->raise();QTimer::singleShot( 2000, Qt::CoarseTimer, QDia, SLOT(accept()) );

void MainWindow::dialogSaveAsFilename()
{
        QDir dd;
        QString ss;
        bool ret = false;
        QLabel *msgBox3;
        QDialog *QDia;

        // save the current active one in case somebody wrote something and we must not lose data.
        if( !libraryFilename.isEmpty() )
            libraryWrite( &libraryFilename, libraryDescription );

        // I think you can set the QInputDialog for multiple inputs or else
        //  better: https://doc.qt.io/qt-6/qfiledialog.html
        // I used the standard dialogs Example dialogs.cpp.

        //QString select = tr("Text files (*)");

        QString fileName = QFileDialog::getSaveFileName(this,
                                                        tr("New Estimate File Name"),
                                                        estimateDirectory,
                                                        tr("Estimate Files (*);;Text Files (*.txt)")
                                                        );

        if( !fileName.isEmpty() ) //estimateDirectory is the global name in MainWindow.h
        {
            ss = dd.filePath( fileName );
            linuxMsWindowsSlashFixes(&ss);

            int sepint = ss.lastIndexOf( dd.separator() ); // '\' or '/'
            if( sepint != -1)
                ss.truncate( sepint );
            estimateDirectory = ss;

            ;//qDebugA() << "User selected directory:" << estimateDirectory;
            ;//qDebugA() << "User selected filename:" << fileName;

            dd.mkpath( estimateDirectory );
            dd.setCurrent( estimateDirectory );

            ss = estimateDirectory + "/" + estDotlib;
            if( !QFileInfo::exists(ss))
            {
                dd.mkdir( ss ); // make a est.lib library separate from estimates.
                loadComboboxLibs( ss ); // clears and reloads the list if any is there
                loadByIndexLibrarynameOfComboboxLibs(0);
            }

            if( oneHeader == nullptr )
                oneHeader = new DialogHeaderOfEstimate(this);

            // delete the file if it already exists
            remove( fileName.toStdString().c_str() );
            // create an empty estimate file
            ret = oneHeader->estimateWriteHeader( &fileName );

            // this is the global name in MainWindow.h
            estimateFilename = fileName;

            if( ret == true )
            {
               /* Writing the estimate body might not always
                * be needed but it won't hurt.
                * There is a case where I use this function
                * to do SaveAs where I do want to write the
                * body and not just start a new filename header.
                */
                if( estimateWrite( &estimateFilename ) == true )
               {
                    estimateWasChanged = false;

                    QString ss = SUCCESSFULLYSAVED;
                    QTIMERMSG3
                }
            }
        }

        return;
}

bool MainWindow::dialogSaveFilename_sub( bool fromDeconstructor )
{
    QMessageBox msgBox, msgBox2;
    QLabel *msgBox3;
    QDialog *QDia;
    bool ret = false;

    /*
     * MONITORING ESTIMATE CHANGES for USER to NOT LOSE WORK
     * -----------------------------------------------------
     * The user can type on a blank estimate with
     * estimateFilename.isEmpty()==true
     *
     * clearEst() sets estimateFilename.isEmpty()==true
     * by doing this: estimateFilename="";
     *
     * Next see load2() has editingFinished connect() set
     * as a signal for all QEditLines to call doCalc().
     *
     * editingFinished signal occurs (see QLineEdit documentation)
     * only when a QLineEdit has changes of text contents.
     * Clicking on a QLineEdit is FocusIN and FocusOut
     * but does not emit a signal to call doCalc().
     *
     * So I inserted a variable inside doCalc() to monitor
     * if an estimate has typing done inside it.
     * So ask the user if he wants to save his changes:
     */
    if( estimateFilename.isEmpty() )
    {
        /* Do not ask if blank estimate and not changed.
         * That would be an idiot's question.
         */
        if( estimateWasChanged == true )
        {
            /* ask user whether to save estimate; no request to do so was made.*/

            msgBox.setText("The current estimate has been modified.");
            msgBox.setInformativeText("Do you want to save your changes first?");
            msgBox.setStandardButtons(QMessageBox::Save |
                                      QMessageBox::Discard |
                                      QMessageBox::Cancel );
            msgBox.setDefaultButton(QMessageBox::Save);
            int ans = msgBox.exec();
            switch (ans)
            {
            case QMessageBox::Save:
                dialogSaveAsFilename();
#ifdef OLDandDELETESOMEDAYsoon
                // false means: do not clearEst() before creating
                if( dialogNewFilename_sub( false ) == false )
                {
                    // let function do its own msgs.
                    //msgBox2.setText("Failed to create new estimate.");
                    //msgBox2.exec();
                    ret = false;
                }
                else
                {
                    ret = estimateWrite( &estimateFilename );
                    if( ret == false )
                    {
                        msgBox2.setText("Fixing blank filename failed to save estimate:\n\""+estimateFilename+"\"" );
                        msgBox2.exec();
                    }
                    else
                    {
                        QString ss = SUCCESSFULLYSAVED;
                        QTIMERMSG3
                        estimateWasChanged = false;
                    }
                }
#endif
                break;
            case QMessageBox::Discard:

                estimateWasChanged = false;// signal done
                ret = true;// signal done
                break;
            case QMessageBox::Cancel:

                // do not do this. We do not know. estimateWasChanged = true;  // signal not done
                ret = false;   // signal not done
                break;
            default:
                // should never be reached
                break;
            }
        }
        else
        {
            ret = true;
        }
    }
    else if( estimateFilename.contains("noname") )
    //user pressed Save but noname yet
    {
        ;//qDebugA() << "contains:"<< estimateFilename;

        /* I thought of using SaveAs instead of
         * dialogNewFilename_sub() because it asks
         * the name and does the estimateWrite()
         * whereas NewFilename only asks the name.
         * Maybe I can delete
         * dialogNewFilename_sub() in the msgBox2 above.
         * Think about it later, Mar 7, 2024.
         */
        dialogSaveAsFilename();

#ifdef OLDandDELETESOMEDAYsoon
        // false means: do not clearEst() before creating
        if( dialogNewFilename_sub( false ) == false )
        {
            // let function do its own msgs.
            //msgBox2.setText("Failed to create new estimate.");
            //msgBox2.exec();
            ret = false;
        }
        else
        {
            ;//qDebugA() << "user said:"<< estimateFilename;

            ret = estimateWrite( &estimateFilename );
            if( ret == false )
            {
                msgBox2.setText("Fixing blank filename failed to save estimate:\n\""+estimateFilename+"\"" );
                msgBox2.exec();
            }
            else
            {
                QString ss = SUCCESSFULLYSAVED;
                QTIMERMSG3
                    estimateWasChanged = false;
            }
        }
#endif
    }
    else // we have a filename and so we do New/Open->save to disk ...
    {
        if( estimateWasChanged==false )
        {
            if( fromDeconstructor == false ) //At Quit don't bother user with another popup.
            {

                // then do not save.
                // User probably saved or closed already
                // and that is why estimateWasChanged is NO
                QString ss = SUCCESSFULLYSAVED;
                QTIMERMSG3
            }
            ret = true;  // signal good exit
        }
        else if( estimateWasChanged==true )
        {
            // we call this from New/Open/->Close or Quit/Deconstructor

            // ?? in case QEditLine failed to signal editLineFinished
            // ?? such as backspace fails to signal a change.

            ret = estimateWrite( &estimateFilename );
            if( ret == false )
            {
                msgBox2.setText("Did not save estimate to\n"+estimateFilename );
                msgBox2.exec();
                ;//qDebugA() << "ret=false; est clearEst() not done; closeFilename_sub()";
            }
            else
            {
                estimateWasChanged = false;

                QString ss = SUCCESSFULLYSAVED;
                QTIMERMSG3
            }
        }
    }

    if( !libraryFilename.isEmpty() )
    {
        libraryWrite( &libraryFilename, libraryDescription );
    }

    if( ret == true )
    {
        estimateWasChanged = false;//we saved or don't care so continue as fresh.

        //;//qDebugA() << "ret=true; est clearEst() done and closeFilename_sub()";
    }
    ;//qDebugA() << "dialogSaveFilename_sub, ret=" << ret << "estimateWasChanged=" << estimateWasChanged;

    return(ret);
}

void MainWindow::mainQuit()
{
    writeSettings(QDir::homePath());
    if( dialogSaveFilename_sub( FROMDECONSTRUCTOR ) == true )
        exit(0);

    return;
}



bool MainWindow::libraryRead( QString *filename, QString *setComboPartialText, bool comboBoxTextOnly )
{
        bool ret = false;
        unsigned block, lg;
        char buf[51];
        QString ss, defaultRefcode;

        comboBoxTextOnly=false;
        /* This variable is for speeding up loadComboboxLibs()
         * The program functions well without speedup; ie. read comboBoxTextOnly=false.
         * I permanently set it to =false because there is a bug and I am tired
         * of looking for it. Jan 19, 2024.
         * The bug is the libraryWrite() is overwriting blanks on top of estimates
         * when I tried to turn off libraryRead()-full and do fast-TextOnly.
         * I left the variable comboBoxTextOnly in the code so you can fix the bug
         * some day, just for the sake of speeding up loadComboLibs().
         *
         */

        if( !comboBoxTextOnly )
            clearLib(nullptr);

        if( filename->isEmpty() )
        {
            ;//qDebugA() << "lib read filename: isEmpty()";
            return(ret);
        }
        else
            ;//qDebugA() << "libraryRead filename:" << *filename;

        QFile fp(*filename);

        /* This is data structures from the original 1987 C program.
           I want to get the file read()s and offsets correct so I maintain
           the old files as input for testing in Dec 2022; DavidP.
        */

        ;//qDebugA() << "inside libraryRead(), filename:" << *filename;

#ifdef SKIP
    Library has takeoffs and library item lists inside one subdirectory est.lib/
        The takeoffs are 680 chars fixed size files.
        The ones named 0 or WD or B are the main library side lists,
        which I will parse below. The takeoffs are simple. Skip parsing them.
            fp.read( takeoff, 680);  Done.

        The ones filenamed with a dot: B.B01 are the corresponding
            takeoff-detailed-Qty calculation
            saved as 680 fixed chars, no format. 17 x 40 = 680. 17 chars per line
            where the filename: B is the library list and B01 is the line item
            library code in that B list.

            When an estimate is written, calling up ctrl-Q would use
            one stored on disk with the estimate OR use the library one
            by auto-constructing a filename: est.lib/B.B01
            where B is the active library
            and the estimate line has the B01 typed in.
            and read the ctrl-Q takeoff into use.

            Their purpose is to provide typical Qty arithmetic, if any.
            For example, how many bricks? = wall length x height/40 per square foot.

            Maybe ctrl-Q automatically made the read decision based on
            no historical one in the estimate itself, and any changes
            were stored to the library also. I have to look in the manual
            to see when save/read to/from library was distinguished from save
            to estimate; probably a ctrl-key, since there was no mouse in those days.

          I looked in EST1A.C loderef() The struct for main libraries is simply
          3 char library code and a pointer to 3+" "+24 char description=28.
          However, the first line is 2 chars longer, but all are \n endings.
          fgets() would work. Read to end of line.
          refch[3], like B01, is a repeat of the *refpt, like this:
          B01 This is a brick------->24chars 5forhrs 5forcost2junkchars\n
          refch[3]=B01

          example:
          *0  MASONRY
          B04 block  4i                 1.85  0.16
          B06 block  6i                 2.75  0.00
          B08 Blocks  8i                1.46  0.08
          B10 fnd. blk.10               3.50  0.00
          B12 fnd. blk.                1.900 .0900

        #define RIZE 300

        struct stref {
                char refch[3];  // this is a memory reference not a filename item.
                char *refpt;
                }ref[RIZE];
#endif
        struct libraryold {
            char title[45];
            // first line of file has the "*WD and Bricks Title", eg..
            char libcode[4];  // library 3 char code library line. Never blank.
            char description[25];
            char hours[6];
            char cost[6];
        }lib;

        if( fp.open(QIODeviceBase::ReadOnly) == false)
              return(ret);
        else
        {
            if( fp.readLine( buf, 50) <= 0 )
            {
              ;//qDebugA() <<"lib readLine 0 bytes";
              fp.close();
              return(ret);
            }
            else
                ret = true;
        }
        sprintf( lib.title, "%.44s", buf);
        ss = QString("%1").arg(buf,-44,' ');
        *setComboPartialText = ss.sliced(4,40).trimmed();
        ;//qDebugA() << "\nread est.lib first line for combobox" << *setComboPartialText;
        if( comboBoxTextOnly == true ) // quit early
        {
            clearLib(nullptr);

            fp.close();
            return(ret);
        }

        // we are not putting the title in the libRow array

        /* Let's display the first 2 chars which is the filename and refcode/libcode2char
         * in the cyan pushbutton under the Cost Library drop-down menu.
         */
        ui->libcode2char->setText(" "+ss.sliced(1,2)+" ");

        ss = QString("%1").arg(lib.title,-3,' ');
        ss.truncate(3);
        defaultRefcode = ss.remove('*').trimmed();

        clearLib(defaultRefcode);

        for( block=0,lg=LIBROWCOUNT-1; lg>0; --lg)
        {
            if( fp.readLine(buf,50) <= 0)
            {
                ;//qDebugA() << "End of file is good too.";
                break;
            }
            // libRow starts at 1, not 0, so increment now
            ++block;

            sprintf( lib.libcode, "%.3s", buf );
            sprintf( lib.description, "%.24s", buf+4 );

            sprintf( lib.hours,"%.5s", buf+30 );     // these sizes and offsets
            sprintf( lib.cost, "%.5s", buf+35 );

            ;//qDebugA() << lib.libcode << lib.description << '|' << lg;

            libRow[block]->col[1]->setText( stripEnd(lib.libcode) );
            libRow[block]->col[2]->setText( stripEnd(lib.description) );
            libRow[block]->col[3]->setText( stripEnd(lib.hours) );
            libRow[block]->col[4]->setText( stripEnd(lib.cost) );

            //ss = QString("%1").arg(lib.title,-3,' ');
            //ss.truncate(3);
            libRow[block]->refcode = defaultRefcode;

            /* These are the same for all lines. The first line of the file
             * has this refcode.
             * See my "observation" about design mistake libTOest.cpp
             */
        }

        //vvvvv
        ;//qDebugA() << "\n**********\nRead this many lines:" << block << " plus 1 title line. The End of libraryRead()\n";

        fp.close();

        return(ret);
}
    // I wrote this so I can use all the buffer to lineEdits sprintf
    // statements when I do the reverse copy of the lineedits to buffer.
    // This way I can have fewer typing mistakes and if I change the
    // read function then I can copy and paste into the write function
    // and only change the sprintf() to sprintfreverse();
    void sprintfreverse( QLineEdit *arg1, char *arg2, char *arg3 )
    {
        sprintf( arg3, arg2, arg1->text().toStdString().c_str() );
    }


#ifdef EXTRAS
std::string MainWindow::openFile()
{
    //https://doc.qt.io/qt-6/qfiledialog.html

    QString filename = QFileDialog::getOpenFileName(
                nullptr,
                QObject::tr("Open Document"),
                QDir::currentPath(),
                QObject::tr("Document files (*.doc *.rtf);;All files (*.*)"));
    /*
    QString dir = QFileDialog::getExistingDirectory(this, tr("Open Directory"),
                                                    "/home",
                                                    QFileDialog::ShowDirsOnly
                                                    | QFileDialog::DontResolveSymlinks);
     */

    return filename.toStdString();
}
#endif

char * MainWindow::stripEnd( char *ee )
{
    char *bb = ee;

    while( *ee != '\0') // find end of string
        ++ee;

    // if empty string was sent to us, do nothing

    // strip spaces from end, quitting at not-a-space or quit at beginning

        for( --ee; ee>=bb && *ee == ' ' ; --ee )
            *ee = '\0';

    return(bb);
}


/************* Below are the write to disk functions **************/

// I wrote this so I can use all the buffer to lineEdits sprintf
// statements when I do the reverse copy of the lineedits to buffer.
// This way I can have fewer typing mistakes and if I change the
// read function then I can copy and paste into the write function
// and only change the sprintf() to sprintfreverse();
void MainWindow::sprintfreverse( QLineEdit *arg1, char *arg2, char *arg3 )
{
    if( arg1 != nullptr )
        sprintf( arg3, arg2, arg1->text().toStdString().c_str() );
    //else
        //sprintf( arg3, arg2, "ZZ" );

}



bool MainWindow::libraryWrite( QString *filename, QString comboBoxPartialText )
{
    bool ret = false;
    unsigned lg, lgtotal, xx,yy;
    char *buf, buffer[51];
    buf = buffer;

    QString qfnp, ss;

    if( filename->isEmpty() )
    {
        ;//qDebugA() << "lib write to filename: isEmpty()";
        return(ret);
    }
    else if( filename->contains('*') || filename->contains('?'))
    {
        QString msg = "Library names with * or ? are disallowed:\n";
        ;//qDebugA() << msg << *filename;
        QMessageBox msgbox;
        msgbox.setText( msg + *filename );
        msgbox.exec();

        return(ret);
    }
    QFile fp(*filename);

    /* This is data structures from the original 1987 C program.
       I want to get the file read()s and offsets correct so I maintain
       the old files as input for testing in Dec 2022; DavidP.
    */

    ;//qDebugA() << "libraryWrite() saving to filename:" << *filename;

#ifdef SKIP
Library has takeoffs and library item lists inside one subdirectory est.lib/
    The takeoffs are 680 chars fixed size files.
    The ones named 0 or WD or B are the main library side lists,
    which I will parse below. The takeoffs are simple. Skip parsing them.
        fp.read( takeoff, 680);  Done.

    The ones filenamed with a dot: B.B01 are the corresponding
        takeoff-detailed-Qty calculation
        saved as 680 fixed chars, no format. 17 x 40 = 680. 17 chars per line
        where the filename: B is the library list and B01 is the line item
        library code in that B list.

        When an estimate is written, calling up ctrl-Q would use
        one stored on disk with the estimate OR use the library one
        by auto-constructing a filename: est.lib/B.B01
        where B is the active library
        and the estimate line has the B01 typed in.
        and read the ctrl-Q takeoff into use.

        Their purpose is to provide typical Qty arithmetic, if any.
        For example, how many bricks? = wall length x height/40 per square foot.

        Maybe ctrl-Q automatically made the read decision based on
        no historical one in the estimate itself, and any changes
        were stored to the library also. I have to look in the manual
        to see when save/read to/from library was distinguished from save
        to estimate; probably a ctrl-key, since there was no mouse in those days.

      I looked in EST1A.C loderef() The struct for main libraries is simply
      3 char library code and a pointer to 3+" "+24 char description=28.
      However, the first line is 2 chars longer, but all are \n endings.
      fgets() would work. Read to end of line.
      refch[3], like B01, is a repeat of the *refpt, like this:
      B01 This is a brick------->24chars 5forhrs 5forcost2junkchars\n
      refch[3]=B01

      example:
      *0  MASONRY
      B04 block  4i                 1.85  0.16
      B06 block  6i                 2.75  0.00
      B08 Blocks  8i                1.46  0.08
      B10 fnd. blk.10               3.50  0.00
      B12 fnd. blk.                1.900 .0900

    #define RIZE 300

    struct stref {
            char refch[3];  // refcode is a memory reference not a filename item.
            char *refpt;
            }ref[RIZE];
#endif
    struct libraryold {
        QLineEdit *title;  // first line of file has the "*WD and Bricks Title", eg..
        QLineEdit *libcode;  // library 3 char code library line. Never blank.
        QLineEdit *description;
        QLineEdit *hours;
        QLineEdit *cost;
    }lib;

    if( fp.open(QIODeviceBase::WriteOnly) == false)
          return(false);
    else
    {
        qfnp = *filename;
        qfnp = stripFilename( &qfnp );

        // the old format needs "*WD description",
        // where * means code for 2 digit lib list name.
        // The new format will allow 2 char est.lib names (libcode2char is what I am newly saying)
        // and no * because we're using a mouse and don't need keyboard hotkeys.
        // The description is stored in the first line of the file still:
        ss = QString("*%1 %2\n").arg(qfnp,-2,' ').arg(comboBoxPartialText,-42,' ');
        ss.truncate(47);
        strncpy( buf, ss.toStdString().c_str(),48 );
  ;//qDebugA() << ss << "<-write library first line";
        /*sprintf( buf, (char *)"*% -2.2s % -42.42s\n",
                            qfnp.toStdString().c_str(),
                            comboBoxPartialText.toStdString().c_str() ); */
        for( yy=0; yy<42; ++yy)
            if( buf[yy] == '\0' )
                buf[yy] = ' ';
        buf[42]='\r';
        buf[43]='\n';
        ;//qDebugA() << buf;

        if( fp.write( buf, 44) < 44 )
        {
          ;//qDebugA() <<"Error: Cannot save.";
          fp.close();
          return(false);
        }
    }

    // find the last line of the library
    for( lg=LIBROWCOUNT; --lg>0; )
    {
        /* refcode is stored as the first 2 chars of the filename
         * and is not saved into the file body. This means
         * refcode should not have been in the libRow[row].refcode
         * struct, but it does not hurt anything; minor design error, 2023.
         */
        lib.libcode = libRow[lg]->col[1];
        lib.description = libRow[lg]->col[2];
        lib.hours = libRow[lg]->col[3];
        lib.cost = libRow[lg]->col[4];

        if( lib.libcode->text().isEmpty() &&
            lib.description->text().isEmpty() &&
            lib.hours->text().isEmpty() &&
            lib.cost->text().isEmpty()
                )
            continue;
        else
            break;
    }
    lgtotal = lg;
    ;//qDebugA() << "We are going to save to lib this many lines:" << lgtotal;
//zzzzz
    // libRow starts at 1, not 0. 0 is left out.
    for( xx=lgtotal, lg=1; xx>0; --xx, ++lg)
    {
        //sprintf( buf, "% 43s", "" ); // warning undefined behaviour
        memset( buf,' ', 43); // fill with spaces.
        buf[43] = '\0';

        qleToUpper( libRow[lg]->col[1] );

        lib.libcode = libRow[lg]->col[1];
        lib.description = libRow[lg]->col[2];
        lib.hours = libRow[lg]->col[3];
        lib.cost = libRow[lg]->col[4];

        sprintfreverse( lib.libcode, (char *)"%.3s", buf );
        sprintfreverse( lib.description, (char *)"%.24s", buf+4 );

        sprintfreverse( lib.hours,(char *)"%5.5s", buf+30 );
        sprintfreverse( lib.cost, (char *)"%5.5s", buf+35 );

        //;//qDebugA() << lib.libcode->text() << lib.description->text() << '|';

        for( yy=0; yy<42; ++yy)
            if( buf[yy] == '\0' )
                buf[yy] = ' ';
        buf[42]='\r';
        buf[43]='\n';

        if( fp.write(buf,44) < 44)
        {
            ;//qDebugA() << "failed write lib.";
            break;
        }
    }

    //libRow starts at 1, so the count ends 1 too high by for(++lg)
    // so:
    --lg;
    ;//qDebugA() << "\n**********\nWrote:" << lg << "library lines, plus 1 title";

    if( lg == lgtotal)
           ret = true;
    fp.close();

    return(ret);
}

    void MainWindow::qleToUpper( QLineEdit *qle )
    {
        QString ss;
        qsizetype xx;

        ss = qle->text();
        for( xx=ss.size(); xx>0; --xx )
            ss.data()[xx-1] = ss.data()[xx-1].toUpper();

        qle->setText(ss);

        return;
    }

    float MainWindow::tediously_meticulously1987compatibility( char *str, qtymath *qtytemp )
    {
        float ret = 0.0;
        double output;
        char *startAt;

        startAt = qtytemp->noErrorHelper( str );

        if( qtytemp->arith( startAt, &output, false ) == true )
            ret = (float)output;

        return(ret);
    }
