
/*  This file is part of PaceMaker Estimator 2.0.
    PaceMaker Estimator 2.0 is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
    PaceMaker Estimator 2.0 is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with PaceMaker Estimator 2.0. If not, see <https://www.gnu.org/licenses/>.
*/

#include "mainwindow.h"
#include "qtimer.h"
#include "totalsubtotal.h"
#include <QDialog>

static bool programaticSignalsActive2 = false;
double prefHoursMultiplier=32.0;
double prefCostsMultiplier=1.15;

/*This one happens only when Cost Libraries roll-down menu changed.
 *See makeNewLibrary() for when user inside Cost Library title does a <enter> EditingFinished.
 */
void MainWindow::on_comboBox_currentIndexChanged(int index)
{
    int hh,ww;
    QLineEdit *qle;
    QString combotext;
    //QFileInfo nm;
    QMessageBox msgBox;
    QString estlib = estimateDirectory + "/" + estDotlib;
    QString *sspt, ss;
;//qDebugA() << "firsty";
    /* I discovered program actions trigger this slot as well
     * as user clicks. So I am filtering them, like I
     * learned to do in qtymath combobox.
     *
     * However there is something wrong with typing a new library
     * name to create a new library. It ??? the current and new
     * libraries. Mar 10, 2023. Edit: I am adding a signal via
     * connect() QLineEdit() of the combobox's editline and its
     * EditingFinished.
     */
    if( programaticSignalsActive2 == true )
        return;

    //QPoint p = MainWindow::mapFromGlobal(QCursor::pos());
    //;//qDebugA() << "This Cost Library needs to be loaded:" << index << p.x() << p.y();

    //filename = QFileDialog::getOpenFileName( this,
    //    tr("Cost Library"), "/home/david/msdos/PaceMakerEstimator/est.lib", tr("Cost Library (* *.jpg *.bmp)"));

    //filename is now the QVariant, first two chars of comboBox text.
    QVariant vv = ui->comboBox->itemData( index, Qt::UserRole );
    sspt = (QString *)vv.data();

    if( sspt->isEmpty() == true )
    {
;//qDebugA() << "qvariant blank on indexChngd means its newname:" << vv;
        /*must have been typed and user pressed enter and it's new.
         * So let makenewlibrary() handle it when its connect()
         * calls it upon editingFinished.
         */

    }
    else //if( sspt->isEmpty() == false )
    {
        //sspt = (QString *)vv.data();

        if( libraryDescription.isEmpty() == true )
            libraryDescription = " ";

        /* save any data changes for the current library
         * before clearing it. */
        if( !libraryFilename.isEmpty() )
            libraryWrite( &libraryFilename, libraryDescription );

        libraryFilename = estlib  + "/" + *sspt;
  ;//qDebugA() << "!!!!!!!!!!" << libraryFilename;
        if( libraryRead( &libraryFilename, &combotext, false ) == true )
        {
            if( combotext.isEmpty() == false )
                libraryDescription = combotext; // used for storing later
            else
                libraryDescription = " ";
            ss = sspt->leftJustified(2,' ',true);
            ui->comboBox->setItemText( index,
                                   ss+cleanQString(combotext,14) );
            ;//qDebugA() <<"Success reading the lib file";

            // make at least one line visible,which can be invisible
            // because the lib side is scrolled down too far.

            qle = libRow[1]->col[1];
            hh = qle->frameGeometry().height();
            ww = qle->frameGeometry().width();
            ui->scrollArea_2->ensureWidgetVisible( qle, ww, hh);
        }
        else
        {
            /* I moved this to makeNewLibrary() because it
             * is triggered by its own connect() instead
             * of this function which is triggered
             * too often, on all index changes.
             */
#ifdef TESTRTRTRTRT
            // try to make the new library
            // if( failure was due to not exist )
            if( nm.exists(libraryFilename) == false )
            {
                if( libraryFilename.contains('.') == true)
                // skip names like B.BO1 and /..
                {
                    msgBox.setText("Choose a name without a '.' dot.");
                    msgBox.exec();
     unused         loadComboboxLibs( estlib ); // clears and reloads the list
                }
                else
                {
                    QFile fp(libraryFilename);
                    if( fp.open(QIODevice::WriteOnly) )
                    {
                        fp.close();
     unused             loadComboboxLibs();not needed because comboBox does itself:
                    }
                    else
                        ;//qDebugA() <<"Error creating new library" << libraryFilename;
                }
            }
            else
                ;//qDebugA() <<"lib file exists but libraryRead() failed. Maybe no lines in it." << libraryFilename;
#endif
        }
    }

    return;
}

#ifdef EVERYKEYSTROKESONOTGOODSEECONNECT
void MainWindow::on_comboBox_editTextChanged(const QString &arg1text )
{
    if( programaticSignalsActive2 == true ) // only let user changes leak through
        return;

    QString estlib = estimateDirectory + "/" + estDotlib + "/";

    /* Clear the lib side and read the filename in, if it already exists.
     * That is the same action as simply clicking on it in the list.
     *
     * If the name is new then create a blank library filename,
     * and reload the library combo box, or else we will lose
     * the name that got over-written by the user.
     * Simply calling loadComboboxLibs() should do.
     */
    QString aa = arg1text;

    QFile fp( estlib + aa );

    if( fp.open(QIODevice::ReadOnly) )
    {
        ;
    }
    else if( fp.open(QIODevice::WriteOnly) )
    {
        ;
    }
    fp.close();
   // on_comboBox_currentIndexChanged(0);

   // loadComboboxLibs( estlib ); // clears and reloads the list

    return;
}
#endif

bool MainWindow::loadComboboxLibs( QString estlib )
{
    QDir dir;
    QStringList nameFilters;
    int xx, ii;
    QString nm, ss;
    QFileInfoList list;
    int saveLastComboBoxIndex;

    programaticSignalsActive2 = true;

    ;//qDebugA() << "estlib loadComboboxLibs:" << estlib;

    dir.setPath( estlib );

    if( dir.exists() == true )
    {
        nameFilters << "*";  // "*","*.";
        dir.setNameFilters( nameFilters);
        list = dir.entryInfoList( nameFilters, QDir::Files, QDir::Name ); //sort by name

        saveLastComboBoxIndex = ui->comboBox->currentIndex();

        ui->comboBox->clear(); // delete all, if any

        //if( list.size() == 0)
            ;//qDebugA() << "0 list of libraries in" << estlib;

        for ( xx = 0; xx < list.size(); ++xx )
        {
            QFileInfo fileInfo = list.at(xx);
            nm = fileInfo.fileName();
            ;//qDebugA() << "\nfileName() returned:" << nm ;

            // remove the prefix directory path (Feb15,2024 I think this is unnecessary)
            ss = dir.relativeFilePath( nm );
            ss=nm;
            if( ss.contains('.') == false
                && ss.contains('?') == false
                && ss.contains('/') == false
                && ss.contains('\\') == false
                && ss.contains('*') == false )
            // skip names like B.BO1 and /.. and * and ?? and ?o
            {
                //;//qDebugA() << "before ssTrunc";

                // 2 char filename is in the QVariant of each comboBox
                QString ssTrunc = ss.leftJustified(2,' ',true).trimmed().toUpper();
                //;//qDebugA() << "after ssTrunc 1";
                QVariant ssV( ssTrunc );
                //;//qDebugA() << "after ssTrunc 2";
                int id;
                ss = ssTrunc.leftJustified(2,' ',true);
                //;//qDebugA() << "after ssTrunc 3" << ss;
                id=ui->comboBox->findText( ss, Qt::MatchStartsWith );
                if( id >= 0 )
                {
                    QString combotext;
                    libraryFilename = estlib + "/" + ssTrunc;
                ;//qDebugA() << "QString filename = estlib / ssTrunc;" << libraryFilename;

                   /* Regarding this code coming up: ((xx+1)<list.size()(?true:false))
                    * We have to load the library as more than comboBoxOnly==true
                    * for the last one because we have to display it or libraryWrite it properly.
                    * We do not know what happens next. So no short cut on the final libraryRead().
                    * The bug I saw was writing a clearLib() library to ZB, the last one,
                    * because I did not do this libraryRead() before Jan 19, 2024.
                    */
                    //if( libraryRead( &libraryFilename, &combotext, true ) == true )
                    if( libraryRead( &libraryFilename, &combotext, ((xx+1)<list.size())?true:false) == true )
                        ui->comboBox->setItemText( id,
                                   ss+cleanQString(combotext,14) );
                    else // Does this ever happen?
                        ui->comboBox->setItemText( id, ss+"  " );
                }
                else
                {
        /* Regarding this code coming up:
         * ((xx+1)<list.size())?true:false))
         * We have to load the library as more complex
         * than (the third argument) comboBoxTextOnly==true
         * The last one must be false because we have to display it or
         * libraryWrite it properly. We do not know what happens next.
         * So no short cut on the final libraryRead(). The bug I saw
         * was writing a clearLib() library to ZB, the last in the list.
         * It happened because I did not do this libraryRead()
         * before Jan 19, 2024.
         *
         */
                    QString combotext;
                    libraryFilename = estlib + "/" + ssTrunc;
                    ;//qDebugA() << "libraryRead():" <<libraryFilename << "and  ((xx+1)<list.size())?true:false) == true is:" <<  (xx+1) << "<" << list.size();

                    if( libraryRead( &libraryFilename, &combotext, ((xx+1)<list.size())?true:false) == true )
                    {
                        ;//qDebugA() << "------>|"<<ss<<"|";
                        ui->comboBox->addItem(
                             ss+cleanQString(combotext,14), ssV );
                    }
                    else
                    {
                        ;//qDebugA() << "crashes 0 item 1 Feb15, 2024";
                        ui->comboBox->addItem( ss+"  ", ssV ); // new item
                    }
                }

                /* fetch leftJustified/truncated2char ss is thusly
                 * QVariant vv = ui->comboBox->currentData( Qt::UserRole );
                 * (QString ) vv.data();
                 * for when we libraryRead.
                 * This means WD40 overwrites WD and overwrites WDoctor.
                 * User has to change first two chars to get a different
                 * filename. This is the 2 char refcode.
                 */
    //;//qDebugA() << "after ssTrunc 4";

            }
        }

        ui->comboBox->setCurrentIndex( saveLastComboBoxIndex );

#ifdef WORKSBUTNOSORTINGbyname
        //QDirIterator it(dir);
        for( xx=0; it.hasNext(); ++xx)
        {
            nm = it.next();

            ss = dir.relativeFilePath( nm ); // remove the prefix directory path
            if( ss.contains('.') == false) // skip names like B.BO1 and /..
            {
                ui->comboBox->addItem( ss );
            }
        }
#endif
    }
    else
    {
        // should never happen
        //qDebug() << "comboBox directory does not exist" << estlib;
    }

    programaticSignalsActive2 = false;

    return(true);
}

static QLineEdit *comboboxLibLineEdit = nullptr;

/*This function changes the external combotext variable to what is inside file
 * and sets the current combobox to the idx 0 through 10+.
 * 2char_filename+description (a space between might be nice.Apr18)
 */
void MainWindow::loadByIndexLibrarynameOfComboboxLibs(int idx)
{
    QString combotext;
    QString estlib = estimateDirectory + "/" + estDotlib;
    QString *ss;

    if( idx < 0 )
        return;

    // This is the filename:
    QVariant vv = ui->comboBox->itemData( idx, Qt::UserRole );
    ss = (QString *)vv.data();


    if( ss->isEmpty() == false ) // if good filename
    {
        ;//qDebugA() << "qvariant:" << vv;

        /* save any data changes for the current library
         * before clearing it. I got this from on_comboBox_currentIndexChanged(int index).
         * I added this Jan13, 2024. I was losing data
         * when I called this loadByIndexLibrarynameOfComboboxLibs() function to change libraries.
         */
        if( !libraryFilename.isEmpty() )
            libraryWrite( &libraryFilename, libraryDescription );

        libraryFilename = estlib  + "/" + *ss;
        if( libraryRead( &libraryFilename, &combotext, false ) == true )
        {
            if( combotext.isEmpty() == false )
                libraryDescription = combotext; // used for storing later
            else
                libraryDescription = " ";

            QString sspad = ss->leftJustified(2,' ',true);

            // save the filename and description into the idx Text
            ui->comboBox->setItemText(idx,
                         cleanQString(sspad+libraryDescription,2+14));
        }
        ui->comboBox->setCurrentIndex(idx);

        /* Clear highlighted library line if any, since this is a new library.
         * verticalSlider.cpp does this for lib and est sides. FYI.
         *
         * I added this Jan 12, 2024 because after switching to a new library
         * which had 5 items, then the old 343 line library had me at line 233
         * for example, and that is not an appropriate place to add
         * an estimate line to a 5 item library when pressing the Lib<Est button.
         *
        eraseHighlights( &keyboardnow.litLibButton, 0)
        keyboardnow.litLibButton = 0L;
        keyboardnow.currentLibbuttonrow = 0L;
        keyboardnow.rowL = 0L;
        keyboardnow.colL = 0L;
        show_linenums_showHighlights( &keyboardnow.litLibButton, &keyboardnow.litEstButton, keyboardnow.row, keyboardnow.row );
        */
    }
    else
    {
       ;//qDebugA()<<"qvariant_sspt failed with no filename.idx:"<<idx;
    }

/* mainwindow.cpp calls this function early with func(idx=0)
 * and sets this connect() up before the user does anything.
 */
#ifndef NEW_IDEA
    /* We need this EditingFinished signal because the textChanged
     * slot wasn't good because it is signaling on every keystroke.
     *
     * Set up the connect() once because it persists until user quits; ie. thus nullptr test
     */
    if( comboboxLibLineEdit == nullptr )
    {
        comboboxLibLineEdit = ui->comboBox->lineEdit();
        if( comboboxLibLineEdit != nullptr )
            connect( comboboxLibLineEdit,
                    &QLineEdit::editingFinished,
                    this,
                    &MainWindow::makeNewLibrary );
        else
        {
            ;//qDebugA() << "lib combobox is not giving us its QLineEdit()";
        }

   /* Allowing adds to the combobox list will trigger
    * index changed signals which we must handle differently
    * for new library names. I was seeing old names'
    * content erased. (I think fixed that bug).
    *
    * Does this add trigger makeNewLibrary() upon editfinished,
    * before or after index change function gets called?
    */
    //ui->comboBox->setInsertPolicy( QComboBox::NoInsert );
    //ui->comboBox->setInsertPolicy( QComboBox::InsertAtTop );
      ui->comboBox->setInsertPolicy( QComboBox::InsertAlphabetically );
// See makeNewLibrary() for special creation of new blank libraries.
    }
#endif

    return;
}


/* All of this is more complex than you might expect because I am using a
 * comboBox widget as a menu selection and a list editor for the user to edit lists
 * and I am changing the comboBox functionality so it erases duplicates after the fact.
 * I cannot remember the reason why I had to leave duplicates turned on as a comboBox feature
 * although I do not want dups. I decided on one libraryFilename with two chars since
 * that is the structure of operating filesystems.
 *
 * I remember something: The names are 2+14 chars
 * but the old EstimateMaker had only 2chars as the library code. That means a combobox with
 * a setting of NO DUPLICATES would still have ZZapple and ZZpear as unique, no duplicates,
 * whereas from my point of view the two ZZ starting library codes are called a duplicate
 * and I have to eliminate them myself. I probably do have comboBox set to NO DUPLICATES
 * and it does not work for my situation!
 *
 * I should have done my comboBox with two fields: I took a wrong fork in the road.
 * I remember thinking about using a more versatile widget but the other widget looked
 * more cumbersome.
 *
 * Two fields could have been: ZZ and Description, like apple and pear, because the programming
 * for this simple comboBox widget is longer after all,
 * and might not seem reasonable. It is reasonable. Modularized functions are doing
 * reloads from disk directory to eliminate duplicates and then resetting the index so
 * the user sees the right library list; the one he expects to see. This design strategy
 * is everywhere in the library functions, not just this makeNewLibrary(). Keep it in mind
 * for easy understanding of the coding. I put notes.
 */
void MainWindow::makeNewLibrary()
{
    int hh,ww,index;
    QLineEdit *qle;
    QString combotext;
    QFileInfo nm;
    QMessageBox msgBox;

    QString estlib = estimateDirectory + "/" + estDotlib;
    QString *sspt, ss, ssOriginal;
    int id; //saveLastComboBoxIndex;

    /* January 13, 2024: I think there might be a bug here. What if the user modified the items
     * in a library. Do those items get saved into the library before starting this new library?
     * I thought of this because I had a bug where Quit sometimes lost the library items.
     * The user typed a new estimate name and the loadComboboxLibs() was called before saving
     * the data and it happened to have changed the active library, thus data lost!
     * I fixed the bug by doing this in some, but not all places, before loadComboboxLibs():
     *     // save the current active one in case somebody wrote something and we must not lose data.
    if( !libraryFilename.isEmpty() )
        libraryWrite( &libraryFilename, libraryDescription );

     * If no new estimate was created then the old library name got saved because no reload of libraries
     * happened, which was a better outcome, but it made it harder to find the intermittent bug.
     *
     * I will go away and see where makeNewLibrary() is called.
     * Edit: It is in this file as a connect() signal and nowhere else.
     * That means the user can type library items and change the library code for the list
     * and yes it was a bug such that the below code was reloading the library with the same
     * or different library code and discarded the items! I tried it. It was broken.
     * Items were lost, which were for the old library list, and the new list is empty.
     *
     * Also: The user sees the items and sees his title with the library code he typed
     * so we should say the 2char library code is new and ask if he wants a new blank library
     * or start with the current items which he sees, which are from the old 2char library code.
     * Maybe he wants two libraries with the same content, like a backup.
     */

    // New Jan 13, 2024 saves the current data items to the old 2char libcode
    // since the old libraryFilename is untouched this far into this function
    // and also ask user a 3-way question at msgBox().
    if( !libraryFilename.isEmpty() )
    {
        libraryWrite( &libraryFilename, libraryDescription );

        /* Save the original libraryFilename first 2char library code
         * in case user chooses Discard and then we can put the
         * original library back up on their library side screen
         * like nothing happened from their typing error.
         */
        // ss=filename[2]+libraryDescription[14]
        ssOriginal = libraryFilename;
        stripFilename( &ssOriginal ); // look at the end, not the subdirs
    }
    else
        ssOriginal = "  ";

    //;//qDebugA() << "second";

//;//qDebugA()<< "User triggered new library-make by ENTER or mouse cursor exit";
//;//qDebugA()<< "triggering EditingFinished in libraries comboBox";
//;//qDebugA()<< "Do library creation if first 2 chars are unique filename.";

    ss = cleanQString( ui->comboBox->currentText(),2+14);
    ;//qDebugA() << "new library is created by user." << ss;
    if( ss.size() < 3 )
      ss.resize(3,' ');
    libraryDescription = ss.sliced(2); // ss=filename[2]+libraryDescription[14]
    // 2 char filename is in the QVariant of each comboBox
    QString ssTrunc = ss.leftJustified(2,' ',true).trimmed().toUpper();

    libraryFilename = estlib  + "/" + ssTrunc;

    if( nm.exists(libraryFilename) == false )
    {
      index = ui->comboBox->currentIndex();
      if( index >= 0 )
      {
        //zzzz
        QVariant ssV( ssTrunc );

        ui->comboBox->setItemData( index, ssV );
        ui->comboBox->setItemText( index, ss );

        // try to make the new library
        // if( failure was due to not exist )
        QString str = libraryFilename;
        stripFilename( &str ); // look at the end, not the subdirs
        if( str.contains('.') == true)
            // skip names like B.BO1 and /..
        {
            msgBox.setText("Choose a name without a '.' dot:"+libraryFilename);
            msgBox.exec();
            loadComboboxLibs( estlib );
            // clears and reloads the combo list

            // Next try to load the specific 2char filename if it happens to exist
            // but I do not care if this fails. No new library is going to be created
            // because the filename is disallowed with the dot. User knows something failed.

            str=str.leftJustified(2,' ',true);
            id=ui->comboBox->findText(str, Qt::MatchStartsWith );
            ;//qDebugA() << id<<"<-id makeNewLibrary() str is:" << str;

            loadByIndexLibrarynameOfComboboxLibs( (id>=0)?id:0 );
        }
        else
        {
           /* if new filename then let's ask user
            * if he wants a new blank library for the given 2char code or save the contents
            * to the new code. This is new Jan 13, 2024. Previously the dialog question was not asked
            * and you see the coding creates a new filename as long as there is no dot and then
            * opens and closes on the new filename and puts nothing in it.
            * New Jan 13, 2024 is: we ask the question and if yes, then we are going to
            * do a new libraryWrite().
            */
            msgBox.setText("You typed a new 2char library code->"+ssTrunc);
            msgBox.setInformativeText("Save All library items OR\nSave library as BLANK.");

            msgBox.setStandardButtons(QMessageBox::SaveAll | QMessageBox::Save | QMessageBox::Discard);
            msgBox.setDefaultButton(QMessageBox::SaveAll);

            int ans = msgBox.exec();

            QFile fp(libraryFilename);

            switch (ans)
            {
            case QMessageBox::Save:  //Save as BLANK

                /* inferior method because the Description is not set inside the fp file
                 * so do clearLib() and use libraryWrite() instead.
                 */
                //if( fp.open(QIODevice::WriteOnly) )
                //{
                  //  fp.close();
                  //  /* loadComboboxLibs();not needed because comboBox already going.
                  //   * User is typing in it
                  //   * to trigger this makeNewLibrary().
                  //   */
                //}
                //else
                //{
                    //msgBox.setText("The new library failed to be created:"+libraryFilename);
                    //msgBox.exec();
                //}

                clearLib(ssTrunc);
                if( !libraryFilename.isEmpty() )
                    libraryWrite( &libraryFilename, libraryDescription );

                loadComboboxLibs( estlib ); // clears and reloads list

                id=ui->comboBox->findText(ssTrunc, Qt::MatchStartsWith );

                loadByIndexLibrarynameOfComboboxLibs( (id>=0)?id:0 );
                break;

            case QMessageBox::SaveAll:

                /* We are creating a new library
                 * which is not empty, by using the current library buffer.
                 */

                if( !libraryFilename.isEmpty() )
                    libraryWrite( &libraryFilename, libraryDescription );

                loadComboboxLibs( estlib ); // clears and reloads list
                // the last one is not it, so do this next:

                id=ui->comboBox->findText(ssTrunc, Qt::MatchStartsWith );

                loadByIndexLibrarynameOfComboboxLibs( (id>=0)?id:0 );
                break;

            default: // Discard
                fp.remove();
                // .. and then loadComboboxLibs() will Discard it from list too
                // since it is not on the disk drive.
                loadComboboxLibs( estlib ); // clears and reloads list

                //;//qDebugA() << "\n" << ssOriginal << "<-ssOriginal";
                id=ui->comboBox->findText(ssOriginal, Qt::MatchStartsWith );

                loadByIndexLibrarynameOfComboboxLibs( (id>=0)?id:0 );
                break;
            }
           /* end of Jan 13, 2024 addition */
        }
      }
    }
    else if( libraryRead( &libraryFilename, &combotext, false ) == true )
    {
    /* Why would a new name be readable you ask?
     * Multiple comboBoxText can share one two-letter filename.
     * If so, we must erase the multiple names, though.
     * The comboBox creates the multiple names when the user types.
     * (I cannot control how simple comboBox works).
     *
     * I will overwrite filename readable ones and create
     * new ones and after that I will erase the entire combox
     * contents and reload it from the hard storage subdirectory
     * thus cleaning up multiple names for one filename because the subdirectory
     * is different behaviour than comboBox in that subdirectories disallow duplicate names!
     */
            //Ignore comboxtext which we just loaded.
            //Over-write it with new libraryDescription user typed.
            libraryWrite( &libraryFilename, ss.sliced(2) );

            // The snag is if libraryRead() succeeded then we
            // have two combo items pointing
            // at one filename, which thankfully now
            // has the new libraryDescription.

            // Ideally we should remove the old comboBox item.
            // We could search all the items for same
            // filename. But below is easier and is less error prone.
            // It removes duplicates by reading freshly from the
            // directory which is cleaner than the comboBox.

            loadComboboxLibs( estlib ); // clears and reloads list
            // the last one is not it, so do this next:

            int id;
            ss=ss.leftJustified(2,' ',true);
            id=ui->comboBox->findText(ss, Qt::MatchStartsWith );
            ;//qDebugA() << id<<"<-id makeNewLibrary() ss is:" << ss;

            loadByIndexLibrarynameOfComboboxLibs( (id>=0)?id:0 );

            /* This method does work too?
             *
         int xx;
         for( xx=0; xx < ui->comboBox->count(); ++xx)
         {
             vv = ui->comboBox->itemData( xx, Qt::UserRole );
             sspt = (QString *)vv.data();
             if( *sspt == libraryFilename )
                 break;
         }
       */
    }
    /* skip this
    else
      ;//qDebugA()<<"lib file exists but libraryRead() failed."
          <<"makeNewLibrary() did nothing."
              << libraryFilename;
     */

    return;
}

void MainWindow::on_actionLibrariesDelete_triggered()
{
    bool ok;
    QStringList items;
    int xx;
    QString estlib = estimateDirectory + "/" + estDotlib;
    QDir dir( estlib );

    // save the current active one in case somebody wrote something and we must not lose data.
    if( !libraryFilename.isEmpty() )
        libraryWrite( &libraryFilename, libraryDescription );

    for( xx=0; xx < ui->comboBox->count(); ++xx)
         items << ui->comboBox->itemText(xx);

    QInputDialog qiDia;
    qiDia.setFont(globalFont2);
    qiDia.setGeometry( ui->scrollArea_3->geometry() );
    qiDia.adjustSize();
    QString item = qiDia.getItem( ui->scrollArea_3, tr("Choose"),
                                          tr(".  Library to Archive:  ."), items, 0, false, &ok);
    if (ok && !item.isEmpty())
    {
         if( item.size() < 2 )
            item.resize(2,' ');
         QFile file( estlib + "/" + item.sliced(0,2).trimmed() );
        QString ss;
        ss = estlib + "/archive/"+ item.sliced(0,2).trimmed();

        dir.mkpath( estlib + "/archive" );
        file.rename( ss );
        ;//qDebugA() << item.sliced(0,2).trimmed() << "moved to archive" << ss;

        loadComboboxLibs( estlib ); // clears and reloads the list
        loadByIndexLibrarynameOfComboboxLibs( 0 );
    }

    return;
}

void MainWindow::on_actionLibrariesRecover_triggered()
{
    bool ok;
    QStringList items;
    int xx, one;
    QString nm, ss;
    QString estlib = estimateDirectory + "/" + estDotlib;
    QDir dir( estlib + "/archive" ), dd;

    QDirIterator it(dir);

    // save the current active one in case somebody wrote something and we must not lose data.
    if( !libraryFilename.isEmpty() )
        libraryWrite( &libraryFilename, libraryDescription );

    for( xx=0, one=0; it.hasNext(); ++xx)
    {
        nm = it.next();

        ss = dd.filePath( nm );

        ;//qDebugA() << "on_actionLibrariesRecover dd.filePath(nm)" << ss;

        stripFilename( &ss );

        if( ss.contains('.') == false
            && ss.contains('?') == false
            && ss.contains('/') == false
            && ss.contains('\\') == false
            && ss.contains('*') == false)
        // skip names like B.BO1 and /..
        {
            ++one;
            items << ss;
        }
    }
    if( !one )
        items << "none";

    QInputDialog qiDia;

    qiDia.setFont(globalFont2);
    qiDia.setGeometry( ui->scrollArea_3->geometry() );
    qiDia.adjustSize();

    QString item = qiDia.getItem( ui->scrollArea_3, tr("Choose"),
                   tr(".  Recover Library:    ."), items, 0, false, &ok);

    if (ok && !item.isEmpty() && item!="none")
    {
        if( item.size() < 2 )
            item.resize(2,' '); //fix nullptr "" QStrings or sliced crashes
        ss = estlib + "/" + item.sliced(0,2).trimmed();
        QFile file( estlib + "/archive/" + item.sliced(0,2).trimmed() );

        file.rename( ss );
        ;//qDebugA() << item.sliced(0,2).trimmed() << "moved from archive" << ss;

        loadComboboxLibs( estlib ); // clears and reloads the list

        int id;
        stripFilename( &ss ); // look at the end, not the subdirs

        ss=ss.leftJustified(2,' ',true);
        id=ui->comboBox->findText(ss,
                                    Qt::MatchStartsWith );
        ;//qDebugA() << id<<"<-id and truncated ss is:" << ss;

        loadByIndexLibrarynameOfComboboxLibs( (id>=0)?id:0 );
    }

    return;
}

QString MainWindow::stripFilename( QString *ss )
{
    QDir dd;
    int sepint;

    linuxMsWindowsSlashFixes(ss);
    /* Qt does correct file opens with \ or / no matter which OS you are in
     * but the next line of code will not work if we did not correct
     * the ss, above, to have the right slash type which matches dd.separator().
     */
    sepint = ss->lastIndexOf( dd.separator() ); // '\' or '/'

    if( sepint != -1 && (ss->size() > (sepint+1) ))
    {
        *ss = ss->sliced( sepint+1 );
        //;//qDebugA() << "sliced stripFilenameOfDirectories" << *ss;
    }

    return ( *ss );
}

void MainWindow::dialogChooseFont() //ffff
{
    bool ret;
    //QFont font2 = property("font").value<QFont>();
    int fontPtSz;
    int chose;
    QDialog *QDia;
    QMessageBox msgBox;

    ;//qDebugA() << "Opening point size" << globalFont2.pointSize();

    fontPtSz = globalFont2.pointSize();

    QInputDialog qiDia;
    qiDia.setFont(globalFont2);
    //qiDia.setGeometry( ui->scrollArea_2->geometry() );  // this does not move it to the right spot
    //qiDia.adjustSize(); // this does not move it to the right spot. I tried.

    chose = qiDia.getInt( ui->label_3, tr("Choose"),
                                      tr(". Font:  ."), fontPtSz, 2*fontPtSz/5, 5*fontPtSz/2, 1, &ret);

    //globalFont2.setPointSize(QFontInfo(font).pointSize() * 3/2);
    //QFontMetrics fontMetrics(font);
    //return QSize(fontMetrics.horizontalAdvance(text), fontMetrics.height());

    if (ret && chose>0 )
    {
        // Put up the wait dialog in events...cpp
        waitLoading( 0, (int)(152*(1.0+ESTROWCOUNT/800.0)) ); //reset it.
        ;//qDebugA() << "target:" << (int)(152*(1.0+ESTROWCOUNT/800.0));
        /* WRONG:
         * I think I need setfocus() to start spontaneous() in events..cpp.
         * Anyway it is nice to see the cursor blink inside est side
         * top row so I know the font change has nearly completed.
         * Bug:It highlights but does not blink.
         * ---
         *It takes 70s 140s to load a font change at 999 lines of estRow.
         * 50s is for loading PaceMakerEstimator2.
         * I estimate it was 892 usually or always, instead of 390.
         * Is it when going from 16 to 18 font and 430 is from 20 to 16?
         * Yikes, this is risky for freezing or ending wait-sign too soon
         * if I pick the wrong number.
         *
         * I saw a font 20 take over ~5 minutes and the 892 was triggered
         * but the kill shot took 3 more minutes and Qt was stuck at 928.
         * I just did nothing and eventually it unfroze and it deleted
         * the wait-signal exactly on time.
         *
         * So I presume some variable extra activity is called.
         * Is it setStylesheet() that is extremely slow?
         */
        QTimer::singleShot(100, estRow[1]->col[2], SLOT(setFocus()));

        fontPtSz = chose;
        // This is done in load2() ---> globalFont2 = globalFont2;

        //too limited to not all widgets?: QString ss = QString( "QLineEdit { Font : %1pt }").arg(fontPtSz);
        //ss1 = QString( "%1" ).arg(fontPtSz);

        /* Do not do this here because load2() does
         * thousands of QLineEdit and doing
         * some twice, font and then border, is slow.
         * I moved this to load2.cpp. Feb 13, 2024.
         */
        //ss = QString( "QLabel,QMenuBar,QMenu,QLineEdit { Font : %1pt }").arg(ss1);
        //;//qDebugA() << "Stylesheet string with font point size:\n" << ss;
        //qApp->setStyleSheet(QString( "QLabel,QMenuBar,QMenu,QLineEdit { font: %1pt }").arg(fontPtSz));

        globalFont2.setPointSize(fontPtSz);
        writeSettings(QDir::homePath());

        /* I notice resize sometimes fails to shrink enough on some fonts
         * but is more reliable on increases of screen size
         * so let's go smaller and then bigger when load2() is called.
         *
         * Edit: That worked sometimes, not all the time.
         * I see in load2() that the font might not be selected if a near one
         * is set by the stylesheet.
         * So I changed load2() to check and reload pointsize
         * before setting the QMainWindow geometry. See variable "trail" in load2().
         * I found a bug in stackexchange.com. See resizeBug.txt
         * https://stackoverflow.com/questions/48831898/qt-window-incorrect-size-until-user-event

Great MCVE, exactly what's needed to easily investigate the issue.
Looks like this FlowLayout class was not designed to have it's minimum size change on user action. Layout gets updated 'by chance' by QWidget kernel when the window is moved.
I could make it work smartly by modifying FlowLayout::minimumSize() behaviour, here are the changes I did: see resizeBug.txt for more.

         */

        load2( globalFont2 ); // This is called in one other spot: mainwindow.cpp

        /*
        hide();
        resize( globalFont2.pointSize()*11, globalFont2.pointSize()*7);
        show();
        resize( globalFont2.pointSize()*112, globalFont2.pointSize()*70);
        //QSize qs = QGuiApplication::primaryScreen()->availableGeometry().size();
        //move(qs.width()/3, qs.height()/3 );
        //move(qs.width()/2 - sizeHint().width()/2, qs.height()/2 - sizeHint().height()/2 );
        show();
        */

        setDRAGtotalsBottomSubtotalRow(); // just updates the DRAG buttons

        /* kill the header too or else it keeps the old font
         * in the pushButtons:
         */
        if( oneHeader != nullptr )
            oneHeader->~DialogHeaderOfEstimate();
        if( archiveQtyFrame != nullptr )
            archiveQtyFrame->~qtymath();

        msgBox.setText("Manually resize by grabbing the corner, if necessary\nor exit and restart and the screen will resize perfectly.");
        msgBox.exec();
    }

    return;
}

void MainWindow::dialogChooseEstimateDirectory()
{
    bool ret;
    QString text;
    QMessageBox msgBox;
    QInputDialog qiDia;
    qiDia.setFont(globalFont2);
    qiDia.setGeometry( ui->scrollArea_3->geometry() );  // this does not move it to the right spot
    qiDia.adjustSize(); // this does not move it to the right spot. I tried.

    /* Should not be needed
     * because readSettings() sets the slashes correctly:
     * linuxMsWindowsSlashFixes( &estimateDirectory );
     */

    text = qiDia.getText(this, tr("Choose Estimate Directory"),
                                        tr(".                    Estimate Directory:                    ."),
                                            QLineEdit::Normal,
                                         estimateDirectory, &ret);
    if ( ret && !text.isEmpty() )
    {
        linuxMsWindowsSlashFixes( &text );

        QDir dd;
        QString estlib = text + "/" + estDotlib;

        if( !dd.mkpath( estlib ) )
        {
            //QString ss = "QMessageBox { min-width: %1px; min-height: %2px; }.arg(fontPtSz*estlib.length()).arg(10*fontPtSz)";
            //setStyleSheet( ss );
            msgBox.setText("Cannot create "+estlib);
            msgBox.exec();
        }
        else
        {
            /* If the new dir exists and is different than the old dir
             * then save the new settings to the old location
             * so if the user starts from the old location
             * then the new estimateDirectory is the starting
             * point for reading files.
             */
            if( QFileInfo::exists(text) && text!=estimateDirectory )
            {
                QString oldEstimateDirectory = estimateDirectory;
                estimateDirectory = text;
                // discern homePath()
                int sepint = oldEstimateDirectory.lastIndexOf( dd.separator() ); // '\' or '/'
                if( sepint != -1)
                {
                    oldEstimateDirectory.truncate( sepint );
                    writeSettings(oldEstimateDirectory);

                    /* reload libraries like we do at mainwindow.cpp startup */
                    loadComboboxLibs( estimateDirectory + "/" + estDotlib );
                    loadByIndexLibrarynameOfComboboxLibs(0);
                }
            }

            dd.setCurrent( estimateDirectory );
            writeSettings(dd.homePath());

            /* reload libraries like we do at mainwindow.cpp startup */
            loadComboboxLibs( estimateDirectory + "/" + estDotlib );
            loadByIndexLibrarynameOfComboboxLibs(0);
        }
    }

   return;
}

bool MainWindow::writeSettings( QString usually_homePath )
{
    QFile *fp;
    unsigned twice;
    QDir dd;
    bool ret=true;
    QString ss;
    /* for Linux and MSwindows and MacOS: */
    QString qss2 = usually_homePath+"/"+ ESTIMATEMAKER+"/"+ESTCONFIG;

    /* only for Linux: */
    QString dirPrefix = QDir::homePath()+"/"+DOTCONFIG +"/"+ ESTIMATEMAKER;
    QString qss = dirPrefix+"/"+ESTCONFIG;

    linuxMsWindowsSlashFixes( &qss );
    linuxMsWindowsSlashFixes( &qss2 );
    linuxMsWindowsSlashFixes( &dirPrefix );

    if( QDir::separator()=='/' && dd.mkpath( dirPrefix ) )
    {
     // linux has $HOME/.config/PaceMakerEstimator subdirectory and $HOME
        twice = 2;
    }
    else
        twice = 1;

    for( unsigned xx=1 ;xx<=twice; ++xx)
    {
        // write it to estimateDirectory and
        // .config so alternate start up sequences find the data
        if( xx == 1 )
        {
            fp = new QFile(qss2);
            qDebug() << "writeSettings2" << qss2;
        }
        else
        {
            delete fp;
            fp = new QFile(qss);
            qDebug() << "writeSettings1" << qss;
        }

        if( fp->open(QIODeviceBase::WriteOnly) == false )
            ret=false;
        else
        {
            ss = QString("<font>%1\r\n").arg(globalFont2.pointSize());
            if( fp->write( ss.toStdString().c_str() ) != -1 )
            {
                ;//qDebugA() << "write:" << ss;

                ss = QString("<timesHours>%1\r\n").arg(prefHoursMultiplier);
                if( fp->write( ss.toStdString().c_str() ) != -1 )
                {
                    ss = QString("<timesCosts>%1\r\n").arg(prefCostsMultiplier);
                    if( fp->write( ss.toStdString().c_str() ) != -1 )
                    {
                        ss = QString("<estimateDirectory>%1\r\n").arg(estimateDirectory); 

                        /* What if the user has an active estimateFilename?
                         * Those names are not like /home/user/PaceMakerEstimator/estimateFilename
                         * See dd.setCurrent( estimateDirectory );
                         */

                        if( fp->write( ss.toStdString().c_str() ) != -1 )
                        {
                            ss = QString("<selectCount>%1\r\n").arg(prefSelectCount);

                            if( fp->write( ss.toStdString().c_str() ) != -1 )
                            {

                            }

                        }
                    }
                }
            }
            fp->close();
        }
    }
    delete fp;
    return(ret);
}

bool MainWindow::readSettings(QFont *font)
{
    bool ret=true;
    char ss[80];
    QFont font2 = property("font").value<QFont>();
    int fontPtSz = font2.pointSize();
    QString eD;
    QString qss1, qss2;
    QDir dd;

    /* if estimatemaker is started with argv(1) then
     * estimateDirectory == nullptr never happens
     * and PaceMakerEstimator.config is writeSettings() into the data area
     * and not into home/user/.config and also readSettings() from data area.
     */
    if( estimateDirectory == nullptr )
    {
        qss1 = dd.homePath()+"/"+DOTCONFIG +"/"+ ESTIMATEMAKER;
        dd.mkpath( qss1 ); //create path /home/user/.config/PaceMakerEstimator subdir
        estimateDirectory = qss1;  // same for now

        /* estimateDirectory will be modified below if .config/PaceMakerEstimator subdir
         * has a saved config name or if using home/user/PaceMakerEstimator is successful below.
         */
    }

    QString fff = estimateDirectory+"/"+ESTCONFIG;
    // in case "estimateDir/" creates estimateDir//estconfig"
    linuxMsWindowsSlashFixes( &fff );

    QFile fp(fff);

    if( fp.exists()==false || fp.open(QIODeviceBase::ReadOnly) == false )
    {
        ret=false;

        /* There is no .config, like .config/PaceMakerEstimator/PaceMakerEstimator.config
         * so set home subdirectory for estimates to /home/user/PaceMakerEstimator
         */
        qss2 = dd.homePath()+"/"+ ESTIMATEMAKER;
        if( !dd.mkpath( qss2 ) )
            estimateDirectory = qss1; // failed so revert to this lesser choice
        else
            estimateDirectory = qss2;//create home/PaceMakerEstimator for estimates

        linuxMsWindowsSlashFixes(&estimateDirectory);
    }
    else
    {
        while(  fp.readLine(ss, sizeof(ss)) != -1 )
        {
            ;//qDebugA() << ss;
            if( strncmp( ss,"<font>",6)==0 )
            {
                int dd = QString(ss).sliced(6).toInt();

                if( dd>=fontPtSz/20 && dd<=20*fontPtSz )
                // ignore it if the value is ridiculous range
                {
                    font2.setPointSize(dd);
                    *font = font2;
                }
            }
            else if( strncmp( ss, "<timesHours>",12 )==0 )
            {
                prefHoursMultiplier = QString(ss).sliced(12).toDouble();
            }
            else if( strncmp( ss, "<timesCosts>",12 )==0 )
            {
                prefCostsMultiplier = QString(ss).sliced(12).toDouble();
            }
            /* The estimateDirectory is in the first .config location
             * or the second config and those can reset the estimateDirectory
             * to a final value here. It is a chain depending on any failures
             * so as to select some location for a user. User can change it
             * in the Settings menu.
             */
            else if( strncmp( ss, "<estimateDirectory>",19 )==0 )
            {
                eD = QString(ss).sliced(19).simplified();
                linuxMsWindowsSlashFixes(&eD);

                QDir dir;
                dir.setPath( eD );
                if( dir.exists() )
                {
                   estimateDirectory = eD;
                }
            }
            else if( strncmp( ss, "<selectCount>",13 )==0 )
            {
                prefSelectCount = QString(ss).sliced(13).toInt();
            }
         }
        fp.close();
    }

    /* We have to be sure the est.lib exist */
    dd.mkpath( estimateDirectory + "/" + estDotlib );

    return(ret);
}


void MainWindow::action_xHours()
{
    bool ret;
    double pref;
    QString chose;
    QInputDialog qiDia;
    QString currentvalue = QString("%1").arg(prefHoursMultiplier,6,'f',2);

    qiDia.setGeometry( ui->scrollArea_3->geometry() );
    qiDia.adjustSize();

    chose = qiDia.getText( ui->scrollArea_3, tr("Subtotal Hours times"), tr("Example: 32.00 per Hour"), QLineEdit::Normal, currentvalue, &ret);

    pref = chose.trimmed().toDouble();
    if (ret && pref<0.00001 )
    {
        prefHoursMultiplier = 0.01;
    }
    else if (ret && pref>100000.0 )
    {
        prefHoursMultiplier = 100.0;
    }
    else if( ret )
        prefHoursMultiplier = pref;

    return;
}

void MainWindow::action_xCosts()
{
    bool ret;
    QString chose;
    double pref;
    QInputDialog qiDia;
    QString currentvalue = QString("%1").arg(prefCostsMultiplier,6,'f',2);

    qiDia.setGeometry( ui->scrollArea_3->geometry() );
    qiDia.adjustSize();

    chose = qiDia.getText( ui->scrollArea_3, tr("Subtotal Costs times"), tr("Example: 1.15 is 15% mark-up of cost."), QLineEdit::Normal, currentvalue, &ret);

    pref = chose.trimmed().toDouble();
    if (ret && pref<0.00001 )
    {
        prefCostsMultiplier = 0.01;
    }
    else if (ret && pref>100000.0 )
    {
        prefCostsMultiplier = 100.0;
    }
    else if( ret )
        prefCostsMultiplier = pref;

    return;
}

QString MainWindow::cleanQString(QString ss, int sz)
{
    QString result;

    // if sz <= 0 then sz limit on the result is ignored.

    for(int ii = 0; ii < ss.size(); ++ii)
    {
        QChar qc = ss.at(ii);
        unsigned char cc = *(unsigned char*)(&qc);
        if(cc >= 127)
        {
            ; //result.append("?");
        }
        else if (QChar(cc).isPrint())
        {
            result.append(QChar(cc));
            if( --sz == 0 )
                break;
        }
    }

    return(result);
}

void MainWindow::linuxMsWindowsSlashFixes( QString *eD )
{
    if( QDir::separator()=='/')
    {
        // doubles to singles
        while( eD->contains("//") )
            eD->replace(QString("//"), QString("/"));
        while( eD->contains("\\\\") )
            eD->replace(QString("\\\\"), QString("/"));

        eD->replace(QString("\\"), QString("/")); // linux

        //qDebug() << "separator /" << QDir::separator();
    }
    else /*  '\'  */
    {
        // doubles to singles
        while( eD->contains("//") )
            eD->replace(QString("//"), QString("/"));
        while( eD->contains("\\\\") )
            eD->replace(QString("\\\\"), QString("\\"));

        eD->replace(QString("/"), QString("\\")); // msWindows

        //qDebug() << "separator \\" << QDir::separator();
    }

    return;
}
