
/*  This file is part of PaceMaker Estimator 2.0.
    PaceMaker Estimator 2.0 is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
    PaceMaker Estimator 2.0 is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with PaceMaker Estimator 2.0. If not, see <https://www.gnu.org/licenses/>.
*/

#include "mainwindow.h"
#include "ui_qtymath.h"

//#define TEST_ERRORMSG errormsg=1;
#define TEST_ERRORMSG ;

#define CHARS 40
// add 2 to CHARS to make CHARS2
#define CHARS2 42

qtymath::qtymath(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::qtymath)
{
    QString estlib;

    ui->setupUi(this);

    estlib = ((MainWindow *)parent)->estimateDirectory + "/" + ((MainWindow *)parent)->estDotlib;
    filenameRefcodeLibcode = "__.___";
    ui->libcode2char->setText("    ");
    ui->libcode2char->setToolTip( "2char library code for this estimate line." );

    setStyleSheet("QLineEdit { background-color: beige } ");

    // All these are failed experiments to set the qtymath.ui
    // to point at the Qty line. I do not know how. I gave up.
    // I put it relative to scrollArea_4 instead.
    //https://doc.qt.io/qt-6/application-windows.html#window-geometry
    //X11 provides no standard or easy way to get the frame geometry once the window is decorated. Qt solves this problem with nifty heuristics and clever code that works on a wide range of window managers that exist today. Don't be surprised if you find one where QWidget::frameGeometry() returns wrong results though.
    /* Edit Feb20: I am not going to do this, but
     * I know the answer to fix failed experiments:
     * You have to add up the 3 vertical heights:
     * scrollArea_4
     * scrollArea_x (scrollable estimate area)
     * QEditLine row[anycol]
     * That sum is where to put the arrow at the selected Qty box.
     */

    /* We turn the frame off/on in the constructor/destructor
     * but we have to do it for estimate side FocusIn FocusOut
     * mouse clicks, not just for when we construct/deconstruct
     * the qtyDetail. See eventMouseKeyboard.cpp.
     */
    //turn on the Qty MainWindow ->col highlight
    long row = ((MainWindow *)parent)->keyboardnow.litEstButton;
    if( row > 0 )
    {
        QLineEdit *qle = ((MainWindow *)parent)->estRow[ row ]->col[3];
        //qle->setFrame( (((MainWindow *)parent)->globalFrame==true)?false:true );
        qle->setStyleSheet( "background-color: beige" );

        ((MainWindow *)parent)->saveQtyMathHighlightRow = row;

        int hh,ww;
        hh = qle->frameGeometry().height();
        ww = qle->frameGeometry().width();
        ((MainWindow *)parent)->ui->scrollArea->ensureWidgetVisible( qle, ww, hh);

        /* I cannot understand why this crashes when row == 0 because
         * I set estRow[0]->fileQtyBlock = nullptr in mainwindow.cpp
         * Was it row < 0 sometimes? I doubt it. I initialize
         * keyboardnow to 0L and 0 in mainwindow.h
         * Edit Feb 28, 2023: I am adding initialize keyboardnow to
         * mainwindow.cpp nullptr for() loop in case I am misunderstanding
         * something about mainwindow.h "struct mousekeyboard_now"
         * as initialization to 0L and 0;
         * Edit: Feb 28. That idea failed to fix this crashable line
         * when row<=0;
        ;//qDebugA()<<"loadUI row=" << row;
         * loadUI row= 0
21:05:46: Qt/build-PaceMakerEstimator-Desktop-Debug/PaceMakerEstimator crashed.
         *
         * Oh, I figured it out. estRow[0] == 0 too, of course.
         */
        loadUiFromFileQtyBlock (
                ((MainWindow *)parent)->estRow[ row ]->fileQtyBlock
                                );

        ui->libcode2char->setText(
            " " + ((MainWindow *)parent)->estRow[row]->refcode + " "
                                    );
        if(((MainWindow *)parent)->estRow[row]->col[2]->text() == "")
           ui->libcodeDescription->setText("Estimate Line Description");
        else
            ui->libcodeDescription->setText(
                ((MainWindow *)parent)->estRow[row]->col[1]->text() + " " +
                ((MainWindow *)parent)->estRow[row]->col[2]->text()
                                           );
        //ui->label->setText("right column");

        QString Arefcode = ((MainWindow *)parent)->estRow[row]->refcode;
        Arefcode.resize(2,'_'); //fix nullptr "" QStrings or sliced crashes
        QString Alibcode = ((MainWindow *)parent)->estRow[row]->col[1]->text();
        //;//qDebugA() << "ouch"<<Alibcode<<"ouch";
        if( Alibcode.isEmpty() )
        {
            //filenameRefcodeLibcode = Arefcode.sliced(0,2).trimmed().leftJustified(2,'_');
            filenameRefcodeLibcode = Arefcode.leftJustified(2,'_') + QString(".%1").arg( "___",-3,(QChar)'_');
            //;//qDebugA() << "isEmpty"<<Alibcode<<"ouch";
        }
        else
        {
            /* I trimmed refcode in estimateRead() instead.
             * old: filenameRefcodeLibcode
             *  = Arefcode.sliced(0,2).trimmed().leftJustified(2,'_');
             */
            //filenameRefcodeLibcode = Arefcode.leftJustified(2,'_') + "." + Alibcode;
            filenameRefcodeLibcode = Arefcode.leftJustified(2,'_') + QString(".%1").arg( Alibcode,-3,(QChar)'_');
            //;//qDebugA() << "notisEmpty"<<Alibcode<<"ouch";
        }
    }

    /* I set the default value for filenameRefcodeLibcode to "__._" in case row!>0
     * so "__._*" is the search and not "*" which would include all of them. I want only one list
     * type for one estimate item at a time and __._ is the same as typing no 3char lib code
     * into the estimate and also not using a library item Lib>Est, but just manually
     * typing the estimate line. In other words, pressing Calc button with nothing
     * in the estimate line requires a careful default in case user goes back and fills
     * in the estimate line and expects to see his Archive of Calcs. He won't see them if I
     * store them to something other than __._ which means blank 2char list and blank 3char item.
     */
    QString search = filenameRefcodeLibcode + "*";
    loadComboboxLibMenu( estlib, search );

    ((MainWindow *)parent)->qtymathRequestsQtyColBoxHighlight = true;

    sz = ((MainWindow *)parent)->qtyDetailBoxMonospaceFont.pointSize();
    //sz = ((MainWindow *)parent)->globalFont2.pointSize();
    scra = ((MainWindow *)parent)->ui->scrollArea_4; // scrollArea_4;
    pb = ((MainWindow *)parent)->ui->qtymath;

    geoscra = scra->geometry();
    geopb = pb->geometry();

    ui->libcode2char->setStyleSheet(" QPushButton {background-color: cyan;border-style: outset;border-width: 1px;border-radius: 3px;border-color: black;font: bold 16px;padding-bottom: 1px; } QPushButton:pressed {background-color: cyan;border-style: inset;}");
        //setStyleSheet(" QPushButton {background-color: cyan;border-style: outset;border-width: 1px;border-radius: 3px;border-color: black;font: bold 16px;padding: 2px; } QPushButton:pressed {background-color: rgb(56%, 40%, 29%);border-style: inset;}");
    ui->libcode2char->setMaximumWidth( 2*sz );
    ui->libcode2char->setMinimumWidth( sz );

    //geoFrame = scra->frameGeometry();
    //qtydetail->setGeometry( geoFrame.x()+geo.x(), geoFrame.y()+geo.y(), (13*sz)+(5*sz), 13*sz );
    //qtydetail->setGeometry( geoFrame.x(), geoFrame.y(), (13*sz)+(5*sz), 13*sz );
    //qtydetail->resize( 32 * globalFont2.pointSize(), 26 * globalFont2.pointSize());
    //qtydetail->setGeometry( 14*sz, (27*sz*row)/20, (13*sz)+(5*sz), 13*sz );

    /* You have to use the setting in
     * void MainWindow::dialogChooseFont() menuEdit.cpp
     * But load2() is called twice, so do I have to set globalFont2 in load2()? New Jan31,2024: I erased the globalFont2=??? in load2()
     */
    // MainWindow* parent2 = qobject_cast<MainWindow*>(this->parent());
    // I think this works too:
    // ((MainWindow *)parent) instead of parent2

    /* This setGeometry() works but maybe it conflicts
     * with the widgets inside the frame, below.
     * I tried omitting it and then it is
     * almost automatically the correct size.
     * I added #defines instead.
     */
    /*  this was for the globalfont sz, not
     *  ((MainWindow *)parent)->qtyDetailBoxMonospaceFont
     */


#define HH1 24
#define WW1 23
#define WW2  4
#define WW3  4
#define WW4  1
    int ww = ((WW1+WW2+WW3)*sz) + (24*24)/sz;//I can tell QtCreator is adding some spacers which are not proportional to the font. I add an inverse-to-font-sz to compensate for what I think is a fixed spacer between the 3 horz objects. Thus a 12 font gets 48 pixels addition and a 48 font gets 12 pixels addition. If I knew their math I could be more precise.

    this->setGeometry(
                (geoscra.x() + geopb.x()) - ww + (WW4*sz/2),
                geoscra.y(),
                ww,
                22 + (HH1*sz)
                    );

    ui->qtyCalcOK->setFont( ((MainWindow *)parent)->globalFont2 );
    ui->cancelButton->setFont( ((MainWindow *)parent)->globalFont2 );
    ui->deleteButton->setFont( ((MainWindow *)parent)->globalFont2 );
    // I do not like the qtymath Calc Box to be red button outlines, like the global:
    ui->deleteButton->setStyleSheet( "QPushButton:default {border-color: black;}" );
    ui->cancelButton->setStyleSheet( "QPushButton:default {border-color: black;}" );
    ui->qtyCalcOK->setStyleSheet( "QPushButton:default {border-color: black;}" );

    ui->qtyDetail->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail2->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail3->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail4->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail5->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail6->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail7->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail8->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail9->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail10->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail11->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail12->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail13->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail14->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail15->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail16->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail17->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );

    ui->qtyDetail_2->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail2_2->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail3_2->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail4_2->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail5_2->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail6_2->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail7_2->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail8_2->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail9_2->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail10_2->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail11_2->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail12_2->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail13_2->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail14_2->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail15_2->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail16_2->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );
    ui->qtyDetail17_2->setFont( ((MainWindow *)parent)->qtyDetailBoxMonospaceFont );

    ui->qtyDetail->setMaxLength(CHARS);
    ui->qtyDetail2->setMaxLength(CHARS);
    ui->qtyDetail3->setMaxLength(CHARS);
    ui->qtyDetail4->setMaxLength(CHARS);
    ui->qtyDetail5->setMaxLength(CHARS);
    ui->qtyDetail6->setMaxLength(CHARS);
    ui->qtyDetail7->setMaxLength(CHARS);
    ui->qtyDetail8->setMaxLength(CHARS);
    ui->qtyDetail9->setMaxLength(CHARS);
    ui->qtyDetail10->setMaxLength(CHARS);
    ui->qtyDetail11->setMaxLength(CHARS);
    ui->qtyDetail12->setMaxLength(CHARS);
    ui->qtyDetail13->setMaxLength(CHARS);
    ui->qtyDetail14->setMaxLength(CHARS);
    ui->qtyDetail15->setMaxLength(CHARS);
    ui->qtyDetail16->setMaxLength(CHARS);
    ui->qtyDetail17->setMaxLength(CHARS);

    ui->qtyDetail_2->setMaxLength(7);
    ui->qtyDetail2_2->setMaxLength(7);
    ui->qtyDetail3_2->setMaxLength(7);
    ui->qtyDetail4_2->setMaxLength(7);
    ui->qtyDetail5_2->setMaxLength(7);
    ui->qtyDetail6_2->setMaxLength(7);
    ui->qtyDetail7_2->setMaxLength(7);
    ui->qtyDetail8_2->setMaxLength(7);
    ui->qtyDetail9_2->setMaxLength(7);
    ui->qtyDetail10_2->setMaxLength(7);
    ui->qtyDetail11_2->setMaxLength(7);
    ui->qtyDetail12_2->setMaxLength(7);
    ui->qtyDetail13_2->setMaxLength(7);
    ui->qtyDetail14_2->setMaxLength(7);
    ui->qtyDetail15_2->setMaxLength(7);
    ui->qtyDetail16_2->setMaxLength(7);
    ui->qtyDetail17_2->setMaxLength(7);

    ui->qtyDetail->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail2->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail3->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail4->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail5->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail6->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail7->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail8->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail9->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail10->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail11->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail12->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail13->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail14->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail15->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail16->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail17->setFrame( ((MainWindow *)parent)->globalFrame );

    ui->qtyDetail_2->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail2_2->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail3_2->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail4_2->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail5_2->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail6_2->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail7_2->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail8_2->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail9_2->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail10_2->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail11_2->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail12_2->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail13_2->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail14_2->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail15_2->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail16_2->setFrame( ((MainWindow *)parent)->globalFrame );
    ui->qtyDetail17_2->setFrame( ((MainWindow *)parent)->globalFrame );

    /*
     * https://forum.qt.io/topic/35999/solved-qplaintextedit-how-to-change-the-font-to-be-monospaced/4
       QPlainTextEdit monoEdit;
       QFont ff("unexistent");
       ff.setStyleHint(QFont::Monospace);
       monoEdit.setFont(ff);
     */

    //ui->qtyDetail->setMinimumHeight(HH1*sz);
    //ui->qtyDetail_2->setMinimumHeight(HH1*sz);

    //ui->qtyDetail->setMaximumHeight(HH1*sz);
    //ui->qtyDetail_2->setMaximumHeight(HH1*sz);

    ui->qtyDetail->setMinimumWidth(WW1*sz);
    ui->qtyDetail->setMaximumWidth(WW1*sz);
    ui->qtyDetail2->setMinimumWidth(WW1*sz);
    ui->qtyDetail2->setMaximumWidth(WW1*sz);
    ui->qtyDetail3->setMinimumWidth(WW1*sz);
    ui->qtyDetail3->setMaximumWidth(WW1*sz);
    ui->qtyDetail4->setMinimumWidth(WW1*sz);
    ui->qtyDetail4->setMaximumWidth(WW1*sz);
    ui->qtyDetail5->setMinimumWidth(WW1*sz);
    ui->qtyDetail5->setMaximumWidth(WW1*sz);
    ui->qtyDetail6->setMinimumWidth(WW1*sz);
    ui->qtyDetail6->setMaximumWidth(WW1*sz);
    ui->qtyDetail7->setMinimumWidth(WW1*sz);
    ui->qtyDetail7->setMaximumWidth(WW1*sz);
    ui->qtyDetail8->setMinimumWidth(WW1*sz);
    ui->qtyDetail8->setMaximumWidth(WW1*sz);
    ui->qtyDetail9->setMinimumWidth(WW1*sz);
    ui->qtyDetail9->setMaximumWidth(WW1*sz);
    ui->qtyDetail10->setMinimumWidth(WW1*sz);
    ui->qtyDetail10->setMaximumWidth(WW1*sz);
    ui->qtyDetail11->setMinimumWidth(WW1*sz);
    ui->qtyDetail11->setMaximumWidth(WW1*sz);
    ui->qtyDetail12->setMinimumWidth(WW1*sz);
    ui->qtyDetail12->setMaximumWidth(WW1*sz);
    ui->qtyDetail13->setMinimumWidth(WW1*sz);
    ui->qtyDetail13->setMaximumWidth(WW1*sz);
    ui->qtyDetail14->setMinimumWidth(WW1*sz);
    ui->qtyDetail14->setMaximumWidth(WW1*sz);
    ui->qtyDetail15->setMinimumWidth(WW1*sz);
    ui->qtyDetail15->setMaximumWidth(WW1*sz);
    ui->qtyDetail16->setMinimumWidth(WW1*sz);
    ui->qtyDetail16->setMaximumWidth(WW1*sz);
    ui->qtyDetail17->setMinimumWidth(WW1*sz);
    ui->qtyDetail17->setMaximumWidth(WW1*sz);

    ui->qtyDetail_2->setMinimumWidth(WW2*sz);
    ui->qtyDetail_2->setMaximumWidth(WW2*sz);
    ui->qtyDetail2_2->setMinimumWidth(WW2*sz);
    ui->qtyDetail2_2->setMaximumWidth(WW2*sz);
    ui->qtyDetail3_2->setMinimumWidth(WW2*sz);
    ui->qtyDetail3_2->setMaximumWidth(WW2*sz);
    ui->qtyDetail4_2->setMinimumWidth(WW2*sz);
    ui->qtyDetail4_2->setMaximumWidth(WW2*sz);
    ui->qtyDetail5_2->setMinimumWidth(WW2*sz);
    ui->qtyDetail5_2->setMaximumWidth(WW2*sz);
    ui->qtyDetail6_2->setMinimumWidth(WW2*sz);
    ui->qtyDetail6_2->setMaximumWidth(WW2*sz);
    ui->qtyDetail7_2->setMinimumWidth(WW2*sz);
    ui->qtyDetail7_2->setMaximumWidth(WW2*sz);
    ui->qtyDetail8_2->setMinimumWidth(WW2*sz);
    ui->qtyDetail8_2->setMaximumWidth(WW2*sz);
    ui->qtyDetail9_2->setMinimumWidth(WW2*sz);
    ui->qtyDetail9_2->setMaximumWidth(WW2*sz);
    ui->qtyDetail10_2->setMinimumWidth(WW2*sz);
    ui->qtyDetail10_2->setMaximumWidth(WW2*sz);
    ui->qtyDetail11_2->setMinimumWidth(WW2*sz);
    ui->qtyDetail11_2->setMaximumWidth(WW2*sz);
    ui->qtyDetail12_2->setMinimumWidth(WW2*sz);
    ui->qtyDetail12_2->setMaximumWidth(WW2*sz);
    ui->qtyDetail13_2->setMinimumWidth(WW2*sz);
    ui->qtyDetail13_2->setMaximumWidth(WW2*sz);
    ui->qtyDetail14_2->setMinimumWidth(WW2*sz);
    ui->qtyDetail14_2->setMaximumWidth(WW2*sz);
    ui->qtyDetail15_2->setMinimumWidth(WW2*sz);
    ui->qtyDetail15_2->setMaximumWidth(WW2*sz);
    ui->qtyDetail16_2->setMinimumWidth(WW2*sz);
    ui->qtyDetail16_2->setMaximumWidth(WW2*sz);
    ui->qtyDetail17_2->setMinimumWidth(WW2*sz);
    ui->qtyDetail17_2->setMaximumWidth(WW2*sz);

    ui->cancelButton->setMaximumWidth(WW3*sz);
    ui->qtyCalcOK->setMaximumWidth(WW3*sz);
    ui->deleteButton->setMaximumWidth(WW3*sz);


    return;
}

/* Caution: This signal can be the user clicking
 * OR if you do additem() and clear(), a bunch of these signals are issued
 * because the index changes during those activities.
 * I had a difficult bug where I did not know this. The bug looked like
 * user clicking on the last item in the list when I know I did not click,
 * but the wrong data kept loading into the Qty window!
 * void QComboBox::currentIndexChanged(int index)
    * This signal is sent whenever the currentIndex in the combobox changes either through user interaction or programmatically. The item's index is passed or -1 if the combobox becomes empty or the currentIndex was reset.
 * textActivated() might be a better signal. Documentation says it is the user.
 * void QComboBox::textActivated(const QString &text)
    * This signal is sent when the user chooses an item in the combobox. The item's text is passed. Note that this signal is sent even when the choice is not changed. If you need to know when the choice actually changes, use signal currentIndexChanged() or currentTextChanged().
 * or see: void QComboBox::highlighted(int index)
 *
 * March 9, 2023: I decided to throw up an external flag to ignore
 * comboBox signals while I programatically delete filenames
 * and reload the combobox. See programaticSignalsActive
 */
static bool programaticSignalsActive = false;

void  qtymath::on_comboBox_currentIndexChanged( int index )
{
    QString estlib, sumStr;
    MainWindow *parent = qobject_cast<MainWindow*>(this->parent());
    char buf[680];
    qtymathHidden *deats;
    int xx,yy;
    QMessageBox msgBox(ui->comboBox), msgBox2(ui->comboBox);
    QString fromList;
    QString search = filenameRefcodeLibcode + "*";

    int countComboItems;
    QString date;
    QString addadot;
    QString estlibplus;

    if( programaticSignalsActive == true )
        return;

    msgBox2.setGeometry( ui->libcode2char->geometry() );

    deats = new qtymathHidden();

    estlib = parent->estimateDirectory + "/" + parent->estDotlib + "/" ;

    ;//qDebugA() << "\nQty combo box index:" << index;

    // the default limit on combobox is 10 visible. So maybe limit files to 10.
    if(index==0) // "Archive of Calcs"
    {
        // reload all the directory of items in case some were added:

        // index 0 is already set to 0 after I put
        // reload in index 1 and index 2. So nothing to do here.
    }
    else if(index==1) // " Add this Calc  "
    {
        // save the fileQtyBlock->block to est.lib/refcode_libcode-date-time-stamp
        deats->block[0] = ui->qtyDetail->text();
        deats->block[1] = ui->qtyDetail2->text();
        deats->block[2] = ui->qtyDetail3->text();
        deats->block[3] = ui->qtyDetail4->text();
        deats->block[4] = ui->qtyDetail5->text();
        deats->block[5] = ui->qtyDetail6->text();
        deats->block[6] = ui->qtyDetail7->text();
        deats->block[7] = ui->qtyDetail8->text();
        deats->block[8] = ui->qtyDetail9->text();
        deats->block[9] = ui->qtyDetail10->text();
        deats->block[10] = ui->qtyDetail11->text();
        deats->block[11] = ui->qtyDetail12->text();
        deats->block[12] = ui->qtyDetail13->text();
        deats->block[13] = ui->qtyDetail14->text();
        deats->block[14] = ui->qtyDetail15->text();
        deats->block[15] = ui->qtyDetail16->text();
        deats->block[16] = ui->qtyDetail17->text();

        QString ss;
        long sum, mul;
        unsigned indx, xx;

        for( sum=0L, mul=1L, indx=0; indx < 17; ++indx )
        {
            ss = deats->block[indx].leftJustified(40, ' ');
            strncpy( buf+(indx*40), ss.toStdString().c_str(), 40 );

            // This hash is just for forming a unique name later.
            // Use folding on a string, summed 4 bytes at a time:
            // https://opendsa-server.cs.vt.edu/ODSA/Books/Everything/html/HashFuncExamp.html
            // If the Qty box is the same data then the same filename
            // will be generated so as to overwrite the old filename
            // instead of creating a new filename.

            for( xx = 0; xx<40; ++xx )
            {
                mul = ((xx % 4) == 0) ? 1 : mul * 256;

                /* Skip all spaces except the first one
                 * so "(1+2)" is different than "    (1+2)".
                 * Skipping most spaces reduces the total
                 * since 680 chars will probably be mostly
                 * spaces. A lower total is so modulo 100000
                 * is more assuredly unique.
                 */
                if(xx!=0 && ss[xx]==' ')
                    ;
                else
                {
                    long dd = (long)(ss[xx].unicode());
                    sum += (dd * mul);

                    //sum += ( (dd * mul) + indx );
                    /* I could have added indx because think about
                     * identical lines rearranged in a different order.
                     * Without "+ indx", then the same hash number results.
                     * But now I think that is a good thing since
                     * the qty total is the same, so overwrite the same file.
                     */
                }
            }
        }
        sum = sum % 100000L;
        sumStr = QString("%1").arg( sum );
        ;//qDebugA() << "\nHash output:" << sum << "and modulo" << sumStr.rightJustified(5,'0');
    //--------------------------------

        // I decided to only include month so as to not have so many filenames
        QDate current;
        //QString date = current.currentDate().toString("yyMMdd");
        //QString date = current.currentDate().toString("MM");


        msgBox2.setText("Old archive names for this\nlibrary list and library item\ncan be deleted." /*+estlibplus*/ );
    msgBox2.setInformativeText("Delete all old before saving current Calc Box?\n(Usually choose No)");
msgBox2.setStandardButtons( QMessageBox::No | QMessageBox::Yes );
    msgBox2.setDefaultButton(QMessageBox::No);

    int ans = msgBox2.exec();
    switch (ans)
    {
          case QMessageBox::No:
                // It is not a date anylonger, just a count:
                countComboItems = ui->comboBox->count();
                if( countComboItems > 0)
                {
                    // the first items are "Archive of Calcs" and "Add a Calc" and we do not want 2
                    // as our first file number. We want 1, so decrement.
                    --countComboItems;
                }
                date = QString("%1").arg( countComboItems,2,10,(QChar)'0');
                addadot = (filenameRefcodeLibcode.indexOf('.') == -1)? ".":"";

                estlibplus = estlib +
                                     filenameRefcodeLibcode +
                                     addadot +
                                     "_" + date +
                                     sumStr.rightJustified(5,'0');

                ;//qDebugA() << "\nwrite new estlib filename:" << estlibplus;
              break;

          case QMessageBox::Yes:
              programaticSignalsActive = true;

              ;//qDebugA() << "\nYes, qty box delete:" << search;

              xx = ui->comboBox->count();
              for( yy=2; yy<xx; ++yy )
              {
                  ui->comboBox->setCurrentIndex(yy);
                  fromList = estlib + ui->comboBox->currentText();
                  remove( fromList.toStdString().c_str() );
              }

              // It is not a date anylonger, just a count:
              date = QString("%1").arg( 1,2,10,(QChar)'0');
              addadot = (filenameRefcodeLibcode.indexOf('.') == -1)? ".":"";

              estlibplus = estlib +
                           filenameRefcodeLibcode +
                           addadot +
                           "_" + date +
                           sumStr.rightJustified(5,'0');
              programaticSignalsActive = false;

#ifdef ANOTHERWAY
              programaticSignalsActive = true;

              ;//qDebugA() << "\nYes, qty box delete:" << search;
              //userClickedOn = estlib + ui->comboBox->currentText();

              xx = ui->comboBox->count();
              for( yy=2; yy<xx; ++yy )
              {
                  QString dd;
                  dd = estlib + search;
                  remove( dd.toStdString().c_str() );
              }
              programaticSignalsActive = false;
#endif
              break;

          default:
                // should not happen
            ;
              break;
    }

    //--------------------------------
    // write current Qty box into est.lib for future estimates to reuse calcs.

        QFile fp(estlibplus);

        if( fp.open(QIODeviceBase::WriteOnly) == false )
        {
            msgBox.setText("Failed to create a qty archive\n"+estlibplus);
            msgBox.exec();
        }
        else
        {
            if( fp.write( buf, 680) == 680 )
            {
                  /* too much chatter
                    msgBox.setText("Created a qty archive\n"+estlibplus);
                    msgBox.exec();
                   */
                // do a reload of the menu because it changed
                loadComboboxLibMenu( estlib, search );
                ui->label->setStyleSheet("QLabel { background-color : green; color : white; }");
                ui->label->setText("Saved"); //Created a qty archive\n"+estlibplus);
            }
            fp.close();
        }

    }
    else if(index >= 2 ) // default action is to read the clicked-on filename and load it into qty box
    {
            QString userClickedOn;

            userClickedOn = estlib + ui->comboBox->currentText();

    ;//qDebugA() << "\nQty combo box filename:" << userClickedOn;

            // read the pick, est.lib/currentText(), into
            // a temporary qtymathHidden, called deats, and then load
            // deats into the Ui which is qtymath, but neither of these
            // are fileQtyBlock->block yet. User can press Insert to do that.

            QFile fp(userClickedOn);

            if( fp.open(QIODeviceBase::ReadOnly) == false )
                ;   //ret = false; // let user decide since not able to open.
            else
            {
                if( fp.read( buf, 680) == 680 )
                {
                    char spare[41];

                    for( unsigned indx=0; indx < 17; ++indx )
                    {
                        strncpy( spare, buf+(40*indx), 40);
                        *(spare + 40) = '\0';

                        QString ss;  // = QString("%1").arg(spare); same
                        ss = ss.fromStdString( spare );
                        while( ss.endsWith(" ") ) // instead of trimmed()
                            ss.chop(1);
                        deats->block[indx] = ss;
                    }
                    loadUiFromFileQtyBlock( deats );

                }

                fp.close();
            }
            // do a reload of the menu not because it changed, but because
            // I want the top menu item to reappear and erase " Saved"
            QString search = filenameRefcodeLibcode + "*";
            loadComboboxLibMenu( estlib, search );
    }

    delete deats;

    return;
}

bool qtymath::loadComboboxLibMenu( QString estlib, QString search )
{
    QDir dir;
    QStringList nameFilters;
    int xx, ii;
    QString nm, ss;
    QFileInfoList list;
    //int saveLastComboBoxIndex;

    programaticSignalsActive = true;

    //;//qDebugA() << "search"<<search<<"ouch";

    ui->label->setStyleSheet("QLabel { background-color : white; color : black; }");
    ui->comboBox->setFont( ((MainWindow *)parent())->globalFont2 );

    ui->label->setText(" "); // erase saved
    ui->comboBox->clear(); // delete all, if any
    ui->comboBox->addItem( "Archive of Calcs" );
    ui->comboBox->addItem( " Add this Calc  " );

    dir.setPath( estlib );
    //;//qDebugA() << estlib << "loading Library Help menu. Search:" << search;

    if( dir.exists() == true )
    {
        //nameFilters << "*.*";  instead probably "B_* for example or B_.B08*
        nameFilters << search;

        dir.setNameFilters( nameFilters);
        list = dir.entryInfoList( nameFilters, QDir::Files, QDir::Name ); //sort by name

        //saveLastComboBoxIndex = ui->comboBox->currentIndex();

        for ( xx = 0; xx < list.size(); ++xx )
        {
            QFileInfo fileInfo = list.at(xx);
            nm = fileInfo.fileName();

            ss = dir.relativeFilePath( nm ); // remove the prefix directory path
            if( ss.contains('_') == true || ss.contains('.') == true)
            // include only names like B.BO1 or 0_ or B_.B08
            {
                ui->comboBox->addItem( ss );
            }
            //;//qDebugA() << ss << "item loading Library Help menu";
        }
        //ui->comboBox->setCurrentIndex( saveLastComboBoxIndex );
        //ui->comboBox->setCurrentIndex( 0 );

#ifdef WORKSBUTNOSORTINGbyname
        //QDirIterator it(dir);
        for( xx=0; it.hasNext(); ++xx)
        {
            nm = it.next();

            ss = dir.relativeFilePath( nm ); // remove the prefix directory path
            if( ss.contains('.') == false) // skip names like B.BO1 and /..
            {
                ui->comboBox->addItem( ss );
            }
        }
#endif
    }

    programaticSignalsActive = false;

    return(true);
}

#define DOUBLECLICKFEATURE
#ifdef DOUBLECLICKFEATURE
bool ignoreQtyPress = false;
#endif

qtymath::~qtymath()
{
    ignoreQtyPress = false;

    // archiveQtyFrame is used to kill the qtymath instance if
    // void MainWindow::on_qtymath_pressed() is pressed again
    // so only one is active at a time.
    // Set it to nullptr so as to not kill it again.
    //MainWindow* parent2 = qobject_cast<MainWindow*>(this->parent());

    MainWindow *parent = qobject_cast<MainWindow*>(this->parent());

    /* We turn the frame off/on in the constructor/destructor
     * but we have to do it for estimate side FocusIn FocusOut
     * mouse clicks, not just for when we construct/deconstruct
     * the qtyDetail. See eventMouseKeyboard.cpp.
     */
    // turn off the Qty MainWindow ->col highlight
    parent->qtymathRequestsQtyColBoxHighlight = false;
    if( parent->saveQtyMathHighlightRow > 0 )
    {
        QLineEdit *qle = parent->estRow[ parent->saveQtyMathHighlightRow ]->col[3];
        //qle->setFrame( (((MainWindow *)parent)->globalFrame==true)?true:false );
        //qle->setStyleSheet( "background-color: white" ); //It is NOT white
        qle->setStyleSheet("background-color: palette()" );  // apparently this resets it per https://stackoverflow.com/questions/23029415/how-to-set-default-system-background-color
    }

    delete ui;

    parent->archiveQtyFrame = nullptr;

    return;
}


void qtymath::on_cancelButton_clicked()
{
    ignoreQtyPress = false;
    qtymath::~qtymath();
}

void MainWindow::on_qtymath_pressed()
{

#ifdef DOUBLECLICKFEATURE
    if( ignoreQtyPress == true )
    {
        ignoreQtyPress = false;
        if( archiveQtyFrame != nullptr )
        {
            archiveQtyFrame->hide();
            archiveQtyFrame->~qtymath();
        }
    }
    else
    {
        ignoreQtyPress = true;

#endif
        if( archiveQtyFrame != nullptr )
            archiveQtyFrame->~qtymath();

        qtymath *qtydetail = new qtymath( this );

        archiveQtyFrame = qtydetail;
        // saved for deleting from the menuEdit.cpp dialogFont selection menu
        // since it needs redrawing if the font was changed.

        qtydetail->show();
        qtydetail->raise();
        qtydetail->activateWindow();
#ifdef DOUBLECLICKFEATURE
    }
#endif
}

void MainWindow::on_qtymath_released()
{

    return;
}

/* See constructor of visible qtymath calls this method
 * so as to load qtymath with 17 lines from its qtymathHidden estRow[row].
 * If no qtymathHidden  for the row then nothing happens until saveUi...()
 * because user might change the row before saveUi...()
 *
 * ADD SOME CODE TO CALC AND SUM THE LINES. THIS LOADS ONLY!!!!!!
 * Also set sumOf17lines qtymath variable.
 */
bool qtymath::loadUiFromFileQtyBlock( qtymathHidden *fileQtyBlock )
{
    bool ret = false;

    /* 680 char blocks are loaded into memory as an object containing
     *
     * long row
     * QString
     * double sumOf17line
     *  (See qtymathHidden.h and .cpp).
     *
     * qtymathHidden *fileQtyBlock, one per est line, is in mainwindow.h.
     * All new estimates have these preset to nullptr by mainwindow.cpp
     *
     * They are set to a 680 char new blank memory chunk if the
     * user pressed the Qty detail button to request a qtymath object
     * on an estimate line. See qtymath constructor function for
     * new qtymathHidden 680 chunk creation as I just described.
     * One qtymath is for one est line. See deleteButton for destruction
     * of a 680 char chunk.
     *
     * If 680 chunks come into existence from disk, that way is not
     * here. See the estimate read function in menuNewOpen.cpp.
     * This loadUiFromFileQtyBlock() function ONLY PARSES qtymath object
     * from these blobs that are already in memory as *fileQtyBlock s.
     *
     * I did it this way because the qtymath object only lets one
     * at a time exist. I wrote a secondary object
     * called qtymathHidden, 100s hidden or perhaps none in existence.
     * I set it to contain one QString of 680 chars
     * and one sumOf17lines so I do not have to redo the math calc often.
     *
     * It could have been 17 QString of 40 chars each,
     * instead of 17x40=680, one QString=680
     * It is easy to change.
     *
     */

    if( fileQtyBlock != nullptr )
    {
        ui->qtyDetail->setText(fileQtyBlock->block[0] );
        ui->qtyDetail2->setText(fileQtyBlock->block[1] );
        ui->qtyDetail3->setText(fileQtyBlock->block[2] );
        ui->qtyDetail4->setText(fileQtyBlock->block[3] );
        ui->qtyDetail5->setText(fileQtyBlock->block[4] );
        ui->qtyDetail6->setText(fileQtyBlock->block[5] );
        ui->qtyDetail7->setText(fileQtyBlock->block[6] );
        ui->qtyDetail8->setText(fileQtyBlock->block[7] );
        ui->qtyDetail9->setText(fileQtyBlock->block[8] );
        ui->qtyDetail10->setText(fileQtyBlock->block[9] );
        ui->qtyDetail11->setText(fileQtyBlock->block[10] );
        ui->qtyDetail12->setText(fileQtyBlock->block[11] );
        ui->qtyDetail13->setText(fileQtyBlock->block[12] );
        ui->qtyDetail14->setText(fileQtyBlock->block[13] );
        ui->qtyDetail15->setText(fileQtyBlock->block[14] );
        ui->qtyDetail16->setText(fileQtyBlock->block[15] );
        ui->qtyDetail17->setText(fileQtyBlock->block[16] );

        // Fill in the right hand columns; false means do not report errors
        on_qtyDetail_byRow( ui->qtyDetail , ui->qtyDetail_2, false );
        on_qtyDetail_byRow( ui->qtyDetail2, ui->qtyDetail2_2, false );
        on_qtyDetail_byRow( ui->qtyDetail3, ui->qtyDetail3_2, false );
        on_qtyDetail_byRow( ui->qtyDetail4, ui->qtyDetail4_2, false );
        on_qtyDetail_byRow( ui->qtyDetail5, ui->qtyDetail5_2, false );
        on_qtyDetail_byRow( ui->qtyDetail6, ui->qtyDetail6_2, false );
        on_qtyDetail_byRow( ui->qtyDetail7, ui->qtyDetail7_2, false );
        on_qtyDetail_byRow( ui->qtyDetail8, ui->qtyDetail8_2, false );
        on_qtyDetail_byRow( ui->qtyDetail9, ui->qtyDetail9_2, false );
        on_qtyDetail_byRow( ui->qtyDetail10, ui->qtyDetail10_2, false );
        on_qtyDetail_byRow( ui->qtyDetail11, ui->qtyDetail11_2, false );
        on_qtyDetail_byRow( ui->qtyDetail12, ui->qtyDetail12_2, false );
        on_qtyDetail_byRow( ui->qtyDetail13, ui->qtyDetail13_2, false );
        on_qtyDetail_byRow( ui->qtyDetail14, ui->qtyDetail14_2, false );
        on_qtyDetail_byRow( ui->qtyDetail15, ui->qtyDetail15_2, false );
        on_qtyDetail_byRow( ui->qtyDetail16, ui->qtyDetail16_2, false );
        on_qtyDetail_byRow( ui->qtyDetail17, ui->qtyDetail17_2, false );

        sumOf17Lines = qtymath_sumOf17Lines();
        ret = true;
    }

    return( ret );
}

/* This saves qtymath contents to the current saveQtyMathHighlightRow qtymathHidden
 * and this method also creates a qtymathHidden object if none exists
 * as yet for that estRow[saveQtyMathHighlightRow] (see ==nullptr)
 */
bool qtymath::saveUiToFileQtyBlock( long row, QWidget *parent )
{
  bool ret = false;

  if( row > 0 )
  {
    qtymathHidden *fqb = ((MainWindow *)parent)->estRow[row]->fileQtyBlock;

    if( fqb == nullptr )
    {
       //new qtymathHidden because estimate file
       // did not load one for this est line.

       ;//qDebugA() << "Initialize a new qtymathHidden because qtymathHidden does not exist for row:" << row;
       // initialize it by using its constructor, the same as menuNewOpen.cpp does.

        fqb = new qtymathHidden();

        ((MainWindow *)parent)->estRow[ row ]->fileQtyBlock =  fqb;

       /* delete ((qtymathHidden *)fileQtyBlock);
        * Do the object deletes when closing the estimate
          or else it is a memory leak!
            Edit: See clearLibEst.cpp for one delete call
            and qtymath.cpp has another one for the qtymath Delete button.
        */
    }

    if( fqb != nullptr )
    {
        fqb->block[0] = ui->qtyDetail->text();
        fqb->block[1] = ui->qtyDetail2->text();
        fqb->block[2] = ui->qtyDetail3->text();
        fqb->block[3] = ui->qtyDetail4->text();
        fqb->block[4] = ui->qtyDetail5->text();
        fqb->block[5] = ui->qtyDetail6->text();
        fqb->block[6] = ui->qtyDetail7->text();
        fqb->block[7] = ui->qtyDetail8->text();
        fqb->block[8] = ui->qtyDetail9->text();
        fqb->block[9] = ui->qtyDetail10->text();
        fqb->block[10] = ui->qtyDetail11->text();
        fqb->block[11] = ui->qtyDetail12->text();
        fqb->block[12] = ui->qtyDetail13->text();
        fqb->block[13] = ui->qtyDetail14->text();
        fqb->block[14] = ui->qtyDetail15->text();
        fqb->block[15] = ui->qtyDetail16->text();
        fqb->block[16] = ui->qtyDetail17->text();

        sumOf17Lines = qtymath_sumOf17Lines();

        ret = true;
    }
  }

  return( ret );
}

void qtymath::on_deleteButton_clicked()
{
    MainWindow *parent = qobject_cast<MainWindow*>(this->parent());
    long row = parent->saveQtyMathHighlightRow;

    ;//qDebugA() << "Before delete the qtymathHidden object";

    if( row > 0 )
    {
        // most estimate lines have no qty details and are just a number
        if( parent->estRow[row]->fileQtyBlock != nullptr )
        {
            ;//qDebugA() << "Before/mid/After delete the qtymathHidden object";

            // if it exists, free it because user said so.

            delete (parent->estRow[row]->fileQtyBlock);
            parent->estRow[row]->fileQtyBlock = nullptr;

            parent->estRow[row]->col[3]->clear();  //Qty column blanked

            ;//qDebugA() << "After delete the qtymathHidden object";

            /* If it exists then Qty QLineEdit() manual editing
             * should be prevented? No. Let them edit.

             *Cursor is blocked for readOnly so this is not the way:
             * QLineEdit().readOnly(false) when qtymath.delete
             * QLineEdit().readOnly(true) when qtymath.Qty\nInsert
             *
             * This would work: Let the user edit the Qty.
             * Do not care if the qtymath object 17 lines total
             * does not match because it can be reasserted by
             * pressing the [Qty request] and [Qty\Insert].
             */
        }
        else
        {
            parent->estRow[row]->col[3]->clear();  //Qty column blanked

            ;//qDebugA() << "No fileQtyBlock. Delete does clear()-only Qty column.";
        }
    }

    // Next do same as cancel button:
    //qtymath::~qtymath();
    on_cancelButton_clicked();

    return;
}

void qtymath::on_qtyCalcOK_pressed() //now called InsertQTY button
{
    MainWindow *parent = qobject_cast<MainWindow*>(this->parent());
    long row = parent->saveQtyMathHighlightRow;
    if( row > 0 )
    {
        saveUiToFileQtyBlock( row, parent );

        parent->estRow[row]->col[3]->selectAll();
        parent->estRow[row]->col[3]->insert( QString("%1").arg(sumOf17Lines));
        // do not use setText() because it has no undo/redo via QEditLine
        //parent->estRow[row]->col[3]->setText( QString("%1").arg(sumOf17Lines));
    }

    return;
}

void qtymath::on_qtyCalcOK_released() //now called InsertQTY button
{
    MainWindow *parent = qobject_cast<MainWindow*>(this->parent());

    long row = parent->saveQtyMathHighlightRow;
    if( row > 0 )
    {
        // in case the Qty changed then do the est line math again.
        parent->estRow[row]->col[3]->setFocus( Qt::TabFocusReason );
        parent->doCalc();
    }

    // same as: qtymath::~qtymath();
    on_cancelButton_clicked();

    return;
}


/* This is a situation from 1989 when I coded this to run
 * on a 8088 cpu. The code was accepting typing on a
 * tty screen and recalculating the totals inside (example: [3*4])
 * waiting for the closing ]
 * It had to be fast to keep up with the typist.
 * So the code has an external variable
 * and it does potentially infinite function recursion for each
 * math operation - + * / ^.
 * I wasted too much time trying to understand and get
 * the ++arithstr working as it recurs. It looks like
 * spaghetti code BUT IT IS NOT! Read more below.
 *
 * I opted to leave it alone because its recursion works amazingly.
 * So you see a static char *arithstr just like 1989.
 */
static char *arithstr;
static char noErrorHelperbuf[CHARS2];

void qtymath::on_qtyDetail_byRow( QLineEdit *in, QLineEdit *out, bool errdisplay )
{
    double output;
    QString sub;
    char *startAt;

    QString save_out = out->text();

    if( in->text().trimmed().isEmpty() == true)
    {
        out->setText( "      " );
    }
    else
    {
        char buf[CHARS2];
        strncpy( buf, in->text().toStdString().c_str(), CHARS );
        startAt = noErrorHelper( buf );

        if( arith( startAt, &output, errdisplay ) == true )
        {
            /* Only the first if() is what I expect as a quantity
             * but allow for other creative users to see something
             * other than =00000 and =99999 mispresenting 9,999,999;
             * which is bigger than, and also possibly tiny numbers.
             */

            sub = formatQtymathOutput( output );

            /* This if() is also from 1989 where the program would
             * allow the user to insert a = after the last ]
             * and then the math result would be inserted
             * inside the window but not be used in the Qty summation.
             * The intent was to give the user a calculator
             * scratch area and have a diary of how the calc
             * was done, likely producing a preliminary
             * but unused result. The next line
             * would likely use the number for calculating
             * the Qty into the 17 sums in the right column.
             * How do you get a number into the right column?
             * You simply omit the '=':
             */
            ;//qDebugA() << "insert QTY calc arithstr() value to right column->" << arithstr << "<-\n";

            if( noErrors( startAt, 1 ) == false )
                out->setText( "=error ");
            else if( *arithstr==']' && *(arithstr+1)=='=' )
            {
                /* append the number after the '=', in qtyDetail
                 * which is the "in" QLineEdit not "out" QLineEdit.
                 */
                QString sub2 = sub.remove('=').trimmed();
                //;//qDebugA() <<"sub2 qtymath test|"<<sub2<<"|";

                // like mouse does a select for a replace
                in->setSelection( arithstr-noErrorHelperbuf, sub2.length()+2 );
                in->insert( "="+sub2+" " );
                // sub="=    45" becomes "=45 " and "=123456" becomes "=123456 "
                out->setText( "=     ");
            }
            else
                out->setText( sub );
        }
        else
            out->setText( "= error" );
    }

    if( save_out != out->text() )
    {
        /* if change of one line's right column total
         * then recalculate the sum of right column: sumOf17Lines
         */
        sumOf17Lines = qtymath_sumOf17Lines();
    }

    ui->label->setStyleSheet("QLabel { background-color : white; color : black; }");
    ui->label->setText(" "); // erase "saved" word since user started typing again.

    return;
}

double qtymath::qtymath_sumOf17Lines()
{
    /* We completed the line calculation.
     * Next sum all the right hand column of 17 lines.
     */
    double dd = 0.0;
    QString ss, sub;

    ss = ui->qtyDetail_2->text();
    dd += ss.replace('=',' ').toDouble();
    ss = ui->qtyDetail2_2->text();
    dd += ss.replace('=',' ').toDouble();
    ss = ui->qtyDetail3_2->text();
    dd += ss.replace('=',' ').toDouble();
    ss = ui->qtyDetail4_2->text();
    dd += ss.replace('=',' ').toDouble();
    ss = ui->qtyDetail5_2->text();
    dd += ss.replace('=',' ').toDouble();
    ss = ui->qtyDetail6_2->text();
    dd += ss.replace('=',' ').toDouble();
    ss = ui->qtyDetail7_2->text();
    dd += ss.replace('=',' ').toDouble();
    ss = ui->qtyDetail8_2->text();
    dd += ss.replace('=',' ').toDouble();
    ss = ui->qtyDetail9_2->text();
    dd += ss.replace('=',' ').toDouble();
    ss = ui->qtyDetail10_2->text();
    dd += ss.replace('=',' ').toDouble();
    ss = ui->qtyDetail11_2->text();
    dd += ss.replace('=',' ').toDouble();
    ss = ui->qtyDetail12_2->text();
    dd += ss.replace('=',' ').toDouble();
    ss = ui->qtyDetail13_2->text();
    dd += ss.replace('=',' ').toDouble();
    ss = ui->qtyDetail14_2->text();
    dd += ss.replace('=',' ').toDouble();
    ss = ui->qtyDetail15_2->text();
    dd += ss.replace('=',' ').toDouble();
    ss = ui->qtyDetail16_2->text();
    dd += ss.replace('=',' ').toDouble();
    ss = ui->qtyDetail17_2->text();
    dd += ss.replace('=',' ').toDouble();

    sub = formatQtymathOutput( dd );

    ss = QString("Insert\nQTY\n\n\n%1").arg(sub);
    ss.replace('=',' ');
    ui->qtyCalcOK->setText(ss);

    return( dd );
}

QString qtymath::formatQtymathOutput( double output )
{
    QString sub;

    // fits in 6 chars without exponents:
    if( (output <= 999999 && output >= .0001) || output==0.0)
        sub = QString("=%1").arg( output,6);

    // fits 6 chars but needs a minus sign; no exponents
    else if( (output >= -99999.4 && output <= -1.0 ) )
        sub = QString("=%1").arg( output,6);

    // right justify, but needs a minus sign; no exponents
    else if( (output > -1.0 && output <= -.001 ) )
        sub = QString("=%1").arg( output,6, 10, -1,' ');

    // every tiny and huge neg num is left over (do exponents)
    else if( output < 0.0 )
        // make room for negative sign by reducing 1precision
        sub = QString("%1").arg( output,6, 'g', 1,' ');

    // every tiny and huge pos num is left over (do exponents)
    else
        sub = QString("%1").arg( output,7, 'g', 2,' ');

    return(sub);
}

/* In the old days: sometimes we did not want the error msgs
 * since the typing was incomplete and real time.
 * Now we wait for carriage return or tab. So maybe
 * we always want error messages and we could remove the errormsg
 * argument.
 */

bool qtymath::arith( char *presetArithstr, double *output, bool errormsg )
{
    double xx;
    double adble;
    double bdble;
    double cdble;
    char atofdone,op,sp;
    QMessageBox msgBox(this);

    if( presetArithstr != nullptr )
        arithstr = presetArithstr;

    /* This arith() coding is beyond impressive from 1989.
     * It looks ugly-but-fast 8088 cpu coding needing an external.
     * But it does a recursive call of arith() calls arith()
     * calls arith(), below, depending on greater than one
     * arithmetic operators, such as + - / * ^, and depending
     * on nested square brackets to order the operations
     * properly. Wow! I never guessed that simple code
     * would make this work: [[3-2+4]-[3*3]+[10-[10/5]] = 4
     * It does do the operations non-sequentially, as it should.
     *
     * Simpler example: 3 operations: [2+3-15/2].
     * It disobeys order of operations.
     * It should do (2+3)-(15/2) = -2.5.
     * Instead it brain-dead does ((2+3)-15)/2 = -5
     * because the coding is doing one operation via arith()
     * at a time in simple order (2+3) = 5-15/2
     * and then (5-15) = -10/2
     * and then -10/2 = -5.
     *
     * However, it does obey brackets:
     *          [[3-2+4]-[3*3]+[10-[10/5]] = 4
     * In other words, it is doing any depth of square brackets
     * within square brackets. Wow!
     *
     * Now that I wrote all that, I can see a way of adding
     * order of operations without damaging this elegant
     * recursive arith() function, which works.
     *
     * Create a prescan which inserts the square brackets
     * around multiply and divide operators if there are no
     * brackets, ignoring add/subtract.
     * Then run arith() against the prescan invented line
     * instead of the actual line.
     *
     * A stupid second way is: Break the line into shorter line segments,
     * like (2-3+5) - (4*3/2) or (4-3/5) becomes (4) - (3/5)
     * and run arith() against each line.
     * Next combine them into one or more lines with operators.
     * Rerun arith() against each line.
     * Combine these shorter lines into one line eventually.
     * Rerun arith() and you get the final result.
     * It helps to know arith() does evaluate (4)
     * with no operator as a result of 4. So short lines like
     * (4) will give a correct result.
     *
     * As of Feb 24, 2023 the code is not doing order of
     * arithmetic operators as I propose to improve.
     * It's doing operations in sequence, if no brackets'='.
     */

    atofdone=sp=0;
    op='+';
    xx=0;

    for(++arithstr;*arithstr;++arithstr)
    {

;//qDebugA() << "arithstr() values" << arithstr << "\n";

        if(*arithstr==' '||*arithstr=='?')
            sp=1;
        else if(isdigit(*arithstr)||*arithstr=='.')
        {
            if(atofdone)
            {
                if(sp)
                {
                    if(errormsg)
                    {
                        msgBox.setText("Space between digits?");
                        msgBox.exec();
                    }
                    *(arithstr-1)='?';
                    return(false);
                }
            }
            else
            {
OK:
                atofdone=1;

                if( do_op(xx,lookahead(arithstr), op, &cdble, errormsg) == false )
                        return(false);
                xx = cdble;
                op=sp=0;
            }
        }
        else if(*arithstr=='[')
        {
            if( arith( nullptr, &adble, errormsg) == false )
                return(false);

            if( do_op(xx, adble, op, &bdble, errormsg) == false )
                return(false);
            xx = bdble;
            op=sp=0;
        }
        else if(*arithstr==']')
        {
            *output = xx;
            return(true);
        }
        else if(*arithstr=='-'||*arithstr=='+'||*arithstr=='*'||*arithstr=='/'||*arithstr=='^')
        {
            sp=atofdone=0;
            if(op)
            {
                // My 2023 note: I think 1989 code is trying to
                // allow [ 4 *   -3 ] = -12
                // such that the - is not an operator.
                if((*arithstr=='-'||*arithstr=='+')&&(op=='^'||op=='*'||op=='/'))
                    goto OK;
                else if(errormsg)
                {
                    msgBox.setText("Ambiguous [ ].\nAdd some brackets around the - + * / ^ to clarify.");
                    msgBox.exec();
                    return(false);
                }
                else
                    return(false);
            }
            else
                op=*arithstr;
        }
        else
            sp=0;/*release space test if letters are being used inside*/
    } /*for*/

    *output = xx;
    return(true);
}

bool qtymath::do_op(double xx, double xx2, char op, double *output, short errormsg )
{
    QMessageBox msgBox(this);

    TEST_ERRORMSG;

    ;//qDebugA() << "do_op() values" << xx << " " << xx2;

    if(!op)
    {
        if(errormsg)
        {
            QString say;
            say = QString("[%1]is presumed your intent.\nOtherwise, there is a missing\noperation sign between\n%2 and %3").arg(xx2).arg(xx).arg(xx2);
            msgBox.setText(say);
            msgBox.exec();
            *output = xx2;
        }
        return(true);
    }
    else if(op=='+')
        xx=xx+xx2;
    else if(op=='-')
        xx=xx-xx2;
    else if(op=='/')
    {
        if(xx2==0)
        {
            if(errormsg)
            {
                msgBox.setText("Division by zero [ /0 ] is not possible.");
                msgBox.exec();
            }
            return(false);
        }
        xx=xx/xx2;
    }
    else if(op=='*')
        xx=xx*xx2;
    else if(op=='^')
        xx=pow(xx,xx2);

    *output = xx;
    return(true);
}

double qtymath::lookahead(char *str)
{
    double yy;

    for(yy=::atof(str);*str;++str)
    {
        if(*str=='%')
        {
            yy=yy/100;
            break;
        }
        else if(!isdigit(*str) && *str!='.' && *str!='E')
            break;
    }
    ;//qDebugA() << "lookahead() value" << yy;

    return(yy);
}

bool qtymath::noErrors(char *str, short errormsg)
{
    QMessageBox msgBox(this);

    TEST_ERRORMSG;

    bool ret = true;
    int cc;
    char *firstSquare,*pt1, *secondSquare,test;

    firstSquare=strchr(str,'[');
    secondSquare=strrchr(str,']');

    if( !firstSquare && !secondSquare )/*be quiet if no []*/
        ret=true;
    else if(!firstSquare)  // we already know *secondSquare == ']'
    {
        if(errormsg)
        {
            msgBox.setText("Missing '[' ?");
            msgBox.exec();
        }
        ret=false;
    }
    else if(!secondSquare) // we already know *firstSquare == '['
    {
        if(errormsg)
        {
            msgBox.setText("Missing ']' ?");
            msgBox.exec();
        }
        ret=false;
    }
    else if(firstSquare>secondSquare)
    {
        if(errormsg)
        {
            msgBox.setText("] [ is backwards.");
            msgBox.exec();
        }
        ret=false;
    }
    else
    {
        test=0;
        for(cc=0, pt1=firstSquare; pt1<=secondSquare; ++pt1)
        {
            if(*pt1=='[')
                ++cc;
            else if(*pt1==']')
                --cc;

            if(cc<0)
            {
                if(errormsg)
                {
                    msgBox.setText("Missing '[' ?");
                    msgBox.exec();
                }
                ret=false;
                test = 0; // force ignore it. One msg is enough.
                break;
            }
            else if(cc==0 && pt1!=secondSquare)  /*check for [3+2]/[2-3] - error. Should be: [[3+2]/[2-3]] */
                test=1;
        }

        if(test)
        {
            if(errormsg)
            {
                QString makemsg;
                makemsg = QString("The whole calculation must be bracketted.\nExample:[ %1 ]").arg( str );

                msgBox.setText( makemsg);
                msgBox.exec();
            }
            ret=false;
        }
        else if(cc>0)
        {
            if(errormsg)
            {
                msgBox.setText("The brackets [ ] must be of equal number.");
                msgBox.exec();
            }
            ret = false;
        }
    }

    if( ret == true ) // one last error check
    {
        /* secondSquare is the last ]
         * What if some arithmetic operators are outside
         * the last ]? Warn the user but ret=true, not false
         * so the idea is allowed.
         *
         * secondSquare=strrchr(str,']');
         */
        if( secondSquare == nullptr )
            ;
        else if( strchr(secondSquare,'+') != nullptr
                || strchr(secondSquare,'-') != nullptr
                || strchr(secondSquare,'*') != nullptr
                || strchr(secondSquare,'/') != nullptr
                || strchr(secondSquare,'^') != nullptr
        )
        {
            msgBox.setText("Some + - * / ^ are AFTER the [] brackets\nand so are excluded from the calculation.");
            msgBox.exec();
            ret=true;
        }

        if( firstSquare != nullptr && firstSquare > str)
        {
            char part[firstSquare-str+1];

            sprintf( part, "%.*s", (int)(firstSquare-str), str );

            //yes it happens. Here is an example:
            //  4*3  - [4-2]
            //;//qDebugA() << part << "does firstSquare test ever happen?";

            if( strrchr(part,'+') != nullptr
                    || strrchr(part,'-') != nullptr
                    || strrchr(part,'*') != nullptr
                    || strrchr(part,'/') != nullptr
                    || strrchr(part,'^') != nullptr
            )
            {
                QString makemsg;
                makemsg = QString("Some + - * / ^ are BEFORE the [] brackets\nand are included only if sufficient operation signs are present: %1").arg( part );

                msgBox.setText(makemsg);
                msgBox.exec();

                ret=true;
            }
        }

    }

    return(ret);
}

// I added a begin space because the arith() skipped
// the first char demanding it always be a ( or a [
// But I discovered the code works without brackets too
// if I add a space and let it skip the space.
// This is a minor improvement over 1989.
// I cannot modify arith() char-skip. Doing that damages
// the algorithm. So I modify buf.

char *qtymath::noErrorHelper( char *str )
{
    char *startAt = noErrorHelperbuf;

    noErrorHelperbuf[0] = ' ';
    strncpy( noErrorHelperbuf+1, str, CHARS );
    noErrorHelperbuf[CHARS+1] = '\0';

/* the old 1989 code does not like me adding a space if it
 * is doing a [2+4]=  calculation; ie. buf[1] == '['
 * because it fails to skip the first '[' and then
 * misses the '=' at the end of the line, if it is there.
 * It is not a math error, but the '=' signal to put the
 * total in the right column or the left column is missed.
 * So move startAt up a little if '[' is up ahead.
 */
for( char *pt=noErrorHelperbuf; *pt != '\0'; ++pt )
{
    if( *pt == ' ' )
        ;
    else if( *pt == '[' )
    {
        startAt = pt;
        break;
    }
    else if( isdigit(*pt) == true )// quit early if numeric chars
                                 // but quitting will make ]= fail per above note.
        break;
}

/* The original documentation demanded the user only use
 * square brackets and only the part inside the first
 * set of squares was calculated. The next one or more
 * were ignored. That idea is not intuitive.
 * I typed [2+3]*[4-2]=5. It had to be [[2+3]*[4-2]].
 * I am could add some code here to assure the
 * first instance =10 because reading instructions
 * about brackets should not be required.
 * However, see noErrors() which warns the user
 * and refuses the line until the surrounding brackets are
 * done.
 */

    return(startAt);
}


#ifdef OLD1989_CODE_ESTIMATEMAKER_QTYBOX

#include "filesize"

static char *arithstr;
extern char errormsg,*towstr2;

calc(where)
unsigned where;
{
    extern int yy;
    extern unsigned segds;
    extern char linechanged,stored;
    extern char create_equals;		/*used in estoff.c*/

    double arith();
    char *index(),str[41],ret;

    stored='n';/*changed so it will store*/
    linechanged='n';/*undo linechanged for next run*/

    take.tot=0;
    ret=1;

    if(!setjmp(0))
    {
        if(errchk(str))
        {
            arithstr=index(str,'[');
            take.tot=(float)arith();
            }
        putsn(towstr2,24,29,50);/*replace menu if no errors*/
        }
    else
    {
        take.tot=ret=0;
        tone(320,2);
        putsn("       <Ctrl-E> turns error messages ON/OFF.      ",24,29,50);
        if(!errormsg)
        {
            ret=1;
            goto OK;
            }
        }

    if(ret && ((*arithstr==']' && *(arithstr+1)=='=')||create_equals))
        moveanswer(where);
    else
    {
OK:
        strncpy(take.ch,str,40);
        highlight();
        _lmove(sizeof(struct takeoff),&take,segds,0,where);

        addalltake();
        }

    return(ret);
    }

double do_op(xx,xx2,op)
double xx,xx2;
char op;
{
    extern char *escape;
    char msg();
    double pow();

    if(!op)
    {
        if(errormsg)
            msg("Missing operation sign.",escape,0);
        longjmp(0,1);
        }
    else if(op=='+')
        xx=xx+xx2;
    else if(op=='-')
        xx=xx-xx2;
    else if(op=='/')
    {
        if(xx2==0)
        {
            if(errormsg)
                msg("Division by zero [ /0 ] is not possible.",escape,0);
            longjmp(0,1);
            }
        xx=xx/xx2;
        }
    else if(op=='*')
        xx=xx*xx2;
    else if(op=='^')
        xx=pow(xx,xx2);

    return(xx);
    }

errchk(str)
char *str;
{
    extern char *escape;
    int cc;
    char *rindex(),*index(),*pt1,*pt2,msg(),test;
    char *mleft="Missing '[' ?";
    char *entire="The whole calculation must be bracketted. ie. [ 5/[2-4] ]";

    strncpy(str,take.ch,40);
    *(str+40)=0;

    pt1=index(str,'[');
    pt2=rindex(str,']');

    if(!pt1&&!pt2)/*be quiet if no []*/
        return(0);
    else if(!pt1)
    {
        if(errormsg)
            msg(mleft,escape,0);
        longjmp(0,1);
        }
    else if(!pt2)
    {
        if(errormsg)
            msg("Missing ']' ?",escape,0);
        longjmp(0,1);
        }

    if(pt1>pt2)
    {
        if(errormsg)
            msg("] [ What?",escape,0);
        longjmp(0,1);
        }

    test=0;
    for(cc=0;pt1<=pt2;++pt1)
    {
        if(*pt1=='[')
            ++cc;
        else if(*pt1==']')
            --cc;
        if(cc<0)
        {
            if(errormsg)
                msg(mleft,escape,0);
            longjmp(0,1);
            }
        else if(cc==0 && pt1!=pt2)  /*check for [3+2]/[2-3] - error. Should be: [[3+2]/[2-3]] */
            test=1;
        }

    if(test)
    {
        if(errormsg)
            msg(entire,escape,0);
        longjmp(0,1);
        }

    if(cc)
    {
        if(errormsg)
            msg("The brackets [ ] must be of equal number.",escape,0);
        longjmp(0,1);
        }

    return(1);
    }

double arith()
{
    extern char *escape;
    double xx,do_op(),lookahead();
    char atofdone,op,sp,msg(),*index();

    atofdone=sp=0;
    op='+';
    xx=0;

    for(++arithstr;*arithstr;++arithstr)
    {
        if(*arithstr==' '||*arithstr=='?')
            sp=1;
        else if(isdigit(*arithstr)||*arithstr=='.')
        {
            if(atofdone)
            {
                if(sp)
                {
                    if(errormsg)
                        msg("Space between digits?",escape,0);
                    *(arithstr-1)='?';
                    longjmp(0,1);
                    }
                }
            else
            {
OK:
                atofdone=1;
                xx=do_op(xx,lookahead(arithstr),op);
                op=sp=0;
                }
            }
        else if(*arithstr=='[')
        {
            xx=do_op(xx,arith(),op);
            op=sp=0;
            }
        else if(*arithstr==']')
            return(xx);
        else if(*arithstr=='-'||*arithstr=='+'||*arithstr=='*'||*arithstr=='/'||*arithstr=='^')
        {
            sp=atofdone=0;
            if(op)
            {
                if((*arithstr=='-'||*arithstr=='+')&&(op=='^'||op=='*'||op=='/'))
                    goto OK;
                else if(errormsg)
                {
                    msg("Error between [ ]?",escape,0);
                    longjmp(0,1);
                    }
                else
                    longjmp(0,1);
                }
            else
                op=*arithstr;
            }
        else
            sp=0;/*release space test if letters are being used inside*/
        } /*for*/

    return(xx);
    }

highlight()
{
    extern int yy;
    char *rindex(),*index(),*pt1,*pt2,save;

    save=take.ch[40];
    take.ch[40]=0; /*assure zero termination*/

    putsn(take.ch,yy,30,40);
    scr_ch_atr(yy,30,40,7);

    pt1=index(take.ch,'[');
    pt2=rindex(take.ch,']');
    if(pt1&&pt2)
        scr_ch_atr(yy,30+pt1-take.ch,pt2-pt1+1,15);

    take.ch[40]=save;
    return;
    }

double lookahead(str)
char *str;
{
    double yy,atof();

    for(yy=atof(str);*str;++str)
    {
        if(*str=='%')
        {
            yy=yy/100;
            break;
            }
        else if(!isdigit(*str) && *str!='.' && *str!='E')
            break;
        }

    return(yy);
    }

moveanswer(where)
unsigned where;
{
    extern int qdec,yy;
    extern unsigned segds;
    extern char linechanged,stored;

    extern char *escape;
    unsigned strlen();
    char str[41],ss[35],*rindex(),*pt,msg(),*ftos(),dump[7],*ptt;

    strncpy(str,take.ch,40);
    str[40]=0;

    pt=rindex(str,']');
    if(!pt)
        return;
    else if(pt-str>32)
    {
        msg("Insufficient room after the last ']' to write: '=ANSWER'",escape,0);
        return;
        }

    *(++pt)='=';
    ++pt;

    ptt=ftos((double)take.tot,ss,5-qdec,qdec);
    if(strlen(ptt)>6)
    {
      overflow(dump,ptt,6);
        strncpy(pt,dump,6);
        }
    else
        strncpy(pt,ptt,6);

    take.tot=0;
    strncpy(take.ch,str,40);
    highlight();
    _lmove(sizeof(struct takeoff),&take,segds,0,where);
    addalltake();
    return;
    }

    This part is from rapid ftos() 2002,
    which added the X-windows unix output

    char *ftos( double num, char *dest, char ccldec, char ccrdec )
    {
      int ldec = ccldec;
      int rdec = ccrdec;
      char buf[35], pad[35];
      int sz=0;
      char *targ, *src, *pt;
      int uptodecimal_sz;

      /* Convert float to string. Match the characteristics of
         the DOS function.
       */

      sprintf( buf, "%*.*f", ldec, rdec, num );

      pt = strchr( buf,'.' );
      if( pt != (char *)0 )
         uptodecimal_sz = pt - buf;
      else
         uptodecimal_sz = strlen(buf);

      /* requested digits: */
      sz = ldec+rdec+1;

      /* But, do not truncate digits left of decimal
         because then we would lose accuracy!
       */
      if( uptodecimal_sz > sz )
         sz = uptodecimal_sz;

      if( dest == (char *)0 )
      {
         /* right justify the string and pad with ' ' */
         sprintf( pad, "% *.*s", sz, sz, buf );

    #ifdef GD
         if( *plotter_pngsz == '\0' )
         {
    #endif
            XSetForeground( display, gc1, colour_seg[BRIGHT_SEG].colour[ scr_attr & 0xF ] );
            if( (scr_attr & 128) == 0 ) /* 7th bit=0 indicates black backgrnd */
               XSetBackground( display, gc1, colour_seg[DARK_SEG].colour[ 0 ] );  /* black */
            else
               XSetBackground( display, gc1, colour_seg[DARK_SEG].colour[ (scr_attr>>4) & 0x7 ] );

            XDrawImageString( display, win, gc1, (currentcol16*pixelspercol)>>4,
                                              ((currentrow16*pixelsperrow)>>4)+pixelsperrow*8/10,
                                              pad,
                                              sz
                          );


    #ifdef GD
         }
    #endif
    #ifdef GD
         else if( plotter_isopen == 'y' )
         {
           if( currentrow != 0 )/*don't print overlapping mess on top line*/
           {
            if( imagesz('w')<290 || imagesz('h')<170 )
            {
            gdImageString( plotfd, gdFontSmall,
                                    (currentcol16*pixelspercol)>>4,
                        (currentrow16*pixelsperrow)>>4,
                    pad,
                    colour_seg[BRIGHT_SEG].gd_colour[ scr_attr & 0xF ]
                     );
            }
            else if( imagesz('w')<=430 || imagesz('h')<=320 )
            {
            gdImageString( plotfd, gdFontTiny,
                                    (currentcol16*pixelspercol)>>4,
                        (currentrow16*pixelsperrow)>>4,
                    pad,
                    colour_seg[BRIGHT_SEG].gd_colour[ scr_attr & 0xF ]
                     );
            }
            else if( imagesz('w')<890 || imagesz('h')<670 )
            {
            gdImageString( plotfd, gdFontMediumBold,
                                    (currentcol16*pixelspercol)>>4,
                        (currentrow16*pixelsperrow)>>4,
                    pad,
                    colour_seg[BRIGHT_SEG].gd_colour[ scr_attr & 0xF ]
                     );
            }
            else
            {
            gdImageString( plotfd, gdFontGiant,
                                    (currentcol16*pixelspercol)>>4,
                        (currentrow16*pixelsperrow)>>4,
                    pad,
                    colour_seg[BRIGHT_SEG].gd_colour[ scr_attr & 0xF ]
                     );
            }
           }

         }
    #endif

         /* save to virtual screen */
         if( currentrow<maxrow)
         {
        /*didn't fix the overlap: sz = strlen( pad );  added Nov/2001 DP */
            for( targ=VirtualChars+(currentrow*maxcol+currentcol), src=pad; --sz>=0 && currentcol<maxcol; ++currentcol, currentcol16 += 16 )
               *targ++ = *src++;
         }
      }
      else
      {
         /* right justify the string and pad with ' ' */
         sprintf( dest, "% *.*s", sz, sz, buf );
      }

      return( dest );
    }

    A Qt6 version of ftos(): untested

        char *ftos( double num, char *dest, char ccldec, char ccrdec )
        {
          int ldec = ccldec;
          int rdec = ccrdec;
          char buf[35], pad[35];
          int sz=0;
          char *targ, *src, *pt;
          int uptodecimal_sz;

          /* Convert float to string. Match the characteristics of
             the DOS function.
           */

          sprintf( buf, "%*.*f", ldec, rdec, num );

          pt = strchr( buf,'.' );
          if( pt != (char *)0 )
             uptodecimal_sz = pt - buf;
          else
             uptodecimal_sz = strlen(buf);

          /* requested digits: */
          sz = ldec+rdec+1;

          /* But, do not truncate digits left of decimal
             because then we would lose accuracy!
           */
          if( uptodecimal_sz > sz )
             sz = uptodecimal_sz;

          /* right justify the string and pad with ' ' */

          /* This gives a warning in Qt6, disliking the %space
           * sprintf( dest, "% *.*s", sz, sz, buf );
           * So ... memset first:
           */
          memset( dest, ' ', sz);

          short offset = sz - strlen(buf);
          if( offset <= 0 )
              sprintf( dest, "%.*s", sz, buf );
          else
              sprintf( dest+offset, "%.*s", sz-offset, buf );

          return( dest );
        }


#endif


/* The estimate and library QLineEdits were
 * done with connect() calls in load2().
 * Here I do 17 slots instead for these qtymath QLineEdits.
 *
 * Feb24: I learned returnPressed also generates
 * an editingFinished signal. So these are commented out.
 */

void qtymath::on_qtyDetail_returnPressed()
{
    ui->qtyDetail2->setFocus( Qt::TabFocusReason );

    //on_qtyDetail_byRow( ui->qtyDetail, ui->qtyDetail_2 );
}
void qtymath::on_qtyDetail2_returnPressed()
{
    ui->qtyDetail3->setFocus( Qt::TabFocusReason );
    //on_qtyDetail_byRow( ui->qtyDetail2, ui->qtyDetail2_2 );
}
void qtymath::on_qtyDetail3_returnPressed()
{
    ui->qtyDetail4->setFocus( Qt::TabFocusReason );
    //on_qtyDetail_byRow( ui->qtyDetail3, ui->qtyDetail3_2 );
}
void qtymath::on_qtyDetail4_returnPressed()
{
    ui->qtyDetail5->setFocus( Qt::TabFocusReason );
    //on_qtyDetail_byRow( ui->qtyDetail4, ui->qtyDetail4_2 );
}
void qtymath::on_qtyDetail5_returnPressed()
{
    ui->qtyDetail6->setFocus( Qt::TabFocusReason );
    //on_qtyDetail_byRow( ui->qtyDetail5, ui->qtyDetail5_2 );
}
void qtymath::on_qtyDetail6_returnPressed()
{
    ui->qtyDetail7->setFocus( Qt::TabFocusReason );
    //on_qtyDetail_byRow( ui->qtyDetail6, ui->qtyDetail6_2 );
}
void qtymath::on_qtyDetail7_returnPressed()
{
    ui->qtyDetail8->setFocus( Qt::TabFocusReason );
    //on_qtyDetail_byRow( ui->qtyDetail7, ui->qtyDetail7_2 );
}
void qtymath::on_qtyDetail8_returnPressed()
{
    ui->qtyDetail9->setFocus( Qt::TabFocusReason );
    //on_qtyDetail_byRow( ui->qtyDetail8, ui->qtyDetail8_2 );
}
void qtymath::on_qtyDetail9_returnPressed()
{
    ui->qtyDetail10->setFocus( Qt::TabFocusReason );
    //on_qtyDetail_byRow( ui->qtyDetail9, ui->qtyDetail9_2 );
}
void qtymath::on_qtyDetail10_returnPressed()
{
    ui->qtyDetail11->setFocus( Qt::TabFocusReason );
    //on_qtyDetail_byRow( ui->qtyDetail10, ui->qtyDetail10_2 );
}
void qtymath::on_qtyDetail11_returnPressed()
{
    ui->qtyDetail12->setFocus( Qt::TabFocusReason );
    //on_qtyDetail_byRow( ui->qtyDetail11, ui->qtyDetail11_2 );
}
void qtymath::on_qtyDetail12_returnPressed()
{
    ui->qtyDetail13->setFocus( Qt::TabFocusReason );
    //on_qtyDetail_byRow( ui->qtyDetail12, ui->qtyDetail12_2 );
}
void qtymath::on_qtyDetail13_returnPressed()
{
    ui->qtyDetail14->setFocus( Qt::TabFocusReason );
    //on_qtyDetail_byRow( ui->qtyDetail13, ui->qtyDetail13_2 );
}
void qtymath::on_qtyDetail14_returnPressed()
{
    ui->qtyDetail15->setFocus( Qt::TabFocusReason );
    //on_qtyDetail_byRow( ui->qtyDetail14, ui->qtyDetail14_2 );
}
void qtymath::on_qtyDetail15_returnPressed()
{
    ui->qtyDetail16->setFocus( Qt::TabFocusReason );
    //on_qtyDetail_byRow( ui->qtyDetail15, ui->qtyDetail15_2 );
}
void qtymath::on_qtyDetail16_returnPressed()
{
    ui->qtyDetail17->setFocus( Qt::TabFocusReason );
    //on_qtyDetail_byRow( ui->qtyDetail16, ui->qtyDetail16_2 );
}
void qtymath::on_qtyDetail17_returnPressed()
{
    ui->qtyDetail->setFocus( Qt::TabFocusReason );
    //on_qtyDetail_byRow( ui->qtyDetail17, ui->qtyDetail17_2 );
}


void qtymath::on_qtyDetail_editingFinished()
{
    on_qtyDetail_byRow( ui->qtyDetail, ui->qtyDetail_2, true );
}
void qtymath::on_qtyDetail2_editingFinished()
{
    on_qtyDetail_byRow( ui->qtyDetail2, ui->qtyDetail2_2, true );
}
void qtymath::on_qtyDetail3_editingFinished()
{
    on_qtyDetail_byRow( ui->qtyDetail3, ui->qtyDetail3_2, true );
}
void qtymath::on_qtyDetail4_editingFinished()
{
    on_qtyDetail_byRow( ui->qtyDetail4, ui->qtyDetail4_2, true );
}
void qtymath::on_qtyDetail5_editingFinished()
{
    on_qtyDetail_byRow( ui->qtyDetail5, ui->qtyDetail5_2, true );
}
void qtymath::on_qtyDetail6_editingFinished()
{
    on_qtyDetail_byRow( ui->qtyDetail6, ui->qtyDetail6_2, true );
}
void qtymath::on_qtyDetail7_editingFinished()
{
    on_qtyDetail_byRow( ui->qtyDetail7, ui->qtyDetail7_2, true );
}
void qtymath::on_qtyDetail8_editingFinished()
{
    on_qtyDetail_byRow( ui->qtyDetail8, ui->qtyDetail8_2, true );
}
void qtymath::on_qtyDetail9_editingFinished()
{
    on_qtyDetail_byRow( ui->qtyDetail9, ui->qtyDetail9_2, true );
}
void qtymath::on_qtyDetail10_editingFinished()
{
    on_qtyDetail_byRow( ui->qtyDetail10, ui->qtyDetail10_2, true );
}
void qtymath::on_qtyDetail11_editingFinished()
{
    on_qtyDetail_byRow( ui->qtyDetail11, ui->qtyDetail11_2, true );
}
void qtymath::on_qtyDetail12_editingFinished()
{
    on_qtyDetail_byRow( ui->qtyDetail12, ui->qtyDetail12_2, true );
}
void qtymath::on_qtyDetail13_editingFinished()
{
    on_qtyDetail_byRow( ui->qtyDetail13, ui->qtyDetail13_2, true );
}
void qtymath::on_qtyDetail14_editingFinished()
{
    on_qtyDetail_byRow( ui->qtyDetail14, ui->qtyDetail14_2, true );
}
void qtymath::on_qtyDetail15_editingFinished()
{
    on_qtyDetail_byRow( ui->qtyDetail15, ui->qtyDetail15_2, true );
}
void qtymath::on_qtyDetail16_editingFinished()
{
    on_qtyDetail_byRow( ui->qtyDetail16, ui->qtyDetail16_2, true );
}
void qtymath::on_qtyDetail17_editingFinished()
{
    on_qtyDetail_byRow( ui->qtyDetail17, ui->qtyDetail17_2, true );
}

