
/*  This file is part of PaceMaker Estimator 2.0.
    PaceMaker Estimator 2.0 is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
    PaceMaker Estimator 2.0 is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with PaceMaker Estimator 2.0. If not, see <https://www.gnu.org/licenses/>.
*/
#include "mainwindow.h"
#include "tabs_plusEstReadWrite.h"
//#include "qthread.h"
#include "qtimer.h"
#include "totalsubtotal.h"
#include "dialogheaderofestimate.h"
#include "ui_dialogheaderofestimate.h"
#include <QPrinter>
#include <QPrintDialog>
#include <QAbstractPrintDialog>
#include <QPainter>
#include <QTextEdit>
#include <QPainter>

#define AsPromised /*NEVER undefine this. I am putting it to amplify importance.*/

static tabPage *tab[TABCOUNT] = {nullptr}; //initialize at least the first one as nullptr so tab[1] = new tabPage(this) gets allocated somewhere soon below.
static int indexOld = 1; //active Tab number, 1to10 for this QLineEditArray{}

    /* struct of one tab is in tabs_plusEstReadWrite.h
     *
     *  and from mainwindow.h, here are the details similar to mainwindow widget
     *  because they are going to be interchanged like offline storage,
     *  to expand the pages of an estimate without overwhelming Qt's layout limit:
     *  One has QStrings and the other has QLineEdits and QPushButtons.
     *
     *   typedef struct estTab
     *   {
            qtymathHidden *fileQtyBlock;
            QString refcode;
            QString *col[MAXCOL_libest];

            } *estTabRow[ESTROWCOUNT];

        and the QLineEdits look similar, like this:

    struct QLineEditArray {

            qtymathHidden *fileQtyBlock;
            QString refcode;
            QLineEdit *col[MAXCOL_libest];
            QPushButton *pushbutton;

            } *libRow[ESTROWCOUNT], *estRow[ESTROWCOUNT];
      */

/* This is the slot which is called by mouse click on a Tab widget: index 0 to 9 */
void MainWindow::on_tabWidget_currentChanged(int indexNew )
{
    ui->tabWidget->setEnabled(false);
    ++indexNew; // 1 to 10 instead of 0 to 9 so as to match Tab 1 screen names.

    if( indexNew < TABCOUNT && indexNew >= 1 )
    {
        if( indexOld != indexNew )
        /*  1 is the initialized value for indexOld and I chose
         *  to skip this step at startup since nothing is in
         *  the QLineEditsArray when on_tabWidget_currentChanged()
         *  is called by the Qt setupUi().
         *
         *  We are ready for the first Tab click, if it ever happens.
         *
         *  I did this skip-when-equal because maybe
         *  ...currentChanged()  event is thrown before
         *  the QLineEdits start and I do not know what junk is
         *  going to be copied to the tabPage.
         *  I will start doing copies into the tabPages only if
         *  Tab 2 is clicked, or higher.
         *
         *  The QLineEditsArray will always be the primary source
         *  since I can see it on the screen and more likely discern a bug.
         *
         *  estimateWrite will do a unloadQLineEditsToTab() before
         *  writing the estimate to filename from all tabPages.
         *
         *  estimateRead() will copy all estimate filename into tabPages
         *  and then do a copy to tab[1].loadTabtoQLineEdits()
         *  before handing control to the user to do work on the QLineEdits.
         *  Search "AsPromised" to see that code line everwhere.
         *
         *  User does not work on Tabs directly. He works in QWidgets QLineEdits.
         *
         *  Also, I block user doing Tab clicks during estimateRead()
         *  and estimateWrite() so activity here on Tabs is not happening
         *  during filename to/from Tabs.
         *  Search ui->tabWidget->setEnabled(false)
         */
        {
            ;//qDebugA( << "doing Tab copies new:old "<<indexNew<<":"<<indexOld;
            if( tab[indexOld] == nullptr )
            {
                ;//qDebugA( << "index Tab allocating tab[indexOld]" << indexOld;
                tab[indexOld] = new tabPage(this);
            }
            if( tab[indexNew] == nullptr )
            {
                ;//qDebugA( << "index Tab allocating tab[indexNew]" << indexNew;
                tab[indexNew] = new tabPage(this);
            }

#ifdef ARTIFICIALtrial_TESTallROWs
        /* See mainwindow.cpp where I filled all rows and columns
         * with values for a Quality Assurance arithmetic
         * test of all the rows before release of the software.
         */
        tab[indexNew]->unloadQLineEditsToTab();
        tab[indexNew]->loadTabtoQLineEdits();
#endif

        tab[indexOld]->unloadQLineEditsToTab();
        clearEstStack();/*Undo/Redo buttons that were done on another Tab*/

#ifdef AsPromised
            tab[indexNew]->loadTabtoQLineEdits();
#endif
            // scroll up to top so user sees estimate. He might be on line 230.
            scroll_est_into_view( estRow[1]->col[2] );
            indexOld = indexNew; // save it to MainWindow.h with the QLineEditArray
        }
    }
    ui->tabWidget->setEnabled(true);
    return;
}

void tabPage::unloadQLineEditsToTab()
{
    int rr,cc;

    for( rr=1; rr<=ESTROWCOUNT-1; ++rr)
    {
        for( cc=1; cc<MAXCOL_libest; ++cc)
        {
            ;//qDebugA() << "\nrr/cc" << rr<<cc;

            // Next we do QLineEdits text() saved as QString.
            if( parent2->estRow[rr]->col[cc] != nullptr )
                estTabRow[rr]->col[cc] = parent2->estRow[rr]->col[cc]->text();
            else
                estTabRow[rr]->col[cc] = "";
        }

        /* free memory before using this row because memory leaks are bad
         * and we do not need to remember anything in Tab because QLineEditArray
         * is user work area, not the Tab. Tab is an old version of his
         * work which he picked up into the QLineEdits upon his arrival
         * on this Tab, and now he is leaving this Tab and storing his
         * changes back to it.
         */
        if( estTabRow[rr]->fileQtyBlock != nullptr )
        {
            delete estTabRow[rr]->fileQtyBlock;
            estTabRow[rr]->fileQtyBlock = nullptr;
        }

        if( parent2->estRow[rr]->fileQtyBlock != nullptr )
        {
            // estTabRow will retain this Tab memory pointer.
            estTabRow[rr]->fileQtyBlock = parent2->estRow[rr]->fileQtyBlock;

            /* QLineEditArray here, will no longer point at it.
             * We initialize it to nullptr as it prepares for the new Tab.
             * No memory is lost since estTabRow is pointing at it.
             *
             * We have to unload or else estTabRow and estRow might accidentally
             * point at the same fileQtyBlock. The next loadTabtoQLineEdits()
             * will be responsible for setting these fileQtyBlock pointers of
             * QLineEditsArray.
             * loadTabtoQLineEdits() must always by called, AsPromised,
             * before passing control to user who operates in the QLineEdit realm.
             */
            parent2->estRow[rr]->fileQtyBlock = nullptr;
        }

        // This is easiest. nullptr and QString are passed with an equal sign.
        estTabRow[rr]->refcode = parent2->estRow[rr]->refcode;
    }
    return;
}

void tabPage::loadTabtoQLineEdits()
{
    int rr,cc;

    for( rr=1; rr<=ESTROWCOUNT-1; ++rr)
    {

        for( cc=1; cc<MAXCOL_libest; ++cc)
        {
            ;//qDebugA() << "did we get rr cc" << rr << cc << estTabRow[rr];
            // Do QLineEdits text() copied from QStrings.
            if( estTabRow[rr]->col[cc] != nullptr )
                parent2->estRow[rr]->col[cc]->setText( estTabRow[rr]->col[cc] );
            else
                parent2->estRow[rr]->col[cc]->setText(""); //I think this never happens since Tabs get filled from QLineEditArray and those are all initialized to "" strings.
        }

        if( parent2->estRow[rr]->col[1]->text().indexOf('=') >= 0 )
        // =  ==  === subtotal total blocks
        {
            parent2->estRow[rr]->col[1]->setReadOnly(true);
            parent2->estRow[rr]->col[2]->setReadOnly(true);
        }
        else
        {
            parent2->estRow[rr]->col[1]->setReadOnly(false);
            parent2->estRow[rr]->col[2]->setReadOnly(false);
        }


        parent2->estRow[rr]->fileQtyBlock = estTabRow[rr]->fileQtyBlock;

        /* QLineEditArray estRow now points at it.
         * No memory is lost since estRow is pointing at it.
         * It is harmless to set estTabRow because QLineEditArray is driving
         * the action for this Tab until a new Tab is selected.
         */
        estTabRow[rr]->fileQtyBlock = nullptr;

        // This is easy too. nullptr and QString are passed with an equal sign.
        parent2->estRow[rr]->refcode = estTabRow[rr]->refcode;
    }
    parent2->setDRAGtotalsBottomSubtotalRow();
    ;//qDebugA() << "here is what tabsthinks bottomRow" << parent2->bottomSubtotalRow;

    return;
}

void MainWindow::clearTabPages()
{
    for( int xx=1; xx < TABCOUNT; ++xx)
    {
        delete tab[xx];
        tab[xx] = nullptr;
    }
    ui->tabWidget->setCurrentIndex(0);

    return;
}

/*************************************************************************/

/* These cannot be [4] although long might be 4 because memory boundary
 * is not obeyed when you do *ptL = 2322L (for example)(bug fix Mar2023)
 * I think I can move these to local now that I understand this.
 */
static char ptLchar[20];
static char ptUchar[20];
static unsigned *ptU, uu;

struct takeoff {
    char ch[40];
    char ref[2];
    unsigned ext;
    float tot;  // 4 bytes long
}; //48 bytes
static struct takeoff take1, *take2;

#define NEWtabsFeatureFeb2024
/* I added Tabs as a feature because ESTROWCOUNT
 * had to be set to 231, whereas the old EstimateMaker-1987 was 800 lines
 * of estimate. The new GUI library is sluggish with too many
 * QLineEdit widgets for each cell.
 * Search for this line of code "if( block < ESTROWCOUNT )"
 * It cuts off reading long estimates.
 *
 * If the estimate is long then the NEWtabsFeatureFeb2024
 * will accomodate and not quit because all the data is
 * read into tabsPage[TABCOUNT] before putting ONE tabPage[] into
 * the QLineEdits which you see in the code as estRow[]->.
 *
 * The old version reads/writes estimate directly to/from the QLineEdits.
 *
 * I left the old version estimateRead() estimateWrite()
 * after the #else since it is filename compatible but truncates estimates
 * and if there is a bug in the new functions, I will revert
 * to see if it is a new or old problem so I know to look in Tabs
 * coding or not. Tabs adds to bug hunt complexity since it is
 * paralleling the QLineEdits data, which I never wanted to do
 * although I considered such, before I decided on the
 * simpler design pre-Feb2024.
 */

#ifdef NEWtabsFeatureFeb2024
bool MainWindow::estimateRead( QString *filename )
{
    bool ret = false;
    qint64 er,readsz,xc;
    long takelg,lg, maxest,*ptL, block, filelength;
    //int qdec,mdec,tdec,ttdec,tmdec;
    char startofmem[81], *spare=startofmem;
    long lastSubtotalBlock = -1L;
    QChar takeoffflag[ESTROWCOUNT * (TABCOUNT-1)]; // this is a list of which rows have Qty details

    //ui->tabWidget->setCurrentIndex(0); Do not do this because clearEst() does this.
    ui->tabWidget->setEnabled(false);
    //turn off user mouse clicks while we copy data into Tabs

    if( oneHeader == nullptr )
        oneHeader = new DialogHeaderOfEstimate(this);

    // do some error checking and leave the estimate not cleared; clearEst()
    if( oneHeader->estimateReadHeader( filename ) == false )
    {
        ;//qDebugA() << "failed to read" << *filename;

        ui->tabWidget->setEnabled(true);
        return(ret);
    }
    else
    {
        // clear and reload now that we know there are no basic errors
        clearEst();
        oneHeader->estimateReadHeader( filename );
    }

    QFile fp(*filename);

    /* This is the original 1987 C program. I am changing the
       destination memory but using the same names.
       I want to get the file read()s and offsets correct so I maintain
       the old estimate files as input for testing in Dec 2022; DavidP.
     */

    ;//qDebugA() << "again, filename:" << *filename;

    if( fp.open(QIODeviceBase::ReadOnly) == false)
    {
        ui->tabWidget->setEnabled(true);
        return(ret);
    }

    filelength = fp.bytesAvailable();

    // 407 bytes are header of estimate, plus 4 bytes of long, totaling 411 bytes.
    if( !fp.seek(407L) )  //2022 ctrl-Q "takeoffs" start here.
    {
        ;//qDebugA() <<"Error: Cannot seek to 407, near end of header";

        fp.close();
        ui->tabWidget->setEnabled(true);

        return(ret);
    }

    if( fp.read( ptLchar, 4) != 4 )
    {
        ;//qDebugA() << "failed to read 4 chars for long calc.";
        fp.close();
        ui->tabWidget->setEnabled(true);

        return(false);
    }

    ptL=(long *)ptLchar;
    takelg = *ptL;
    ;//qDebugA() << "estread ctrl-Q calc area, size in bytes:"<<takelg<<"takeoff-lines:" << takelg/48L;
    if( filelength < takelg )
    {
        ;//qDebugA() << "The crtl-Q size is bigger than the filelength:" <<filelength;
        ;//qDebugA() << "Quit now \n and then exit/reload the program fixes it (bug) Mar2023.";
        fp.close();
        ui->tabWidget->setEnabled(true);

        return(false);
    }

    /* Some old 1987 NOTES about Data File Format for Estimate Maker
     * I will maintain this format in the year 2022:

     * MSDOS est memory format is the same 80 char length as
     * the disk file for each estimate line and no format difference.
     *
     * This is the description which I found in the 1987
     * C program:
     *
     * byte 1 indicates if it has a QTY-take-off item (byte='\2')
     *        or has a regular quantity item (byte='\1')
     * (edit 2022: I am calling it "QChar takeoff" in the struct below)
     *
     *	49 bytes screen area
     *	8 bytes are 2 float nos. to store totals
     *	all 58 above is in MSDOS main memory ds seg (segds)
     *  all below is ext (seg est22)
     *
     *	12 bytes of additional description
     *      (Edit me 2022: unused, I am pretty sure. Future use.)
     *	2 bytes unsigned segment address of 1st take-off item
     *          (is offset 70, but important on disk only
     *          if it 0 or not-0 because memory targets change
     *          every time the program
     *          loads a filename.
     *          So 2022 we use 0 or not-0 to, not-read or read,
     *          respectively, the qtymath box 17 max lines.
     *	2 bytes unused
     *	2 bytes of refcode  is offset 74 and is important to know on
     *              disk also, regarding which library list sourced
     *              the estimate line.
     *              Library names were *03. Only 03 meant refcode.
     *              and the asterisk only meant is was a listname, not a
     *              library line code name, which is 3 chars.
     *	4 bytes unused      is offset 76+4=80 chars total
     */

    struct estimateold {
        float totalhrs;
        float totalcost;
        unsigned memtake;
        QChar takeoff;
        char libcode[4];  // library 3 char code library line, if any; sometimes blank.
        char description[65];// this is actually 24+1 chars below, I see.
        char quantity[8];
        char hours[8];
        char cost[8];
        char refcode[3]; // tells which library list, like *AA was the source.
            // "AA\0" is three chars. Asterisk is omitted.
            // It is not the library code. It is the list name flag "*".
    }est;

    //QChar takeoffflag[ESTROWCOUNT * (TABCOUNT-1)]; // this is a list of which ones have Qty details
        // which we oddly call 'takeoffs' 1=no, 2=yes.

    /* 2022 note: parsing the description was complex because the
     * old estimate maker displayed the description two ways,
     * big and small, and stored the big end into a different
     * memory area, external to msdos to squeeze more ram into use.
     * Here is estline() clipped out:
     * 	if(big)
    {
    putsn("|",yy,29,1);
    putsn(pt+1,yy,30,16);
        uu=ramseges;
    ramseges=est22;
    xx=((pt-startofmem)/58)*22;
    extputsn(xx,yy,46,12);
    ramseges=uu;
    putsn(" ",yy,58,1);
    putsn(pt+30,yy,59,20);
    }
    else
    {
    putsn("|",yy,29,1);
    putsn(pt+1,yy,30,49);
    }
    I am only interested in 'big'. putsn last arg is number of chars.
    Previous to that are row,col on the VGA monitor. First arg is a (char *), eg."hello".
    So I have to see how that was read from disk: esttostr() into memory
    and then interpret the above, which is the actual display on screen
    and piece it together in the same coherence, which I do next.
    The putsn are chars size: 1,16,12,1,20 = 50 chars, starting near centre of display.
    */

    /*
      The answer to the puzzle was simple. The disk storage is already
      coherent and in order. Ignore all the above memory tucking.
      Simply parse the first 5 as "|",libcode, " ",
      and then description through to qty,cost,hrs is the next
      45 chars = 50 chars total.
    */

    /* Why do I fp.seek() ahead and read the estimate lines first?
     * Why do I fp.seek() back when I could have done no seeks if I read
     * qtymath (takoffs) first?
     * Answer: The estimate lines have flags at the beginning of each line
     * which say yes or no to whether they have qtymath (takeoffs)
     * Those qtymath boxes have to be assigned to the correct est lines
     * which have the yes/no flags. So I read the est lines, and then
     * make a list qtymath flagged est lines, and then fp.seek() back
     * and read the takeoffs until each block hits a 00 marker.
     * Mar 6, 2023. It could have been done by saving the qtymathHidden
     * blocks in a sequence and assigning them as the est lines were read
     * and then the disk read order would be more natural.
     *
     * However, I wrote the code for est lines first because it was
     * the most important and interesting, starting Dec 1, 2022, and
     * so the logic fell out of my head to jump back and fetch, as
     * the next step a month later.
     */
    lg = filelength - takelg - 411L;  // 2022 version
    //lg = filename->filelength(fileno(fp))-takelg-411L; /*estlist size  1987*/

    if( !fp.seek( takelg + 411L ) )  //2022
    {
        fp.close();
        ui->tabWidget->setEnabled(true);

        return(ret);
    }

    maxest = lg/80L;

    /* This is if maxest==0 and then none would be new tabPage(this); */
    if( tab[1] == nullptr )
        tab[1] = new tabPage(this);

    for( block=0,lg=maxest; lg>0; --lg)
    {
        ++block; // estimate rows are 1 to 231 for example, not 0 to 230.

        if( fp.read(spare,80) != 80 )
        {
            ;//qDebugA() << "failed to read the next 80 chars. That should not happen because we checked the filelength before starting this.";
            break;
        }
        takeoffflag[block] = est.takeoff = *spare;
        sprintf( est.libcode, "%.3s", spare+1 );
        sprintf( est.description, "%.24s", spare+5 );

        // I think offset spare+58 is on the disk but is unused; for future use.
        // I see it is not printed to screen in the if(big) C code snippet.
        //sprintf( est.description, "%.49s<>%.12s<", spare+1, spare+58 );

        sprintf( est.quantity, "%.7s", spare+29 ); // I triple checked
        sprintf( est.hours, "%.5s", spare+37 );     // these sizes and offsets
        sprintf( est.cost, "%.5s", spare+43 );

        float *fl;
        fl = (float *)(spare+50);
        est.totalhrs  = *fl;
        fl = (float *)(spare+54);
        est.totalcost  = *fl;

        sprintf( est.refcode, "%.2s", spare+74 );
        /* this was not displayed for each est line in big, above,
         * but was in the header as you scrolled down/up.
         * "0 " means main one, maybe. "WD" was in BIGTRADE
         * and I saw an est.lib/WD filename to match, and est.lib/0.
         */

        /* ;//qDebugA() << est;  raw data has non-printable chars,
         * so skip 1 char The skipped char is the flag to
         * indicate ctrl-Q; takeoffs == char '\2'
         * for the estimate line; ie. detailed Quantity
         * calculations. NOT==char '\1'
         */

        ;//qDebugA() << est.libcode << est.description << est.quantity<<est.hours<<est.cost<< '|' << est.totalhrs << '|' << est.totalcost<< '|' << est.refcode<<'|' << block<<'|';
        if( block <= (ESTROWCOUNT-1)*(TABCOUNT-1) )
            ;
        else
        {
            ;//qDebugA() << "estimateRead() lines is longer than" << ESTROWCOUNT-1 << "\nprobably because a longer estimate was made with a different version of PaceMakerEstimator.";
            break;
        }

        /* The next are floats for fast summing. We can ignore
         * those because we are going to use doubles or bigdecimal()
         * for recalculating everything, instead of floats.
         * They seem to be failing so far anyway because
         * maybe float * to a char string is not on a correct
         *  4 byte boundary or it is a differing float system
         * from Desmet compiler 1986.
         *
         * The memory offset of the first takoff item is
         * unused by us because we are not using MSDOS memory.
         * So skip it too.
         */

        /* "block" is in a for() of 1 to maxest.
         * The coding before Tabs was 1 to 231 rows: estRow[block].setText()
         * The new counts are 1 to 10 tabs of 1 to 231 rows
         * and the math converting block into that outcome,
         * where block is now 1 to 2310 rows, is thus:
         */
        if( tab[1+((block-1)/(ESTROWCOUNT-1))] == nullptr )
            tab[1+((block-1)/(ESTROWCOUNT-1))] = new tabPage(this);
        ;//qDebugA( << "\n1+((block-1)%(ESTROWCOUNT-1))  1+((block-1)/(ESTROWCOUNT-1))" << 1+((block-1)%(ESTROWCOUNT-1)) << 1+((block-1)/(ESTROWCOUNT-1));
        ;//qDebugA() << block << 1+((block-1)%(ESTROWCOUNT-1)) << 1+((block-1)/(ESTROWCOUNT-1));

        QString sslibcode;
        sslibcode = QString("%1").arg(est.libcode,-2,' ');// must be "=  " not "=" for example. I wondered why -2 and not -3. Answer: -2 is the minimum, not the maximum. So 3 happens if libcode is 3chars.
        if( sslibcode.simplified().isEmpty() ) // I want isEmpty() to work later, so "", not " ", "  ", not "   "
            sslibcode = "";

        tab[1+((block-1)/(ESTROWCOUNT-1))]->estTabRow[1+((block-1)%(ESTROWCOUNT-1))]->col[1] = sslibcode;
        tab[1+((block-1)/(ESTROWCOUNT-1))]->estTabRow[1+((block-1)%(ESTROWCOUNT-1))]->col[2] = stripEnd(est.description);
        tab[1+((block-1)/(ESTROWCOUNT-1))]->estTabRow[1+((block-1)%(ESTROWCOUNT-1))]->col[3] = stripEnd(est.quantity);
        tab[1+((block-1)/(ESTROWCOUNT-1))]->estTabRow[1+((block-1)%(ESTROWCOUNT-1))]->col[4] = stripEnd(est.hours);
        tab[1+((block-1)/(ESTROWCOUNT-1))]->estTabRow[1+((block-1)%(ESTROWCOUNT-1))]->col[5] = stripEnd(est.cost);

// I moved this code block to #ifndef CALC_THEONELINES below
#ifdef HAStobeMovedOutSideOfTabsLoopBecauseFunctionsAreQLineEditcoding
        // New March 2023. "===" libcodes are subtotal/total text areas
        if( est.libcode[0] == '=' )
        {
            QString ss = estRow[block]->col[1]->text();
            // update the subtotal total since the file has no sums nor multiplication
            if( ss.indexOf( uniqueID_TOTALSblock ) == 0 ) // end of subtotal grouping
            {
                bottomSubtotalRow = ESTROWCOUNT-1; // prevent TOTALS creation
                insertTotalSubtotal( block );
            }
            lastSubtotalBlock = block;
        }
        else if( *ss.data() != '=' )
            calc_one_estline( block ); // this uses double-float numbers
#endif

        QString ss = QString("%1").arg( est.refcode,-2,' ');
        tab[1+((block-1)/(ESTROWCOUNT-1))]->estTabRow[1+((block-1)%(ESTROWCOUNT-1))]
            ->refcode = ss.trimmed();
    }
    if( block != maxest )
    {
        fp.close();
        ;//qDebugA() << "Should NOT ever happen. block=" << block << "should equal maxest, not this:" << maxest;
        ui->tabWidget->setEnabled(true);

        return(false);
    }
    ;//qDebugA() << "\n**********\nReading " << maxest << " estimate lines succeeded.\nNext read the ctrl-Q details, if any.\n";

#ifdef SKIP1987
    if(!loadtake(fp,takelg))
        er=0;
    ----------------
                         loadtake(fp,lg)
                         FILE *fp;
    long lg;
    {
#endif
        /* I copied old 1987 loadtake() into here, as follows:
          unsigned wh,*ptu,dummy;
          char exttakebegin[], *exttake;

       ctrl-Q Qty calculation windows are called qtymath, 2022AD ver..
       One of them: 48 bytes each in memory malloc() as needed, I guess.

        That ext WAS a memory pointer to the next one. If zero then
        that is end of list.
        The value is meaningless in the file. In the file it only
        matters if the first char of the estimate line was a \1 or \2.
        That means I can monitor that spot as the quit and go to
        next estimate line
        while I am loading the BIGTRADE filename into the program.

???When I create a BIGTRADE file I must store non-zero
        unsigned values in that spot until end of list of ctrl-Qty
        items for that estimate line; last one is 0 ?
???I think write ALL 0s.

        Library has takeoffs?
           Yes, but I looked in EST1A.C loderef() (a deliberate
           misspelling to distinguish it from old code which was loadref)

           The struct is simply 3 char library code
                        and a pointer, *refpt,
                        to a 28 char description.

           The library takeoffs (now named qtymath) are different
           library filenames. Read detail at the function which
           does library filename parsing.
           17 lines of qtymath each have QEditLine of 40 chars,
                                     and QEditLine of 7 chars,
           similar to old code "struct takeoff{} take;
                    has ch[40]; and
                    has float tot;
           for each line.

           So maximum 17 "take" memory allocations had a char *refpt
           pointing at the first line which is one
           "struct takeoff( } take" and then the next line of 16 more
           may or may not exist so as to limit memory use.
           I think a pointer points at the next line in memory,
           if the pointer is not null.

           HOWEVER, I notice the filesize of est.lib/B.B08 as a qtymath
           item is always 680=17*40. That means only the first indicator
           on the disk matters to use. If it is non-zero then
           read 680 bytes. We are not going to use memory pointers
           after that data is read into our object, qtymath.
           Our coding in 2022 is Qt objects like QLineEdit,
           not pointers to raw memory.

           The confusing part to me was that there can be many of these
           680 char blocks depending on how many estimate lines
           set a ctrl-Q qty detail which we will now call qtymath object.
           Why are they all in sequence on the disk?
           How do you match them up with each estimate line?
           The answer is you wrote all the estimate lines first
           and then you went back in memory and wrote all the
           680 byte-blocks. So when we read, we read the estimate
           lines first, as above, and then we run through the estimate
           sequentially looking for a flag which as of Feb 27, 2023
           does not exist for each estimate line.

           I will probably name it est[row]->qtymath in
           public mainwindow.h available for all estimate lines.
           It will be a QTYMATH *qtymath = new QTYMATH in mainwindow.h
           public. Also see qtymath.h and qtymath.cpp.
           It will be nullptr if the estimate line has no details,
           which in the old says we called up with keystroke ctrl-Q
           and flagged yes/no by using the first byte in the est line.
           So although we will use est[row]->qtymath, we must
           also set the 1st char of estimate line as
           non-zero '\1' = none or '\2' = yes, on the disk file during
           estimate to disk writes.
    See bool MainWindow::estimateWrite( QString *filename )


            #define RIZE 300

            struct stref {
                    char refch[3];
                    char *refpt;
                    }ref[RIZE];

            struct takeoff {
                char ch[40];
                char ref[2];
                unsigned ext;
                float tot;
                }take;

            I think the RSIZE must mean there was a limit of 300
            ctrl-Q used in one estimate; ie. 300 lines can have
            Qty like this although an estimate could have had 640+ lines
            depending on msdos memory available and so some
            were plain non-ctrl-Q lines.

            The 2022 version will have a limit equal to the
            number of maximum estimate lines, instead of less than.
            Where do we put the estimate line pointer to the
            ctrl-qty? I am still not sure where it was in the 1987 ver..
            I found it. See above, offset 70 in each estimate line:
            "     *	2 bytes unsigned segment address of 1st take-off item
                      (is offset 70, "
            We will not use that pointer to memory.
???Do we have to
            write it to maintain old code compatibility? I think not,
            because it was just a convenience to keep the data
            structures less confusing; ie. memory matched disk.
???I must test this theory.
            I still have the old msdos program.

            Also see est[row]->qtymath in mainwindow.h
            which I mentioned in more notes above, in this function.

            And:
        2 bytes of ref is offset 74 and is important to know on
            disk also, regarding which library list/line sourced
            the estimate line because the qtymath is
            archived in the library as separate filenames
            so that math can be used in other new estimates. wow!

            Do we write the new est.lib/B.B08 in the qtymath object?
            No. It was used for F5 read/ F6 save while in the Qty box.
            It was for future estimate line use, like a library item.

            This will change the library but have no effect on
            historical estimates because they have their own copies.

            It makes sense because a type of brick, for example,
            would have the same qty math.
            The filenames were X.B08 or XX.B10 where B08 for the
            line from the library list, which was called X or XX.
            Those library names are 1 or 2 characters coded as ref[2]
            in the old 1987 program.

            B08 and B10 are the estimate line or library
            line 3 char codes column one: est[row]->col[1]
                                          lib[row]->col[1]
            in mainwindow.h

            Those ctr-qty "takeoffs" library files are always
            exactly 680 chars, which is 17x40=680.

            The same qty takeoffs in the estimate filename is stored
            as 48 chars, length, not 40; see the takeoff structure above.

            The number of lines of 48 bytes used take.ext==0 as the
            no-next-line indicator. So 48 bytes could be 1x up to 17x in the file.

            You cannot just read a block of 680 chars off from the
            estimate file if a takeoff (qtymath) is indicated
            for an est line. That is 40 chars per line x 17 = 680.

            I did fast code in 1987 for slow 8088 CPUs
            and a lot of features were accomplished.
            It is too bad I had to quit to work for a medical lab
            because revenues were insufficient. Small contractors were
            not ready to buy computers and software. I picked
            the wrong niche. Only large contractors would have worked.
            My coding example sold myself as a capable programmer though.


        struct takeoff {
            char ch[40];
            char ref[2];
            unsigned ext;
            float tot;  // 4 bytes long
            }take; //48 bytes

    // This for() reads the file from after the header 411
    // into external 1987 msdos memory.
    // and the next for() assigns all the est[] pointers to point
    // at all the spots if any have a '\2' char which indicates
    // a takeoff ctrl-Q Qty. '\1' means none.
    //
    // Why are we reading it from two spots, out of order?
    // The takeoffs are after the 411 bytes of header.
    // The estlines are after the 411header + takeoff area -> here tail.
// So the tail of the file is read first, to find out about the \2 and \1.
// Next the earlier part is read, which is the takeoffs which are assigned
// in order, without any knowledge of which est[] line they belong to,
// except for the fact they were saved in order. So read them in order.

// I look at the BIGTRADE file. They are fixed @48 chars, variable
// number of these 48s, is the correct answer.
// Position 42 *unsigned ext is 0 signaling terminate after a bunch of
// them, indicating go to next est[] line.

     * I learned all of this from ESTOFFA.C storetake() in the 1987 code.
     * Here is an important item for when we store an estimate.
     * This code is writing a 4 char size long-int at 407 offset
     * which tells the 1x48 thru 17x48 number of qty lines.
     * That xc*48 is not for one estline qty-detail.
     * It is for ALL qty-details and the separation of each
     * one for each est line is the many take.ext==0 terminators.
     *
        fseek(fp,407L,(int)0);
        ptL=(long *)spare;
        *ptL=(long)xc*48L;
        if(fwrite(spare,sizeof(char),(int)4,fp)!=4)
                return(0);

        loadest() and loadtake() inside EST1A.C
        read that 4 byte long and use it to read the
        full block of ALL qtymath lines and
        afterwards distribute them to their
        corresponding est lines.
     */

        if( !fp.seek(411L) )  //2022 ctrl-Q "takeoffs" start here.
        {
            ;//qDebugA() <<"Error: Cannot seek to 411, near end of header";

            fp.close();
            ui->tabWidget->setEnabled(true);

            return(ret);
        }
        // takelg was already read far above. It is takeoff bytes total.
        if( takelg > 0 )
        {
            char *pt;
            char *pttakeoff = (char *)malloc( takelg );
            if( fp.read( pttakeoff,takelg ) != takelg )
            {
                free(pttakeoff);

                ;//qDebugA() <<"Error: Cannot read qtymath details.";

                fp.close();
                ui->tabWidget->setEnabled(true);

                return(ret);
            }

            pt = pttakeoff;
            /*
             * xc=1 is our estRow[1] variable in mainwindow.h
               xc=2 is our estRow[2] variable in mainwindow.h
               The takeoffflag[xc] got set above and
                  starts at xc=1,
                    not xc=0 because we do not use row=0 in estRow[row]
             */
            for(xc=1;xc<=maxest;++xc)
            {
                if( takeoffflag[xc] == '\2' )// THIS IS THE TAKE-OFF ctrl-Q flag
                {
                    tab[1+((xc-1)/(ESTROWCOUNT-1))]->estTabRow[1+((xc-1)%(ESTROWCOUNT-1))]
                        ->fileQtyBlock = new qtymathHidden();
                    ;//qDebugA() << "\n1+((xc-1)%(ESTROWCOUNT-1)) 1+((xc-1)/(ESTROWCOUNT-1))" << 1+((xc-1)%(ESTROWCOUNT-1)) << 1+((xc-1)/(ESTROWCOUNT-1));

                /* this is '\1' or '\2' in est line first char 1987 coding
                   but it is '\0' or a non-Zero pointer fileQtyBlock

                   new qtymathHidden()
                 */

                /* delete qtymath1;
                 * Do the deletes when closing the estimate
                     or else it is a memory leak!
                 *
                   Mar2, 2023: done! See here for deletes.
                   void qtymath::on_deleteButton_clicked()

                   Also, I delete qtymathHidden in clearLibEst.cpp
                 */
                    for( unsigned indx=0; takelg >= 48; takelg -= 48L, ++indx )
                    {
                        strncpy( spare, pt, 40);
                        *(spare + 40) = '\0';

                        QString str = str.fromStdString( spare );

                        take2 = (struct takeoff *)pt;

                        ;//qDebugA() << "take2.ch:" << str << "row:" << xc << "take2.ref:|" << take2->ref[0] << take2->ref[1] << "|\n";

                        if( indx < 17 )
                        {
                            tab[1+((xc-1)/(ESTROWCOUNT-1))]->estTabRow[1+((xc-1)%(ESTROWCOUNT-1))]->fileQtyBlock
                                ->block[indx] = str.trimmed();
                        }

                        //if( take2->ext == 0 ) // does not work

                        //maybe this technique will work like above
                        // ptL=(long *)ptLchar;
                        //takelg = *ptL;

                        /* YES IT DOES WORK! How will I remember this trick?
                     * I think the char string has to be at the start,
                     * not the middle of what you are going to point at as the
                     * unsigned or float or double, et cetera.
                     * Maybe the middle (not the start) is randomly misaligned
                     * in operating system memory which is an issue for numerics.
                     * Make char[20], not char[4] for misalignment room.
                     */

                        ptUchar[0] = *(pt+42);
                        ptUchar[1] = *(pt+43);
                        ptU=(unsigned *)ptUchar;
                        unsigned uu = *ptU;
                        if( uu == 0 )
                        //if( *(pt+43) == 0 ) <this works but is very bad because 1char.
                        {
                            pt+=48;
                            break;
                        }
                        pt+=48;
                    }
                }
            }
            free(pttakeoff);
        }

        ret = true;

        fp.close();

    /* I do AsPromised: estimateRead() will copy all estimate filename
     * into tabPages and then do a load to
     * tab[indexOld].loadTabtoQLineEdits()
     * before handing control to the user to do work on the QLineEdits.
     * We use tab[1] because Tab 1 is start up default
     * and set in clearEst() no matter what the user selected.
     */
    ;//qDebugA() << "did we get this far?";
    for(int xx=TABCOUNT-1;xx>=1;--xx)
    {
        ;//qDebugA() << "did we get xx" << xx;

        /* calc_one_estline() has to be done on each line.
         * However, I could have written, but did not write, a estTabRow version
         * instead of using estRow version of calc_one_estline()
         * and left this code block up above. But the second function,
         * insertTotalSubtotal(), was too tedious and error prone
         * to write a second version of it. So I do the work
         * in QLineEditArray instead of in tabPages although
         * the data is the same.
         * Find: HAStobeMovedOutSideOfTabsLoopBecauseFunctionsAreQLineEditcoding
         *
         * Doing the subtotals is necessary too.
         * The loop does Tab 10 down to Tab 2
         * before returning control to the user ending with QLineEditArray
         * loaded with Tab 1.
         * tabPages are all the same size to fit the one QLineEditArray.
         *
         */
        if( tab[xx] != nullptr )
        {
            /* some tab[] are empty, but xx==1 is never empty
             * since it is the first page of all estimates.
             */
            tab[xx]->loadTabtoQLineEdits();

#ifndef CALC_THEONELINES
            for( block=1; block<=ESTROWCOUNT-1; ++block)
            {
                // New March 2023. "===" libcodes are subtotal/total text areas
                QString ss = estRow[block]->col[1]->text();

                if( ss.indexOf( uniqueID_TOTALSblock ) == 0 )
                // detects both is1987 "=  " first line,
                //    and new subtotal blocks "=  " 6th line
                {
                    bottomSubtotalRow = ESTROWCOUNT-1; // prevent TOTALS creation
                    if( insertTotalSubtotal( block ) == false )
                        break;
                    lastSubtotalBlock = block;
                }
                else
                    calc_one_estline( block ); // this uses double-float numbers
            }
            tab[xx]->unloadQLineEditsToTab();
#endif
        }
    }
    ;//qDebugA() << "did we get loadTabto" ;

#ifdef AsPromised
    if( tab[1] != nullptr )
    tab[1]->loadTabtoQLineEdits(); // AsPromised, user is given control at
                                   // active Tab and loadTabtoQLineEdits is
                                   // this last step or else QLineEditArray
                                   // is not the active Tab data when user gains
                                   // control in QLineEdits realm.
#endif

    /* Next, I put this below in, although I did all the
     * insertTotalSubtotal( block ); above, because the TOTAL in the
     * description is wrong until I did a recalculation of all of them from
     * line 1 to the end by using this call, which is also used elsewhere to do stupid slow
     * but accurate. I almost gave up looking for why TOTAL is overwritten by DRAG and gets the
     * right number but gets it wrong when I
     * insertTotalSubtotal() here, which is THE SAME AS DRAG!
     * Edit: I think this is the problem which prevents TOTAL: if("row>=bottomSubtotalRow)")
     * Am I doing it on the wrong block? (block is row)
     * NO. Because I looked at DRAG and it is on ANY "===" line.
     * I do note that DRAG is always on a FULL set of subtotals-block.
     * I am doing it on an incomplete block, because I did it on "=   SUBTOTALS" which is unique.
     * whereas bottomSubtotalRow = row; is set on the last line "===                ---- ----"
     *
     * I thought of a trick to simply set   bottomSubtotalRow = block;
     * up above (above is 0) instead of letting it be set inside by
     * insertTotalSubtotal(). Wow! Let us try it:
     * Edit: The trick bottomSubtotalRow = 0 worked.
     */
    ;//qDebugA() << "did we get pre insertTotalSubt" ;

        bottomSubtotalRow = 0L; // force TOTALS creation by assuring this is top of estimate.
        insertTotalSubtotal( lastSubtotalBlock ); // last one always needs a TOTAL
        //searchForSubtotalTotalUpdateThem(1);  // <--this works but is slower by redoing them all.

        // scroll up to top so user sees estimate. He might be on line 230.
        scroll_est_into_view( estRow[1]->col[2] );

        ui->tabWidget->setEnabled(true);

        return( ret );
    }



/****************************************************************/
    /* See the sister function: estimateRead()
     * for detailed notes about the file structure.
     * I wrote notes while I was doing the read from disk function
        of the old 1987 Desmet C language code.
     */
    char howmanyLongChar[20];  //not [4]
    long howmany=0L, *howmanyLongpt;

    char totFloatChar[20]; //not [4]
    float totFloat, *totFloatpt;

    bool MainWindow::estimateWrite( QString *filename )
    {
        bool ret = false;
        qint64 er,readsz,xc;
        unsigned block, yy;
        long takelg, lg, maxest,*ptL, xx;
        //int qdec,mdec,tdec,ttdec,tmdec;
        char *buf, buffer[81];
        buf = buffer;

    /* Here is a very difficult bug which I had to solve:
     * The "estimateChanged" variable was always switching back to
     * true AFTER I saved the estimate and meticulously set it to false.
     * Also, an estimate was maybe missing the calculation of a line, which
     * was done after saving. I never saw that bug, but it is the same bug.
     *
     * These problems occurred because docalc() only gets called upon
     * focusOut, which is in events... .cpp. When you press Save or Open or New
     * buttons that causes a focusOut event which is processed
     * AFTER estimateWrite() for some reason and that docalc() changes
     * the estimateChanged flag back to true after I set it false here.
     *
     * So here we must turn off the keyboard cursor
     * before estimateWrite() and before setting estimateChanged false
     * so that the keyboard cursor stops receiving FocusIn and FocusOut.
     * I think the msgBox OK button returns focus to the keyboard cursor
     * and generates the extra FocusOuts. Perhaps it is not the order of
     * events that is surprising but the extra FocusOut.
     * Skip more thinking. Let's do some solutions without determining
     * those two possibilities.
     */
    /* Solution 1: Call the same function as clearSlider button because I know that turns off
     * keyboard row and col before continuing to write the estimate:
     * on_clearSlider_pressed();
     * THAT FAILED because I see the cursor blinking on the same est line after doing that idea.
     * Solution 2: How to turn off cursor? I know I am correct about this because I manually
     * turned off the cursor before using the Open button and the estimateWrite was done first
     * because estimateChanged was true and then I pressed CANCEL and Open again
     * and estimateChanged remained set to false and estimateWrite was not redone!
     * ie. the FocusOut did not do another docalc() and did not set estimateChanged to true.
     * So how to stop the blinking cursor programmatically?
     * Doing it with the mouse is obvious: click on any NOT-QLineEdit widget.
     * https://stackoverflow.com/questions/35762856/how-can-i-move-the-keyboard-cursor-focus-to-a-qlineedit
     *  has:
QPoint pos(line_edit->width()-5, 5);
QMouseEvent e(QEvent::MouseButtonPress, pos, Qt::LeftButton, Qt::LeftButton, 0);
qApp->sendEvent(line_edit, &e);
QMouseEvent f(QEvent::MouseButtonRelease, pos, Qt::LeftButton, Qt::LeftButton, 0);
qApp->sendEvent(line_edit, &f);
or
https://stackoverflow.com/questions/526761/set-qlineedit-focus-in-qt
QTimer::singleShot(0, line, SLOT(setFocus()));
     */
        QTimer::singleShot(0, ui->clearSlider, SLOT(setFocus())); // docalc() got called, presumably.
            // See eventMouseKeyboard.cpp keyboard FocusOut.
        // Wow, that worked! The blinking cursor is not on the QEditLine.
        // January 23, 2024. And clearSlider was not clicked.

    /* Write the estimate header first. It does its own fp.open/fp.close
     * because it is also called from the dialogheader... .
     */

        ui->tabWidget->setEnabled(false);
        linuxMsWindowsSlashFixes( filename );

        if( oneHeader != nullptr )
        {
            ret = oneHeader->estimateWriteHeader( filename );
            if( ret == false )
            {
                //QMessageBox msgBox;
                //msgBox.setText("Estimate header failed to write to disk;\nOut of disk space?");
                //msgBox.exec();
                ui->tabWidget->setEnabled(true);

                return(ret);
            }
        }
        else
        {
            ;
      /*Jan 23, 2024. I think this should quit and return failure if there is no header.
       *I do not know if there is a reason to continue saving the body so I won't
       *insert a return(ret); here for now.
       *The answer is likely that there is always a oneHeader so putting a return here
       *would never happen anyway.
       */
        }

        QFile fp(*filename);

        if( fp.open(QIODeviceBase::ReadWrite) == false)
        {
            ui->tabWidget->setEnabled(true);

            return(ret);
        }

    /* write the qtymathHidden qtymath details for all est lines
     * before the estimate lines are written.
     * 1987 C coding called qtymath takeoffs.
     */

#define TAKEOFFS
#ifdef TAKEOFFS
        if( !fp.seek( 411L) )
        {
            fp.close();
            QMessageBox msgBox;
            msgBox.setText("fp.seek failed inside estimateWrite");
            msgBox.exec();
            ui->tabWidget->setEnabled(true);

            return(ret);
        }

        qtymath *qtytemp = new qtymath( this );

        int li,cnt;
        QString fill40spaces;
        fill40spaces.fill( ' ', 40 );
        howmany = 0L;
        qtymathHidden *fqb;

        /* This is trickier than I thought, for Tabs.
         * Do I know which Tab QLineEditArray is on?
         * I need to unload its contents into its correct Tab
         * before writing the estimate or else the user's current
         * QLineEdits will be lost. Oh yes, indexOld.
         */
        if( tab[indexOld] == nullptr ) // be sure one is allocated
            tab[indexOld] = new tabPage(this);

        tab[indexOld]->unloadQLineEditsToTab();
        /* Next all the Tabs, 1 through 10, are accurate data
         * but QLineEdits is in an unloaded state which must be corrected
         * AsPromised, before "return(ret);".
         */

    for( int tt=1; tt<TABCOUNT; ++tt)
    {
        if( tab[tt]==nullptr ) // skip unallocated, blank ones
            continue;

        for( xc=1; xc < ESTROWCOUNT; ++xc )
        {
            fqb = tab[tt]->estTabRow[xc]->fileQtyBlock;
            if( fqb != nullptr )
            {
                ;//qDebugA() << "\nXCxcxcxcxcxc" << xc;

                /* first find the last non-blank line
                 * because we are not writing end blanks. */
                for( li = 17; --li>=0; )
                {
                    if( fqb->block[li].isEmpty() )
                        ;
                    // strip the prefix and suffix blanks if any.
                    else if( fqb->block[li].trimmed().isEmpty() )
                        ;
                    else
                    {
                        break;
                    }
                }
                /* Now we know li is the last not-Empty block line.
                 * write from 0 up to and including li
                 * and then we will quit writing. If li==0 is blank then
                 * we will have written one blank line and no more.
                 */
                ;//qDebugA() << "Number of lines to write is:" << li+1 <<"|";

                /* Write this line to the estimate in this format:
                 * struct takeoff {
                    char ch[40];
                    char ref[2];
                    unsigned ext;
                    float tot;  // 4 bytes long
                    }; //48 bytes
                    static struct takeoff take1;
                 */

                for( cnt=0; cnt<=li; ++cnt)
                {
                    //take1.ch, "%.40s",

                        QString str;

                    // left justified and padded out with blanks to 40 chars.
                    if( fqb->block[cnt].isEmpty() )
                        str.fill( ' ', 40 );
                    else
                        str = QString("%1").arg( fqb->block[cnt], -40, ' ');

                    if( fp.write( str.toStdString().c_str(), 40) != 40 )
                    {
                        ;//qDebugA() << "failed to write the next 40 chars.";
                        goto HARDEXIT;
                    }
                    ;//qDebugA() << "\nWrote 40. cnt=" << cnt << "|" << str <<"|";

                    //TEMPORARY     take1.ref[0] = 'Z';
                    take1.ref[1] = 'Z';
                    if( fp.write( "ZZ", 2) != 2 )
                    {
                        ;//qDebugA() << "failed to write the next 2 chars.";
                        goto HARDEXIT;
                    }
                    if(cnt==li) // last line
                    {
                        if( fp.write( "\0", 1) != 1 )
                        {
                            ;//qDebugA() << "failed to write the next 1 chars.";
                            goto HARDEXIT;
                        }
                        if( fp.write( "\0", 1) != 1 )
                        {
                            ;//qDebugA() << "failed to write the next 1 chars.";
                            goto HARDEXIT;
                        }
                    }
                    else
                    {
                        if( fp.write( "\1", 1) != 1 )
                        {
                            ;//qDebugA() << "failed to write the next 1 chars.";
                            goto HARDEXIT;
                        }
                        if( fp.write( "\1", 1) != 1 )
                        {
                            ;//qDebugA() << "failed to write the next 1 chars.";
                            goto HARDEXIT;
                        }
                    }

                    /* In this 2023 version of Estimate Maker I decided
                     * to not use the take1.tot and to not save it
                     * into the qtymathHidden object as a variable
                     * because the calculations for each line might
                     * become out of sync with the total of all lines.
                     * It does not matter for the estimate because Qty
                     * controls the old estimates. However, I can run
                     * the calculation for the entire box unlike 1987
                     * when this take1.tot was a speed up idea for slow CPUs.
                     *
                     */
                    // this maintains the old 1987 msdos file format:
                    char buf1987[42];
                    sprintf( buf1987, "%.40s", str.toStdString().c_str() );
                    char *ptbuf1987 = buf1987;
                    totFloat = tediously_meticulously1987compatibility( ptbuf1987, qtytemp );
                    // offset 44 is unused variable for 2023 version

                    /* Write Qty, old code: take.tot = (float) *(spare+44)
                     * This is offset 44.
                     */
                    if( fp.write( (char *)&totFloat, 4) != 4 )
                    {
                        ;//qDebugA() << "failed to write the next 4 chars.";
                        goto HARDEXIT;
                    }
                    else
                        ;//qDebugA() << "success wrote take1.tot=" << totFloat;

                    ++howmany;
                }
            }
        }
    }

        if( !fp.seek( 407L) )
        {
            delete qtytemp;

            fp.close();
            QMessageBox msgBox;
            msgBox.setText("fp.seek 407L failed in estimateWrite");
            msgBox.exec();

            tab[indexOld]->loadTabtoQLineEdits();
            ui->tabWidget->setEnabled(true);

            return(ret);
        }

        howmanyLongpt = (long *)howmanyLongChar;
        *howmanyLongpt = howmany*48L;
        ;//qDebugA() << "\nTotal takeoff byte count=|" << *howmanyLongpt <<"|"<<howmany*48L;

        if( fp.write( howmanyLongChar, 4) != 4 )
        {
            ;//qDebugA() << "failed to write the next 4, as long.";
            goto HARDEXIT;
        }

        if( !fp.seek( *howmanyLongpt + 411L)  )
        {
            delete qtytemp;

            fp.close();
            QMessageBox msgBox;
            msgBox.setText("fp.seek 407L failed in estimateWrite");
            msgBox.exec();

            tab[indexOld]->loadTabtoQLineEdits();
            ui->tabWidget->setEnabled(true);

            return(ret);
        }
        // ready to write the estimate body

        delete qtytemp;
        if(0)
        {
        HARDEXIT:  // skip writing qtymathHidden. Try to continue with the estimate.

            delete qtytemp;

            if( !fp.seek( 411L) )
            {
                fp.close();
                QMessageBox msgBox;
                msgBox.setText("fp.seek failed and Qty boxes failed to save; estimateWrite");
                msgBox.exec();

                tab[indexOld]->loadTabtoQLineEdits();
                ui->tabWidget->setEnabled(true);

                return(ret);
            }
        }
#endif

        QMessageBox msgBox;

        //Next write the estimate lines, after determining how many non-blank

        struct estimateold {
            QString totalhrs;
            QString totalcost;
            //unsigned memtake;
            char takeoff;   // See qtymathHidden *fqb;  in mainwindow.h

            QString libcode;  // library 3 char code library line, if any; sometimes blank.
            QString description;
            QString quantity;
            QString hours;
            QString cost;
            QString refcode; // tells which library list, like *AA was the source.
                // "AA\0" is three chars. Asterisk is omitted.
                // It is not the library code. It is the list name flag "*".
        }est;

    /* Find the last line of the estimate side.
     * Reverse search will work with no crashing, even if load2.cpp is
     * set to test-mode #define FASTtruncate compiled,
     * because isEmpty() is both nullptr and "".
     */

#ifdef DEADWRONGbecause
        because the lines are not estimateRead()
            as blank lines, but rather are set to "" after fp.gets().
            Therefore this additional test is unnecessary.

    /* This is a typical blank line which we write as filler.
     * It is as useless as "isEmpty()" but we have to catch these too
     * so set up a test for them.
     */
    memset( buf,' ', 80); // fill with spaces
    buf[0] = '\1';
    buf[78]='\r';
    buf[79]='\n';
    buf[80] = '\0';
    // skip the '\1' \2' qtymath flag and skip the \r\n. 77 spaces is proof.
    QString ssfiller = QString("%1").arg(buf+1,77);
    // skip the '\1' \2' qtymath flag and skip the \r\n

    if( (!est.libcode.contains(ssfiller) && !est.libcode.isEmpty()) ||
        (!est.description.contains(ssfiller) && !est.description.isEmpty()) ||
        (!est.quantity.contains(ssfiller) && !est.quantity.isEmpty()) ||
        (!est.hours.contains(ssfiller) && !est.hours.isEmpty()) ||
        (!est.cost.contains(ssfiller) && !est.cost.isEmpty()) ||
        (!est.totalhrs.contains(ssfiller) && !est.totalhrs.isEmpty()) ||
        (!est.totalcost.contains(ssfiller) && !est.totalcost.isEmpty())
        )
    {
        ;//break;
    }
#endif

    for( int tt=1; tt<TABCOUNT; ++tt)
    {
        if( tab[tt]==nullptr ) // skip unallocated, blank ones
            continue;

        for( maxest=ESTROWCOUNT; --maxest>=1; )
        {
            est.libcode = tab[tt]->estTabRow[maxest]->col[1];
            est.description = tab[tt]->estTabRow[maxest]->col[2];
            est.quantity = tab[tt]->estTabRow[maxest]->col[3];
            est.hours = tab[tt]->estTabRow[maxest]->col[4];
            est.cost = tab[tt]->estTabRow[maxest]->col[5];
            est.totalhrs = tab[tt]->estTabRow[maxest]->col[6];
            est.totalcost = tab[tt]->estTabRow[maxest]->col[7];

            if( !est.libcode.isEmpty() ||
                !est.description.isEmpty() ||
                !est.quantity.isEmpty() ||
                !est.hours.isEmpty() ||
                !est.cost.isEmpty() ||
                !est.totalhrs.isEmpty() ||
                !est.totalcost.isEmpty()
                )
            {
                ;//qDebugA() << "detected nonblank" << est.libcode;
                ;//qDebugA() << "detected nonblank" << est.description;
                ;//qDebugA() << "detected nonblank" << est.quantity;
                ;//qDebugA() << "detected nonblank" << est.hours;
                ;//qDebugA() << "detected nonblank" << est.cost;
                ;//qDebugA() << "detected nonblank" << est.totalhrs;
                ;//qDebugA() << "detected nonblank" << est.totalcost;
                break;
            }
        }


    ;//qDebugA() << "We are going to save to EST this many lines:" << maxest;

        for( xx=maxest, lg=1; xx>0; --xx, ++lg)
        {
            //;//qDebugA() << "\nXCxcxcxcxcxc" << lg << tab[tt]->estTabRow[lg]->col[2];

            //sprintf( buf, "% 80s", "" ); // warning undefined behaviour
            memset( buf,' ', 80); // fill with spaces
            buf[80] = '\0';

            est.libcode = tab[tt]->estTabRow[lg]->col[1];
            est.description = tab[tt]->estTabRow[lg]->col[2];
            est.quantity = tab[tt]->estTabRow[lg]->col[3];
            est.hours = tab[tt]->estTabRow[lg]->col[4];
            est.cost = tab[tt]->estTabRow[lg]->col[5];
            est.totalhrs = tab[tt]->estTabRow[lg]->col[6];
            est.totalcost = tab[tt]->estTabRow[lg]->col[7];

            est.refcode = tab[tt]->estTabRow[lg]->refcode;

        /* sprintfreverse( est.takeoff = buf[0];
         * takeoff is now in tab[tt]->estTabRow[row]->fileQtyBlock
         * and it points to an object called qtymathHidden
         * and there is potentially one for each estimate line.
         * Old 1987 flag was:  We will maintain the file format:

            est.takeoff = '\2';  YES
        else
            est.takeoff = '\1';  NO
         *
         * New 2023 flag in c++ code is:
            est.qtymathHidden = not-nullptr;  YES =estRow[lg]->qtymathHidden
        else
            est.qtymathHidden = nullptr;      NO  =estRow[lg]->qtymathHidden
         */

            buf[0] = (tab[tt]->estTabRow[lg]->fileQtyBlock==nullptr)?'\1':'\2';

            sprintfreverseTabversion( est.libcode, (char *)"%.3s", buf+1 );
            sprintfreverseTabversion( est.description, (char *)"%.24s", buf+5 );
            sprintfreverseTabversion( est.quantity, (char *)"%7.7s", buf+29 );//I triple checked
            sprintfreverseTabversion( est.hours, (char *)"%5.5s", buf+37 );//these sizes and offsets
            sprintfreverseTabversion( est.cost, (char *)"%5.5s", buf+43 );

            sprintfreverseTabversion( est.totalhrs, (char *)"%6.6s", buf+49 );
            sprintfreverseTabversion( est.totalcost, (char *)"%6.6s", buf+56 );

        /* These two floats are not used in the 2023 version.
         * However, I am writing them to maintain backwards compatibility,
         * which is useful for testing the new 2023 Estimate Maker against
         * the old 1987 Estimate Maker.
         */

            //                          totalhrs float 4 chars
            //                          totalcost float 4 chars

            //Mar31,2023. I think this is wrong because the totFloatChar might
            //not be the real beginning of the float when you use a ptFloat to point.
            //See memory bug BIGTRADE.disaster zzz

            // Putting them into a buf has tricky memory boundary issues:
            /*
        totFloatpt = (float *)totFloatChar; //msdos compatibility trick
        totFloat = est.totalhrs->text().toFloat();
        *totFloatpt = totFloat;
        strncpy( buf+62, totFloatChar, 4 );

        totFloatpt = (float *)totFloatChar; //msdos compatibility trick
        totFloat = est.totalcost->text().toFloat();
        *totFloatpt = totFloat;
        */

            totFloat = est.totalhrs.toFloat();
            strncpy( buf+62,(char *)&totFloat, 4 );

            totFloat = est.totalcost.toFloat();
            strncpy( buf+66, (char *)&totFloat, 4 );

            // skip the asterisk; eg. "*WD"
            strncpy( buf+74, est.refcode.toStdString().c_str(), 2);

            ;//qDebugA() << est.libcode->text() << est.description->text() << '|';

            for( yy=0; yy<78; ++yy)
                if( buf[yy] == '\0' )
                    buf[yy] = ' ';
            buf[78]='\r';
            buf[79]='\n';

            if( fp.write(buf,80) < 80)
            {
                ;//qDebugA() << "failed to write the next 80 chars.";
                break;
            }
        }

        /* This pads the tabPages with blank lines, except the last tabPage
         * won't because we do not need a spacer for a next tabPage to align
         * because there is no next tabPage; ie. tab[tt]->nextTab(tt+1)==false.
         */
        if( tt < TABCOUNT && tab[tt]->nextTab(tt+1) && maxest!=0)
        {
            ;//qDebugA( << "save to EST this many blank lines:" << ESTROWCOUNT-maxest-1 << "tt+1:" << (tt+1);

            for( xx=ESTROWCOUNT-maxest-1; xx>0; --xx )
            {
                //sprintf( buf, "% 80s", "" ); // warning undefined behaviour
                memset( buf,' ', 80); // fill with spaces.
                buf[80] = '\0';
                buf[0] = '\1';  //

                for( yy=0; yy<78; ++yy)
                    if( buf[yy] == '\0' )
                        buf[yy] = ' ';
                buf[78]='\r';
                buf[79]='\n';

                if( fp.write(buf,80) < 80)
                {
                    ;//qDebugA() << "failed to write the next 80 chars.";
                    break;
                }
            }
        }
    }
        ;//qDebugA() << "\n******\nWrote" << maxest << "estimate lines succeeded.\n";

        ret = true;

        fp.resize( fp.pos() ); // when I shrank the file trouble ensued.Mar2023
        fp.close();

#ifdef AsPromised
        tab[indexOld]->loadTabtoQLineEdits();
#endif
        ui->tabWidget->setEnabled(true);

        return( ret );
    }

    bool tabPage::nextTab( int tt )
    {
        bool ret = false;

        // check ALL the next tabs, not just the adjacent next one.
        for( ; tt<TABCOUNT; ++tt)
        {
            if(tab[tt]!=nullptr)
            {
                ret = true;
                break;
            }
        }
        return(ret);
    }

#else // pre-March 2024 version is file compatible but will truncate! to 231 est lines

bool MainWindow::estimateRead( QString *filename )
{
    bool ret = false;
    qint64 er,readsz,xc;
    long takelg,lg, maxest,*ptL, block, filelength;
    //int qdec,mdec,tdec,ttdec,tmdec;
    char startofmem[ESTROWCOUNT], *spare=startofmem;
    long lastSubtotalBlock = -1L;

    ui->tabWidget->setEnabled(false);
    //turn off user mouse clicks while we copy data into Tabs

    if( oneHeader == nullptr )
        oneHeader = new DialogHeaderOfEstimate(this);

    // do some error checking and leave the estimate not cleared; clearEst()
    if( oneHeader->estimateReadHeader( filename ) == false )
    {
        ;//qDebugA() << "failed to read" << *filename;

        ui->tabWidget->setEnabled(true);
        return(ret);
    }
    else
    {
        // clear and reload now that we know there are no basic errors
        clearEst();
        oneHeader->estimateReadHeader( filename );
    }

    QFile fp(*filename);

    /* This is the original 1987 C program. I am changing the
       destination memory but using the same names.
       I want to get the file read()s and offsets correct so I maintain
       the old estimate files as input for testing in Dec 2022; DavidP.
     */

    ;//qDebugA() << "again, filename:" << *filename;

    if( fp.open(QIODeviceBase::ReadOnly) == false)
    {
        ui->tabWidget->setEnabled(true);
        return(ret);
    }

    filelength = fp.bytesAvailable();

    // 407 bytes are header of estimate, plus 4 bytes of long, totaling 411 bytes.
    if( !fp.seek(407L) )  //2022 ctrl-Q "takeoffs" start here.
    {
        ;//qDebugA() <<"Error: Cannot seek to 407, near end of header";

        fp.close();
        ui->tabWidget->setEnabled(true);

        return(ret);
    }

    if( fp.read( ptLchar, 4) != 4 )
    {
        ;//qDebugA() << "failed to read 4 chars for long calc.";
        fp.close();
        ui->tabWidget->setEnabled(true);

        return(false);
    }

    ptL=(long *)ptLchar;
    takelg = *ptL;
    ;//qDebugA() << "estread ctrl-Q calc area, size in bytes:"<<takelg<<"takeoff-lines:" << takelg/48L;
    if( filelength < takelg )
    {
        ;//qDebugA() << "The crtl-Q size is bigger than the filelength:" <<filelength;
        ;//qDebugA() << "Quit now \n and then exit/reload the program fixes it (bug) Mar2023.";
        fp.close();
        ui->tabWidget->setEnabled(true);

        return(false);
    }

    /* Some old 1987 NOTES about Data File Format for Estimate Maker
     * I will maintain this format in the year 2022:

     * MSDOS est memory format is the same 80 char length as
     * the disk file for each estimate line and no format difference.
     *
     * This is the description which I found in the 1987
     * C program:
     *
     * byte 1 indicates if it has a QTY-take-off item (byte='\2')
     *        or has a regular quantity item (byte='\1')
     * (edit 2022: I am calling it "QChar takeoff" in the struct below)
     *
     *	49 bytes screen area
     *	8 bytes are 2 float nos. to store totals
     *	all 58 above is in MSDOS main memory ds seg (segds)
     *  all below is ext (seg est22)
     *
     *	12 bytes of additional description
     *      (Edit me 2022: unused, I am pretty sure. Future use.)
     *	2 bytes unsigned segment address of 1st take-off item
     *          (is offset 70, but important on disk only
     *          if it 0 or not-0 because memory targets change
     *          every time the program
     *          loads a filename.
     *          So 2022 we use 0 or not-0 to, not-read or read,
     *          respectively, the qtymath box 17 max lines.
     *	2 bytes unused
     *	2 bytes of refcode  is offset 74 and is important to know on
     *              disk also, regarding which library list sourced
     *              the estimate line.
     *              Library names were *03. Only 03 meant refcode.
     *              and the asterisk only meant is was a listname, not a
     *              library line code name, which is 3 chars.
     *	4 bytes unused      is offset 76+4=80 chars total
     */

    struct estimateold {
        float totalhrs;
        float totalcost;
        unsigned memtake;
        QChar takeoff;
        char libcode[4];  // library 3 char code library line, if any; sometimes blank.
        char description[65];// this is actually 24+1 chars below, I see.
        char quantity[8];
        char hours[8];
        char cost[8];
        char refcode[3]; // tells which library list, like *AA was the source.
            // "AA\0" is three chars. Asterisk is omitted.
            // It is not the library code. It is the list name flag "*".
    }est;

    QChar takeoffflag[ESTROWCOUNT]; // this is a list of which ones have Qty details
        // which we oddly call 'takeoffs' 1=no, 2=yes.

    /* 2022 note: parsing the description was complex because the
     * old estimate maker displayed the description two ways,
     * big and small, and stored the big end into a different
     * memory area, external to msdos to squeeze more ram into use.
     * Here is estline() clipped out:
     * 	if(big)
    {
    putsn("|",yy,29,1);
    putsn(pt+1,yy,30,16);
        uu=ramseges;
    ramseges=est22;
    xx=((pt-startofmem)/58)*22;
    extputsn(xx,yy,46,12);
    ramseges=uu;
    putsn(" ",yy,58,1);
    putsn(pt+30,yy,59,20);
    }
    else
    {
    putsn("|",yy,29,1);
    putsn(pt+1,yy,30,49);
    }
    I am only interested in 'big'. putsn last arg is number of chars.
    Previous to that are row,col on the VGA monitor. First arg is a (char *), eg."hello".
    So I have to see how that was read from disk: esttostr() into memory
    and then interpret the above, which is the actual display on screen
    and piece it together in the same coherence, which I do next.
    The putsn are chars size: 1,16,12,1,20 = 50 chars, starting near centre of display.
    */

    /*
      The answer to the puzzle was simple. The disk storage is already
      coherent and in order. Ignore all the above memory tucking.
      Simply parse the first 5 as "|",libcode, " ",
      and then description through to qty,cost,hrs is the next
      45 chars = 50 chars total.
    */

    /* Why do I fp.seek() ahead and read the estimate lines first?
     * Why do I fp.seek() back when I could have done no seeks if I read
     * qtymath (takoffs) first?
     * Answer: The estimate lines have flags at the beginning of each line
     * which say yes or no to whether they have qtymath (takeoffs)
     * Those qtymath boxes have to be assigned to the correct est lines
     * which have the yes/no flags. So I read the est lines, and then
     * make a list qtymath flagged est lines, and then fp.seek() back
     * and read the takeoffs until each block hits a 00 marker.
     * Mar 6, 2023. It could have been done by saving the qtymathHidden
     * blocks in a sequence and assigning them as the est lines were read
     * and then the disk read order would be more natural.
     *
     * However, I wrote the code for est lines first because it was
     * the most important and interesting, starting Dec 1, 2022, and
     * so the logic fell out of my head to jump back and fetch, as
     * the next step a month later.
     */
    lg = filelength - takelg - 411L;  // 2022 version
    //lg = filename->filelength(fileno(fp))-takelg-411L; /*estlist size  1987*/

    if( !fp.seek( takelg + 411L ) )  //2022
    {
        fp.close();
        ui->tabWidget->setEnabled(true);

        return(ret);
    }

    maxest = lg/80L;

    for( block=0,lg=maxest; lg>0; --lg) //80x32=2560, 1987 code, block reads were faster
    {
        if( fp.read(spare,80) != 80 )
        {
            ;//qDebugA() << "failed to read the next 80 chars. That should not happen because we checked the filelength before starting this.";
            break;
        }
        ++block;
        takeoffflag[block] = est.takeoff = *spare;
        sprintf( est.libcode, "%.3s", spare+1 );
        sprintf( est.description, "%.24s", spare+5 );

        // I think offset spare+58 is on the disk but is unused; for future use.
        // I see it is not printed to screen in the if(big) C code snippet.
        //sprintf( est.description, "%.49s<>%.12s<", spare+1, spare+58 );

        sprintf( est.quantity, "%.7s", spare+29 ); // I triple checked
        sprintf( est.hours, "%.5s", spare+37 );     // these sizes and offsets
        sprintf( est.cost, "%.5s", spare+43 );

        float *fl;
        fl = (float *)(spare+50);
        est.totalhrs  = *fl;
        fl = (float *)(spare+54);
        est.totalcost  = *fl;

        sprintf( est.refcode, "%.2s", spare+74 );
        /* this was not displayed for each est line in big, above,
         * but was in the header as you scrolled down/up.
         * "0 " means main one, maybe. "WD" was in BIGTRADE
         * and I saw an est.lib/WD filename to match, and est.lib/0.
         */

        /* ;//qDebugA() << est;  raw data has non-printable chars,
         * so skip 1 char The skipped char is the flag to
         * indicate ctrl-Q; takeoffs == char '\2'
         * for the estimate line; ie. detailed Quantity
         * calculations. NOT==char '\1'
         */

        ;//qDebugA() << est.libcode << est.description << est.quantity<<est.hours<<est.cost<< '|' << est.totalhrs << '|' << est.totalcost<< '|' << est.refcode<<'|' << block<<'|';
        if( block < ESTROWCOUNT )
            ;
        else
        {
            ;//qDebugA() << "estimateRead() lines is longer than" << ESTROWCOUNT-1 << "\nprobably because a longer estimate was made with a different version of PaceMakerEstimator.";
            break;
        }

        /* The next are floats for fast summing. We can ignore
         * those because we are going to use doubles or bigdecimal()
         * for recalculating everything, instead of floats.
         * They seem to be failing so far anyway because
         * maybe float * to a char string is not on a correct
         *  4 byte boundary or it is a differing float system
         * from Desmet compiler 1986.
         *
         * The memory offset of the first takoff item is
         * unused by us because we are not using MSDOS memory.
         * So skip it too.
         */

        // must be "=  " not "=" for example
        estRow[block]->col[1]->setText( QString("%1").arg( est.libcode,-2,' ') );
        estRow[block]->col[2]->setText( stripEnd(est.description) );
        estRow[block]->col[3]->setText( stripEnd(est.quantity) );
        estRow[block]->col[4]->setText( stripEnd(est.hours) );
        estRow[block]->col[5]->setText( stripEnd(est.cost) );

        // New March 2023. "===" libcodes are subtotal/total text areas
        if( est.libcode[0] == '=' )
        {
            QString ss = estRow[block]->col[1]->text();
            // update the subtotal total since the file has no sums nor multiplication
            if( ss.indexOf( uniqueID_TOTALSblock ) == 0 ) // end of subtotal grouping
            {
                bottomSubtotalRow = ESTROWCOUNT-1; // prevent TOTALS creation
                insertTotalSubtotal( block );
            }
            lastSubtotalBlock = block;
        }
        else
            calc_one_estline( block ); // this uses double-float numbers

        QString ss = QString("%1").arg( est.refcode,-2,' ');
        estRow[block]->refcode = ss.trimmed();
    }
    if( block != maxest )
    {
        fp.close();
        ;//qDebugA() << "Should NOT ever happen. block=" << block << "should equal maxest, not this:" << maxest;
        ui->tabWidget->setEnabled(true);

        return(false);
    }
    ;//qDebugA() << "\n**********\nReading " << maxest << " estimate lines succeeded.\nNext read the ctrl-Q details, if any.\n";

#ifdef SKIP1987
    if(!loadtake(fp,takelg))
        er=0;
    ----------------
                         loadtake(fp,lg)
                         FILE *fp;
    long lg;
    {
#endif
        /* I copied old 1987 loadtake() into here, as follows:
          unsigned wh,*ptu,dummy;
          char exttakebegin[], *exttake;

       ctrl-Q Qty calculation windows are called qtymath, 2022AD ver..
       One of them: 48 bytes each in memory malloc() as needed, I guess.

        That ext WAS a memory pointer to the next one. If zero then
        that is end of list.
        The value is meaningless in the file. In the file it only
        matters if the first char of the estimate line was a \1 or \2.
        That means I can monitor that spot as the quit and go to
        next estimate line
        while I am loading the BIGTRADE filename into the program.

???When I create a BIGTRADE file I must store non-zero
        unsigned values in that spot until end of list of ctrl-Qty
        items for that estimate line; last one is 0 ?
???I think write ALL 0s.

        Library has takeoffs?
           Yes, but I looked in EST1A.C loderef() (a deliberate
           misspelling to distinguish it from old code which was loadref)

           The struct is simply 3 char library code
                        and a pointer, *refpt,
                        to a 28 char description.

           The library takeoffs (now named qtymath) are different
           library filenames. Read detail at the function which
           does library filename parsing.
           17 lines of qtymath each have QEditLine of 40 chars,
                                     and QEditLine of 7 chars,
           similar to old code "struct takeoff{} take;
                    has ch[40]; and
                    has float tot;
           for each line.

           So maximum 17 "take" memory allocations had a char *refpt
           pointing at the first line which is one
           "struct takeoff( } take" and then the next line of 16 more
           may or may not exist so as to limit memory use.
           I think a pointer points at the next line in memory,
           if the pointer is not null.

           HOWEVER, I notice the filesize of est.lib/B.B08 as a qtymath
           item is always 680=17*40. That means only the first indicator
           on the disk matters to use. If it is non-zero then
           read 680 bytes. We are not going to use memory pointers
           after that data is read into our object, qtymath.
           Our coding in 2022 is Qt objects like QLineEdit,
           not pointers to raw memory.

           The confusing part to me was that there can be many of these
           680 char blocks depending on how many estimate lines
           set a ctrl-Q qty detail which we will now call qtymath object.
           Why are they all in sequence on the disk?
           How do you match them up with each estimate line?
           The answer is you wrote all the estimate lines first
           and then you went back in memory and wrote all the
           680 byte-blocks. So when we read, we read the estimate
           lines first, as above, and then we run through the estimate
           sequentially looking for a flag which as of Feb 27, 2023
           does not exist for each estimate line.

           I will probably name it est[row]->qtymath in
           public mainwindow.h available for all estimate lines.
           It will be a QTYMATH *qtymath = new QTYMATH in mainwindow.h
           public. Also see qtymath.h and qtymath.cpp.
           It will be nullptr if the estimate line has no details,
           which in the old says we called up with keystroke ctrl-Q
           and flagged yes/no by using the first byte in the est line.
           So although we will use est[row]->qtymath, we must
           also set the 1st char of estimate line as
           non-zero '\1' = none or '\2' = yes, on the disk file during
           estimate to disk writes.
    See bool MainWindow::estimateWrite( QString *filename )


            #define RIZE 300

            struct stref {
                    char refch[3];
                    char *refpt;
                    }ref[RIZE];

            struct takeoff {
                char ch[40];
                char ref[2];
                unsigned ext;
                float tot;
                }take;

            I think the RSIZE must mean there was a limit of 300
            ctrl-Q used in one estimate; ie. 300 lines can have
            Qty like this although an estimate could have had 640+ lines
            depending on msdos memory available and so some
            were plain non-ctrl-Q lines.

            The 2022 version will have a limit equal to the
            number of maximum estimate lines, instead of less than.
            Where do we put the estimate line pointer to the
            ctrl-qty? I am still not sure where it was in the 1987 ver..
            I found it. See above, offset 70 in each estimate line:
            "     *	2 bytes unsigned segment address of 1st take-off item
                      (is offset 70, "
            We will not use that pointer to memory.
???Do we have to
            write it to maintain old code compatibility? I think not,
            because it was just a convenience to keep the data
            structures less confusing; ie. memory matched disk.
???I must test this theory.
            I still have the old msdos program.

            Also see est[row]->qtymath in mainwindow.h
            which I mentioned in more notes above, in this function.

            And:
        2 bytes of ref is offset 74 and is important to know on
            disk also, regarding which library list/line sourced
            the estimate line because the qtymath is
            archived in the library as separate filenames
            so that math can be used in other new estimates. wow!

            Do we write the new est.lib/B.B08 in the qtymath object?
            No. It was used for F5 read/ F6 save while in the Qty box.
            It was for future estimate line use, like a library item.

            This will change the library but have no effect on
            historical estimates because they have their own copies.

            It makes sense because a type of brick, for example,
            would have the same qty math.
            The filenames were X.B08 or XX.B10 where B08 for the
            line from the library list, which was called X or XX.
            Those library names are 1 or 2 characters coded as ref[2]
            in the old 1987 program.

            B08 and B10 are the estimate line or library
            line 3 char codes column one: est[row]->col[1]
                                          lib[row]->col[1]
            in mainwindow.h

            Those ctr-qty "takeoffs" library files are always
            exactly 680 chars, which is 17x40=680.

            The same qty takeoffs in the estimate filename is stored
            as 48 chars, length, not 40; see the takeoff structure above.

            The number of lines of 48 bytes used take.ext==0 as the
            no-next-line indicator. So 48 bytes could be 1x up to 17x in the file.

            You cannot just read a block of 680 chars off from the
            estimate file if a takeoff (qtymath) is indicated
            for an est line. That is 40 chars per line x 17 = 680.

            I did fast code in 1987 for slow 8088 CPUs
            and a lot of features were accomplished.
            It is too bad I had to quit to work for a medical lab
            because revenues were insufficient. Small contractors were
            not ready to buy computers and software. I picked
            the wrong niche. Only large contractors would have worked.
            My coding example sold myself as a capable programmer though.


        struct takeoff {
            char ch[40];
            char ref[2];
            unsigned ext;
            float tot;  // 4 bytes long
            }take; //48 bytes

    // This for() reads the file from after the header 411
    // into external 1987 msdos memory.
    // and the next for() assigns all the est[] pointers to point
    // at all the spots if any have a '\2' char which indicates
    // a takeoff ctrl-Q Qty. '\1' means none.
    //
    // Why are we reading it from two spots, out of order?
    // The takeoffs are after the 411 bytes of header.
    // The estlines are after the 411header + takeoff area -> here tail.
// So the tail of the file is read first, to find out about the \2 and \1.
// Next the earlier part is read, which is the takeoffs which are assigned
// in order, without any knowledge of which est[] line they belong to,
// except for the fact they were saved in order. So read them in order.

// I look at the BIGTRADE file. They are fixed @48 chars, variable
// number of these 48s, is the correct answer.
// Position 42 *unsigned ext is 0 signaling terminate after a bunch of
// them, indicating go to next est[] line.

    * I learned all of this from ESTOFFA.C storetake() in the 1987 code.
     * Here is an important item for when we store an estimate.
     * This code is writing a 4 char size long-int at 407 offset
     * which tells the 1x48 thru 17x48 number of qty lines.
     * That xc*48 is not for one estline qty-detail.
     * It is for ALL qty-details and the separation of each
     * one for each est line is the many take.ext==0 terminators.
     *
        fseek(fp,407L,(int)0);
        ptL=(long *)spare;
        *ptL=(long)xc*48L;
        if(fwrite(spare,sizeof(char),(int)4,fp)!=4)
                return(0);

        loadest() and loadtake() inside EST1A.C
        read that 4 byte long and use it to read the
        full block of ALL qtymath lines and
        afterwards distribute them to their
        corresponding est lines.
     */

        if( !fp.seek(411L) )  //2022 ctrl-Q "takeoffs" start here.
        {
            ;//qDebugA() <<"Error: Cannot seek to 411, near end of header";

            fp.close();
            ui->tabWidget->setEnabled(true);

            return(ret);
        }
        // takelg was already read far above. It is takeoff bytes total.
        if( takelg > 0 )
        {
            char *pt;
            char *pttakeoff = (char *)malloc( takelg );
            if( fp.read( pttakeoff,takelg ) != takelg )
            {
                free(pttakeoff);

                ;//qDebugA() <<"Error: Cannot read qtymath details.";

                fp.close();
                ui->tabWidget->setEnabled(true);

                return(ret);
            }

            pt = pttakeoff;
            /*
             * xc=1 is our estRow[1] variable in mainwindow.h
               xc=2 is our estRow[2] variable in mainwindow.h
               The takeoffflag[xc] got set above and
                  starts at xc=1,
                    not xc=0 because we do not use row=0 in estRow[row]
             */
            for(xc=1;xc<=maxest;++xc)
            {
                if( takeoffflag[xc] == '\2' )// THIS IS THE TAKE-OFF ctrl-Q flag
                {
                    estRow[xc]->fileQtyBlock = new qtymathHidden();

                /* this is '\1' or '\2' in est line first char 1987 coding
                   but it is '\0' or a non-Zero pointer fileQtyBlock

                   new qtymathHidden()
                 */

                /* delete qtymath1;
                 * Do the deletes when closing the estimate
                     or else it is a memory leak!
                 *
                   Mar2, 2023: done! See here for deletes.
                   void qtymath::on_deleteButton_clicked()

                   Also, I delete qtymathHidden in clearLibEst.cpp
                 */

                    for( unsigned indx=0; takelg >= 48; takelg -= 48L, ++indx )
                    {
                        strncpy( spare, pt, 40);
                        *(spare + 40) = '\0';

                        QString str = str.fromStdString( spare );

                        take2 = (struct takeoff *)pt;

                        ;//qDebugA() << "take2.ch:" << str << "row:" << xc << "take2.ref:|" << take2->ref[0] << take2->ref[1] << "|\n";

                        if( indx < 17 )
                            estRow[xc]->fileQtyBlock->block[indx] = str.trimmed();

                        //if( take2->ext == 0 ) // does not work

                        //maybe this technique will work like above
                        // ptL=(long *)ptLchar;
                        //takelg = *ptL;

                    /* YES IT DOES WORK! How will I remember this trick?
                     * I think the char string has to be at the start,
                     * not the middle of what you are going to point at as the
                     * unsigned or float or double, et cetera.
                     * Maybe the middle (not the start) is randomly misaligned
                     * in operating system memory which is an issue for numerics.
                     * Make char[20], not char[4] for misalignment room.
                     */

                        ptUchar[0] = *(pt+42);
                        ptUchar[1] = *(pt+43);
                        ptU=(unsigned *)ptUchar;
                        unsigned uu = *ptU;
                        if( uu == 0 )
                        //if( *(pt+43) == 0 ) <this works but is very bad because 1char.
                        {
                            pt+=48;
                            break;
                        }
                        pt+=48;
                    }
                }
            }
            free(pttakeoff);
        }

        ret = true;

        fp.close();

        /* I put this in although I did all the  insertTotalSubtotal( block ); above
     * because the TOTAL in the description is wrong until I did a recalculation of all of them from
     * line 1 to the end by using this call, which is also used elsewhere to do stupid slow
     * but accurate. I almost gave up looking for why TOTAL is overwritten by DRAG and gets the
     * right number but gets it wrong when I
     * insertTotalSubtotal() here, which is THE SAME AS DRAG!
     * Edit: I think this is the problem which prevents TOTAL: if("row>=bottomSubtotalRow)")
     * Am I doing it on the wrong block? (block is row)
     * NO. Because I looked at DRAG and it is on ANY "===" line.
     * I do note that DRAG is always on a FULL set of subtotals-block.
     * I am doing it on an incomplete block, because I did it on "=   SUBTOTALS" which is unique.
     * whereas bottomSubtotalRow = row; is set on the last line "===                ---- ----"
     * I thought of a trick to simply set   bottomSubtotalRow = block;
     * up above instead of letting it be set inside by insertTotalSubtotal().
     * Wow! Let us try it:
     */
        bottomSubtotalRow = 0L; // force TOTALS creation by assuring this is top of estimate.
        insertTotalSubtotal( lastSubtotalBlock ); // last one always has a TOTAL
        //searchForSubtotalTotalUpdateThem(1);  // <--this works but is slower by redoing them all.

        ui->tabWidget->setEnabled(true);

        return( ret );
    }

    /* See the sister function: estimateRead()
     * for detailed notes about the file structure.
     * I wrote notes while I was doing the read from disk function
        of the old 1987 Desmet C language code.
     */
    char howmanyLongChar[20];  //not [4]
    long howmany=0L, *howmanyLongpt;

    char totFloatChar[20]; //not [4]
    float totFloat, *totFloatpt;

    bool MainWindow::estimateWrite( QString *filename )
    {
        bool ret = false;
        qint64 er,readsz,xc;
        unsigned block, yy;
        long takelg, lg, maxest,*ptL, xx;
        //int qdec,mdec,tdec,ttdec,tmdec;
        char *buf, buffer[81];
        buf = buffer;

        /* Here is a very difficult bug which I had to solve:
     * The "estimateChanged" variable was always switching back to
     * true AFTER I saved the estimate and meticulously set it to false.
     * Also, an estimate was maybe missing the calculation of a line, which was done after saving.
     * I never saw that bug, but it is the same bug.
     *
     * These problems occurred because docalc() only gets called upon
     * focusOut, which is in events... .cpp. When you press Save or Open or New
     * buttons that causes a focusOut event which is processed AFTER estimateWrite() for some reason
     * and that docalc() changes the estimateChanged flag back to true after I set it false here.
     *
     * So here we must turn off the keyboard cursor
     * before estimateWrite() and before setting estimateChanged false
     * so that the keyboard cursor stops receiving FocusIn and FocusOut.
     * I think the msgBox OK button returns focus to the keyboard cursor
     * and generates the extra FocusOuts. Perhaps it is not the order of events that is surprising
     * but the extra FocusOut.
     * Skip more thinking. Let's do some solutions without determining those two possibilities.
     */
        /* Solution 1: Call the same function as clearSlider button because I know that turns off
     * keyboard row and col before continuing to write the estimate:
     * on_clearSlider_pressed();
     * THAT FAILED because I see the cursor blinking on the same est line after doing that idea.
     * Solution 2: How to turn off cursor? I know I am correct about this because I manually
     * turned off the cursor before using the Open button and the estimateWrite was done first
     * because estimateChanged was true and then I pressed CANCEL and Open again
     * and estimateChanged remained set to false and estimateWrite was not redone!
     * ie. the FocusOut did not do another docalc() and did not set estimateChanged to true.
     * So how to stop the blinking cursor programmatically?
     * Doing it with the mouse is obvious: click on any NOT-QLineEdit widget.
     * https://stackoverflow.com/questions/35762856/how-can-i-move-the-keyboard-cursor-focus-to-a-qlineedit
     *  has:
QPoint pos(line_edit->width()-5, 5);
QMouseEvent e(QEvent::MouseButtonPress, pos, Qt::LeftButton, Qt::LeftButton, 0);
qApp->sendEvent(line_edit, &e);
QMouseEvent f(QEvent::MouseButtonRelease, pos, Qt::LeftButton, Qt::LeftButton, 0);
qApp->sendEvent(line_edit, &f);
or
https://stackoverflow.com/questions/526761/set-qlineedit-focus-in-qt
QTimer::singleShot(0, line, SLOT(setFocus()));
     */
        QTimer::singleShot(0, ui->clearSlider, SLOT(setFocus())); // docalc() got called, presumably.
            // See eventMouseKeyboard.cpp keyboard FocusOut.
        // Wow, that worked! The blinking cursor is not on the QEditLine. January 23, 2024. And clearSlider was not clicked.

        /* Write the estimate header first. It does its own fp.open/fp.close
     * because it is also called from the dialogheader... .
     */

        ui->tabWidget->setEnabled(false);


        if( oneHeader != nullptr )
        {
            ret = oneHeader->estimateWriteHeader( filename );
            if( ret == false )
            {
                //QMessageBox msgBox;
                //msgBox.setText("Estimate header failed to write to disk;\nOut of disk space?");
                //msgBox.exec();
                ui->tabWidget->setEnabled(true);

                return(ret);
            }
        }
        else
        {
            ;
            /*Jan 23, 2024. I think this should quit and return failure if there is no header.
        *I do not know if there is a reason to continue saving the body so I won't
        *insert a return(ret); here for now.
        *The answer is likely that there is always a oneHeader so putting a return here
        *would never happen anyway.
        */
        }

        QFile fp(*filename);

        if( fp.open(QIODeviceBase::ReadWrite) == false)
        {
            ui->tabWidget->setEnabled(true);

            return(ret);
        }

            /* write the qtymathHidden qtymath details for all est lines
     * before the estimate lines are written.
     * 1987 C coding called qtymath takeoffs.
     */

#define TAKEOFFS
#ifdef TAKEOFFS
        if( !fp.seek( 411L) )
        {
            fp.close();
            QMessageBox msgBox;
            msgBox.setText("fp.seek failed inside estimateWrite");
            msgBox.exec();
            ui->tabWidget->setEnabled(true);

            return(ret);
        }

        qtymath *qtytemp = new qtymath( this );

        int li,cnt;
        QString fill40spaces;
        fill40spaces.fill( ' ', 40 );
        howmany = 0L;
        qtymathHidden *fqb;

        for( xc=1; xc < ESTROWCOUNT; ++xc )
        {
            fqb = estRow[xc]->fileQtyBlock;
            if( fqb != nullptr )
            {
                ;//qDebugA() << "\nXCxcxcxcxcxc" << xc;

                /* first find the last non-blank line
                 * because we are not writing end blanks. */
                for( li = 17; --li>=0; )
                {
                    if( fqb->block[li].isEmpty() )
                        ;
                    // strip the prefix and suffix blanks if any.
                    else if( fqb->block[li].trimmed().isEmpty() )
                        ;
                    else
                    {
                        break;
                    }
                }
                /* Now we know li is the last not-Empty block line.
                 * write from 0 up to and including li
                 * and then we will quit writing. If li==0 is blank then
                 * we will have written one blank line and no more.
                 */
                ;//qDebugA() << "Number of lines to write is:" << li+1 <<"|";

                /* Write this line to the estimate in this format:
                 * struct takeoff {
                    char ch[40];
                    char ref[2];
                    unsigned ext;
                    float tot;  // 4 bytes long
                    }; //48 bytes
                    static struct takeoff take1;
                 */

                for( cnt=0; cnt<=li; ++cnt)
                {
                    //take1.ch, "%.40s",

                        QString str;

                    // left justified and padded out with blanks to 40 chars.
                    if( fqb->block[cnt].isEmpty() )
                        str.fill( ' ', 40 );
                    else
                        str = QString("%1").arg( fqb->block[cnt], -40, ' ');

                    if( fp.write( str.toStdString().c_str(), 40) != 40 )
                    {
                        ;//qDebugA() << "failed to write the next 40 chars.";
                        goto HARDEXIT;
                    }
                    ;//qDebugA() << "\nWrote 40. cnt=" << cnt << "|" << str <<"|";

                    //TEMPORARY     take1.ref[0] = 'Z';
                    take1.ref[1] = 'Z';
                    if( fp.write( "ZZ", 2) != 2 )
                    {
                        ;//qDebugA() << "failed to write the next 2 chars.";
                        goto HARDEXIT;
                    }
                    if(cnt==li) // last line
                    {
                        if( fp.write( "\0", 1) != 1 )
                        {
                            ;//qDebugA() << "failed to write the next 1 chars.";
                            goto HARDEXIT;
                        }
                        if( fp.write( "\0", 1) != 1 )
                        {
                            ;//qDebugA() << "failed to write the next 1 chars.";
                            goto HARDEXIT;
                        }
                    }
                    else
                    {
                        if( fp.write( "\1", 1) != 1 )
                        {
                            ;//qDebugA() << "failed to write the next 1 chars.";
                            goto HARDEXIT;
                        }
                        if( fp.write( "\1", 1) != 1 )
                        {
                            ;//qDebugA() << "failed to write the next 1 chars.";
                            goto HARDEXIT;
                        }
                    }

                    /* In this 2023 version of Estimate Maker I decided
                     * to not use the take1.tot and to not save it
                     * into the qtymathHidden object as a variable
                     * because the calculations for each line might
                     * become out of sync with the total of all lines.
                     * It does not matter for the estimate because Qty
                     * controls the old estimates. However, I can run
                     * the calculation for the entire box unlike 1987
                     * when this take1.tot was a speed up idea for slow CPUs.
                     *
                     */
                    // this maintains the old 1987 msdos file format:
                    char buf1987[42];
                    sprintf( buf1987, "%.40s", str.toStdString().c_str() );
                    char *ptbuf1987 = buf1987;
                    totFloat = tediously_meticulously1987compatibility( ptbuf1987, qtytemp );
                    // offset 44 is unused variable for 2023 version

                    /* Write Qty, old code: take.tot = (float) *(spare+44)
                     * This is offset 44.
                     */
                    if( fp.write( (char *)&totFloat, 4) != 4 )
                    {
                        ;//qDebugA() << "failed to write the next 4 chars.";
                        goto HARDEXIT;
                    }
                    else
                        ;//qDebugA() << "success wrote take1.tot=" << totFloat;

                    ++howmany;
                }
            }
        }

        if( !fp.seek( 407L) )
        {
            fp.close();
            QMessageBox msgBox;
            msgBox.setText("fp.seek 407L failed in estimateWrite");
            msgBox.exec();
            ui->tabWidget->setEnabled(true);

            return(ret);
        }

        howmanyLongpt = (long *)howmanyLongChar;
        *howmanyLongpt = howmany*48L;
        ;//qDebugA() << "\nTotal takeoff byte count=|" << *howmanyLongpt <<"|"<<howmany*48L;

        if( fp.write( howmanyLongChar, 4) != 4 )
        {
            ;//qDebugA() << "failed to write the next 4, as long.";
            goto HARDEXIT;
        }

        if( !fp.seek( *howmanyLongpt + 411L)  )
        {
            fp.close();
            QMessageBox msgBox;
            msgBox.setText("fp.seek 407L failed in estimateWrite");
            msgBox.exec();
            ui->tabWidget->setEnabled(true);

            return(ret);
        }
        // ready to write the estimate body

        delete qtytemp;
        if(0)
        {
        HARDEXIT:  // skip writing qtymathHidden. Try to continue with the estimate.

            delete qtytemp;

            if( !fp.seek( 411L) )
            {
                fp.close();
                QMessageBox msgBox;
                msgBox.setText("fp.seek failed and Qty boxes failed to save; estimateWrite");
                msgBox.exec();
                ui->tabWidget->setEnabled(true);

                return(ret);
            }
        }
#endif

#ifdef JAN2023NOHEADERYET
        else
        {
            er = // first char is always F
                if( fp.write( "F", 1) != 1)
            {
                fp.close();
                ;//qDebugA() << "failed tp write to estimate file" << *filename;
                return(ret);
            }
            else
                headstr[0]=buf[0];

            if(er>0)
                er=fp.read( headstr+1, 400);
            //er=fread(headstr+1,sizeof(char),(int)400,fp); //1987 version
            if(er>0)
            {
                er=fp.read( setstr, 6);
                //er=fread(setstr,sizeof(char),(int)6,fp);
                timecode=setstr[0];
                qdec=(int)setstr[1];
                mdec=(int)setstr[2];
                tdec=(int)setstr[3];
                ttdec=(int)setstr[4];
                tmdec=(int)setstr[5];
            }
            if(er<=0)
            {
                ;//qDebugA() <<"Error: Cannot read the drive.";
                fp.close();
                return(ret);
            }
        }

        if(timecode=='M'||timecode=='m') // minutes or hours for the estimate units
        {
            //headinga=heading1;
            timecode='M';
        }
        else
        {
            //headinga=heading2;
            timecode='H';
        }

        if(qdec<0||qdec>5)
            qdec=0;
        if(mdec<0||mdec>4)
            mdec=2;
        if(tdec<0||tdec>4)
            tdec=2;
        if(ttdec<0||ttdec>5)
            ttdec=1;
        if(tmdec<0||tmdec>5)
            tmdec=2;

        // 407 bytes are header of estimate, plus 4 bytes of long, totaling 411 bytes.

        if( fp.read( ptLchar, 4) != 4 )
        {
            ;//qDebugA() << "failed to read 4 chars for long calc.";
            fp.close();
            return(false);
        }
        //fread(spare,sizeof(char),(int)4,fp);/*at end of header&codes to read the size of takeoff area*/

        ptL=(long *)ptLchar;
        takelg = *ptL;
        ;//qDebugA() << "ctrl-Q calc area, size in bytes:" << takelg << "takeoff-lines:" << takelg/48L;


        struct estimateold {

            unsigned memtake;
            QChar takeoff;
            char libcode[4];  // library 3 char code library line, if any; sometimes blank.
            char description[65];
            char quantity[8];
            char hours[8];
            char cost[8];
            float totalhrs; I do these two as QLineEdit in 2023 version
                float totalcost;

            char refcode[3]; // tells which library list, like *AA was the source.
                // "AA\0" is three chars. Asterisk is omitted.
                // It is not the library code. It is the list name flag "*".
        }est;

        QChar takeoffflag[ESTROWCOUNT]; // this is a list of which ones have Qty details
            // which we oddly call 'takeoffs' 1=no, 2=yes.


        lg = fp.bytesAvailable() - takelg; //already read: 411L;  // 2022 version
        //lg = filename->filelength(fileno(fp))-takelg-411L; /*estlist size  1987*/

        if( fp.seek(/*takelg+*/ 411L) == 0 )  //2022
        {
            fp.close();
            return(ret);
        }

        // write takeoffs first, although read last
        struct takeoff {
            char ch[40];
            char ref[2];
            unsigned terminator;
            float tot;
        }take;

        if( fp.seek(411L) == false )  //2022 ctrl-Q "takeoffs" start here.
        {
            ;//qDebugA() <<"Error: Cannot seek to 411, end of header";

            fp.close();
            return(ret);
        }

        for(xc=0;xc<=maxest;++xc)
        {
            //ss = takeoffflag[xc];
            //ss.truncate(1);
            //if( *ss.data() != '\2' )      // <--- THIS IS THE TAKE-OFF ctrl-Q flag
            if( takeoffflag[xc] != '\2' )      // <--- THIS IS THE TAKE-OFF ctrl-Q flag
                continue;

            while(1)
            {
                if( fp.read(spare,48) <= 0)
                {
                    ;//qDebugA() << "failed on takeoff-area, next 48 chars. That should not happen because the 0 terminator should have ended this.";

                    ret = false;
                    fp.close();
                    return(ret);
                }

                strncpy( take.ch, spare, 40 );
                strncpy( take.ref, spare, 2 );
                take.terminator = (unsigned) *(spare+42);
                take.tot = (float) *(spare+44);

                ss = take.ch;
                ss.truncate(40); // see struct above. First 40 are text.
                ;//qDebugA() << ss;

                //terminator is 0 unsigned in that 42 spot. See struct of takeoffs, above.
                if( take.terminator == 0 )
                {
                    ;//qDebugA() << "...done:" << xc;
                    break;
                }
            }
        }

#endif

        QMessageBox msgBox;

        //Next write the estimate lines, after determining how many non-blank

        struct estimateold {
            QLineEdit *totalhrs;
            QLineEdit *totalcost;
            //unsigned memtake;
            char takeoff;   // See qtymathHidden *fqb;  in mainwindow.h

            QLineEdit *libcode;  // library 3 char code library line, if any; sometimes blank.
            QLineEdit *description;
            QLineEdit *quantity;
            QLineEdit *hours;
            QLineEdit *cost;
            QString refcode; // tells which library list, like *AA was the source.
                // "AA\0" is three chars. Asterisk is omitted.
                // It is not the library code. It is the list name flag "*".
        }est;

        /* Find the last line of the estimate side.
     * Reverse search will work sa crashing, even if load2.cpp is
     * test-mode compiled, because isEmpty() is both nullptr and "".
     */
        for( maxest=ESTROWCOUNT; --maxest>=1; )
        {
            est.libcode = estRow[maxest]->col[1];
            est.description = estRow[maxest]->col[2];
            est.quantity = estRow[maxest]->col[3];
            est.hours = estRow[maxest]->col[4];
            est.cost = estRow[maxest]->col[5];
            est.totalhrs = estRow[maxest]->col[6];
            est.totalcost = estRow[maxest]->col[7];

            if(     !est.libcode->text().isEmpty() ||
                !est.description->text().isEmpty() ||
                !est.quantity->text().isEmpty() ||
                !est.hours->text().isEmpty() ||
                !est.cost->text().isEmpty() ||
                !est.totalhrs->text().isEmpty() ||
                !est.totalcost->text().isEmpty()
                )
                break;
        }

        ;//qDebugA() << "We are going to save to EST this many lines:" << maxest;

        for( xx=maxest, lg=1; xx>0; --xx, ++lg)
        {
            //sprintf( buf, "% 80s", "" ); // warning undefined behaviour
            memset( buf,' ', 80); // fill with spaces.
            buf[80] = '\0';

            est.libcode = estRow[lg]->col[1];
            est.description = estRow[lg]->col[2];
            est.quantity = estRow[lg]->col[3];
            est.hours = estRow[lg]->col[4];
            est.cost = estRow[lg]->col[5];
            est.totalhrs = estRow[lg]->col[6];
            est.totalcost = estRow[lg]->col[7];

            est.refcode = estRow[lg]->refcode;

            /* sprintfreverse( est.takeoff = buf[0];
         * takeoff is now in estRow[row]->fileQtyBlock
         * and it points to an object called qtymathHidden
         * and there is potentially one for each estimate line.
         * Old 1987 flag was:  We will maintain the file format:

            est.takeoff = '\2';  YES
        else
            est.takeoff = '\1';  NO
         *
         * New 2023 flag in c++ code is:
            est.qtymathHidden = not-nullptr;  YES =estRow[lg]->qtymathHidden
        else
            est.qtymathHidden = nullptr;      NO  =estRow[lg]->qtymathHidden
         */

            buf[0] = (estRow[lg]->fileQtyBlock==nullptr)?'\1':'\2';

            sprintfreverse( est.libcode, (char *)"%.3s", buf+1 );
            sprintfreverse( est.description, (char *)"%.24s", buf+5 );
            sprintfreverse( est.quantity, (char *)"%7.7s", buf+29 );//I triple checked
            sprintfreverse( est.hours, (char *)"%5.5s", buf+37 );//these sizes and offsets
            sprintfreverse( est.cost, (char *)"%5.5s", buf+43 );

            sprintfreverse( est.totalhrs, (char *)"%6.6s", buf+49 );
            sprintfreverse( est.totalcost, (char *)"%6.6s", buf+56 );

            /* These two floats are not used in the 2023 version.
         * However, I am writing them to maintain backwards compatibility,
         * which is useful for testing the new 2023 Estimate Maker against
         * the old 1987 Estimate Maker.
         */

            //                          totalhrs float 4 chars
            //                          totalcost float 4 chars

            //Mar31,2023. I think this is wrong because the totFloatChar might
            //not be the real beginning of the float when you use a ptFloat to point.
            //See memory bug BIGTRADE.disaster

            // Putting them into a buf has tricky memory boundary issues:
            /*
        totFloatpt = (float *)totFloatChar; //msdos compatibility trick
        totFloat = est.totalhrs->text().toFloat();
        *totFloatpt = totFloat;
        strncpy( buf+62, totFloatChar, 4 );

        totFloatpt = (float *)totFloatChar; //msdos compatibility trick
        totFloat = est.totalcost->text().toFloat();
        *totFloatpt = totFloat;
        */

            totFloat = est.totalhrs->text().toFloat();
            strncpy( buf+62,(char *)&totFloat, 4 );

            totFloat = est.totalcost->text().toFloat();
            strncpy( buf+66, (char *)&totFloat, 4 );

            // skip the asterisk; eg. "*WD"
            strncpy( buf+74, est.refcode.toStdString().c_str(), 2);

            ;//qDebugA() << est.libcode->text() << est.description->text() << '|';

            for( yy=0; yy<78; ++yy)
                if( buf[yy] == '\0' )
                    buf[yy] = ' ';
            buf[78]='\r';
            buf[79]='\n';

            if( fp.write(buf,80) < 80)
            {
                ;//qDebugA() << "failed to write the next 80 chars.";
                break;
            }
        }
        ;//qDebugA() << "\n******\nWrote" << maxest << "estimate lines succeeded.\n";

        ret = true;

        fp.resize( fp.pos() ); // when I shrank the file trouble ensued.Mar2023
        fp.close();
        ui->tabWidget->setEnabled(true);

        return( ret );
    }

#endif

    void MainWindow::sprintfreverseTabversion( QString arg1, char *arg2, char *arg3 )
    {
        if( arg1 != nullptr )
            sprintf( arg3, arg2, arg1.toStdString().c_str() );
        //else
        //sprintf( arg3, arg2, "ZZ" );

    }

    void MainWindow::dialogPrint()
    {
        QPrinter printer;
        QTextEdit editor;
        QLineEdit *qle;
        QLabel *qla;
        qsizetype width;
        qsizetype xx,itemChars;
        long row, printedrows, rowsOfEstimateLinesOnPaper, linesPerPage;
        long rowsOfHeader, lookAhead=0L, lastEstLine=0L;
        unsigned col;
        bool allempty = true;
        QPainter painter;
        int whichPageIsPrinting, tt=1;
        short fontSz;
        QMessageBox msgBox;

        if( oneHeader == nullptr || estimateFilename == nullptr )
        {
            //;//qDebugA() << "msgBox.setText(There is nothing to print.);";
            msgBox.setText("There is nothing to print.");
            msgBox.exec();
        }
        else
        {
            /*
        enum QPrinter::ColorMode
        This enum type is used to indicate whether QPrinter should print in color or not.

        Constant	Value	Description
        QPrinter::Color	1	print in color if available, otherwise in grayscale.
        QPrinter::GrayScale	0	print in grayscale, even on color printers.
        enum QPrinter::DuplexMode
        This enum is used to indicate whether printing will occur on one or both sides of each sheet of paper (simplex or duplex printing).

        Constant	Value	Description
        QPrinter::DuplexNone	0	Single sided (simplex) printing only.
        QPrinter::DuplexAuto	1	The printer's default setting is used to determine whether duplex printing is used.
        QPrinter::DuplexLongSide	2	Both sides of each sheet of paper are used for printing. The paper is turned over its longest edge before the second side is printed
        QPrinter::DuplexShortSide	3	Both sides of each sheet of paper are used for printing. The paper is turned over its shortest edge before the second side is printed
        enum QPrinter::OutputFormat
        The OutputFormat enum is used to describe the format QPrinter should use for printing.

        Constant	Value	Description
        QPrinter::NativeFormat	0	QPrinter will print output using a method defined by the platform it is running on. This mode is the default when printing directly to a printer.
        QPrinter::PdfFormat	1	QPrinter will generate its output as a searchable PDF file. This mode is the default when printing to a file.
        See also outputFormat(), setOutputFormat(), and setOutputFileName().

        enum QPrinter::PageOrder
        This enum type is used by QPrinter to tell the application program how to print.

        Constant	Value	Description
        QPrinter::FirstPageFirst	0	the lowest-numbered page should be printed first.
        QPrinter::LastPageFirst	1	the highest-numbered page should be printed first.
        */

            /*
        *
        QAbstractPrintDialog::PrintToFile       0x0001	The print to file option is enabled.
        QAbstractPrintDialog::PrintSelection	0x0002	The print selection option is enabled.
        QAbstractPrintDialog::PrintPageRange	0x0004	The page range selection option is enabled.
        QAbstractPrintDialog::PrintShowPageSize	0x0008	Show the page size + margins page only if this is enabled.
        QAbstractPrintDialog::PrintCollateCopies	0x0010	The collate copies option is enabled
        QAbstractPrintDialog::PrintCurrentPage
        */

            QPrintDialog dialog(&printer, this);
            dialog.setWindowTitle(tr("Print Estimate"));

            if( editor.textCursor().hasSelection() )
            {
                dialog.setOption(QAbstractPrintDialog::PrintSelection, true);
                dialog.setOption(QAbstractPrintDialog::PrintToFile, true );
                dialog.setOption(QAbstractPrintDialog::PrintPageRange,true );
                dialog.setOption(QAbstractPrintDialog::PrintShowPageSize,true);
                dialog.setOption(QAbstractPrintDialog::PrintCollateCopies,false);
            }

            if( dialog.exec() != QDialog::Accepted )
                ;
            else
            {
                /* This was for testing. The QPrintDialog() sets these.
                 * See printer above instead of uncommenting these: */
                //printer.setOutputFormat(QPrinter::PdfFormat);
                //printer.setOutputFileName("/home/est/print.pdf");

                if( !painter.begin(&printer) )
                {
                    qWarning("Failed to open file. Is it writable?");
                }
                else
                {
                        if( tab[1]->nextTab(2) == true )
                        {
                            /* if tabPages 2 or any tab up to 10 are used
                             * then this is your place to print.
                             *
                             * Go to next else if only Tab 1 is used.
                             * That else() is the original print function
                             * before Tabs. I could erase it one day?
                             */

                            /* New March 4, 2024: putting print into a for() Tab 1 through 10.
                             */
                            ui->tabWidget->setEnabled(false);

                            /* I need to unload QLineEdits into its correct Tab
                             * before printing estimate or else the user's
                             * current QLineEdits will not print.
                             * I do a different idea in estimateWrite()
                             * where I write from estTabRow[]. Here I write from
                             * estRow[], which are QLineEdits because I want
                             * each Tab to print separately with a header
                             * so the subtotals look the truth, that the
                             * header is the start point for each subtotal.
                             * The estimateWrite() is continuous so I did a
                             * different idea.
                             */
                            tab[indexOld]->unloadQLineEditsToTab();
                            for( tt=1; tt<TABCOUNT; )
                            {
                                if( tab[tt]==nullptr ) //skip holes
                                    continue;
                                if( tt==1 ) // the rest are msgBox() questioned.
                                    tab[tt]->loadTabtoQLineEdits();

                                ;//qDebugA() << tt;

                                if( printer.fromPage() > 0 && printer.toPage() > 0 )
                                {
                            /* This if() had a bug where a blank sheet
                             * is ejected at the end.
                             * I wasted time fixing the .._sub() routine.
                             * The logic is too messy. I tried and failed.
                             * Instead do this: Call dialogPrint_sub()
                             * here with copies set to 1 and surround
                             * this, below, in a for() for the number of copies
                             *  for( int cc=printer.copyCount();cc>0;--cc)
                                    dialogPrint_sub(&painter,&printer,xx,1);
                             * also failed; ie. page eject blank at the end
                             * but the ;//qDebugA() showed NO newPage()!
                             *
                             * I think change _sub() to no newPage()s when
                             * xx!=0 and
                             * simply return; and put a for() loop for
                             * the copyCount() and newPage()s and skip doing
                             * the last newPage() because painter.end() does one?
                             * Yes, that worked.
                             */
                                    ;//qDebugA() << printer.fromPage() << printer.toPage();

                                    // user said something like pages 3 to 7
                                    for( int xx=printer.fromPage();
                                         xx<=printer.toPage(); ++xx )
                                    {
                                        for( int cc=printer.copyCount();cc>0;--cc)
                                        {
                                            dialogPrint_sub(&painter,&printer,xx,1,tt);
                                            if( cc==1 && xx==printer.toPage() && printer.fromPage()!=printer.toPage() )
                                            {
                                                ; // if() skip doing last newPage();
                                            }
                                            else
                                            {
                                                if( !printer.newPage() )
                                                {
                                                    qWarning("failed printing page; storage full or printer failed.");
                                                }
                                            }
                                        }
                                    }
                                }
                                else //print all pages,no collation,no range,no specifics.
                                {
                                    ;//qDebugA() << "print ALL. count:" << printer.copyCount();

                                    // 0 prints all pages
                                    dialogPrint_sub( &painter, &printer, 0, printer.copyCount(), tt);
                                }

                                tab[tt]->unloadQLineEditsToTab();
                                ;//qDebugA() << "painter.end for " << tt;
                                painter.end();

                                int start;
                                for( start=tt; tab[1]->nextTab(tt+1) == true; ++tt)
                                {
                                    if( tab[tt+1] == nullptr )
                                        continue;
                                    else
                                    {
                                        ++tt;
                                        break;
                                    }
                                }
                                if( start == tt )
                                    break;

                               QMessageBox msgBox;
                               QString ss = QString("Print the next batch: Tab %1?").arg(tt);

                                int ans;
                                if( tt < TABCOUNT && tab[tt] != nullptr )
                                {
                                    ;//qDebugA() << "before msgBox" << tt;

                                    tab[tt]->loadTabtoQLineEdits();
                                    msgBox.setText(ss);
                                    //msgBox.setInformativeText("Do you ");
                                    msgBox.setStandardButtons( QMessageBox::Yes | QMessageBox::No );
                                    msgBox.setDefaultButton(QMessageBox::Yes);
                                    ans = msgBox.exec();
                                }
                                else
                                    ans=QMessageBox::No;

                                if(ans == QMessageBox::No)
                                    break;
                                else
                                    if( !painter.begin(&printer) )
                                        break;
                            }
#ifdef AsPromised
                            /* load the Tab which the user sees as active, not tt */
                            tab[indexOld]->loadTabtoQLineEdits();
#endif
                            ui->tabWidget->setEnabled(true);
                        }
                        //if( dialog.testOption(QPrintDialog::PageRange) )
                        else if( printer.fromPage() > 0 && printer.toPage() > 0 )
                        {
                    /* This if() had a bug where a blank sheet
                     * is ejected at the end.
                     * I wasted time fixing the .._sub() routine.
                     * The logic is too messy. I tried and failed.
                     * Instead do this: Call dialogPrint_sub()
                     * here with copies set to 1 and surround
                     * this, below, in a for() for the number of copies
                     *  for( int cc=printer.copyCount();cc>0;--cc)
                            dialogPrint_sub(&painter,&printer,xx,1);
                     * also failed; ie. page eject blank at the end
                     * but the ;//qDebugA() showed NO newPage()!
                     *
                     * I think change _sub() to no newPage()s when
                     * xx!=0 and
                     * simply return; and put a for() loop for
                     * the copyCount() and newPage()s and skip doing
                     * the last newPage() because painter.end() does one?
                     * Yes, that worked.
                     */
                            ;//qDebugA() << printer.fromPage() << printer.toPage();

                            // user said something like pages 3 to 7
                            for( int xx=printer.fromPage();
                                 xx<=printer.toPage(); ++xx )
                            {
                                for( int cc=printer.copyCount();cc>0;--cc)
                                {
                                    dialogPrint_sub(&painter,&printer,xx,1,tt);
                                    if( cc==1 && xx==printer.toPage() && printer.fromPage()!=printer.toPage())
                                        ; // if() skip doing last newPage();
                                    else
                                    {
                                        if( !printer.newPage() )
                                        {
                                            qWarning("failed printing page; storage full or printer failed.");
                                        }
                                    }
                                }
                            }
                            painter.end();
                        }
                        else if( 0 && dialog.testOption(QPrintDialog::PrintSelection) )
                        {
                            // user named a list of pages, like 2,5,6
                            // I cannot find how to do this one.
                            painter.end();
                        }
#ifdef YouWantAnEndlessLoopBUG
                        else if( dialog.testOption(QAbstractPrintDialog::PrintCollateCopies) )
                        {
                            //collate flag means do all the 1s and then all the 2s.
                            //bug Feb2024
                            for( int xx=1; xx<=25; ++xx )//guess page number maximum
                            {
                                dialogPrint_sub( &painter, &printer, xx, printer.copyCount());
                            }
                            painter.end();
                        }
#endif

                        else //print all pages,no collation,no range,no specifics.
                        {
                            ;//qDebugA() << "print ALL. count:" << printer.copyCount();

                            // 0 prints all pages. It ejects no blank page. Best.
                            dialogPrint_sub( &painter, &printer, 0, printer.copyCount(),tt);
                            painter.end();
                        }
                }
            }
        }
        return;
    }

/* 0 prints all pages, or 1,2,3 .. chooses one page.
 * I run through the entire logic for each page although I have no
 * intention of printing most pages if 1,2,or3 is selected.
 * This coding strategy keeps the logic simple
 * for all or one-page printing. I simply skip the textDraw and newPage
 * functions for unprinted pages. I have no way to know which lines
 * are on which pages unless I run through the entire estimate this way.
 * I quit early after doing the one page requested, Feb 2024.
 * I do not quit early when pageAllorNum==0 (ALL pages).
 */
#ifdef NEWPAGE_is_a_failedIdea_forWhenOnePageIsAttempted_SeeELSEbelow
    bool MainWindow::dialogPrint_sub( QPainter *painter,
                                     QPrinter *printer, int pageAllorNum, int countOfPrints )
    {
        QTextEdit editor;
        QLineEdit *qle;
        QLabel *qla;
        QString ss;
        qsizetype width;
        qsizetype xx,itemChars;
        long row, printedrows, rowsOfEstimateLinesOnPaper, linesPerPage;
        long rowsOfHeader, lookAhead=0L, lastEstLine=0L;
        unsigned col;
        bool allempty = true;
        int whichPageIsPrinting;
        short fontSz;
        QMessageBox msgBox;

        // Find the end of the estimate to mark the quit point.
        for( allempty = true,row=ESTROWCOUNT-1; row>0L; --row)
        {

            for(col=1; col<=7;++col)
            {
                qle = estRow[row]->col[col];
                if( qle != nullptr )
                {
                    if( qle->text() != "" )
                    {
                        allempty = false;
                        break;
                    }
                }
            }
            if( allempty==false )
            {
                lastEstLine = row;
                break;
            }
        }

        for( ; countOfPrints>0; --countOfPrints )
        {
            //painter->setPen(Qt::blue); let the QPrintDialog do the settings
            fontSz = 8;
            QFont ff("monospace", fontSz);
            ff.setFixedPitch(true);
            ff.setLetterSpacing(QFont::PercentageSpacing,94);
            painter->setFont( ff );
            /* Print header of the estimate. */

            rowsOfHeader = 1L;
            rowsOfEstimateLinesOnPaper = 1L;
            whichPageIsPrinting=1;
            lookAhead=0L;
            int paperHeightInches = printer->pageRect(QPrinter::Point).height() / printer->resolution();
            //linesPerPage = 568/fontSz;
            linesPerPage = (printer->pageRect(QPrinter::Point).height()*0.675)/fontSz;
            int dpiPerRow = fontSz*3/2;
            printedrows = (linesPerPage)*(whichPageIsPrinting-1);

            ;//qDebugA() << "paperHeightInches" << paperHeightInches << "height" << printer->pageRect(QPrinter::Point).height() << "dpi" << printer->resolution();

            QLayout *layout1 = oneHeader->ui->formLayout->layout();
            QLayout *layout2 = oneHeader->ui->formLayout_2->layout();

            for( int ii=0; ii<layout1->count();++ii)
            {
                itemChars = 24;
                xx = 1;

                qla = qobject_cast<QLabel *>(layout1->itemAt(ii)->widget());
                if( qla )
                {
                    ss = qla->text();
                    ss = QString("%1:").arg(ss,8,' ');
                    //;//qDebugA() << ss;
                    if( pageAllorNum==0 || pageAllorNum==whichPageIsPrinting)
                        painter->drawText( (fontSz-1)*xx, dpiPerRow*(rowsOfHeader), ss );
                }

                qle = qobject_cast<QLineEdit *>(layout1->itemAt(ii)->widget());
                if( qle )
                {
                    ss = qle->text();
                    ss = QString("%1").arg(ss,0-itemChars,' ');
                    //;//qDebugA() << ss;
                    if( pageAllorNum==0 || pageAllorNum==whichPageIsPrinting)
                        painter->drawText( (fontSz-1)*(8+xx), dpiPerRow*(rowsOfHeader), ss );
                    ++rowsOfHeader;
                }
            }

            rowsOfHeader = 1L;

            for( int ii=0; ii<layout2->count();++ii)
            {
                itemChars = 24;
                xx = 8+itemChars+1;

                qla = qobject_cast<QLabel *>(layout2->itemAt(ii)->widget());
                if( qla )
                {
                    ss = qla->text();
                    ss = QString("%1:").arg(ss,8,' ');
                    //;//qDebugA() << ss;
                    if( pageAllorNum==0 ||pageAllorNum==whichPageIsPrinting)
                        painter->drawText( (fontSz-1)*xx, dpiPerRow*(rowsOfHeader), ss );
                }

                qle = qobject_cast<QLineEdit *>(layout2->itemAt(ii)->widget());
                if( qle )
                {
                    ss = qle->text();
                    ss = QString("%1").arg(ss,0-itemChars,' ');
                    //;//qDebugA() << ss;
                    if( pageAllorNum==0 ||pageAllorNum==whichPageIsPrinting)
                        painter->drawText( (fontSz-1)*(8+xx), dpiPerRow*(rowsOfHeader), ss );
                    ++rowsOfHeader;
                }
            }
            ++rowsOfHeader;
            printedrows += rowsOfHeader;

#ifdef NOTWORKINGYETandTHEOTHERWAYABOVEWORKSinstead
            if(0)
            {
                QList<QLineEdit *> lineEdits = oneHeader->ui->formLayout_2->findChildren<QLineEdit *>();
                //QList<QPushButton *> allPButtons = parentWidget.findChildren<QPushButton *>();
                for( QLineEdit *lineEdit : lineEdits )
                {
                    ;//qDebugA() << lineEdit->text();
                }
                QList<QLabel *> labels = oneHeader->ui->formLayout_2->findChildren<QLabel *>();
                for( QLabel *label : labels )
                {
                    ;//qDebugA() << label->text();
                }
            }
            return;
#endif
#ifdef SKIPTHISTOO_because_NotUsingAbsoluteNamesIsEasier_seeForLoopAbove
            qle = oneHeader->ui->clientLineEdit;
            ss = qle->text();
            ;//qDebugA() << ss << "<-printing this clientLineEdit";
            width = 1;

            painter.drawText( (fontSz-1)*width, fontSz*3/2*(rowsOfHeader), ss );

            qle = oneHeader->ui->addressLineEdit;
            ss = qle->text();
            painter.drawText( (fontSz-1)*width, fontSz*3/2*(++rowsOfHeader), ss );

            qle = oneHeader->ui->cityLineEdit;
            ss = qle->text();
            painter.drawText( (fontSz-1)*width, fontSz*3/2*(++rowsOfHeader), ss );

            qle = oneHeader->ui->postalCodeLineEdit;
            ss = qle->text();
            painter.drawText( (fontSz-1)*width, fontSz*3/2*(++rowsOfHeader), ss );

            qle = oneHeader->ui->siteLineEdit;
            ss = qle->text();
            painter.drawText( (fontSz-1)*width, fontSz*3/2*(++rowsOfHeader), ss );

            // COLUMN TWO and start at top row again:

            qle = oneHeader->ui->dateLineEdit;
            ss = qle->text();

            rowsOfHeader = 0L;
            width = qle->maxLength() + 5*fontSz;
            painter.drawText( (fontSz-1)*width, fontSz*3/2*(++rowsOfHeader), ss );

            qle = oneHeader->ui->projectLineEdit;
            ss = qle->text();
            painter.drawText( (fontSz-1)*width, fontSz*3/2*(++rowsOfHeader), ss );

            qle = oneHeader->ui->homePhoneLineEdit;
            ss = qle->text();
            painter.drawText( (fontSz-1)*width, fontSz*3/2*(++rowsOfHeader), ss );

            qle = oneHeader->ui->businessPhoneLineEdit;
            ss = qle->text();
            painter.drawText( (fontSz-1)*width, fontSz*3/2*(++rowsOfHeader), ss );

            qle = oneHeader->ui->contactNameLineEdit;
            ss = qle->text();
            painter.drawText( (fontSz-1)*width, fontSz*3/2*(++rowsOfHeader), ss );

            qle = oneHeader->ui->emailLineEdit;
            ss = qle->text();

#endif

            /* Print body of the estimate. */
            ++printedrows;  // add space under the header

            for( row = 1L; row <= lastEstLine; ++row)
            {
                allempty = true;
                for(col=1; col<=7;++col)
                {
                    qle = estRow[row]->col[col];
                    if( qle != nullptr )
                    {
                        if( qle->text() != "" )
                        {
                            allempty = false;
                            break;
                        }
                    }
                }
                if( !allempty )
                {
                    qsizetype totalwidth=0;

                    for(col=1; col<=7;++col)
                    {
                        qle = estRow[row]->col[col];
                        if( qle != nullptr )
                        {
                            ss = qle->text();
                            width = qle->maxLength();
                            if( col >=3 )
                            {
                                --width; // the page was not fitting left<-->right
                                ss = QString("%1").arg(ss,width,' ');
                            }
                            //else if(col == 2)
                            //    ;//qDebugA() << ss << "printedrows:" << printedrows << "pixelrow" << fontSz*11/8*(rowsOfHeader+(row-rowsOfEstimateLinesOnPaper));

                            if( pageAllorNum==0 ||pageAllorNum==whichPageIsPrinting)
                                painter->drawText( (fontSz-1)*totalwidth, fontSz*11/8*(rowsOfHeader+(row-rowsOfEstimateLinesOnPaper)), ss );
                            totalwidth += width;
                        }
                    }
                }

                if( printedrows < lookAhead )
                // catch up to completing the === block before considering a newPage()
                {
                    ;//qDebugA() << ss << "printedrows:" << printedrows << "lookAhead" << lookAhead << "pixelrow" << fontSz*11/8*(rowsOfHeader+(row-rowsOfEstimateLinesOnPaper)) << "row" << row;
                    ++printedrows; // increment normally and no newPage()
                }
                else
                {
                    /* If the next line is a === SUBTOTAL then skip forward
                     * until === is not true and set lookAhead to that row.
                     */
                    lookAhead = row+1;
                    while( lookAhead<=lastEstLine && estRow[lookAhead]->col[1]->text().indexOf('=') >= 0 )
                        ++lookAhead;

                    /* subtotal/total blocks must not be split printed across two pages.
                 * So if printedrows <= lookAhead we test for needing a newPage()
                 * using printedrows  incrementing it until we hit a needed newPage().
                 */
                    long anticipated_printedrows = printedrows;
                    ++printedrows; //must do this once even if no newpage()

                    for( long rrr=row; rrr < lookAhead; ++rrr ) //ie.lookAhead bigger than printedrows+1 == lookAhead
                    {
                        ;//qDebugA() << estRow[rrr]->col[1]->text() << "anticipated" << anticipated_printedrows << "lookAhead" << lookAhead << "pixelrow" << fontSz*11/8*(rowsOfHeader+(row-rowsOfEstimateLinesOnPaper)) << "row" << row << "(linesPerPage)*whichPageIsPrinting" << (linesPerPage)*whichPageIsPrinting;

                        if( ++anticipated_printedrows >= (linesPerPage)*whichPageIsPrinting )
                        {
                            if(pageAllorNum==0||pageAllorNum==whichPageIsPrinting)
                            {
                                if( !printer->newPage() )
                                {
                                    qWarning("failed printing page; storage full or printer failed.");
                                    return(false);
                                }
                                ;//qDebugA() << "newPage()" << whichPageIsPrinting;
                            }
                            ++whichPageIsPrinting;
                            rowsOfHeader=0L;
                            rowsOfEstimateLinesOnPaper = row;
                            printedrows = anticipated_printedrows;
                            break;
                        }
                    }
                }
            }
            if( pageAllorNum != 0 )
            {
                if( pageAllorNum==whichPageIsPrinting)
                {
                    if( !printer->newPage() )
                    {
                        qWarning("failed printing page; storage full or printer failed.");
                        return(false);
                    }
                    ;//qDebugA() << "newPage()endofloop" << whichPageIsPrinting;
                    ++whichPageIsPrinting;
                }
            }
            else if( countOfPrints > 1 )
            {
                if( !printer->newPage() )
                {
                    qWarning("failed printing page; storage full or printer failed.");
                    return(false);
                }
                ;//qDebugA() << "newPage()endofloopmultipage" << whichPageIsPrinting;
                ++whichPageIsPrinting;
            }
        }

        return(true);
    }

#else  /* This version omits all the newPage() when printing one page;
        * pageAllorNum != 0. I had a bug where a
        * blank page was being ejected at the end because
        * painter.end() does an invisible newPage() call; now fixed.
        */

    bool MainWindow::dialogPrint_sub( QPainter *painter,
                                     QPrinter *printer, int pageAllorNum, int countOfPrints, int whichtt )
    {
        QTextEdit editor;
        QLineEdit *qle;
        QLabel *qla;
        QString ss;
        qsizetype width;
        qsizetype xx,itemChars;
        long row, printedrows, rowsOfEstimateLinesOnPaper, linesPerPage;
        long rowsOfHeader, lookAhead=0L, lastEstLine=0L;
        unsigned col;
        bool allempty = true;
        int whichPageIsPrinting;
        short fontSz;
        QMessageBox msgBox;

        // Find the end of the estimate to mark the quit point.
        for( allempty = true,row=ESTROWCOUNT-1; row>0L; --row)
        {

            for(col=1; col<=7;++col)
            {
                qle = estRow[row]->col[col];
                if( qle != nullptr )
                {
                    if( qle->text() != "" )
                    {
                        allempty = false;
                        break;
                    }
                }
            }
            if( allempty==false )
            {
                lastEstLine = row;
                break;
            }
        }

        for( ; countOfPrints>0; --countOfPrints )
        {
            //painter->setPen(Qt::blue); let the QPrintDialog do the settings
            fontSz = 8;
            QFont ff("monospace", fontSz);
            ff.setFixedPitch(true);
            ff.setLetterSpacing(QFont::PercentageSpacing,94);
            painter->setFont( ff );
            /* Print header of the estimate. */

            rowsOfHeader = 1L;
            rowsOfEstimateLinesOnPaper = 1L;
            whichPageIsPrinting=1;
            lookAhead=0L;
            //int paperHeightInches = printer->pageRect(QPrinter::Point).height() / printer->resolution();
            //linesPerPage = 568/fontSz;
            linesPerPage = (printer->pageRect(QPrinter::Point).height()*0.675)/fontSz;
            int dpiPerRow = fontSz*3/2;
            printedrows = (linesPerPage)*(whichPageIsPrinting-1);

            ;//qDebugA() << "paperHeightInches" << paperHeightInches << "height" << printer->pageRect(QPrinter::Point).height() << "dpi" << printer->resolution();

            QLayout *layout1 = oneHeader->ui->formLayout->layout();
            QLayout *layout2 = oneHeader->ui->formLayout_2->layout();

            for( int ii=0; ii<layout1->count();++ii)
            {
                itemChars = 24;
                xx = 1;

                qla = qobject_cast<QLabel *>(layout1->itemAt(ii)->widget());
                if( qla )
                {
                    ss = qla->text();
                    ss = QString("%1:").arg(ss,8,' ');
                    //;//qDebugA() << ss;
                    if( pageAllorNum==0 || pageAllorNum==whichPageIsPrinting)
                        painter->drawText( (fontSz-1)*xx, dpiPerRow*(rowsOfHeader), ss );
                }

                qle = qobject_cast<QLineEdit *>(layout1->itemAt(ii)->widget());
                if( qle )
                {
                    ss = qle->text();
                    ss = QString("%1").arg(ss,0-itemChars,' ');
                    //;//qDebugA() << ss;
                    if( pageAllorNum==0 || pageAllorNum==whichPageIsPrinting)
                        painter->drawText( (fontSz-1)*(8+xx), dpiPerRow*(rowsOfHeader), ss );
                    ++rowsOfHeader;
                }
            }

            rowsOfHeader = 1L;

            for( int ii=0; ii<layout2->count();++ii)
            {
                itemChars = 24;
                xx = 8+itemChars+1;

                qla = qobject_cast<QLabel *>(layout2->itemAt(ii)->widget());
                if( qla )
                {
                    ss = qla->text();
                    ss = QString("%1:").arg(ss,8,' ');
                    //;//qDebugA() << ss;
                    if( pageAllorNum==0 ||pageAllorNum==whichPageIsPrinting)
                        painter->drawText( (fontSz-1)*xx, dpiPerRow*(rowsOfHeader), ss );
                }

                qle = qobject_cast<QLineEdit *>(layout2->itemAt(ii)->widget());
                if( qle )
                {
                    ss = qle->text();
                    ss = QString("%1").arg(ss,0-itemChars,' ');
                    //;//qDebugA() << ss;
                    if( pageAllorNum==0 ||pageAllorNum==whichPageIsPrinting)
                        painter->drawText( (fontSz-1)*(8+xx), dpiPerRow*(rowsOfHeader), ss );
                    ++rowsOfHeader;
                }
            }
            ++rowsOfHeader;
            printedrows += rowsOfHeader;

#ifdef NOTWORKINGYETandTHEOTHERWAYABOVEWORKSinstead
            if(0)
            {
                QList<QLineEdit *> lineEdits = oneHeader->ui->formLayout_2->findChildren<QLineEdit *>();
                //QList<QPushButton *> allPButtons = parentWidget.findChildren<QPushButton *>();
                for( QLineEdit *lineEdit : lineEdits )
                {
                    ;//qDebugA() << lineEdit->text();
                }
                QList<QLabel *> labels = oneHeader->ui->formLayout_2->findChildren<QLabel *>();
                for( QLabel *label : labels )
                {
                    ;//qDebugA() << label->text();
                }
            }
            return;
#endif
#ifdef SKIPTHISTOO_because_NotUsingAbsoluteNamesIsEasier_seeForLoopAbove
            qle = oneHeader->ui->clientLineEdit;
            ss = qle->text();
            ;//qDebugA() << ss << "<-printing this clientLineEdit";
            width = 1;

            painter.drawText( (fontSz-1)*width, fontSz*3/2*(rowsOfHeader), ss );

            qle = oneHeader->ui->addressLineEdit;
            ss = qle->text();
            painter.drawText( (fontSz-1)*width, fontSz*3/2*(++rowsOfHeader), ss );

            qle = oneHeader->ui->cityLineEdit;
            ss = qle->text();
            painter.drawText( (fontSz-1)*width, fontSz*3/2*(++rowsOfHeader), ss );

            qle = oneHeader->ui->postalCodeLineEdit;
            ss = qle->text();
            painter.drawText( (fontSz-1)*width, fontSz*3/2*(++rowsOfHeader), ss );

            qle = oneHeader->ui->siteLineEdit;
            ss = qle->text();
            painter.drawText( (fontSz-1)*width, fontSz*3/2*(++rowsOfHeader), ss );

            // COLUMN TWO and start at top row again:

            qle = oneHeader->ui->dateLineEdit;
            ss = qle->text();

            rowsOfHeader = 0L;
            width = qle->maxLength() + 5*fontSz;
            painter.drawText( (fontSz-1)*width, fontSz*3/2*(++rowsOfHeader), ss );

            qle = oneHeader->ui->projectLineEdit;
            ss = qle->text();
            painter.drawText( (fontSz-1)*width, fontSz*3/2*(++rowsOfHeader), ss );

            qle = oneHeader->ui->homePhoneLineEdit;
            ss = qle->text();
            painter.drawText( (fontSz-1)*width, fontSz*3/2*(++rowsOfHeader), ss );

            qle = oneHeader->ui->businessPhoneLineEdit;
            ss = qle->text();
            painter.drawText( (fontSz-1)*width, fontSz*3/2*(++rowsOfHeader), ss );

            qle = oneHeader->ui->contactNameLineEdit;
            ss = qle->text();
            painter.drawText( (fontSz-1)*width, fontSz*3/2*(++rowsOfHeader), ss );

            qle = oneHeader->ui->emailLineEdit;
            ss = qle->text();

#endif

            /* Print body of the estimate. */
            ++printedrows;  // add space under the header

            for( row = 1L; row <= lastEstLine; ++row)
            {
                // overwrites the same spot at line 0, which is ok to do.
                if( tab[1]->nextTab(2) == true )
                {
                    // paint a Tab/page number above the header
                    ss = QString("t%1p%2").arg(whichtt).arg(whichPageIsPrinting);
                    //xx = printer->pageRect(QPrinter::Point).width() / printer->resolution();
                    itemChars = 24; // same as "Date:"
                    xx = 8+itemChars+1;
                    if( pageAllorNum==0 || pageAllorNum==whichPageIsPrinting)
                        painter->drawText( (fontSz-1)*xx, dpiPerRow*(0), QString("%1").arg(ss,8,' ') ); /* 0-itemChars is 0-24*/
                }

                allempty = true;
                for(col=1; col<=7;++col)
                {
                    qle = estRow[row]->col[col];
                    if( qle != nullptr )
                    {
                        if( qle->text() != "" )
                        {
                            allempty = false;
                            break;
                        }
                    }
                }
                if( !allempty )
                {
                    qsizetype totalwidth=0;

                    for(col=1; col<=7;++col)
                    {
                        qle = estRow[row]->col[col];
                        if( qle != nullptr )
                        {
                            ss = qle->text();
                            width = qle->maxLength();
                            if( col >=3 )
                            {
                                --width; // the page was not fitting left<-->right
                                ss = QString("%1").arg(ss,width,' ');
                            }
                            //else if(col == 2)
                            //    ;//qDebugA() << ss << "printedrows:" << printedrows << "pixelrow" << fontSz*11/8*(rowsOfHeader+(row-rowsOfEstimateLinesOnPaper));

                            if( pageAllorNum==0 ||pageAllorNum==whichPageIsPrinting)
                                painter->drawText( (fontSz-1)*totalwidth, fontSz*11/8*(rowsOfHeader+(row-rowsOfEstimateLinesOnPaper)), ss );
                            totalwidth += width;
                        }
                    }
                }

                if( printedrows < lookAhead )
                // catch up to completing the === block before considering a newPage()
                {
                    ;//qDebugA() << ss << "printedrows:" << printedrows << "lookAhead" << lookAhead << "pixelrow" << fontSz*11/8*(rowsOfHeader+(row-rowsOfEstimateLinesOnPaper)) << "row" << row;
                    ++printedrows; // increment normally and no newPage()
                }
                else
                {
                    /* If the next line is a === SUBTOTAL then skip forward
                     * until === is not true and set lookAhead to that row.
                     */
                    lookAhead = row+1;
                    while( lookAhead<=lastEstLine && estRow[lookAhead]->col[1]->text().indexOf('=') >= 0 )
                        ++lookAhead;

                    /* subtotal/total blocks must not be split printed across two pages.
                 * So if printedrows <= lookAhead we test for needing a newPage()
                 * using printedrows  incrementing it until we hit a needed newPage().
                 */
                    long anticipated_printedrows = printedrows;
                    ++printedrows; //must do this once even if no newpage()

                    for( long rrr=row; rrr < lookAhead; ++rrr ) //ie.lookAhead bigger than printedrows+1 == lookAhead
                    {
                        ;//qDebugA() << estRow[rrr]->col[1]->text() << "anticipated" << anticipated_printedrows << "lookAhead" << lookAhead << "pixelrow" << fontSz*11/8*(rowsOfHeader+(row-rowsOfEstimateLinesOnPaper)) << "row" << row << "(linesPerPage)*whichPageIsPrinting" << (linesPerPage)*whichPageIsPrinting;

                        if( ++anticipated_printedrows >= (linesPerPage)*whichPageIsPrinting )
                        {
                            if(pageAllorNum==whichPageIsPrinting)
                            {
                                ;//qDebugA() << "newPage()" << whichPageIsPrinting;
                                return(true);
                            }
                            if(pageAllorNum==0)
                            {
                                if( !printer->newPage() )
                                {
                                    qWarning("failed printing page; storage full or printer failed.");
                                    return(false);
                                }
                                ;//qDebugA() << "newPage()" << whichPageIsPrinting;
                            }
                            ++whichPageIsPrinting;
                            rowsOfHeader=0L;
                            rowsOfEstimateLinesOnPaper = row;
                            printedrows = anticipated_printedrows;
                            break;
                        }
                    }
                }
            }
            if( pageAllorNum != 0 )
            {
                if( pageAllorNum==whichPageIsPrinting)
                {
                    ;//qDebugA() << "newPage()endofloop" << whichPageIsPrinting;
                    return(true);
                }
                return(false);
            }
            else if( countOfPrints > 1 )
            {
                if( !printer->newPage() )
                {
                    qWarning("failed printing page; storage full or printer failed.");
                    return(false);
                }
                ;//qDebugA() << "newPage()endofloopmultipage" << whichPageIsPrinting;
                ++whichPageIsPrinting;
            }
        }

        return(true);
    }
#endif
