
/*  This file is part of PaceMaker Estimator 2.0.
    PaceMaker Estimator 2.0 is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
    PaceMaker Estimator 2.0 is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with PaceMaker Estimator 2.0. If not, see <https://www.gnu.org/licenses/>.
*/

#include "mainwindow.h"

void MainWindow::validateField()
{
        QLineEdit *qle;
        QString ss, tt;
        QChar *source, *target;
        unsigned size, newsz;

        ;//qDebugA() << "try to fix rejected keyboard input";
        // find the QLineEdit and repair it to something suitable; eg. 12jds becomes 12 for quantity.

        // My validator from scratch.

        if( (qle = currentKeyboardLineEdit()) != nullptr )
        {
            if( qle->isModified() == true )
            {
                tt = ss = qle->text();
                newsz = size = ss.size();

                ;//qDebugA() << "input to be validated:" << ss;

                switch( keyboardnow.col )
                {
                case 1:  // f^5 becomes "f 5"
                    for( source = ss.data(), target = tt.data(), newsz=0; !source->isNull() ; ++source)
                    {
                        if( source->isLetterOrNumber() == true)
                        {
                            *target = *source;
                            //;//qDebugA() << *target;
                            ++target;
                            ++newsz;
                        }
                    }
                    qleToUpper( qle );

                break;

                case 2: // descriptions untouched for now
                break;

                case 3: // On lib or est side these are numbers and could be decimals (lib has no 5)
                case 4:
                case 5:
#define CALCULATOR_Arith_Feb2024
#ifdef CALCULATOR_Arith_Feb2024
                    if( ss.contains('+') ||
                        ss.contains('-') ||
                        ss.contains('/') ||
                        ss.contains('*') ||
                        ss.contains('^')
                        )
                    {
                        qtymath *qtytemp = new qtymath( this );
                        // this maintains the old 1987 msdos file format:
                        char buf1987[42];
                        sprintf( buf1987, "%.40s", ss.toStdString().c_str() );
                        char *ptbuf1987 = buf1987;
                        float totFloat;
                        totFloat = tediously_meticulously1987compatibility( ptbuf1987, qtytemp );
                        ;//qDebugA() << "arith() calc on this line:" << ss << "=" << totFloat;
                        if( totFloat>999.99 )
                            ss = QString("%1").arg(totFloat, 12,'g', 6, ' ' );
                        else if( totFloat>99.99 )
                            ss = QString("%1").arg(totFloat, 12,'g', totFloat<10.0?4:5, ' ' );
                        else
                            ss = QString("%1").arg(totFloat, 12,'g', 4, ' ');
                        delete qtytemp;
                        ;//qDebugA() << "ss line:" << ss ;
                    }
#endif
                    for( source = ss.data(), target = tt.data(), newsz=0; !source->isNull() ; ++source)
                    {
                        if( source->isNumber() == true || *source == '.' )
                        {
                            *target = *source;
                            //;//qDebugA() << *target;

                            ++target;
                            ++newsz;
                        }
                    }
                break;
                }
                // I truncate in case I skipped invalid chars
                if( size != newsz )
                {
                    tt.resize(newsz);
                }

       /* Do not update unless we did a correction because doing
        * a setText() destroys our ability to do undo/redo
        * See docalc() comments where I had to find this bug.
        * Edit: Later I discovered clear() insert() would
        * work better, even without the: if( tt != ss )
        */
                if( tt != ss )
                {
                    // qle->setText( tt );
                    // These pair keep the undo/redo stack
                    // and otherwise do the same as setText()
                    qle->clear();
                    qle->insert( tt );

                    ;//qDebugA() << "fixed->" << qle->text();
                }
            }
        }

        // I see Qt's Validator remove spaces but letters mixed with numbers leaked through.
        // Maybe I needed a separate validator for each QLineEdit.
        // I quit when I realized I was sharing one and fixup() was not doing enough for me.
        // I can write a validator from scratch faster and share it for all inputs simpler
        // than 11,000 of them (which Qt Creator limitation forces me to do because of widget labeling).
        // and trying to learn this one.
#ifdef PRETTYUSELESS_Validator
        if( keyboardnow.libest == 2 )
        {
            qle = estRow[ keyboardnow.row ]->col[ keyboardnow.col ];
            ss = qle->text();
            ;//qDebugA() << "input rejected" << ss;

            if( qle->validator() != nullptr)
            {
                qle->validator()->fixup( ss );
                ;//qDebugA() << "What we fixup() to:" << ss;
            }
        }
        else
        {
            qle = libRow[ keyboardnow.row ]->col[ keyboardnow.col ];
            ss = qle->text();
            ;//qDebugA() << "input rejected" << ss;

            if( qle->validator() != nullptr)
            {
                    qle->validator()->fixup( ss );
                    ;//qDebugA() << "What we fixup() to:" << ss;
            }
        }
#endif

}

