
/*  This file is part of PaceMaker Estimator 2.0.
    PaceMaker Estimator 2.0 is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
    PaceMaker Estimator 2.0 is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with PaceMaker Estimator 2.0. If not, see <https://www.gnu.org/licenses/>.
*/

#include "mainwindow.h"

// This vertical slider is for scrolling up/dn lib and est sides together.
// lib and est also have their individual vertical scrolling, instead of locked scrolling (verticalSlider).
// I use the last known keyboard cursor square.  I initialize the keyboardnow.row value to 1
// so I think hovering over one side or other is not necessary at least once before scrolling.
// But this needs testing to see if it true. It would be odd if the slider is ignored until
// the user hovers over at least one square.
void MainWindow::on_verticalSlider_valueChanged(int value)
{
    int cur = ui->verticalSlider->value();
    int hh,ww, diff, faster=1; // faster==7 crashes. I do not know why. I don't need this feature.
    QLineEdit *qle =  nullptr;
    long rowl, rowe, lowest, highest;

    //;//qDebugA() << "verticalSlider value changed to:" << cur;

    //savetherow = keyboardnow.row;

    // before moving lib and est windows vertically, finish calculation on that row, and
    // next quit focus on keyboard cursor square, if any is active.
    qle = currentKeyboardLineEdit();
    if( qle != nullptr )
    {
        // I think clearCurrentKeyboardCursor() will emit a signal which calls doCalc()
        // so do not call doCalc()until we check if we have to.
        // Edit: confirmed. Focus clear does emit.
        // doCalc();
        clearCurrentKeyboardCursorFocus();
        qle =  nullptr;
    }

    // we have history and current keyboardnow.row .rowE .rowL (see mainwindow.h notes)
    // clearCurrentKeyboardCursorFocus() throws a FocusOut in events.cpp which set rowE rowL
    // history. EDIT: That means savetherow is unnecessary because it's values are moved
    // to the history by the FocusOut. (see in events.cpp). So it cannot be higher/lower.
    // I left in the coding for saevtherow because if it isn't broken then do not fix it.
    // Find a row est side and row lib side:

    rowe = keyboardnow.rowE;
    rowl = keyboardnow.rowL;

    // Use the same row for both sides if necessary.
    if( !rowe )
        rowe = rowl;
    if( !rowl )
        rowl = rowe;

    if(rowl < rowe)
    {
       lowest = rowl;
       highest = rowe;
    }
    else
    {
       lowest = rowe;
       highest = rowl;
    }
#ifdef NOTANYMORE
    if( savetherow > 1L && savetherow < ESTROWCOUNT-2 )
    {
        if( savetherow < lowest )
           lowest = savetherow;
        if( savetherow > highest )
           highest = savetherow;
    }
#endif

#ifdef SKIPTHIS
    // because equating them was not different after all. Doing lowest/highest is new.
    // Next, we have to have lib and est sides using the same row or else
    // the code below loses the lock when scrolling at the top and the bottom.
    // You can comment out this equate to see the scroll problem.
    if( (rowe != rowl) && savetherow > 1L && savetherow < ESTROWCOUNT-2 )
    {
                        //no: keyboardnow.rowL = keyboardnow.rowE =
        rowe = rowl = savetherow;
        // This is my final addition after writing the code.
        // Maybe this could have been written differently. But just leave it,
        // so you can review the scroll problem just be commenting out this if().
        // ensureWidgetVisible() is the problem. We do not know if it scrolls,
        // before it scrolls. One side scroll the other not, and then we lose lock.
    }
#endif

    /* I love this code outcome. Dec 25,2022 midnight plus 51m.
     * Tell the user to keyboard click on any two est and lib squares
     * that you want to be adjacent and then scrolling
     * with the verticalSlider() will move the two squares
     * together as you move beyond the top or bottom of the page.
     * Also, clicking on two adjacent squares makes most sense because they
     * remain adjacent no matter how small of large you verticalSlider scroll.
     */

    diff = lastSliderValue-cur;

  while( --faster>=0 ) // do it more than once; added Jan 12, 2024.
  {
    if(cur<=7 || cur>=93 ) // was 0 and 99 but those are too close to ends.
        ui->verticalSlider->setValue(50);
    else if( diff <= -1 && diff > -48)
    {
        // scroll lib and est side together up.

        if( lowest > 1 )
        {
            // 0 will never happen. Good. There is no 0.

            /* I commented out all the keyboard cursors because the keyboard
             * is always set if there is a keyboard cursor because it hovers
             * over the same square.
             * So there is no reason to do both.
             * So just use the last keyboard cursor as the starting scroll square.
                    if(keyboardnow.row>1)
                            --keyboardnow.row;
             */
            // crashes when rowL==0: qle =libRow[keyboardnow.rowL]->col[keyboardnow.colL];
            qle = ( keyboardnow.colL == 0)?nullptr :  libRow[rowl]->col[keyboardnow.colL];
            if( qle!=nullptr)
            {
                hh = qle->frameGeometry().height();
                ww = qle->frameGeometry().width();
                ui->scrollArea_2->ensureWidgetVisible( qle, ww, hh);

                --keyboardnow.rowL;
                keyboardnow.row = rowl - 1;
            }

            // crashable: qle =          estRow[keyboardnow.rowE]->col[keyboardnow.colE];
            qle = ( keyboardnow.colE == 0)?nullptr : estRow[rowe]->col[keyboardnow.colE];
            if( qle!=nullptr)
            {
                hh = qle->frameGeometry().height();
                ww = qle->frameGeometry().width();
                ui->scrollArea->ensureWidgetVisible( qle, ww, hh);

                --keyboardnow.rowE;
                keyboardnow.row = rowe - 1;
            }
            show_linenums_showHighlights( &keyboardnow.litLibButton, &keyboardnow.litEstButton, keyboardnow.rowL, keyboardnow.rowE );
        }
    }
    else if( diff >= 1 && diff < 48)
    {
        if( highest < ESTROWCOUNT-2 )
        {
            // ESTROWCOUNT will never happen. Good. There is no ESTROWCOUNT.

            // crashes:            qle = libRow[keyboardnow.rowL]->col[keyboardnow.colL];
            qle = ( keyboardnow.colL == 0)?nullptr : libRow[rowl]->col[keyboardnow.colL];
            if( qle!=nullptr)
            {
                hh = qle->frameGeometry().height();
                ww = qle->frameGeometry().width();
                ui->scrollArea_2->ensureWidgetVisible( qle, ww, hh);

                ++keyboardnow.rowL;
                keyboardnow.row = rowl + 1;
            }
            // crashable: qle =          estRow[keyboardnow.rowE]->col[keyboardnow.colE];
            qle = ( keyboardnow.colE == 0)?nullptr : estRow[rowe]->col[keyboardnow.colE];
            if( qle!=nullptr)
            {
                hh = qle->frameGeometry().height();
                ww = qle->frameGeometry().width();
                ui->scrollArea->ensureWidgetVisible( qle, ww, hh);

                ++keyboardnow.rowE;
                keyboardnow.row = rowe + 1;
            }
            show_linenums_showHighlights( &keyboardnow.litLibButton, &keyboardnow.litEstButton, keyboardnow.rowL, keyboardnow.rowE );
        }
    }
  }
  lastSliderValue = cur;

  return;
}

void MainWindow::on_clearSlider_pressed()
{
    eraseHighlights( &keyboardnow.litLibButton, &keyboardnow.litEstButton);
    keyboardnow.litLibButton = keyboardnow.litEstButton = 0L;
    keyboardnow.rowE = keyboardnow.rowL = keyboardnow.row = 0L;
    setEditSelected_Row();
    keyboardnow.colE = keyboardnow.colL = keyboardnow.col = 0L;
    show_linenums_showHighlights( &keyboardnow.litLibButton, &keyboardnow.litEstButton, keyboardnow.row, keyboardnow.row );
}
